//
// syd: seccomp and landlock based application sandbox with support for namespaces
// benches/bench.rs Benchmarks
//
// Copyright (c) 2023, 2024 Ali Polatel <alip@chesswob.org>
//
// SPDX-License-Identifier: GPL-3.0-or-later

use std::{
    fs::{create_dir_all, remove_dir_all, File},
    os::unix::fs::symlink,
    path::{Path, PathBuf},
};

use criterion::{black_box, criterion_group, criterion_main, Criterion};
use nix::unistd::Pid;
use syd::{
    fs::{canonicalize, path_unsafe, MissingHandling},
    proc::proc_fd,
};
use tempfile::tempdir;

fn setup_paths() -> (PathBuf, PathBuf, PathBuf, PathBuf, PathBuf) {
    let temp_dir = tempdir().expect("Failed to create a temp dir");
    let temp_dir_path = temp_dir.path();

    // Existing path with symlinks
    let existing_path = temp_dir_path.join("existing");
    create_dir_all(&existing_path).expect("Failed to create existing path");
    let symlink_path = temp_dir_path.join("symlink");
    symlink(&existing_path, &symlink_path).expect("Failed to create symlink");

    // Self-referencing loop
    let loop_path = temp_dir_path.join("loop");
    create_dir_all(&loop_path).expect("Failed to create loop path");
    let loop_symlink = loop_path.join("self_loop");
    symlink(&loop_path, &loop_symlink).expect("Failed to create self-referencing symlink");

    // Non-existing path
    let non_existing_path = temp_dir_path.join("non_existing");

    // Complex structure setup
    let complex_base = temp_dir_path.join("syd-test");
    let complex_a = complex_base.join("a");
    let complex_1 = complex_base.join("1");
    let complex_target = complex_a.join("target.txt");
    let complex_link_to_a = complex_1.join("2/3/link_to_a");
    let complex_link_to_1 = complex_a.join("b/c/d/link_to_1");
    let complex_link_to_c = complex_1.join("2/3/link_to_c");

    create_dir_all(complex_a.join("b/c/d/e/f"))
        .expect("Failed to create complex a directory structure");
    create_dir_all(complex_1.join("2/3")).expect("Failed to create complex 1 directory structure");
    File::create(&complex_target).expect("Failed to create target file");

    symlink("../../../a", &complex_link_to_a).expect("Failed to create symlink to a");
    symlink("../../../../1", &complex_link_to_1).expect("Failed to create symlink to 1");
    symlink("../../../a/b/c", &complex_link_to_c).expect("Failed to create symlink to c");

    let complex_path = complex_link_to_a.join("b/c/d/e/f/../../../../../b/c/d/link_to_1/../../syd-test/1/2/3/link_to_c/d/e/f/../../link_to_1/2/../././../a/.././a/target.txt");

    (
        temp_dir_path.to_path_buf(),
        non_existing_path,
        symlink_path,
        loop_symlink,
        complex_path,
    )
}

pub fn canonicalize_benchmark(c: &mut Criterion) {
    let (temp_dir_path, non_existing, symlink, loop_path, complex_path) = setup_paths();

    let pid = Pid::this();
    let modes = [
        MissingHandling::Normal,
        MissingHandling::Existing,
        MissingHandling::Missing,
    ];

    for &mode in &modes {
        c.bench_function(&format!("canonicalize_complex_{:?}", mode), |b| {
            b.iter(|| {
                canonicalize(
                    pid,
                    black_box(&complex_path),
                    None,
                    black_box(true),
                    black_box(false),
                    black_box(mode),
                )
            })
        });

        c.bench_function(&format!("canonicalize_non_existing_{:?}", mode), |b| {
            b.iter(|| {
                canonicalize(
                    pid,
                    black_box(&non_existing),
                    None,
                    black_box(true),
                    black_box(false),
                    black_box(mode),
                )
            })
        });

        c.bench_function(&format!("canonicalize_symlink_{:?}", mode), |b| {
            b.iter(|| {
                canonicalize(
                    pid,
                    black_box(&symlink),
                    None,
                    black_box(true),
                    black_box(false),
                    black_box(mode),
                )
            })
        });

        c.bench_function(&format!("canonicalize_loop_{:?}", mode), |b| {
            b.iter(|| {
                canonicalize(
                    pid,
                    black_box(&loop_path),
                    None,
                    black_box(true),
                    black_box(false),
                    black_box(mode),
                )
            })
        });
    }

    let _ = remove_dir_all(temp_dir_path);
}

pub fn path_unsafe_benchmark(c: &mut Criterion) {
    let paths = vec![
        PathBuf::from("/dev/mem"),
        PathBuf::from("/home/user/safe_file"),
        PathBuf::from("/dev/cpu/0/msr"),
        PathBuf::from("/proc/1/environ"),
        PathBuf::from("/proc/1/task/1/mem"),
        PathBuf::from("/tmp"),
        PathBuf::from("/proc/kallsyms"),
    ];

    c.bench_function("path_unsafe", |b| {
        b.iter(|| {
            for path in &paths {
                let _ = path_unsafe(black_box(path));
            }
        })
    });
}

pub fn proc_fd_benchmark(c: &mut Criterion) {
    let this = Pid::this();
    let that = Pid::from_raw(1);
    let this_path = format!("/proc/{this}/fd/0");
    let that_path = "/proc/1/fd/0";

    c.bench_function("proc_fd_this", |b| {
        b.iter(|| proc_fd(this, black_box(&this_path)))
    });

    c.bench_function("proc_fd_that", |b| {
        b.iter(|| proc_fd(that, black_box(&that_path)))
    });
}

pub fn sandbox_globset_benchmark(c: &mut Criterion) {
    let path = PathBuf::from("/etc/passwd");
    for i in &[10, 100, 1000, 10000, 100000] {
        let mut sandbox = syd::sandbox::Sandbox::default();
        for j in 0..*i {
            sandbox
                .config(if j % 2 == 0 {
                    "allow/read+/etc/***"
                } else {
                    "deny/read+/etc/***"
                })
                .unwrap();
        }
        sandbox.build_globsets().unwrap();
        c.bench_function(&format!("sandbox_match {i}"), |b| {
            b.iter(|| {
                sandbox.match_action(syd::sandbox::Capability::CAP_READ, black_box(&path));
            })
        });
        drop(sandbox);
    }
}

pub fn sandbox_forcemap_benchmark(c: &mut Criterion) {
    for i in &[100, 250, 500, 750, 1000] {
        c.bench_function(&format!("sandbox_forcemap {i}"), |b| {
            b.iter(|| {
                // Create sandbox
                let mut sandbox = syd::sandbox::Sandbox::default();
                sandbox.config("sandbox/force:on").unwrap();

                // Initialize sandbox
                let mut paths = vec![];
                for j in 0..*i {
                    let path = format!("/usr/bin/test-{j}");
                    let hash = "a".repeat(128);
                    let _ = sandbox.config(&format!("force+{path}:{hash}:deny"));
                    paths.push(path);
                }

                // Perform match on non-existing key.
                for _ in 0..*i {
                    let path = black_box(Path::new("/no/such/file"));
                    let _ = sandbox.check_iforce(path);
                }

                // Perform match on existing key.
                for path in paths {
                    let path = black_box(&path);
                    for _ in 0..*i {
                        let _ = sandbox.check_iforce(black_box(&path));
                    }
                }

                // Remove paths from the map.
                for j in 0..*i {
                    let path = format!("/usr/bin/test-{j}");
                    let _ = sandbox.config(&format!("force-{path}"));
                }
            });
        });
    }
}

criterion_group!(
    benches,
    canonicalize_benchmark,
    path_unsafe_benchmark,
    proc_fd_benchmark,
    sandbox_globset_benchmark,
    sandbox_forcemap_benchmark,
);
criterion_main!(benches);
