// Copyright (c) 2018 Levente Kurusa
// Copyright (c) 2020 Ant Group
//
// SPDX-License-Identifier: Apache-2.0 or MIT
//

use std::{error::Error as StdError, fmt};

/// The different types of errors that can occur while manipulating control groups.
#[derive(Debug, Eq, PartialEq)]
pub enum ErrorKind {
    FsError,

    Common(String),

    /// An error occured while writing to a control group file.
    WriteFailed(String, String),

    /// An error occured while trying to read from a control group file.
    ReadFailed(String),

    /// An error occured while trying to remove a control group.
    RemoveFailed,
    /// An error occured while trying to parse a value from a control group file.
    ///
    /// In the future, there will be some information attached to this field.
    ParseError,

    /// You tried to do something invalid.
    ///
    /// This could be because you tried to set a value in a control group that is not a root
    /// control group. Or, when using unified hierarchy, you tried to add a task in a leaf node.
    InvalidOperation,

    /// The path of the control group was invalid.
    ///
    /// This could be caused by trying to escape the control group filesystem via a string of "..".
    /// This crate checks against this and operations will fail with this error.
    InvalidPath,

    InvalidBytesSize,

    /// The specified controller is not in the list of supported controllers.
    SpecifiedControllers,

    /// Using method in wrong cgroup version.
    CgroupVersion,

    /// Using method in wrong cgroup mode.
    CgroupMode,

    /// Subsystems is empty.
    SubsystemsEmpty,

    /// An unknown error has occured.
    Other,
}

impl std::fmt::Display for ErrorKind {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        match self {
            ErrorKind::FsError => write!(f, "fs error"),
            ErrorKind::Common(ref s) => write!(f, "common error: {}", s),
            ErrorKind::WriteFailed(ref file, ref value) => write!(
                f,
                "unable to write to a control group file {}, value {}",
                file, value
            ),
            ErrorKind::ReadFailed(ref file) => {
                write!(f, "unable to read a control group file {}", file)
            }
            ErrorKind::RemoveFailed => write!(f, "unable to remove a control group"),
            ErrorKind::ParseError => write!(f, "unable to parse control group file"),
            ErrorKind::InvalidOperation => write!(f, "the requested operation is invalid"),
            ErrorKind::InvalidPath => write!(f, "the given path is invalid"),
            ErrorKind::InvalidBytesSize => write!(f, "invalid bytes size"),
            ErrorKind::SpecifiedControllers => write!(
                f,
                "specified controller is not in the list of supported controllers"
            ),
            ErrorKind::CgroupVersion => write!(f, "using method in wrong cgroup version"),
            ErrorKind::CgroupMode => write!(f, "using method in wrong cgroup mode."),
            ErrorKind::SubsystemsEmpty => write!(f, "subsystems is empty"),
            ErrorKind::Other => write!(f, "an unknown error"),
        }
    }
}

impl std::error::Error for ErrorKind {}

#[derive(Debug)]
pub struct Error {
    kind: ErrorKind,
    cause: Option<Box<dyn StdError + Send + Sync>>,
}

impl fmt::Display for Error {
    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
        if let Some(cause) = &self.cause {
            write!(f, "{} caused by: {:?}", &self.kind, cause)
        } else {
            write!(f, "{}", &self.kind)
        }
    }
}

impl StdError for Error {
    fn source(&self) -> Option<&(dyn StdError + 'static)> {
        #[allow(clippy::manual_map)]
        match self.cause {
            Some(ref x) => Some(&**x),
            None => None,
        }
    }
}

impl Error {
    pub(crate) fn from_string(s: String) -> Self {
        Self {
            kind: ErrorKind::Common(s),
            cause: None,
        }
    }
    pub(crate) fn new(kind: ErrorKind) -> Self {
        Self { kind, cause: None }
    }

    pub(crate) fn with_cause<E>(kind: ErrorKind, cause: E) -> Self
    where
        E: 'static + Send + Sync + StdError,
    {
        Self {
            kind,
            cause: Some(Box::new(cause)),
        }
    }

    pub fn kind(&self) -> &ErrorKind {
        &self.kind
    }
}

pub type Result<T> = ::std::result::Result<T, Error>;
