//! Read the ELF dependency tree.
//!
//! This does not work like `ldd` in that we do not execute/load code (only read
//! files on disk).
use std::{
    collections::{HashMap, HashSet},
    fs,
    fs::File,
    io::{BufReader, Read},
    path::{Path, PathBuf},
};

use goblin::elf::{
    header::{EI_OSABI, ELFOSABI_GNU, ELFOSABI_NONE},
    Elf,
};
use nix::{
    errno::Errno,
    unistd::{access, AccessFlags, Pid},
};
use smallvec::SmallVec;

use crate::{
    fs::{mask_path, safe_open},
    proc::{proc_cwd, proc_ld_env},
    warn, MACHINE_NAME,
};

mod errors;
pub mod ld_so_conf;

pub use errors::Error;
use ld_so_conf::parse_ld_so_conf;

pub(crate) type BufVec = SmallVec<[PathBuf; crate::config::LDD_MAX_ENTRIES]>;
pub(crate) type StrVec = SmallVec<[String; crate::config::LDD_MAX_ENTRIES]>;

/// A library dependency
#[derive(Debug, Clone)]
pub struct Library {
    /// Library name
    pub name: String,
    /// The path to the library.
    pub path: PathBuf,
    /// The normalized real path to the library.
    pub realpath: Option<PathBuf>,
    /// The dependencies of this library.
    pub needed: StrVec,
    /// Runtime library search paths. (deprecated)
    pub rpath: StrVec,
    /// Runtime library search paths.
    pub runpath: StrVec,
}

impl Library {
    /// Is this library found in filesystem.
    pub fn found(&self) -> bool {
        self.realpath.is_some()
    }
}

/// Library dependency tree
#[derive(Debug, Clone)]
pub struct DependencyTree {
    /// The binary’s program interpreter (e.g., dynamic linker).
    pub interpreter: Option<String>,
    /// A list of this binary’s dynamic libraries it depends on directly.
    pub needed: StrVec,
    /// All of this binary’s dynamic libraries it uses in detail.
    pub libraries: HashMap<String, Library>,
    /// Runtime library search paths. (deprecated)
    pub rpath: StrVec,
    /// Runtime library search paths.
    pub runpath: StrVec,
    /// Libraries specified by LD_AUDIT.
    pub auditpath: StrVec,
    /// Libraries specified by LD_PRELOAD.
    pub preloadpath: StrVec,
}

/// Library dependency analyzer
#[derive(Debug, Clone)]
pub struct DependencyAnalyzer {
    env_ld_paths: StrVec,
    conf_ld_paths: StrVec,
    audit_ld_paths: StrVec,
    preload_ld_paths: StrVec,
    additional_ld_paths: BufVec,
    runpaths: StrVec,
    root: PathBuf,
}

impl Default for DependencyAnalyzer {
    fn default() -> Self {
        Self::new(PathBuf::from("/"))
    }
}

impl DependencyAnalyzer {
    /// Create a new dependency analyzer.
    pub fn new(root: PathBuf) -> DependencyAnalyzer {
        DependencyAnalyzer {
            env_ld_paths: SmallVec::new(),
            conf_ld_paths: SmallVec::new(),
            audit_ld_paths: SmallVec::new(),
            preload_ld_paths: SmallVec::new(),
            additional_ld_paths: SmallVec::new(),
            runpaths: SmallVec::new(),
            root,
        }
    }

    /// Add additional library path
    ///
    /// Additional library paths are treated as absolute paths,
    /// not relative to `root`
    pub fn add_library_path(mut self, path: PathBuf) -> Self {
        self.additional_ld_paths.push(path);
        self
    }

    /// Set additional library paths
    ///
    /// Additional library paths are treated as absolute paths,
    /// not relative to `root`
    pub fn library_paths(mut self, paths: BufVec) -> Self {
        self.additional_ld_paths = paths;
        self
    }

    fn read_rpath_runpath(
        &self,
        elf: &Elf,
        path: &Path,
        pid: Pid,
        cwd: Option<&Path>,
    ) -> Result<(StrVec, StrVec), Error> {
        let mut rpaths = SmallVec::new();
        let mut runpaths = SmallVec::new();
        for runpath in &elf.runpaths {
            if let Ok(ld_paths) = self.parse_ld_paths(runpath, path, pid, cwd, false) {
                runpaths = ld_paths;
            }
        }
        for rpath in &elf.rpaths {
            if let Ok(ld_paths) = self.parse_ld_paths(rpath, path, pid, cwd, false) {
                rpaths = ld_paths;
            }
        }
        Ok((rpaths, runpaths))
    }

    /// Analyze the given binary.
    #[allow(clippy::cognitive_complexity)]
    pub fn analyze(
        mut self,
        file: &mut File,
        size: u64,
        deps: &mut Vec<PathBuf>,
        pid: Pid,
        deny_elf32: bool,
        deny_elf_dynamic: bool,
        deny_elf_static: bool,
    ) -> Result<(), Error> {
        let path = deps[0].clone(); // deps is [program, script].

        // read the file (cap at 64M for safety) and parse.
        let size = size.min(64 * 1024 * 1024);
        let mut reader = BufReader::new(file);
        let mut bytes = Vec::with_capacity(size as usize);
        let mut current_size = 0;
        while current_size < size {
            // SAFETY: Free of side-effects due to the loop conditional.
            #[allow(clippy::arithmetic_side_effects)]
            let nleft = size - current_size;
            let nread = reader.by_ref().take(nleft).read_to_end(&mut bytes)?;
            if nread == 0 {
                break; // EOF
            }
            current_size = bytes.len() as u64;
        }
        bytes.shrink_to_fit();
        let elf = Elf::parse(&bytes)?;

        if deny_elf32 && !elf.is_64 {
            // SAFETY: Deny access to ELF32 as requested.
            warn!("ctx": "access", "cap": 'x', "error": "32", "act": "Kill",
                "sys": "exec", "pid": pid.as_raw(), "path": mask_path(path));
            return Err(Error::Goblin(goblin::error::Error::Malformed(
                "Unsafe ELF32".to_string(),
            )));
        }

        let dynamic = elf.dynamic.is_some();
        if deny_elf_dynamic && dynamic {
            // SAFETY: Deny access to dynamically linked binary as requested.
            warn!("ctx": "access", "cap": 'x', "error": "dynamic", "act": "Kill",
                "sys": "exec", "pid": pid.as_raw(), "path": mask_path(path));
            return Err(Error::Goblin(goblin::error::Error::Malformed(
                "Unsafe ELF_DYNAMIC".to_string(),
            )));
        }
        if deny_elf_static && !dynamic {
            // SAFETY: Deny access to statically linked binary as requested.
            warn!("ctx": "access", "cap": 'x', "error": "static", "act": "Kill",
                "sys": "exec", "pid": pid.as_raw(), "path": mask_path(path));
            return Err(Error::Goblin(goblin::error::Error::Malformed(
                "Unsafe ELF_STATIC".to_string(),
            )));
        }

        if !dynamic {
            // Not a dynamically linked binary, return early.
            return Err(Error::NotDynamic);
        }

        // SAFETY: We call proc_cwd() relatively late only when we actually
        // need it. We do error out on inaccessible CWD because we're
        // going to avoid canonicalize errors further down the path, and
        // here we want to ensure a dynamic library cannot be injected
        // by placing it in an inaccessible CWD.
        let cwd = match proc_cwd(pid) {
            Ok(cwd) => Some(cwd),
            Err(Errno::ENOENT) => {
                // SAFETY: CWD doesn't exist, assume no deps therein, and continue.
                None
            }
            Err(errno) => return Err(Error::Io(std::io::Error::from_raw_os_error(errno as i32))),
        };
        let cwd = cwd.as_deref();
        self.load_ld_paths(&path, pid, cwd)?;
        let (mut rpaths, runpaths) = self.read_rpath_runpath(&elf, &path, pid, cwd)?;
        if !runpaths.is_empty() {
            // If both RPATH and RUNPATH are set, only the latter is used.
            rpaths = SmallVec::new();
        }
        self.runpaths.clone_from(&runpaths);
        self.runpaths.extend(rpaths.clone());

        let needed: StrVec = elf.libraries.iter().map(ToString::to_string).collect();
        let mut libraries = HashMap::new();

        let mut stack = needed.clone();
        while let Some(lib_name) = stack.pop() {
            if libraries.contains_key(&lib_name) {
                continue;
            }
            let library = self.find_library(&elf, &lib_name, pid, cwd)?;
            libraries.insert(lib_name, library.clone());
            stack.extend(library.needed);
        }

        let interpreter = elf.interpreter.map(|interp| interp.to_string());
        if let Some(ref interp) = interpreter {
            if !libraries.contains_key(interp) {
                let interp_path = self.root.join(interp.strip_prefix('/').unwrap_or(interp));
                let interp_name = interp_path
                    .file_name()
                    .expect("missing filename")
                    .to_str()
                    .expect("Filename isn't valid Unicode");
                libraries.insert(
                    interp.to_string(),
                    Library {
                        name: interp_name.to_string(),
                        path: interp_path,
                        realpath: crate::fs::canonicalize(
                            pid,
                            PathBuf::from(interp).as_path(),
                            cwd,   // Respect process' CWD.
                            true,  // Resolve symbolic links.
                            false, // Do not enable dotdot restrictions.
                            crate::fs::MissingHandling::Existing,
                        )
                        .ok(),
                        needed: SmallVec::new(),
                        rpath: SmallVec::new(),
                        runpath: SmallVec::new(),
                    },
                );
            }
        }

        // Add all of binary's dynamic libraries: this includes the
        // path of the original executable, the interpreter (e.g. dynamic
        // linker) and all dependent libraries. realpath is canonicalized if
        // Some and if None it's inaccessible or missing so we skip.
        let mut set = HashSet::new(); // omit duplicates.
        for library in libraries.values() {
            // includes the interpreter.
            if let Some(path) = &library.realpath {
                if set.insert(path.clone()) {
                    deps.push(path.to_path_buf());
                }
            }
        }
        for path in self.audit_ld_paths {
            let path = PathBuf::from(path);
            if set.insert(path.clone()) {
                deps.push(path);
            }
        }
        for path in self.preload_ld_paths {
            let path = PathBuf::from(path);
            if set.insert(path.clone()) {
                deps.push(path);
            }
        }

        Ok(())
    }

    /// Parse the colon-delimited list of paths and apply ldso rules
    fn parse_ld_paths(
        &self,
        ld_path: &str,
        elf_path: &Path,
        pid: Pid,
        cwd: Option<&Path>,
        use_space_delimiter: bool,
    ) -> Result<StrVec, Error> {
        let delim = if use_space_delimiter {
            &[' ', ':'][..]
        } else {
            &[':'][..]
        };
        let mut paths = SmallVec::new();
        for path in ld_path.split(delim) {
            let normpath = if path.is_empty() {
                // The ldso treats empty paths as the current directory
                cwd.map(PathBuf::from)
            } else if path.contains("$ORIGIN") || path.contains("${ORIGIN}") {
                // elf_path is canonicalized.
                // let elf_path = fs::canonicalize(elf_path)?;
                let elf_dir = elf_path.parent().ok_or_else(|| {
                    Error::Io(std::io::Error::from_raw_os_error(nix::libc::ENOENT))
                })?;
                let replacement = elf_dir.to_str().ok_or_else(|| {
                    Error::Io(std::io::Error::from_raw_os_error(nix::libc::EINVAL))
                })?;
                let path = PathBuf::from(
                    path.replace("${ORIGIN}", replacement)
                        .replace("$ORIGIN", replacement),
                );
                crate::fs::canonicalize(
                    pid,
                    path.as_path(),
                    cwd,   // Respect process CWD.
                    true,  // Resolve symbolic links.
                    false, // Do not enable dotdot restrictions.
                    crate::fs::MissingHandling::Existing,
                )
                .ok()
            } else {
                //SAFETY: We respect process' CWD so we do not need the root handling.
                //let path = self.root.join(path.strip_prefix('/').unwrap_or(path));
                crate::fs::canonicalize(
                    pid,
                    Path::new(path),
                    cwd,   // Respect process CWD.
                    true,  // Resolve symbolic links.
                    false, // Do not enable dotdot restrictions.
                    crate::fs::MissingHandling::Existing,
                )
                .ok()
            };
            if let Some(normpath) = normpath {
                paths.push(normpath.display().to_string());
            }
        }
        Ok(paths)
    }

    #[allow(clippy::cognitive_complexity)]
    fn load_ld_paths(
        &mut self,
        elf_path: &Path,
        pid: Pid,
        cwd: Option<&Path>,
    ) -> Result<(), Error> {
        match proc_ld_env(pid) {
            Ok(env) if self.root == Path::new("/") => {
                if let Some(paths) = &env[0] {
                    self.audit_ld_paths = self.parse_ld_paths(paths, elf_path, pid, cwd, true)?;
                }
                if let Some(paths) = &env[1] {
                    self.preload_ld_paths = self.parse_ld_paths(paths, elf_path, pid, cwd, true)?;
                }
                if let Some(paths) = &env[2] {
                    self.env_ld_paths = self.parse_ld_paths(paths, elf_path, pid, cwd, false)?;
                }
            }
            Err(std::env::VarError::NotUnicode(_)) => {
                // SAFETY: Do not let through non-unicode paths,
                // as this may be (ab)used to insert malicious
                // dynamic libraries.
                return Err(Error::Io(std::io::Error::from_raw_os_error(
                    nix::libc::EINVAL,
                )));
            }
            _ => {} // environment variable not present or root != /
        }
        // Load all the paths from a ldso config file
        match find_musl_libc() {
            // musl libc
            Ok(_) => {
                // from https://git.musl-libc.org/cgit/musl/tree/ldso/dynlink.c?id=3f701faace7addc75d16dea8a6cd769fa5b3f260#n1063
                let root_str = self.root.display().to_string();
                let root_str = root_str.strip_suffix('/').unwrap_or(&root_str);
                let patterns = [
                    format!("{root_str}/etc/ld-{MACHINE_NAME}-pc-linux-musl.path"),
                    format!("{root_str}/etc/ld-{MACHINE_NAME}-unknown-linux-musl.path"),
                    format!("{root_str}/etc/ld-musl-{MACHINE_NAME}-pc-linux.path"),
                    format!("{root_str}/etc/ld-musl-{MACHINE_NAME}-unknown-linux.path"),
                ];
                for entry in &patterns {
                    match fs::read_to_string(entry) {
                        Ok(content) => {
                            for line in content.lines() {
                                let line_stripped = line.trim();
                                if !line_stripped.is_empty() {
                                    self.conf_ld_paths
                                        .push(root_str.to_string() + line_stripped);
                                }
                            }
                            break;
                        }
                        Err(e) if e.kind() == std::io::ErrorKind::NotFound => {}
                        Err(e) => return Err(e.into()),
                    }
                }
                // default ld paths
                if self.conf_ld_paths.is_empty() {
                    self.conf_ld_paths.push(root_str.to_string() + "/lib");
                    self.conf_ld_paths
                        .push(root_str.to_string() + "/usr/local/lib");
                    self.conf_ld_paths.push(root_str.to_string() + "/usr/lib");
                }
            }
            // glibc
            _ => {
                // Load up /etc/ld.so.conf
                if let Ok(paths) = parse_ld_so_conf("/etc/ld.so.conf", &self.root) {
                    self.conf_ld_paths = paths;
                }
                // the trusted directories are not necessarily in ld.so.conf
                for path in &["/lib", "/lib64/", "/usr/lib", "/usr/lib64"] {
                    self.conf_ld_paths.push(path.to_string());
                }
            }
        }
        self.conf_ld_paths.dedup();
        Ok(())
    }

    /// Try to locate a `lib` that is compatible to `elf`
    fn find_library(
        &self,
        elf: &Elf,
        lib: &str,
        pid: Pid,
        cwd: Option<&Path>,
    ) -> Result<Library, Error> {
        for lib_path in self
            .runpaths
            .iter()
            .chain(self.env_ld_paths.iter())
            .chain(self.conf_ld_paths.iter())
            .map(|ld_path| {
                self.root
                    .join(ld_path.strip_prefix('/').unwrap_or(ld_path))
                    .join(lib)
            })
            .chain(
                self.additional_ld_paths
                    .iter()
                    .map(|ld_path| ld_path.join(lib)),
            )
        {
            if let Ok((file, size)) = safe_open(&lib_path) {
                // read the file (cap at 64M for safety) and parse.
                let size = size.min(64 * 1024 * 1024);
                let mut reader = BufReader::new(file);
                let mut bytes = Vec::with_capacity(size as usize);
                let mut current_size = 0;
                while current_size < size {
                    // SAFETY: Free of side-effects due to the loop conditional.
                    #[allow(clippy::arithmetic_side_effects)]
                    let nleft = size - current_size;
                    let nread = reader.by_ref().take(nleft).read_to_end(&mut bytes)?;
                    if nread == 0 {
                        break; // EOF
                    }
                    current_size = bytes.len() as u64;
                }
                drop(reader);
                bytes.shrink_to_fit();
                if let Ok(lib_elf) = Elf::parse(&bytes) {
                    if compatible_elfs(elf, &lib_elf) {
                        let needed = lib_elf.libraries.iter().map(ToString::to_string).collect();
                        let (rpath, runpath) =
                            self.read_rpath_runpath(&lib_elf, &lib_path, pid, cwd)?;
                        return Ok(Library {
                            name: lib.to_string(),
                            path: lib_path.to_path_buf(),
                            realpath: crate::fs::canonicalize(
                                pid,
                                lib_path,
                                cwd.as_ref().map(PathBuf::from), // Respect process' CWD.
                                true,                            // Resolve symbolic links.
                                false, // Do not enable dotdot restrictions.
                                crate::fs::MissingHandling::Normal, // We verified file exists.
                            )
                            .ok(),
                            needed,
                            rpath,
                            runpath,
                        });
                    }
                }
            }
        }
        Ok(Library {
            name: lib.to_string(),
            path: PathBuf::from(lib),
            realpath: None,
            needed: SmallVec::new(),
            rpath: SmallVec::new(),
            runpath: SmallVec::new(),
        })
    }
}

/// Find musl libc path
fn find_musl_libc() -> Result<(), Error> {
    access(
        Path::new(&format!("/lib/ld-musl-{MACHINE_NAME}.so.1")),
        AccessFlags::F_OK,
    )
    .map_err(|e| e.into())
}

/// See if two ELFs are compatible
///
/// This compares the aspects of the ELF to see if they're compatible:
/// bit size, endianness, machine type, and operating system.
fn compatible_elfs(elf1: &Elf, elf2: &Elf) -> bool {
    if elf1.is_64 != elf2.is_64 {
        return false;
    }
    if elf1.little_endian != elf2.little_endian {
        return false;
    }
    if elf1.header.e_machine != elf2.header.e_machine {
        return false;
    }
    let compatible_osabis = &[
        ELFOSABI_NONE, // ELFOSABI_NONE / ELFOSABI_SYSV
        ELFOSABI_GNU,  // ELFOSABI_GNU / ELFOSABI_LINUX
    ];
    let osabi1 = elf1.header.e_ident[EI_OSABI];
    let osabi2 = elf2.header.e_ident[EI_OSABI];
    if osabi1 != osabi2
        && !compatible_osabis.contains(&osabi1)
        && !compatible_osabis.contains(&osabi2)
    {
        return false;
    }
    true
}
