//
// syd: seccomp and landlock based application sandbox with support for namespaces
// src/lib.rs: Common utility functions
//
// Copyright (c) 2023, 2024 Ali Polatel <alip@chesswob.org>
// likely and unlikely functions are based on the endorphin crate which is:
//    Copyright (c) 2021 Jun Ryoung Ju (junryoungju@gmail.com)
//    SPDX-License-Identifier: MIT
//
// SPDX-License-Identifier: GPL-3.0-or-later

//! # syd: The ☮ther SⒶndbøx
//!
//! [![Shine On You Crazy Diamond!](https://img.shields.io/badge/Shine%20On%20You%20Crazy%20Diamond!-8A2BE2)](https://en.wikipedia.org/wiki/Syd_Barrett)
//! [![license](https://img.shields.io/crates/l/jja.svg)](https://git.sr.ht/~alip/syd/tree/main/item/COPYING)
//! [![msrv](https://img.shields.io/badge/rustc-1.70%2B-green?style=plastic)](https://blog.rust-lang.org/2023/06/01/Rust-1.70.0.html)
//! [![build status](https://builds.sr.ht/~alip/syd.svg)](https://builds.sr.ht/~alip/syd?)
//! [![maintenance-status](https://img.shields.io/badge/maintenance-actively--developed-brightgreen.svg)](https://git.sr.ht/~alip/syd)
//! [![dependency status](https://deps.rs/repo/sourcehut/~alip/syd/status.svg)](https://deps.rs/repo/sourcehut/~alip/syd)
//! [![repology](https://repology.org/badge/latest-versions/syd.svg)](https://repology.org/project/syd/versions)
//!
//! [![syd](https://git.sr.ht/~alip/syd/blob/main/data/syd.png)](https://todo.sr.ht/~alip/syd)
//! [![GNU](https://web.archive.org/web/20221222061733if_/https://dev.exherbo.org/~alip/images/gnu.png)](https://www.gnu.org/philosophy/philosophy.html)
//! [![Linux](https://chesswob.org/jja/tux.png)](https://www.kernel.org/category/about.html)
//! [![Exherbo](https://web.archive.org/web/20230518155203if_/https://dev.exherbo.org/~alip/images/zebrapig.png)](https://www.exherbolinux.org/docs/gettingstarted.html)
//! [![musl libc](https://www.chesswob.org/jja/musl-inside.png)](https://www.musl-libc.org/)
//! [![libsecc☮mp](https://web.archive.org/web/20221222061720if_/https://dev.exherbo.org/~alip/images/libseccomp.png)](https://github.com/seccomp/libseccomp)
//! [![Paludis](http://paludis.exherbolinux.org/paludis_270.png)](https://paludis.exherbolinux.org)
//!
//! syd is a **seccomp**(2) based sandboxing utility for modern Linux\[\>=5.6\]
//! machines to sandbox unwanted process access to filesystem and network resources.
//! syd requires *no root access* and *no ptrace* rights. All you need is a
//! recent Linux kernel and libsecc☮mp which is available on many different
//! architectures, including **x86**, **x86\_64**, **x32**, **arm**, **aarch64**,
//! **mips**, **mips64**... This makes it very easy for a regular user to use. This is
//! the motto of syd: *bring easy, simple, flexible and powerful access restriction
//! to the Linux user!*
//!
//! The basic idea of syd is to run a command under certain restrictions. These
//! restrictions define which system calls the command is permitted to run and which
//! argument values are permitted for the given system call. The restrictions may be
//! applied via two ways.  *seccomp-bpf* can be used to apply simple Secure Computing
//! user filters to run sandboxing fully on kernel space, and *seccomp-notify*
//! functionality can be used to run sandboxing on kernel space and fallback to user
//! space to dereference pointer arguments of system calls (**See
//! [Security](#security) about `TOCTOU` et. al**), which are one of
//! **[pathname](https://en.wikipedia.org/wiki/Path_(computing))**, **[UNIX socket
//! address](https://en.wikipedia.org/wiki/Unix_domain_socket)**,
//! **[IPv4](https://en.wikipedia.org/wiki/IPv4)** or
//! **[IPv6](https://en.wikipedia.org/wiki/IPv6)** network address, and make dynamic
//! decisions using [Unix shell style patterns](https://docs.rs/globset) such as
//! `allow/write+/home/syd/***`, or `allow/write+/run/user/*/pulse` for
//! **[pathnames](https://en.wikipedia.org/wiki/Path_(computing))**, and using
//! **[CIDR](https://docs.rs/ipnetwork)** notation such as
//! `allow/net/connect+127.0.0.1/8!9050`, or
//! `allow/net/connect+::1/8!9050` for
//! **[IPv4](https://en.wikipedia.org/wiki/IPv4)** and
//! **[IPv6](https://en.wikipedia.org/wiki/IPv6)** addresses and perform an action
//! which is by default denying the system call with an appropriate error, which is
//! usually **access denied**, aka `EACCES`. For default disallowed system calls,
//! such as `ptrace` or `process_vm_writev` (**See [Security](#security) about
//! `TOCTOU` et. al**) syd returns `EACCES` as well.
//!
//! To be able to use syd, you need a recent Linux kernel with the system calls
//! **pidfd_getfd**, **pidfd_send_signal**. The Secure Computing facility of the
//! Linux kernel should support the **SECCOMP_USER_NOTIF_FLAG_CONTINUE** operation.
//! It is recommended to have the **CONFIG_CROSS_MEMORY_ATTACH** kernel option
//! enabled, if this option is not enabled, syd will fallback to reading/writing
//! from `/proc/$pid/mem`. Linux-5.11 or later is recommended.

// We like clean and simple code with documentation.
// Keep in sync with main.rs.
#![deny(missing_docs)]
#![deny(clippy::allow_attributes_without_reason)]
#![deny(clippy::arithmetic_side_effects)]
#![deny(clippy::as_ptr_cast_mut)]
#![deny(clippy::as_underscore)]
#![deny(clippy::assertions_on_result_states)]
#![deny(clippy::borrow_as_ptr)]
#![deny(clippy::branches_sharing_code)]
#![deny(clippy::case_sensitive_file_extension_comparisons)]
#![deny(clippy::cast_lossless)]
#![deny(clippy::cast_possible_truncation)]
#![deny(clippy::cast_possible_wrap)]
#![deny(clippy::cast_precision_loss)]
#![deny(clippy::cast_ptr_alignment)]
#![deny(clippy::cast_sign_loss)]
#![deny(clippy::checked_conversions)]
#![deny(clippy::clear_with_drain)]
#![deny(clippy::clone_on_ref_ptr)]
#![deny(clippy::cloned_instead_of_copied)]
#![deny(clippy::cognitive_complexity)]
#![deny(clippy::collection_is_never_read)]
#![deny(clippy::copy_iterator)]
#![deny(clippy::create_dir)]
#![deny(clippy::dbg_macro)]
#![deny(clippy::debug_assert_with_mut_call)]
#![deny(clippy::decimal_literal_representation)]
#![deny(clippy::default_trait_access)]
#![deny(clippy::default_union_representation)]
#![deny(clippy::derive_partial_eq_without_eq)]
#![deny(clippy::doc_link_with_quotes)]
//#![deny(clippy::doc_markdown)]
#![deny(clippy::explicit_into_iter_loop)]
#![deny(clippy::explicit_iter_loop)]
#![deny(clippy::fallible_impl_from)]
#![deny(clippy::missing_safety_doc)]
#![deny(clippy::undocumented_unsafe_blocks)]

/// Compatibility code for different libcs
pub(crate) mod compat;
/// Static configuration, edit & recompile!
pub mod config;
/// Filesystem utilities
pub mod fs;
/// Utilities for hashing
pub mod hash;
/// Secure computing hooks
#[allow(clippy::as_ptr_cast_mut)]
#[allow(clippy::cast_sign_loss)]
#[allow(clippy::undocumented_unsafe_blocks)]
pub mod hook;
/// Simple logging on standard error using JSON lines
pub mod log;
/// /proc utilities
pub mod proc;
/// Sandbox configuration
pub mod sandbox;

// Vendored crates:
/// Interface to Linux capabilities
#[allow(dead_code)]
#[allow(missing_docs)]
#[allow(clippy::arithmetic_side_effects)]
#[allow(clippy::cast_possible_truncation)]
#[allow(clippy::missing_safety_doc)]
#[allow(clippy::undocumented_unsafe_blocks)]
pub mod caps;
/// Native Rust crate for managing control groups on Linux
#[allow(dead_code)]
#[allow(missing_docs)]
#[allow(clippy::arithmetic_side_effects)]
#[allow(clippy::cast_lossless)]
#[allow(clippy::cast_possible_wrap)]
#[allow(clippy::cast_precision_loss)]
#[allow(clippy::cast_sign_loss)]
#[allow(clippy::cognitive_complexity)]
#[allow(clippy::create_dir)]
#[allow(clippy::default_trait_access)]
#[allow(clippy::disallowed_methods)]
#[allow(clippy::explicit_iter_loop)]
#[allow(clippy::manual_try_fold)]
#[allow(clippy::missing_safety_doc)]
#[allow(clippy::type_complexity)]
#[allow(clippy::undocumented_unsafe_blocks)]
#[allow(clippy::unnecessary_to_owned)]
pub mod cgroups;
/// Key-Value based in-memory cache library which supports Custom Expiration Policies
#[allow(clippy::arithmetic_side_effects)]
#[allow(clippy::cast_possible_truncation)]
#[allow(clippy::cloned_instead_of_copied)]
#[allow(clippy::disallowed_methods)]
#[allow(clippy::extra_unused_type_parameters)]
#[allow(clippy::let_and_return)]
#[allow(clippy::manual_flatten)]
#[allow(clippy::manual_hash_one)]
#[allow(clippy::multiple_bound_locations)]
#[allow(clippy::type_complexity)]
#[allow(clippy::undocumented_unsafe_blocks)]
#[allow(clippy::unwrap_or_default)]
#[allow(dead_code)]
#[allow(missing_docs)]
#[allow(unused_imports)]
#[allow(unused_variables)]
pub mod endorphin;
/// Interface to LandLock LSM
#[allow(dead_code)]
#[allow(missing_docs)]
#[allow(unused_imports)]
#[allow(clippy::as_underscore)]
#[allow(clippy::cast_possible_truncation)]
#[allow(clippy::decimal_literal_representation)]
#[allow(clippy::missing_safety_doc)]
#[allow(clippy::type_complexity)]
#[allow(clippy::undocumented_unsafe_blocks)]
pub mod landlock;
/// Read the ELF dependency tree
#[allow(dead_code)]
#[allow(clippy::disallowed_methods)]
#[allow(clippy::manual_flatten)]
#[allow(clippy::too_many_arguments)]
#[allow(missing_docs)]
#[cfg(feature = "elf")]
pub(crate) mod lddtree;
/// The low-level interface for linux namespaces (containers)
pub mod unshare;

#[cfg(feature = "log")]
use std::ffi::CStr;
use std::{
    arch::asm,
    fs::File,
    io::Result as IOResult,
    os::fd::{AsRawFd, FromRawFd, OwnedFd, RawFd},
    path::{Path, PathBuf},
    process::exit,
};

use landlock::{
    path_beneath_rules, Access, AccessFs, RestrictionStatus, Ruleset, RulesetAttr,
    RulesetCreatedAttr, RulesetError,
};
#[cfg(feature = "elf")]
use lddtree::DependencyAnalyzer;
use libseccomp::{error::SeccompErrno, ScmpAction, ScmpArch, ScmpFilterContext, ScmpSyscall};
#[cfg(feature = "log")]
use libseccomp::{ScmpNotifData, ScmpNotifReq, ScmpNotifResp};
use nix::{
    dir::Dir,
    errno::Errno,
    fcntl::OFlag,
    libc::c_int,
    mount::MsFlags,
    sys::{
        signal::{sigaction, signal, SaFlags, SigAction, SigHandler, SigSet, Signal},
        socket::{socket, AddressFamily, SockFlag, SockType},
        stat::Mode,
        wait::{waitpid, WaitStatus},
    },
    unistd::{fork, ForkResult, Gid, Group, Pid, Uid, User},
};
use once_cell::sync::Lazy;
use serde::{ser::SerializeMap, Serialize, Serializer};

#[cfg(feature = "log")]
use crate::fs::mask_path;
use crate::{
    fs::err2no,
    landlock::{RulesetStatus, ABI},
};

// Macros
/// A macro to quickly create an anyhow::Error from a given argument.
#[macro_export]
macro_rules! anyerr {
    ($arg:expr) => {
        anyhow::Error::new($arg)
    };
}

/// A macro to create an anyhow::Error from the last errno.
#[macro_export]
macro_rules! lasterrno {
    () => {
        anyhow::Error::new(nix::errno::Errno::last())
    };
}

/// A macro to generate a formatted io::Error for syscall seccomp rules.
#[macro_export]
macro_rules! seccomp_error {
    ($syscall:expr, $sysname:expr, $error:expr) => {
        io::Error::new(
            io::ErrorKind::Other,
            format!(
                "Failed to add rule for system call {} ( {} ): {}",
                $syscall, $sysname, $error
            ),
        )
    };
}

/* Data structures */
#[derive(Debug, Eq, PartialEq)]
pub(crate) struct Sydcall(ScmpSyscall, ScmpArch);

impl std::hash::Hash for Sydcall {
    fn hash<H: std::hash::Hasher>(&self, state: &mut H) {
        i32::from(self.0).hash(state);
        match self.1 {
            ScmpArch::X86 => 0,
            ScmpArch::X8664 => 1,
            ScmpArch::X32 => 2,
            ScmpArch::Arm => 3,
            ScmpArch::Aarch64 => 4,
            ScmpArch::Mips => 5,
            ScmpArch::Mips64 => 6,
            ScmpArch::Mips64N32 => 7,
            ScmpArch::Mipsel => 8,
            ScmpArch::Mipsel64 => 9,
            ScmpArch::Mipsel64N32 => 10,
            ScmpArch::Ppc => 11,
            ScmpArch::Ppc64 => 12,
            ScmpArch::Ppc64Le => 13,
            ScmpArch::S390 => 14,
            ScmpArch::S390X => 15,
            ScmpArch::Parisc => 16,
            ScmpArch::Parisc64 => 17,
            ScmpArch::Riscv64 => 18,
            _ => u8::MAX,
        }
        .hash(state);
    }
}

// Define a struct that wraps a CStr to implement Display.
#[cfg(feature = "log")]
pub(crate) struct SydCStr<'a>(pub &'a CStr);

// Implement the Display trait for SydCStr
#[cfg(feature = "log")]
impl<'a> std::fmt::Display for SydCStr<'a> {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        // Convert CStr to a string slice for Debug formatting
        let s = format!("{:?}", self.0);

        // Trim the leading and trailing quotes
        // Let's panic if Debug ever changes format so we can mock rust devs.
        let s = &s[1..s.len().saturating_sub(1)];

        // SAFETY: Mask control characters.
        let s = mask_path(s);

        // Write the trimmed string to the formatter
        write!(f, "{s}")
    }
}

#[cfg(feature = "log")]
pub(crate) struct SydNotifReq(ScmpNotifReq);

#[cfg(feature = "log")]
impl Serialize for SydNotifReq {
    fn serialize<S>(&self, serializer: S) -> Result<S::Ok, S::Error>
    where
        S: Serializer,
    {
        let mut map = serializer.serialize_map(Some(4))?;
        map.serialize_entry("id", &self.0.id)?;
        map.serialize_entry("pid", &self.0.pid)?;
        map.serialize_entry("flags", &self.0.flags)?;
        map.serialize_entry("data", &SydNotifData(self.0.data))?;
        map.end()
    }
}

#[cfg(feature = "log")]
pub(crate) struct SydNotifData(ScmpNotifData);

#[cfg(feature = "log")]
impl Serialize for SydNotifData {
    fn serialize<S>(&self, serializer: S) -> Result<S::Ok, S::Error>
    where
        S: Serializer,
    {
        let mut map = serializer.serialize_map(Some(4))?;
        let name = self
            .0
            .syscall
            .get_name_by_arch(self.0.arch)
            .unwrap_or_else(|_| "?".to_string());
        let arch = format!("{:?}", self.0.arch).to_ascii_lowercase();
        let arch = if arch == { "x8664" } { "x86_64" } else { &arch };
        map.serialize_entry("sys", &name)?;
        map.serialize_entry("arch", &arch)?;
        map.serialize_entry("ip", &self.0.instr_pointer)?;
        map.serialize_entry("args", &self.0.args)?;
        map.end()
    }
}

#[cfg(feature = "log")]
pub(crate) struct SydNotifResp(ScmpNotifResp);

#[cfg(feature = "log")]
impl Serialize for SydNotifResp {
    fn serialize<S>(&self, serializer: S) -> Result<S::Ok, S::Error>
    where
        S: Serializer,
    {
        let mut map = serializer.serialize_map(Some(4))?;
        map.serialize_entry("id", &self.0.id)?;
        map.serialize_entry("val", &self.0.val)?;
        map.serialize_entry("err", &self.0.error)?;
        map.serialize_entry("flags", &self.0.flags)?;
        map.end()
    }
}

#[derive(Debug)]
pub(crate) struct SydWaitStatus(crate::compat::WaitStatus);

impl Serialize for SydWaitStatus {
    fn serialize<S>(&self, serializer: S) -> Result<S::Ok, S::Error>
    where
        S: Serializer,
    {
        use crate::compat::WaitStatus;
        let mut map = serializer.serialize_map(None)?;
        match self.0 {
            WaitStatus::Exited(pid, code) => {
                map.serialize_entry("type", "exited")?;
                map.serialize_entry("pid", &pid.as_raw())?;
                map.serialize_entry("code", &code)?;
            }
            WaitStatus::Signaled(pid, signal, core_dump) => {
                map.serialize_entry("type", "signaled")?;
                map.serialize_entry("pid", &pid.as_raw())?;
                map.serialize_entry("sig", &signal)?;
                map.serialize_entry("core_dump", &core_dump)?;
            }
            WaitStatus::Stopped(pid, signal) => {
                map.serialize_entry("type", "stopped")?;
                map.serialize_entry("pid", &pid.as_raw())?;
                map.serialize_entry("sig", &signal)?;
            }
            WaitStatus::PtraceEvent(pid, signal, event) => {
                map.serialize_entry("type", "ptrace_event")?;
                map.serialize_entry("pid", &pid.as_raw())?;
                map.serialize_entry("sig", &signal)?;
                map.serialize_entry("event", &event)?;
            }
            WaitStatus::PtraceSyscall(pid) => {
                map.serialize_entry("type", "ptrace_syscall")?;
                map.serialize_entry("pid", &pid.as_raw())?;
            }
            WaitStatus::Continued(pid) => {
                map.serialize_entry("type", "continued")?;
                map.serialize_entry("pid", &pid.as_raw())?;
            }
            WaitStatus::StillAlive => {
                map.serialize_entry("type", "interrupted")?;
            }
        }
        map.end()
    }
}

// Execution domain constants, taken from sys/personality.h
const PER_LINUX: c_int = 0;
const PER_LINUX_32BIT: c_int = PER_LINUX | ADDR_LIMIT_32BIT;
const PER_LINUX_FDPIC: c_int = PER_LINUX | FDPIC_FUNCPTRS;
const PER_SVR4: c_int = 1 | STICKY_TIMEOUTS | MMAP_PAGE_ZERO;
const PER_SVR3: c_int = 2 | STICKY_TIMEOUTS | SHORT_INODE;
const PER_SCOSVR3: c_int = 3 | STICKY_TIMEOUTS | WHOLE_SECONDS | SHORT_INODE;
const PER_OSR5: c_int = 3 | STICKY_TIMEOUTS | WHOLE_SECONDS;
const PER_WYSEV386: c_int = 4 | STICKY_TIMEOUTS | SHORT_INODE;
const PER_ISCR4: c_int = 5 | STICKY_TIMEOUTS;
const PER_BSD: c_int = 6;
const PER_SUNOS: c_int = PER_BSD | STICKY_TIMEOUTS;
const PER_XENIX: c_int = 7 | STICKY_TIMEOUTS | SHORT_INODE;
const PER_LINUX32: c_int = 8;
const PER_LINUX32_3GB: c_int = PER_LINUX32 | ADDR_LIMIT_3GB;
const PER_IRIX32: c_int = 9 | STICKY_TIMEOUTS;
const PER_IRIXN32: c_int = 0xa | STICKY_TIMEOUTS;
const PER_IRIX64: c_int = 0x0b | STICKY_TIMEOUTS;
const PER_RISCOS: c_int = 0xc;
const PER_SOLARIS: c_int = 0xd | STICKY_TIMEOUTS;
const PER_UW7: c_int = 0xe | STICKY_TIMEOUTS | MMAP_PAGE_ZERO;
const PER_OSF4: c_int = 0xf;
const PER_HPUX: c_int = 0x10;
const PER_MASK: c_int = 0xff;

// Flag constants, taken from sys/personality.h
const UNAME26: c_int = 0x0020000;
const ADDR_NO_RANDOMIZE: c_int = 0x0040000;
const FDPIC_FUNCPTRS: c_int = 0x0080000;
const MMAP_PAGE_ZERO: c_int = 0x0100000;
const ADDR_COMPAT_LAYOUT: c_int = 0x0200000;
const READ_IMPLIES_EXEC: c_int = 0x0400000;
const ADDR_LIMIT_32BIT: c_int = 0x0800000;
const SHORT_INODE: c_int = 0x1000000;
const WHOLE_SECONDS: c_int = 0x2000000;
const STICKY_TIMEOUTS: c_int = 0x4000000;
const ADDR_LIMIT_3GB: c_int = 0x8000000;

/// A type that wraps personality(2) return value and implements Display.
pub struct SydPersona(pub c_int);

impl std::fmt::Display for SydPersona {
    fn fmt(&self, f: &mut std::fmt::Formatter) -> std::fmt::Result {
        let domain = match self.0 & PER_MASK {
            PER_LINUX => "linux",
            PER_LINUX_32BIT => "linux_32bit",
            PER_LINUX_FDPIC => "linux_fdpic",
            PER_SVR4 => "svr4",
            PER_SVR3 => "svr3",
            PER_SCOSVR3 => "scosvr3",
            PER_OSR5 => "osr5",
            PER_WYSEV386 => "wysev386",
            PER_ISCR4 => "iscr4",
            PER_BSD => "bsd",
            PER_SUNOS => "sunos",
            PER_XENIX => "xenix",
            PER_LINUX32 => "linux32",
            PER_LINUX32_3GB => "linux32_3gb",
            PER_IRIX32 => "irix32",
            PER_IRIXN32 => "irixn32",
            PER_IRIX64 => "irix64",
            PER_RISCOS => "riscos",
            PER_SOLARIS => "solaris",
            PER_UW7 => "uw7",
            PER_OSF4 => "osf4",
            PER_HPUX => "hpux",
            _ => "unknown",
        };

        let flags = [
            (UNAME26, "uname26"),
            (ADDR_NO_RANDOMIZE, "addr-no-randomize"),
            (FDPIC_FUNCPTRS, "fdpic-funcptrs"),
            (MMAP_PAGE_ZERO, "mmap-page-zero"),
            (ADDR_COMPAT_LAYOUT, "addr-compat-layout"),
            (READ_IMPLIES_EXEC, "read-implies-exec"),
            (ADDR_LIMIT_32BIT, "addr-limit-32bit"),
            (SHORT_INODE, "short-inode"),
            (WHOLE_SECONDS, "whole-seconds"),
            (STICKY_TIMEOUTS, "sticky-timeouts"),
            (ADDR_LIMIT_3GB, "addr-limit-3gb"),
        ]
        .iter()
        .filter_map(|&(flag, name)| {
            if self.0 & flag == flag {
                Some(name)
            } else {
                None
            }
        })
        .collect::<Vec<_>>()
        .join(",");

        if flags.is_empty() {
            write!(f, "{domain}")
        } else {
            write!(f, "{domain},{flags}")
        }
    }
}

/// SydMsFlags wraps MsFlags and provides from_name.
/// This is already available in newer bitflags versions.
pub(crate) struct SydMsFlags(MsFlags);

impl SydMsFlags {
    fn from_name(name: &str) -> Option<Self> {
        match name {
            "ro" => Some(SydMsFlags(MsFlags::MS_RDONLY)),
            "nosuid" => Some(SydMsFlags(MsFlags::MS_NOSUID)),
            "nodev" => Some(SydMsFlags(MsFlags::MS_NODEV)),
            "noexec" => Some(SydMsFlags(MsFlags::MS_NOEXEC)),
            "sync" => Some(SydMsFlags(MsFlags::MS_SYNCHRONOUS)),
            "remount" => Some(SydMsFlags(MsFlags::MS_REMOUNT)),
            "mandlock" => Some(SydMsFlags(MsFlags::MS_MANDLOCK)),
            "dirsync" => Some(SydMsFlags(MsFlags::MS_DIRSYNC)),
            "noatime" => Some(SydMsFlags(MsFlags::MS_NOATIME)),
            "nodiratime" => Some(SydMsFlags(MsFlags::MS_NODIRATIME)),
            "bind" => Some(SydMsFlags(MsFlags::MS_BIND)),
            "move" => Some(SydMsFlags(MsFlags::MS_MOVE)),
            "rec" => Some(SydMsFlags(MsFlags::MS_REC)),
            "silent" => Some(SydMsFlags(MsFlags::MS_SILENT)),
            "posixacl" => Some(SydMsFlags(MsFlags::MS_POSIXACL)),
            "unbindable" => Some(SydMsFlags(MsFlags::MS_UNBINDABLE)),
            "private" => Some(SydMsFlags(MsFlags::MS_PRIVATE)),
            "slave" => Some(SydMsFlags(MsFlags::MS_SLAVE)),
            "shared" => Some(SydMsFlags(MsFlags::MS_SHARED)),
            "relatime" => Some(SydMsFlags(MsFlags::MS_RELATIME)),
            "kernmount" => Some(SydMsFlags(MsFlags::MS_KERNMOUNT)),
            "i_version" => Some(SydMsFlags(MsFlags::MS_I_VERSION)),
            "strictatime" => Some(SydMsFlags(MsFlags::MS_STRICTATIME)),
            "lazytime" => Some(SydMsFlags(MsFlags::MS_LAZYTIME)),
            "active" => Some(SydMsFlags(MsFlags::MS_ACTIVE)),
            "nouser" => Some(SydMsFlags(MsFlags::MS_NOUSER)),
            _ => None,
        }
    }

    // Convert MsFlags to a vector of flag names
    #[allow(clippy::cognitive_complexity)]
    fn to_names(&self) -> Vec<&str> {
        let mut names = Vec::new();

        if self.0.contains(MsFlags::MS_RDONLY) {
            names.push("ro");
        }
        if self.0.contains(MsFlags::MS_NOSUID) {
            names.push("nosuid");
        }
        if self.0.contains(MsFlags::MS_NODEV) {
            names.push("nodev");
        }
        if self.0.contains(MsFlags::MS_NOEXEC) {
            names.push("noexec");
        }
        if self.0.contains(MsFlags::MS_SYNCHRONOUS) {
            names.push("sync");
        }
        if self.0.contains(MsFlags::MS_REMOUNT) {
            names.push("remount");
        }
        if self.0.contains(MsFlags::MS_MANDLOCK) {
            names.push("mandlock");
        }
        if self.0.contains(MsFlags::MS_DIRSYNC) {
            names.push("dirsync");
        }
        if self.0.contains(MsFlags::MS_NOATIME) {
            names.push("noatime");
        }
        if self.0.contains(MsFlags::MS_NODIRATIME) {
            names.push("nodiratime");
        }
        if self.0.contains(MsFlags::MS_BIND) {
            names.push("bind");
        }
        if self.0.contains(MsFlags::MS_MOVE) {
            names.push("move");
        }
        if self.0.contains(MsFlags::MS_REC) {
            names.push("rec");
        }
        if self.0.contains(MsFlags::MS_SILENT) {
            names.push("silent");
        }
        if self.0.contains(MsFlags::MS_POSIXACL) {
            names.push("posixacl");
        }
        if self.0.contains(MsFlags::MS_UNBINDABLE) {
            names.push("unbindable");
        }
        if self.0.contains(MsFlags::MS_PRIVATE) {
            names.push("private");
        }
        if self.0.contains(MsFlags::MS_SLAVE) {
            names.push("slave");
        }
        if self.0.contains(MsFlags::MS_SHARED) {
            names.push("shared");
        }
        if self.0.contains(MsFlags::MS_RELATIME) {
            names.push("relatime");
        }
        if self.0.contains(MsFlags::MS_KERNMOUNT) {
            names.push("kernmount");
        }
        if self.0.contains(MsFlags::MS_I_VERSION) {
            names.push("i_version");
        }
        if self.0.contains(MsFlags::MS_STRICTATIME) {
            names.push("strictatime");
        }
        if self.0.contains(MsFlags::MS_LAZYTIME) {
            names.push("lazytime");
        }
        if self.0.contains(MsFlags::MS_ACTIVE) {
            names.push("active");
        }
        if self.0.contains(MsFlags::MS_NOUSER) {
            names.push("nouser");
        }

        names
    }
}

/* Constants */

/// Number of CPUs on the system.
/// Initialized lazily on startup.
pub static NPROC: Lazy<usize> = Lazy::new(num_cpus::get);

/* Utilities */

#[inline]
pub(crate) fn op2name(op: u8) -> &'static str {
    match op {
        0x2 => "bind",
        0x3 => "connect",
        0xb => "sendto",
        0x10 => "sendmsg",
        0x14 => "sendmmsg",
        _ => unreachable!(),
    }
}

#[inline]
pub(crate) fn op2errno(op: u8) -> Errno {
    if op == 0x2 {
        /* bind */
        Errno::EADDRNOTAVAIL
    } else {
        /* connect | sendto | sendmsg | sendmmsg */
        Errno::ECONNREFUSED
    }
}

/// Checks LandLock ABI v3 is supported.
/// Returns:
/// - 0: Fully enforced
/// - 1: Partially enforced
/// - 2: Not enforced
/// - 127: Unsupported
pub fn lock_enabled() -> u8 {
    let abi = ABI::V3;
    let path_ro = vec!["/".to_string()];
    let path_rw = vec!["/".to_string()];

    // A helper function to wrap the operations and reduce duplication
    fn landlock_operation(
        path_ro: &[String],
        path_rw: &[String],
        abi: ABI,
    ) -> Result<RestrictionStatus, RulesetError> {
        let ruleset = Ruleset::default().handle_access(AccessFs::from_all(abi))?;
        let created_ruleset = ruleset.create()?;
        let ro_rules = path_beneath_rules(path_ro, AccessFs::from_read(abi));
        let updated_ruleset = created_ruleset.add_rules(ro_rules)?;
        let rw_rules = path_beneath_rules(path_rw, AccessFs::from_all(abi));
        let final_ruleset = updated_ruleset.add_rules(rw_rules)?;
        final_ruleset.restrict_self().map_err(RulesetError::from)
    }

    match landlock_operation(&path_ro, &path_rw, abi) {
        Ok(status) => match status.ruleset {
            RulesetStatus::FullyEnforced => 0,
            RulesetStatus::PartiallyEnforced => 1,
            RulesetStatus::NotEnforced => 2,
        },
        Err(_) => 127,
    }
}

/// Returns true if we are running under syd.
#[allow(clippy::disallowed_methods)]
pub fn syd_enabled() -> bool {
    // This will not work if the sandbox is locked.
    // Path::new("/dev/syd").exists() || Path::new("/dev/syd").exists()
    // SAFETY: In libc, we trust.
    match unsafe { fork() } {
        Ok(ForkResult::Parent { child, .. }) => {
            match waitpid(child, None) {
                Ok(WaitStatus::Exited(_, code)) => {
                    // Check the child's exit status.
                    // Exit status of 0 means syd is enabled.
                    code == 0
                }
                _ => {
                    // If there's an error waiting on the
                    // child, assume syd is not enabled.
                    false
                }
            }
        }
        Ok(ForkResult::Child) => {
            let mut ctx = match ScmpFilterContext::new_filter(ScmpAction::Allow) {
                Ok(ctx) => ctx,
                Err(_) => exit(1),
            };

            let syscall = ScmpSyscall::new("open");
            if ctx.add_rule(ScmpAction::Notify, syscall).is_err() {
                exit(1);
            }

            if ctx.load().is_err() && Errno::last() == Errno::EBUSY {
                // seccomp filter exists
                // syd is in business.
                exit(0);
            } else {
                // seccomp filter does not exist
                exit(1);
            }
        }
        Err(_) => {
            // If there's an error forking,
            // assume syd is not enabled.
            false
        }
    }
}

/// Returns the name of the libsecc☮mp native architecture.
pub fn seccomp_arch_native_name() -> Option<&'static str> {
    match ScmpArch::native() {
        ScmpArch::X86 => Some("x86"),
        ScmpArch::X8664 => Some("x86_64"),
        ScmpArch::X32 => Some("x32"),
        ScmpArch::Arm => Some("arm"),
        ScmpArch::Aarch64 => Some("aarch64"),
        ScmpArch::Mips => Some("mips"),
        ScmpArch::Mips64 => Some("mips64"),
        ScmpArch::Mips64N32 => Some("mips64n32"),
        ScmpArch::Mipsel => Some("mipsel"),
        ScmpArch::Mipsel64 => Some("mipsel64"),
        ScmpArch::Mipsel64N32 => Some("mipsel64n32"),
        ScmpArch::Ppc => Some("ppc"),
        ScmpArch::Ppc64 => Some("ppc64"),
        ScmpArch::Ppc64Le => Some("ppc64le"),
        ScmpArch::S390 => Some("s390"),
        ScmpArch::S390X => Some("s390s"),
        ScmpArch::Parisc => Some("parisc"),
        ScmpArch::Parisc64 => Some("parisc64"),
        ScmpArch::Riscv64 => Some("riscv64"),
        _ => None,
    }
}

/// Given a `Uid`, return the user name of the user.
/// On any error conditions, return "nobody".
pub fn get_user_name(uid: Uid) -> String {
    match User::from_uid(uid) {
        Ok(Some(user)) => user.name,
        _ => "nobody".to_string(),
    }
}

/// Given a username, return the home directory of the user.
/// On any error conditions, return "/var/empty".
pub fn get_user_home(username: &str) -> String {
    // Fetch user details.
    match User::from_name(username) {
        Ok(Some(user)) => user.dir.to_string_lossy().to_string(),
        _ => "/var/empty".to_string(),
    }
}

// Sets the specified signal to be ignored.
//
// This function utilizes the `sigaction` system call to set the specified signal's action
// to `SIG_IGN`, effectively causing the process to ignore that signal.
//
// # Arguments
//
// * `signal` - The signal number (e.g., `SIGTSTP`, `SIGTTIN`, `SIGTTOU`).
//
// # Returns
//
// * `Result<(), Error>` - Returns `Ok(())` if successful, or an error if the operation fails.
//
// # Example
//
// ```no_run
// use nix::sys::signal::SIGTSTP;
//
// let result = syd::ignore_signal(SIGTSTP);
// assert!(result.is_ok());
// ```
pub(crate) fn ignore_signal(signal: Signal) -> Result<(), Errno> {
    let sig_action = SigAction::new(
        SigHandler::SigIgn, // Set to ignore
        SaFlags::empty(),
        SigSet::empty(),
    );

    // SAFETY: The unsafe call to `sigaction` is used to set the signal's disposition
    // to "ignore". We're not invoking any handlers or performing any operations that
    // could lead to data races or other undefined behaviors. Hence, it's safe to call
    // in this context.
    unsafe { sigaction(signal, &sig_action) }.map(drop)
}

const IOPRIO_CLASS_IDLE: i32 = 3;
const IOPRIO_WHO_PROCESS: i32 = 1;

/// Sets the I/O priority of the current thread to idle.
///
/// This function uses the `ioprio_set` syscall to set the I/O
/// scheduling priority of the current thread to the idle class. The
/// idle I/O class is designed for tasks that should only use disk
/// resources when no other process needs them. When a thread is set to
/// idle, it will not compete with other (non-idle) processes for I/O
/// bandwidth.
///
/// Note that this setting is applied at the thread level in Linux,
/// where each thread is treated as a separate scheduling entity. As a
/// result, calling this function will only affect the I/O priority of
/// the thread from which it is called. If the application is
/// multi-threaded and a global I/O priority change is desired, this
/// function needs to be called from each thread, or specific threads
/// requiring the priority change should be targeted.
///
/// The function does not require any parameters and returns a `Result`:
/// - `Ok(())` on success.
/// - `Err(Errno)` containing Errno.
///
/// # Safety
///
/// This function involves an unsafe block due to the direct system call
/// (`nix::libc::syscall`). The `ioprio_set` syscall is considered
/// unsafe as it directly interfaces with the kernel, bypassing Rust's
/// safety guarantees.  However, the usage in this context is safe given
/// that:
/// - We are specifying `IOPRIO_WHO_PROCESS` with `0`, which correctly
///   targets the current thread.
/// - The `ioprio` value is correctly constructed for the idle I/O
///   class.
///
/// Users of this function do not need to take any special safety precautions.
pub(crate) fn set_io_priority_idle() -> Result<(), Errno> {
    // Set I/O priority: higher bits for the class, lower bits for the priority.
    // IOPRIO_CLASS_IDLE is shifted left by 13 bits to fit the class into higher bits.
    // Priority for idle class is not used, hence set to 0 (lower 13 bits).
    let ioprio = IOPRIO_CLASS_IDLE << 13;

    // SAFETY:
    // The syscall libc::SYS_ioprio_set is used to set the I/O priority
    // of a process. This call is considered unsafe because it involves
    // a direct system call, which bypasses the safety checks and
    // abstractions provided by Rust. However, this usage is safe under
    // the following conditions:
    // 1. The first argument IOPRIO_WHO_PROCESS specifies the target as
    //    a process.
    // 2. The second argument 0 refers to the current process. In the
    //    context of ioprio_set, passing 0 for the 'who' parameter
    //    targets the calling process. This is why getpid() is not
    //    necessary here, as 0 implicitly represents the current
    //    process's PID.
    // 3. The third argument ioprio is correctly constructed with a
    //    valid I/O class and priority, ensuring the syscall behaves as
    //    expected.
    if unsafe { nix::libc::syscall(nix::libc::SYS_ioprio_set, IOPRIO_WHO_PROCESS, 0, ioprio) } == 0
    {
        Ok(())
    } else {
        Err(Errno::last())
    }
}

/// Set the current thread's CPU scheduling policy to 'idle'.
///
/// This function sets the CPU scheduling policy of the current thread
/// to SCHED_IDLE, indicating that the thread should only be scheduled
/// to run when the system is idle.
///
/// # Returns
///
/// * `Ok(())` on successful setting of the scheduling policy and priority.
/// * `Err` on failure, with the specific error indicating the cause of the failure.
pub(crate) fn set_cpu_priority_idle() -> Result<(), Errno> {
    // SAFETY: We zero out the sched_param struct. This is safe because:
    // 1. sched_param is a plain data struct with no invariants related
    //    to its fields.
    // 2. All-zero is a valid representation for this struct in the
    //    context of SCHED_IDLE policy.
    let param: nix::libc::sched_param = unsafe { std::mem::zeroed() };

    // SAFETY: The call to nix::libc::sched_setscheduler is safe because:
    // 1. We are passing valid arguments: a PID of 0 for the current
    //    thread, a valid policy (SCHED_IDLE), and a pointer to a
    //    properly initialized sched_param structure.
    // 2. There are no thread-safety issues since the operation only
    //    affects the current thread.
    if unsafe { nix::libc::sched_setscheduler(0, nix::libc::SCHED_IDLE, std::ptr::addr_of!(param)) }
        == 0
    {
        Ok(())
    } else {
        Err(Errno::last())
    }
}

const SECCOMP_ARCH_LIST: &[ScmpArch] = &[
    ScmpArch::X86,
    ScmpArch::X8664,
    ScmpArch::X32,
    ScmpArch::Arm,
    ScmpArch::Aarch64,
    ScmpArch::Mips,
    ScmpArch::Mips64,
    ScmpArch::Mips64N32,
    ScmpArch::Mipsel,
    ScmpArch::Mipsel64,
    ScmpArch::Mipsel64N32,
    ScmpArch::Ppc,
    ScmpArch::Ppc64,
    ScmpArch::Ppc64Le,
    ScmpArch::S390,
    ScmpArch::S390X,
    ScmpArch::Parisc,
    ScmpArch::Parisc64,
    ScmpArch::Riscv64,
];

/// Print list of libseccomp's supported architectures
/// Used by `syd --arch list`
pub fn print_seccomp_architectures() {
    let native = ScmpArch::native();
    for arch in SECCOMP_ARCH_LIST {
        let mut repr = format!("{arch:?}").to_ascii_lowercase();
        if repr == "x8664" {
            // Fix potential confusion.
            repr = "x86_64".to_string();
        }
        if *arch == native {
            println!("- {repr} [*]")
        } else {
            println!("- {repr}");
        }
    }
}

// List of libseccomp supported architectures for the current system.
#[cfg(target_arch = "x86_64")]
pub(crate) const SCMP_ARCH: &[ScmpArch] = &[ScmpArch::X8664, ScmpArch::X86, ScmpArch::X32];
#[cfg(target_arch = "x86")]
pub(crate) const SCMP_ARCH: &[ScmpArch] = &[ScmpArch::X86];
#[cfg(target_arch = "arm")]
pub(crate) const SCMP_ARCH: &[ScmpArch] = &[ScmpArch::Arm];
#[cfg(target_arch = "aarch64")]
pub(crate) const SCMP_ARCH: &[ScmpArch] = &[ScmpArch::Aarch64, ScmpArch::Arm];
#[cfg(target_arch = "mips")]
pub(crate) const SCMP_ARCH: &[ScmpArch] = &[ScmpArch::Mips];
#[cfg(target_arch = "mips64")]
pub(crate) const SCMP_ARCH: &[ScmpArch] = &[ScmpArch::Mips64, ScmpArch::Mips64N32, ScmpArch::Mips];
//#[cfg(target_arch = "mips64n32")]
//pub(crate) const SCMP_ARCH: &[ScmpArch] = &[ScmpArch::Mips64N32, ScmpArch::Mips, ScmpArch::Mips64];
//#[cfg(target_arch = "mipsel")]
//pub(crate) const SCMP_ARCH: &[ScmpArch] =
//    &[ScmpArch::Mipsel64, ScmpArch::Mipsel64N32, ScmpArch::Mipsel];
//#[cfg(target_arch = "mipsel64")]
//pub(crate) const SCMP_ARCH: &[ScmpArch] =
//    &[ScmpArch::Mipsel64, ScmpArch::Mipsel, ScmpArch::Mipsel64N32];
//#[cfg(target_arch = "mipsel64n32")]
//pub(crate) const SCMP_ARCH: &[ScmpArch] =
//    &[ScmpArch::Mipsel64N32, ScmpArch::Mipsel, ScmpArch::Mipsel64];
#[cfg(target_arch = "powerpc")]
pub(crate) const SCMP_ARCH: &[ScmpArch] = &[ScmpArch::Ppc];
#[cfg(target_arch = "powerpc64")]
pub(crate) const SCMP_ARCH: &[ScmpArch] = &[ScmpArch::Ppc64, ScmpArch::Ppc64Le, ScmpArch::Ppc];
//#[cfg(target_arch = "parisc")]
//pub(crate) const SCMP_ARCH: &[ScmpArch] = &[ScmpArch::Parisc];
//#[cfg(target_arch = "parisc64")]
//pub(crate) const SCMP_ARCH: &[ScmpArch] = &[ScmpArch::Parisc64, ScmpArch::Parisc];
#[cfg(target_arch = "riscv64")]
pub(crate) const SCMP_ARCH: &[ScmpArch] = &[ScmpArch::Riscv64];
#[cfg(target_arch = "s390x")]
pub(crate) const SCMP_ARCH: &[ScmpArch] = &[ScmpArch::S390X, ScmpArch::S390];

// Define MACHINE_NAME based on the target architecture
#[cfg(target_arch = "x86_64")]
pub(crate) const MACHINE_NAME: &str = "x86_64";
#[cfg(target_arch = "x86")]
pub(crate) const MACHINE_NAME: &str = "x86";
#[cfg(target_arch = "arm")]
pub(crate) const MACHINE_NAME: &str = "arm";
#[cfg(target_arch = "aarch64")]
pub(crate) const MACHINE_NAME: &str = "aarch64";
#[cfg(target_arch = "mips")]
pub(crate) const MACHINE_NAME: &str = "mips";
#[cfg(target_arch = "mips64")]
pub(crate) const MACHINE_NAME: &str = "mips64";
//#[cfg(target_arch = "mips64n32")]
//pub(crate) const MACHINE_NAME: &str = "mips64n32";
//#[cfg(target_arch = "mipsel")]
//pub(crate) const MACHINE_NAME: &str = "mipsel";
//#[cfg(target_arch = "mipsel64")]
//pub(crate) const MACHINE_NAME: &str = "mipsel64";
//#[cfg(target_arch = "mipsel64n32")]
//pub(crate) const MACHINE_NAME: &str = "mipsel64n32";
#[cfg(target_arch = "powerpc")]
pub(crate) const MACHINE_NAME: &str = "powerpc";
#[cfg(target_arch = "powerpc64")]
pub(crate) const MACHINE_NAME: &str = "powerpc64";
//#[cfg(target_arch = "parisc")]
//pub(crate) const MACHINE_NAME: &str = "parisc";
//#[cfg(target_arch = "parisc64")]
//pub(crate) const MACHINE_NAME: &str = "parisc64";
#[cfg(target_arch = "riscv64")]
pub(crate) const MACHINE_NAME: &str = "riscv64";
#[cfg(target_arch = "s390x")]
pub(crate) const MACHINE_NAME: &str = "s390x";

/// Add all supported architectures to the given filter.
#[allow(clippy::cognitive_complexity)]
pub fn seccomp_add_architectures(ctx: &mut ScmpFilterContext) -> IOResult<()> {
    // Add architectures based on the current architecture
    for arch in SCMP_ARCH {
        seccomp_add_arch(ctx, *arch)?;
    }
    Ok(())
}

/// Convert libseccomp errno to nix errno.
pub fn seccomp2errno(errno: SeccompErrno) -> Errno {
    match errno {
        SeccompErrno::EACCES => Errno::EACCES,
        SeccompErrno::ECANCELED => Errno::ECANCELED,
        SeccompErrno::EDOM => Errno::EDOM,
        SeccompErrno::EEXIST => Errno::EEXIST,
        SeccompErrno::EFAULT => Errno::EFAULT,
        SeccompErrno::EINVAL => Errno::EINVAL,
        SeccompErrno::ENOENT => Errno::ENOENT,
        SeccompErrno::ENOMEM => Errno::ENOMEM,
        SeccompErrno::EOPNOTSUPP => Errno::EOPNOTSUPP,
        SeccompErrno::ERANGE => Errno::ERANGE,
        SeccompErrno::ESRCH => Errno::ESRCH,
        _ => Errno::ECANCELED,
    }
}

fn seccomp_add_arch(ctx: &mut ScmpFilterContext, arch: ScmpArch) -> IOResult<bool> {
    ctx.add_arch(arch).map_err(|e| {
        if let Some(errno) = e.errno() {
            std::io::Error::from_raw_os_error(seccomp2errno(errno) as i32)
        } else {
            std::io::Error::new(std::io::ErrorKind::Other, "Unknown seccomp error")
        }
    })
}

/// Check if arch is 64-bit or 32-bit.
#[inline]
pub fn scmp_arch_bits(arch: &ScmpArch) -> usize {
    match arch {
        ScmpArch::X8664
        | ScmpArch::Aarch64
        | ScmpArch::Mips64
        | ScmpArch::Mips64N32
        | ScmpArch::Mipsel64
        | ScmpArch::Mipsel64N32
        | ScmpArch::Ppc64
        | ScmpArch::Ppc64Le
        | ScmpArch::Parisc64
        | ScmpArch::Riscv64
        | ScmpArch::S390X => 64,
        ScmpArch::X86
        | ScmpArch::X32
        | ScmpArch::Arm
        | ScmpArch::Mips
        | ScmpArch::Mipsel
        | ScmpArch::Ppc
        | ScmpArch::Parisc
        | ScmpArch::S390 => 32,
        _ => 64, // sane default for non-exhaustive enum.
    }
}

/// A helper function to wrap the Landlock operations and reduce duplication
pub fn landlock_operation(
    path_ro: &[String],
    path_rw: &[String],
    abi: ABI,
) -> Result<RestrictionStatus, RulesetError> {
    let ruleset = Ruleset::default().handle_access(AccessFs::from_all(abi))?;
    let created_ruleset = ruleset.create()?;
    let ro_rules = path_beneath_rules(path_ro, AccessFs::from_read(abi));
    let updated_ruleset = created_ruleset.add_rules(ro_rules)?;
    let rw_rules = path_beneath_rules(path_rw, AccessFs::from_all(abi));
    let final_ruleset = updated_ruleset.add_rules(rw_rules)?;
    final_ruleset.restrict_self().map_err(RulesetError::from)
}

/// Simple human size formatter.
#[allow(clippy::arithmetic_side_effects)]
#[allow(clippy::cast_precision_loss)]
pub fn human_size(bytes: usize) -> String {
    const SIZES: &[char] = &['B', 'K', 'M', 'G', 'T', 'P', 'E'];
    let factor = 1024usize;

    let mut size = bytes as f64;
    let mut i = 0;

    while size > factor as f64 && i < SIZES.len() - 1 {
        size /= factor as f64;
        i += 1;
    }

    format!("{:.2}{}", size, SIZES[i])
}

#[cfg(target_env = "musl")]
pub(crate) type IoctlRequest = nix::libc::c_int;
#[cfg(not(target_env = "musl"))]
pub(crate) type IoctlRequest = nix::libc::c_ulong;

const SIOCGIFFLAGS: IoctlRequest = nix::libc::SIOCGIFFLAGS as IoctlRequest;
const SIOCSIFFLAGS: IoctlRequest = nix::libc::SIOCSIFFLAGS as IoctlRequest;

/// Functionally equivalent to "ifconfig lo up".
pub fn bring_up_loopback() -> Result<(), Errno> {
    // Create a socket
    let sockfd: RawFd = socket(
        AddressFamily::Inet,
        SockType::Stream,
        SockFlag::empty(),
        None,
    )?;

    // SAFETY: socket returns a valid FD.
    let sockfd = unsafe { OwnedFd::from_raw_fd(sockfd) };

    // Prepare the interface request
    let mut ifreq = nix::libc::ifreq {
        #[allow(clippy::cast_possible_wrap)]
        ifr_name: [
            b'l' as nix::libc::c_char,
            b'o' as nix::libc::c_char,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
        ],
        // SAFETY: Manually initialize ifr_ifru
        ifr_ifru: unsafe { std::mem::zeroed() },
    };

    // SAFETY: Get the current flags
    if unsafe { nix::libc::ioctl(sockfd.as_raw_fd(), SIOCGIFFLAGS, &mut ifreq) } != 0 {
        return Err(Errno::last());
    }

    // Modify the flags to bring up the interface
    // SAFETY: We're accessing the field of a union here.
    #[allow(clippy::cast_possible_truncation)]
    unsafe {
        ifreq.ifr_ifru.ifru_flags |=
            (nix::libc::IFF_UP | nix::libc::IFF_RUNNING) as nix::libc::c_short
    };

    // SAFETY: Set the new flags
    if unsafe { nix::libc::ioctl(sockfd.as_raw_fd(), SIOCSIFFLAGS, &mut ifreq) } == 0 {
        Ok(())
    } else {
        Err(Errno::last())
    }
}

/// Enumerate shared object dependencies.
#[cfg(feature = "elf")]
#[allow(clippy::too_many_arguments)]
pub fn ldd<P: AsRef<Path>>(
    file: &mut File,
    size: u64,
    deps: &mut Vec<PathBuf>,
    root: Option<P>,
    pid: Pid,
    deny_elf32: bool,
    deny_elf_dynamic: bool,
    deny_elf_static: bool,
) -> Result<(), Errno> {
    // Analyze the binary.
    let analyzer = DependencyAnalyzer::new(
        root.map(|p| p.as_ref().to_path_buf())
            .unwrap_or(PathBuf::from("/")),
    );
    match analyzer.analyze(
        file,
        size,
        deps,
        pid,
        deny_elf32,
        deny_elf_dynamic,
        deny_elf_static,
    ) {
        Ok(_) => Ok(()),
        Err(lddtree::Error::NotDynamic) => {
            // Statically linked binary, the only dependency is the binary itself.
            Ok(())
        }
        Err(
            lddtree::Error::Goblin(goblin::error::Error::Malformed(_))
            | lddtree::Error::Goblin(goblin::error::Error::BadMagic(_)),
        ) => {
            // SAFETY: Do not pass thru malformed ELF to the kernel.
            Err(Errno::ENOEXEC)
        }
        Err(lddtree::Error::Io(error))
        | Err(lddtree::Error::LdSoConf(lddtree::ld_so_conf::LdSoConfError::Io(error)))
        | Err(lddtree::Error::Goblin(goblin::error::Error::IO(error))) => Err(err2no(error)),
        _ => Err(Errno::ENOSYS),
    }
}

/// Check for file descriptor leaks above the standard input, output, and error.
///
/// This function examines the `/proc/self/fd` directory to identify open file descriptors.
/// It prints any open file descriptors other than the standard input (0), output (1), and error (2),
/// indicating potential resource leaks.
///
/// # Parameters
/// - `fd_max`: An optional parameter that sets a maximum file descriptor number to check.
///             If not specified, only the standard file descriptors are considered normal.
///
/// # Returns
/// Returns `true` if leaks are found, otherwise `false`.
pub fn check_fd_leaks(fd_max: Option<RawFd>) -> u32 {
    let proc_fd_path = Path::new("/proc/self/fd");
    let mut dir = match Dir::open(proc_fd_path, OFlag::O_RDONLY, Mode::empty()) {
        Ok(d) => d,
        Err(e) => {
            eprintln!("Failed to open /proc/self/fd: {e}");
            return u32::MAX;
        }
    };

    let mut leaks_found: u32 = 0;
    let dir_fd = dir.as_raw_fd();
    let fd_limit = fd_max.unwrap_or(2); // Default limit only std fds

    for entry in dir.iter() {
        let entry = match entry {
            Ok(e) => e,
            Err(_) => continue,
        };

        let fd_str = entry.file_name().to_string_lossy(); // Use lossy conversion
        let fd = match fd_str.parse::<RawFd>() {
            Ok(fd) => fd,
            Err(_) => continue,
        };

        // Ignore standard file descriptors and the directory stream FD itself
        if fd <= fd_limit || fd == dir_fd {
            continue;
        }

        // Create a PathBuf from the string representation of the file descriptor
        let link_path = proc_fd_path.join(fd_str.into_owned()); // Convert Cow<str> into a String and then into a PathBuf

        match std::fs::read_link(&link_path) {
            Ok(target_path) => {
                eprintln!("!!! Leaked file descriptor {fd} -> {target_path:?} !!!");
                leaks_found = leaks_found.saturating_add(1);
            }
            Err(error) => {
                eprintln!("Failed to read link for FD {fd}: {error}");
            }
        }
    }

    leaks_found
}

/// Parse the given string into a UID.
/// 1. use getpwnam_r(3)
/// 2. parse as integer
pub(crate) fn parse_user(name: &str) -> Result<Uid, Errno> {
    if name.chars().all(|c| c.is_ascii_digit()) {
        Ok(Uid::from_raw(
            name.parse::<nix::libc::uid_t>()
                .map_err(|_| Errno::EINVAL)?,
        ))
    } else if let Some(user) = User::from_name(name)? {
        Ok(user.uid)
    } else {
        Err(Errno::ENOENT)
    }
}

/// Parse the given string into a GID.
/// 1. use getpwnam_r(3)
/// 2. parse as integer
pub(crate) fn parse_group(name: &str) -> Result<Gid, Errno> {
    if name.chars().all(|c| c.is_ascii_digit()) {
        Ok(Gid::from_raw(
            name.parse::<nix::libc::gid_t>()
                .map_err(|_| Errno::EINVAL)?,
        ))
    } else if let Some(group) = Group::from_name(name)? {
        Ok(group.gid)
    } else {
        Err(Errno::ENOENT)
    }
}

/// Return the GNU build id of the current executable.
#[cfg(feature = "elf")]
pub fn current_exe_id() -> Result<String, Errno> {
    let bytes = std::fs::read("/proc/self/exe").map_err(err2no)?;
    let elf = goblin::elf::Elf::parse(&bytes).map_err(|_| Errno::EINVAL)?;
    if let Some(notes) = elf.iter_note_headers(&bytes) {
        for note in notes.flatten() {
            if note.name == "GNU" && note.n_type == 3 {
                return Ok(const_hex::encode(note.desc));
            }
        }
    }
    Err(Errno::ENOENT)
}

#[cfg(target_arch = "x86")]
#[inline(always)]
/// Fork fast.
///
/// # Safety
///
/// Unsafe to be fast!
pub unsafe fn fork_fast() {
    asm!(
        "mov eax, 0x2", // 0x2 is the syscall number for fork on x86
        "int 0x80",     // Interrupt to make the syscall
        out("eax") _,
    );
}

#[cfg(target_arch = "x86_64")]
#[inline(always)]
/// Fork fast.
///
/// # Safety
///
/// Unsafe to be fast!
pub unsafe fn fork_fast() {
    // Inline assembly for x86-64
    asm!(
        "mov rax, 57", // 57 is the syscall number for fork on x86-64
        "syscall",
        out("rax") _,
    );
}

#[cfg(target_arch = "aarch64")]
#[inline(always)]
/// Fork fast.
///
/// # Safety
///
/// Unsafe to be fast!
pub unsafe fn fork_fast() {
    asm!(
        "mov x0, 17",  // SIGCHLD
        "mov x1, 0",   // child_stack (null, not recommended)
        "mov x8, 220", // syscall number for clone
        "svc 0",
        options(nostack),
    );
}

#[cfg(target_arch = "arm")]
#[inline(always)]
/// Fork fast.
///
/// # Safety
///
/// Unsafe to be fast!
pub unsafe fn fork_fast() {
    asm!(
        "mov r7, #2", // 2 is the syscall number for fork on ARM
        "swi #0",     // Software interrupt to make the syscall
        out("r0") _,
        options(nostack),
    );
}

#[cfg(not(any(
    target_arch = "arm",
    target_arch = "x86",
    target_arch = "x86_64",
    target_arch = "aarch64"
)))]
#[inline(always)]
/// Fork fast.
///
/// # Safety
///
/// Unsafe to be fast!
pub unsafe fn fork_fast() {
    let _ = fork();
}

/// Set SIGPIPE handler to default.
pub fn set_sigpipe_dfl() -> Result<(), Errno> {
    // SAFETY: The nix::sys::signal::signal function is unsafe because
    // it affects the global state of the program by changing how a
    // signal (SIGPIPE in this case) is handled. It's safe to call here
    // because changing the SIGPIPE signal to its default behavior will
    // not interfere with any other part of this program that could be
    // relying on a custom SIGPIPE signal handler.
    unsafe { signal(Signal::SIGPIPE, SigHandler::SigDfl) }.map(drop)
}

#[allow(dead_code)]
#[inline]
#[cold]
fn cold() {}

#[allow(dead_code)]
#[inline]
pub(crate) fn likely(b: bool) -> bool {
    if !b {
        cold()
    }
    b
}

#[allow(dead_code)]
#[inline]
pub(crate) fn unlikely(b: bool) -> bool {
    if b {
        cold()
    }
    b
}

/// Write the message to the invalid fd -42.
/// The idea is to look for it in strace logs.
/// Only works in debug mode, noop in release.
pub fn t(_msg: &str) {
    #[cfg(feature = "log")]
    let _ = nix::unistd::write(-42, _msg.as_bytes());
}

#[cfg(feature = "log")]
/// Write a formatted message to an invalid fd.
#[macro_export]
macro_rules! t {
    ($($arg:tt)*) => {{
        syd::t(&format!($($arg)*));
    }}
}

#[cfg(not(feature = "log"))]
/// Write a formatted message to an invalid fd.
#[macro_export]
macro_rules! t {
    ($($arg:tt)*) => {{}};
}
