//
// syd: seccomp and landlock based application sandbox with support for namespaces
// src/syd-path.rs: write integrity force rules based for binaries under PATH
//
// Copyright (c) 2024 Ali Polatel <alip@chesswob.org>
//
// SPDX-License-Identifier: GPL-3.0-or-later

use std::{
    env,
    fs::{canonicalize, read_dir, File},
    io::BufReader,
    path::Path,
    process::ExitCode,
};

use getargs::{Opt, Options};
use nix::unistd::{access, AccessFlags};
use syd::hash::HashAlgorithm;

fn main() -> ExitCode {
    #[allow(clippy::disallowed_methods)]
    syd::set_sigpipe_dfl().expect("SIGPIPE");

    let args = env::args().skip(1).collect::<Vec<_>>();
    let mut opts = Options::new(args.iter().map(String::as_str));

    let mut dsyd = false;
    let mut path = env::var("PATH").unwrap_or_default();
    let mut func = HashAlgorithm::Sha512;
    let mut action = "deny";
    let mut limit = 0usize; // 0 means no limit.
    let mut count = 0usize;

    // SAFETY: We panic on parse errors.
    #[allow(clippy::disallowed_methods)]
    while let Some(opt) = opts.next_opt().expect("next opt") {
        match opt {
            Opt::Short('h') => {
                help();
                return ExitCode::SUCCESS;
            }
            Opt::Short('1') => {
                func = HashAlgorithm::Sha1;
            }
            Opt::Short('2') => {
                func = HashAlgorithm::Sha256;
            }
            Opt::Short('3') => {
                func = HashAlgorithm::Sha384;
            }
            Opt::Short('5') => {
                func = HashAlgorithm::Sha512;
            }
            Opt::Short('m') => {
                func = HashAlgorithm::Md5;
            }
            Opt::Short('d') => {
                action = "deny";
            }
            Opt::Short('k') => {
                action = "kill";
            }
            Opt::Short('w') => {
                action = "warn";
            }
            Opt::Short('p') => {
                path = match opts.value() {
                    Ok(value) => value.to_string(),
                    Err(_) => {
                        eprintln!("-p requires an argument!");
                        return ExitCode::FAILURE;
                    }
                };
            }
            Opt::Short('l') => {
                limit = match opts.value() {
                    Ok(value) => match value.parse::<usize>() {
                        Ok(value) => value,
                        Err(_) => {
                            eprintln!("Invalid argument given to -l!");
                            return ExitCode::FAILURE;
                        }
                    },
                    Err(_) => {
                        eprintln!("-l requires an argument!");
                        return ExitCode::FAILURE;
                    }
                };
            }
            Opt::Short('s') => {
                dsyd = true;
            }
            _ => {
                eprintln!("Unknown option: {opt:?}!");
                return ExitCode::FAILURE;
            }
        }
    }

    let dirs = path.split(':');
    for dir in dirs {
        if !Path::new(dir).is_dir() {
            continue;
        }
        if let Ok(entries) = read_dir(dir) {
            for entry in entries.flatten() {
                // Ensure the file is executable
                let path = entry.path();
                if path.is_file() && access(&path, AccessFlags::X_OK).is_ok() {
                    if let Ok(path) = canonicalize(path) {
                        if let Ok(file) = File::open(&path) {
                            let reader = BufReader::new(file);
                            if let Ok(key) = syd::hash::hash(reader, func) {
                                let prefix = if dsyd { "/dev/syd/" } else { "" };
                                println!(
                                    "{}force+{}:{}:{}",
                                    prefix,
                                    path.to_string_lossy(),
                                    const_hex::encode(key),
                                    action
                                );
                                if limit > 0 {
                                    count += 1;
                                    if count >= limit {
                                        return ExitCode::SUCCESS;
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
    }

    ExitCode::SUCCESS
}

fn help() {
    println!("Usage: syd-path [-h12dklpsw]");
    println!("Write Integrity Force rules based for binaries under PATH.");
    println!("Use -m to calculate MD5 instead of SHA3-512 (\x1b[91minsecure\x1b[0m, \x1b[96mPortage\x1b[0m/\x1b[95mPaludis\x1b[0m vdb compat).");
    println!("Use -1 to calculate SHA1 instead of SHA3-512 (\x1b[91minsecure\x1b[0m).");
    println!("Use -2 to calculate SHA3-256 instead of SHA3-512.");
    println!("Use -d for deny, -k for kill, -w for warn.");
    println!("Specify alternative PATH with -p.");
    println!("Use -l <num> to limit by number of entries.");
    println!("Use -s to prefix rules with /dev/syd.");
}
