//
// syd: seccomp and landlock based application sandbox with support for namespaces
// src/syd.rs: Main entry point
//
// Copyright (c) 2023, 2024 Ali Polatel <alip@chesswob.org>
//
// SPDX-License-Identifier: GPL-3.0-or-later

//! syd: seccomp and landlock based application sandbox with support for namespaces
//! Main entry point.

// We like clean and simple code with documentation.
#![deny(missing_docs)]
#![deny(clippy::allow_attributes_without_reason)]
#![deny(clippy::arithmetic_side_effects)]
#![deny(clippy::as_ptr_cast_mut)]
#![deny(clippy::as_underscore)]
#![deny(clippy::assertions_on_result_states)]
#![deny(clippy::borrow_as_ptr)]
#![deny(clippy::branches_sharing_code)]
#![deny(clippy::case_sensitive_file_extension_comparisons)]
#![deny(clippy::cast_lossless)]
#![deny(clippy::cast_possible_truncation)]
#![deny(clippy::cast_possible_wrap)]
#![deny(clippy::cast_precision_loss)]
#![deny(clippy::cast_ptr_alignment)]
#![deny(clippy::cast_sign_loss)]
#![deny(clippy::checked_conversions)]
#![deny(clippy::clear_with_drain)]
#![deny(clippy::clone_on_ref_ptr)]
#![deny(clippy::cloned_instead_of_copied)]
#![deny(clippy::cognitive_complexity)]
#![deny(clippy::collection_is_never_read)]
#![deny(clippy::copy_iterator)]
#![deny(clippy::create_dir)]
#![deny(clippy::dbg_macro)]
#![deny(clippy::debug_assert_with_mut_call)]
#![deny(clippy::decimal_literal_representation)]
#![deny(clippy::default_trait_access)]
#![deny(clippy::default_union_representation)]
#![deny(clippy::derive_partial_eq_without_eq)]
#![deny(clippy::doc_link_with_quotes)]
#![deny(clippy::doc_markdown)]
#![deny(clippy::explicit_into_iter_loop)]
#![deny(clippy::explicit_iter_loop)]
#![deny(clippy::fallible_impl_from)]
#![deny(clippy::missing_safety_doc)]
#![deny(clippy::undocumented_unsafe_blocks)]

use std::{
    env,
    env::VarError,
    ffi::{CStr, CString, OsStr, OsString},
    fs::File,
    io::Write,
    os::unix::ffi::OsStringExt,
    path::{Path, PathBuf},
    process::{exit, ExitCode},
    str::FromStr,
};

use anyhow::{anyhow, bail, Context};
use getargs::{Opt, Options};
use lexis::ToName;
use libseccomp::{ScmpAction, ScmpFilterContext, ScmpSyscall, ScmpVersion};
use nix::{
    errno::Errno,
    fcntl::{open, OFlag},
    libc::setdomainname,
    mount::{mount, MsFlags},
    sched::{clone, CloneFlags},
    sys::{
        signal::Signal,
        stat::Mode,
        utsname::uname,
        wait::{waitpid, WaitStatus},
    },
    unistd::{close, fexecve, getgid, getuid, sethostname},
    NixPath,
};
use syd::{
    bring_up_loopback, caps,
    config::*,
    error,
    fs::mask_path,
    hash::{hash_str, NoHashSet},
    hook::{ExportMode, Supervisor},
    info,
    log::JsonLinesLogger,
    proc::proc_limit_userns,
    sandbox::Sandbox,
    unshare::{GidMap, UidMap},
    SydPersona,
};

// Set global allocator to mimalloc if mimalloc feature is enabled (default).
#[cfg(feature = "mimalloc")]
#[global_allocator]
static GLOBAL: mimalloc2_rust::GlobalMiMalloc = mimalloc2_rust::GlobalMiMalloc;

// This line will include the content of "esyd.sh" at compile-time
const ESYD: &str = include_str!("esyd.sh");

const NONE: Option<&PathBuf> = None::<PathBuf>.as_ref();

#[allow(clippy::cognitive_complexity)]
fn main() -> anyhow::Result<ExitCode> {
    // Guard against CVE-2019-5736:
    // Copy /proc/self/exe in an anonymous fd (created via memfd_create), seal it and re-execute it.
    // See:
    // - https://github.com/opencontainers/runc/commit/0a8e4117e7f715d5fbeef398405813ce8e88558b
    // - https://github.com/lxc/lxc/commit/6400238d08cdf1ca20d49bafb85f4e224348bf9d
    // Note: syd's procfs protections is another layer of defense against this.
    pentacle::ensure_sealed().context("Failed to seal /proc/self/exe. Is /proc mounted?")?;

    // Set SIGPIPE handler to default.
    syd::set_sigpipe_dfl().context("Failed to set SIGPIPE signal handler to default.")?;

    // Initialize logging.
    JsonLinesLogger::init(&None, "warn")?;

    // Parse CLI arguments
    let mut export: Option<ExportMode> = None;
    let mut sandbox: Sandbox = Sandbox::default();
    let mut cmd_arg0: Option<String> = None;
    let mut cmd_argv: Vec<String> = vec![];
    let mut cmd_envp: NoHashSet = NoHashSet::default();

    // Note: If someone ever needs/wants non-UTF-8 CLI arguments,
    // we can consider changing this, but till then we consider
    // this safe and sound.
    #[allow(clippy::disallowed_methods)]
    let mut argv = argv::iter()
        .map(|os| os.to_str().expect("Argument is not valid UTF-8!"))
        .peekable();

    // Initialize Options.
    // SAFETY: Safe to assume argv has at least one element.
    // On the off chance that it's empty, we panic.
    #[allow(clippy::disallowed_methods)]
    let is_login = argv.next().expect("arg0").starts_with('-');
    let user_done = if is_login || argv.peek().is_none() {
        sandbox
            .parse_profile("user")
            .context("Error parsing user profile!")?;
        true
    } else {
        false
    };
    let mut user_parse = false;
    let mut opts = Options::new(argv);

    // Local options handled by this function.
    let sh = env::var(ENV_SH).unwrap_or(SYD_SH.to_string());

    while let Some(opt) = opts.next_opt().context("Failed to call Options::next!")? {
        match opt {
            /*
             * Basic options
             */
            Opt::Short('h') | Opt::Long("help") => {
                help();
                exit(0);
            }
            Opt::Short('V') | Opt::Long("version") => {
                version();
                exit(0);
            }

            /*
             * Sandbox options
             */
            Opt::Long("sh") => {
                println!("{ESYD}");
                exit(0);
            }
            Opt::Short('E') => {
                export = Some(ExportMode::from_str(
                    opts.value().context("-E requires an argument!")?,
                )?);
                sandbox
                    .parse_profile("lib")
                    .context("Error parsing lib profile!")?;
                if export == Some(ExportMode::PseudoFiltercode) {
                    // This variable makes setup_seccomp_parent print rules.
                    env::set_var("SYD_SECX", "1");
                } else {
                    env::remove_var("SYD_SECX");
                }
            }
            Opt::Short('x') => {
                env::set_var(ENV_NO_SYSLOG, "1");
                sandbox.set_trace(true);
            }
            Opt::Short('m') => {
                let cmd = opts.value().context("-m requires an argument!")?;
                if sandbox.locked() {
                    bail!("Failed to execute magic command `{cmd}': sandbox locked!");
                } else {
                    sandbox
                        .config(cmd)
                        .context(format!("Failed to execute magic command `{cmd}'."))?;
                }
            }
            Opt::Short('f') => {
                // Login shell compatibility:
                // Parse user profile as necessary.
                user_parse = true;
            }
            Opt::Short('l') | Opt::Long("login") => {
                // Login shell compatibility:
                // Parse user profile as necessary.
                user_parse = true;
            }
            Opt::Short('c') => {
                // When multiple -c arguments are given,
                // only the first one is honoured and
                // the rest is ignored in consistency
                // with how bash and dash behaves.
                user_parse = true;
                if cmd_argv.is_empty() {
                    cmd_argv.push(sh.clone());
                    cmd_argv.push("-c".to_string());
                    cmd_argv.push(
                        opts.value()
                            .context("-c requires an argument!")?
                            .to_string(),
                    );
                }
            }
            Opt::Short('P') => {
                let path = opts.value().context("-f requires an argument!")?;
                if sandbox.locked() {
                    error!("ctx": "init", "op": "configure_sandbox",
                        "path": mask_path(path),
                        "error": "sandbox is locked", "errno": Errno::EAGAIN as i32);
                } else {
                    sandbox
                        .parse_config_file(path)
                        .with_context(|| format!("Failed to parse configuration file `{path}'."))?;
                }
            }
            /* We keep --profile for syd-1 compatibility.
             * It's undocumented. */
            Opt::Short('p') | Opt::Long("profile") => {
                let profile = opts.value().context("-p requires an argument!")?;
                if sandbox.locked() {
                    error!("ctx": "init", "op": "configure_sandbox",
                        "profile": mask_path(profile),
                        "error": "sandbox is locked", "errno": Errno::EAGAIN as i32);
                } else {
                    sandbox.parse_profile(profile).with_context(|| {
                        format!("Failed to parse configuration profile `{profile}'.")
                    })?;
                }
            }

            /*
             * Unshare options
             */
            Opt::Short('a') => {
                let val = opts
                    .value()
                    .context("-a requires an argument!")?
                    .to_string();
                cmd_arg0 = Some(val);
            }
            Opt::Short('e') => {
                let value = opts.value().context("-e requires an argument!")?;
                match value.split_once('=') {
                    Some((var, val)) => {
                        cmd_envp.insert(hash_str(var));
                        if !val.is_empty() {
                            // This way we give the user the chance to pass-through
                            // denylisted environment variables e.g.
                            //      syd -eLD_LIBRARY_PATH= cmd
                            // is equivalent to
                            //      syd -eLD_LIBRARY_PATH=$LD_LIBRARY_PATH cmd
                            env::set_var(var, val);
                        }
                    }
                    None => {
                        cmd_envp.remove(&hash_str(value));
                        env::remove_var(value);
                    }
                }
            }

            Opt::Short(c) => {
                bail!("Invalid option `-{c}'!");
            }
            Opt::Long(c) => {
                bail!("Invalid option `--{c}'!");
            }
        }
    }

    if user_parse && !user_done && !sandbox.locked() {
        sandbox
            .parse_profile("user")
            .context("Error parsing user profile!")?;
    }

    // Prepare the command to execute, which may be a login shell.
    cmd_argv.extend(opts.positionals().map(|s| s.to_string()));
    let argv0 = if !cmd_argv.is_empty() {
        Some(cmd_argv.remove(0))
    } else {
        None
    };

    let argv0 = match (export.is_some(), argv0, is_login) {
        (true, _, _) => "true".to_string(),
        (false, Some(argv0), false) => argv0.to_string(),
        (false, None, false) | (false, _, true) => {
            if cmd_arg0.is_none() {
                // Allow user to override with -a
                // SAFETY: unwrap is safe as `sh` is UTF-8.
                #[allow(clippy::disallowed_methods)]
                let sh = Path::new(&sh)
                    .file_name()
                    .map(|s| s.to_str().unwrap())
                    .unwrap_or("-");
                cmd_arg0 = Some(format!("-{sh}"));
            }
            sh.to_string()
        }
    };

    // Handle namespace re-exec
    let init = env::var("SYD_INIT")
        == Err(VarError::NotUnicode(OsString::from_vec(vec![
            0xde, 0xad, 0xca, 0x11,
        ])));
    let fd = if !init {
        // Prepare /proc/self/exe fd for rexec.
        open(
            "/proc/self/exe",
            OFlag::O_PATH | OFlag::O_CLOEXEC,
            Mode::empty(),
        )
        .context("Failed to open /proc/self/exe. Is /proc mounted?")?
    } else {
        return Supervisor::run(&argv0, cmd_arg0, cmd_argv, cmd_envp, sandbox, export)
            .map(ExitCode::from);
    };

    // root/map:1 implies unshare/user:1
    // We do this here rather than in sandbox.rs,
    // because the user may toggle this flag many times.
    let map_root = sandbox.map_root();
    if map_root {
        sandbox.set_unshare_user(true);
    }

    // Set unshare-mount if we've any bind mounts.
    if sandbox.has_bind_mounts() {
        sandbox.set_unshare_mount(true);
    }

    // root: implies unshare/mount:1 and unshare/pid:1.
    if sandbox.root.is_some() {
        sandbox.set_unshare_mount(true);
        sandbox.set_unshare_pid(true);
    }

    // SAFETY: We cannot support NEWPID without NEWNS.
    // ie, pid namespace must have its own private /proc.
    if sandbox.unshare_pid() {
        sandbox.set_unshare_mount(true);
    }

    // Set up Linux namespaces if requested. Note,
    // we set it up here before spawning the child so as to
    // include the syd process into the pid namespace as well
    // such that the sandbox process and syd have the identical
    // view of /proc.
    let mut namespaces = CloneFlags::empty();
    for namespace in sandbox.namespaces() {
        namespaces |= syd::unshare::namespace::to_clone_flag(namespace);
    }
    if namespaces.is_empty() {
        let _ = close(fd); // /proc/self/exe fd is no longer necessary.

        // No namespace arguments passed, run normally.
        return Supervisor::run(&argv0, cmd_arg0, cmd_argv, cmd_envp, sandbox, export)
            .map(ExitCode::from);
    }

    let id_buf = if sandbox.unshare_user() {
        // create the UID and GID mappings.
        let uid = getuid().as_raw();
        let gid = getgid().as_raw();

        let uid_buf = {
            let uid_maps = vec![
                UidMap {
                    inside_uid: if map_root { 0 } else { uid },
                    outside_uid: uid,
                    count: 1,
                }, // Map the current user.
            ];
            let mut buf = Vec::new();
            for map in uid_maps {
                writeln!(
                    &mut buf,
                    "{} {} {}",
                    map.inside_uid, map.outside_uid, map.count
                )
                .context("Failed to format uidmap!")?;
            }
            buf
        };

        let gid_buf = {
            let gid_maps = vec![
                GidMap {
                    inside_gid: if map_root { 0 } else { gid },
                    outside_gid: gid,
                    count: 1,
                }, // Map the current group.
            ];
            let mut buf = Vec::new();
            for map in gid_maps {
                writeln!(
                    &mut buf,
                    "{} {} {}",
                    map.inside_gid, map.outside_gid, map.count
                )
                .context("Failed to format gidmap!")?;
            }
            buf
        };
        Some((uid_buf, gid_buf))
    } else {
        None
    };

    // Tell the kernel to keep the capabilities after the unshare call.
    // This is important because unshare() can change the user
    // namespace, which often leads to a loss of capabilities.
    caps::securebits::set_keepcaps(true).context("Failed to set keep capabilities flag!")?;

    // Prepare command to reexec syd!
    #[allow(clippy::disallowed_methods)]
    let mut argv: Vec<CString> = vec![CString::new("syd").unwrap()];
    #[allow(clippy::disallowed_methods)]
    argv.extend(env::args().skip(1).map(|arg| CString::new(arg).unwrap()));
    let argv: Vec<&CStr> = argv.iter().map(|arg| arg.as_c_str()).collect();

    // Prepare environment for fexecve, add the SYD_INIT variable.
    assert_eq!(env::var("SYD_INIT"), Err(VarError::NotPresent), "SYD_INIT");
    let syd_init_value = vec![0xde, 0xad, 0xca, 0x11];
    let mut syd_init = Vec::from("SYD_INIT=".as_bytes());
    syd_init.extend(&syd_init_value);
    // SAFETY: We ensure that the bytes are valid for CString except for
    // NUL, which we know is not present.
    let syd_init = unsafe { CString::from_vec_unchecked(syd_init) };

    #[allow(clippy::disallowed_methods)]
    let mut envp: Vec<CString> = env::vars()
        .map(|(key, value)| CString::new(format!("{}={}", key, value)).unwrap())
        .collect();
    envp.push(syd_init);
    let envp: Vec<&CStr> = envp.iter().map(|env_var| env_var.as_c_str()).collect();

    const STACK_SIZE: usize = 72 * 1024; // 72KB stack size for the new process
    let mut stack: [u8; STACK_SIZE] = [0; STACK_SIZE];
    #[allow(clippy::disallowed_methods)]
    let cb = Box::new(|| {
        // Child process is executing here.

        // Set host and domain name for uts namespace.
        // Ignore errors.
        if sandbox.unshare_uts() {
            // SAFETY: Domain name must not contain a nul-byte.
            #[allow(clippy::disallowed_methods)]
            let domainname = CString::new(sandbox.domainname.clone()).unwrap();

            // SAFETY: There's no setdomainname wrapper in nix.
            let _ = unsafe { setdomainname(domainname.as_ptr() as *const _, domainname.len()) };
            let _ = sethostname(&sandbox.hostname);
        }

        // Bring up loopback device for net namespace.
        if sandbox.unshare_net() {
            // Set up the loopback interface.
            // Ignore errors.
            match bring_up_loopback() {
                Ok(_) => info!("ctx": "init", "op": "bring_up_loopback"),
                Err(errno) => {
                    error!("ctx": "init", "op": "bring_up_loopback", "errno": errno as i32)
                }
            }
        }

        // Set mount propagation on the root filesystem for mount namespace.
        if sandbox.unshare_mount() {
            if let Some(flags) = sandbox.propagation {
                mount(Some("none"), "/", NONE, flags, NONE)
                    .expect("Failed to change root filesystem propagation");
            }

            // Mount private procfs as necessary.
            // The target directory may be under the chroot directory.
            // Use hidepid=2 to hide pid=1.
            if sandbox.unshare_pid() {
                if let Some(ref proc) = sandbox.proc {
                    let flags = MsFlags::MS_NOSUID | MsFlags::MS_NOEXEC | MsFlags::MS_NODEV;
                    mount(Some("proc"), proc, Some("proc"), flags, Some("hidepid=2"))
                        .expect("Failed to mount proc filesystem");
                }
            }
        }

        // Write uid/gid map for user namespace.
        if let Some((ref uid_buf, ref gid_buf)) = id_buf {
            // SAFETY: Limit maximum user namespaces to 1.
            // We'll enter into a user subnamespace in run().
            proc_limit_userns().expect("Failed to limit maximum user namespaces");

            // Write "deny" to /proc/self/setgroups before writing to gid_map.
            File::create("/proc/self/setgroups")
                .and_then(|mut f| f.write_all(b"deny"))
                .expect("Failed to write to /proc/self/setgroups");
            File::create("/proc/self/gid_map")
                .and_then(|mut f| f.write_all(&gid_buf[..]))
                .expect("Failed to write to /proc/self/gid_map");
            File::create("/proc/self/uid_map")
                .and_then(|mut f| f.write_all(&uid_buf[..]))
                .expect("Failed to write to /proc/self/uid_map");

            // Set inheritable mask and ambient caps to retain caps after execve(2).
            caps::securebits::set_keepcaps(true).expect("Failed to set keep capabilities flag");
            let permitted_caps = caps::read(None, caps::CapSet::Permitted)
                .expect("Failed to read permitted capabilities");
            caps::set(None, caps::CapSet::Inheritable, &permitted_caps)
                .expect("Failed to set inheritable capabilities mask");
            // Set the same capabilities as ambient, if necessary.
            for cap in permitted_caps {
                caps::raise(None, caps::CapSet::Ambient, cap)
                    .expect("Failed to raise capability to ambient!");
            }
        }

        // Rexecute syd in the new namespace.
        match fexecve(fd, &argv, &envp) {
            Ok(_) => {
                // SAFETY: fexecve does not return on success.
                unsafe { std::hint::unreachable_unchecked() }
            }
            Err(errno) => errno as isize,
        }
    });

    let child = clone(
        cb,
        &mut stack,
        namespaces,
        Some(Signal::SIGCHLD as nix::libc::c_int),
    )
    .context("Failed to unshare!")?;

    let _ = close(fd); // /proc/self/exe fd is no longer necessary.

    // SAFETY: Set up a Landlock sandbox to disallow all access.
    let p = vec![];
    let _ = syd::landlock_operation(&p, &p, LANDLOCK_ABI);

    // SAFETY: Set up a seccomp filter which only allows
    // 1. read, write and exit
    // 2. wait4 and waitpid
    // 3. memory allocation functions
    // 4. unlinkat to clean up temporary directories.
    let mut filter = ScmpFilterContext::new_filter(ScmpAction::KillProcess)
        .context("Failed to create seccomp filter!")?;
    let allow_call = [
        "read",
        "write",
        "exit",
        "exit_group",
        "wait4",
        "waitpid",
        "brk",
        "madvise",
        "mmap",
        "mmap2",
        "mprotect",
        "munmap",
        "sigaction",
        "sigaltstack",
        "sigpending",
        "sigprocmask",
        "sigsuspend",
        "sigreturn",
        "rt_sigaction",
        "rt_sigpending",
        "rt_sigprocmask",
        "rt_sigqueueinfo",
        "rt_sigreturn",
        "rt_sigtimedwait",
        "rt_sigtimedwait_time64",
    ];
    for call in allow_call {
        filter.add_rule(ScmpAction::Allow, ScmpSyscall::new(call))?;
    }
    filter.load().context("Failed to load seccomp filter!")?;

    loop {
        match waitpid(child, None) {
            Ok(WaitStatus::Exited(_, code)) => {
                #[allow(clippy::cast_possible_truncation)]
                #[allow(clippy::cast_sign_loss)]
                break Ok(ExitCode::from(code as u8));
            }
            Ok(WaitStatus::Signaled(_, signal, _)) => {
                break Ok(ExitCode::from(128_u8.saturating_add(signal as u8)));
            }
            Ok(WaitStatus::StillAlive) | Err(Errno::EINTR) => {}
            Err(error) => {
                break Err(anyhow!("Failed to wait: {error}"));
            }
            _ => {}
        }
    }
}

fn help() {
    println!(
        "syd [-acefhlmpxEPV] [--] {{command [arg...]}}
syd --sh
syd-cat profile-name|list
syd-chk
syd-env pid [-i] [name=value]... {{command [arg...]}}
syd-err number|name-regex
syd-exec {{command [arg...]}}
syd-ldd
syd-lock
syd-log
syd-ls set
syd-mem [-HV] [pid]
syd-norm path
syd-read path
syd-run pid {{command [arg...]}}
syd-size size|human-size
syd-stat [pid]
syd-sys [-a list|native|x86|x86_64|aarch64...] number|name-regex
syd-test [<name-regex>|<number>|<number>..<number>]..
syd-tty [pid]"
    );
}

/// Print version information.
fn version() {
    #[allow(clippy::disallowed_methods)]
    let major = env!("CARGO_PKG_VERSION_MAJOR")
        .parse::<u64>()
        .expect("CARGO_PKG_VERSION_MAJOR");
    #[allow(clippy::disallowed_methods)]
    let minor = env!("CARGO_PKG_VERSION_MINOR")
        .parse::<u64>()
        .expect("CARGO_PKG_VERSION_MINOR");
    #[allow(clippy::disallowed_methods)]
    let patch = env!("CARGO_PKG_VERSION_PATCH")
        .parse::<u64>()
        .expect("CARGO_PKG_VERSION_PATCH");
    let hex_version = (major << 16) | (minor << 8) | patch;
    let code_name = hex_version
        .to_name()
        .split('_')
        .map(|word| {
            let mut c = word.chars();
            match c.next() {
                None => String::new(),
                Some(f) => f.to_uppercase().collect::<String>() + c.as_str(),
            }
        })
        .collect::<Vec<String>>()
        .join(" ");
    #[cfg(feature = "elf")]
    let build_id = syd::current_exe_id().unwrap_or("Unknown".to_string());
    #[cfg(not(feature = "elf"))]
    let build_id = "Unavailable";

    println!("syd {} ({})", *VERSION, code_name);
    println!("Author: Ali Polatel");
    println!("License: GPL-3.0-or-later");

    let feat = [
        #[cfg(debug_assertions)]
        "+debug",
        #[cfg(not(debug_assertions))]
        "-debug",
        #[cfg(feature = "elf")]
        "+elf",
        #[cfg(not(feature = "elf"))]
        "-elf",
        #[cfg(feature = "log")]
        "+log",
        #[cfg(not(feature = "log"))]
        "-log",
        #[cfg(feature = "mimalloc")]
        "+mimalloc[secure]",
        #[cfg(not(feature = "mimalloc"))]
        "-mimalloc",
        #[cfg(feature = "oci")]
        "+oci",
        #[cfg(not(feature = "oci"))]
        "-oci",
        #[cfg(feature = "uring")]
        "+uring",
        #[cfg(not(feature = "uring"))]
        "-uring",
    ];
    println!("Features: {}", feat.join(", "));

    let state = match syd::lock_enabled() {
        0 => "fully enforced",
        1 => "partially enforced",
        2 => "not enforced",
        _ => "unsupported",
    };
    println!("Landlock: {state}");

    let libapi = libseccomp::get_api();
    match ScmpVersion::current() {
        Ok(libver) => {
            println!(
                "LibSeccomp: v{}.{}.{} api:{}",
                libver.major, libver.minor, libver.micro, libapi
            );
        }
        Err(error) => {
            println!("LibSeccomp: ? (error: {error})");
        }
    }

    let uname = match uname() {
        Ok(info) => OsStr::to_str(info.release()).unwrap_or("?").to_string(),
        Err(_) => "?".to_string(),
    };
    println!("Id (build): {build_id}");
    println!("Host (build): {}", env!("SYD_BUILDHOST"));
    println!(
        "Host (target): {uname} {}",
        syd::seccomp_arch_native_name().unwrap_or("?")
    );
    println!("Target Environment: {}", env!("SYD_TARGET_ENV"));

    // SAFETY: In libc we trust.
    let pers = match unsafe { nix::libc::personality(0xFFFFFFFF) } {
        n if n < 0 => format!("? (error: {})", Errno::last()),
        n => SydPersona(n).to_string(),
    };
    println!("Target Persona: {pers}");

    println!("Target Pointer Width: {}", env!("SYD_TARGET_POINTER_WIDTH"));
    println!(
        "Target CPU Count: {} ({} physical)",
        num_cpus::get(),
        num_cpus::get_physical()
    );
    println!("Target CPU Endian: {}", env!("SYD_TARGET_ENDIAN"));
    println!("Target CPU Features: {}", env!("SYD_TARGET_FEATURE"));
}
