//
// syd: seccomp and landlock based application sandbox with support for namespaces
// src/test/test.rs: Integration tests
//
// Copyright (c) 2023, 2024 Ali Polatel <alip@chesswob.org>
//
// SPDX-License-Identifier: GPL-3.0-or-later

#![allow(non_snake_case)]
#![allow(clippy::disallowed_methods)]

use std::{
    env,
    fs::{create_dir_all, metadata, remove_file, File},
    io::{BufReader, Read, Write},
    os::{
        fd::{AsRawFd, FromRawFd},
        unix::{fs::PermissionsExt, process::ExitStatusExt},
    },
    path::Path,
    process::{Command, Stdio},
    thread::sleep,
    time::Duration,
};

use nix::{
    sys::stat::{umask, Mode},
    unistd::{close, pipe, unlink, Pid, Uid},
};
use syd::hash::HashAlgorithm;

use crate::{
    assert, assert_eq, assert_status_code, assert_status_code_matches, assert_status_killed,
    assert_status_not_ok, assert_status_not_supported, assert_status_ok,
    assert_status_ok_if_supported, fixup, ignore, util::*,
};

const EX_SIGIOT: i32 = 128 + nix::libc::SIGIOT;
const EX_SIGSEGV: i32 = 128 + nix::libc::SIGSEGV;

/// Represents a test case.
pub type Test<'a> = (&'a str, fn() -> TestResult);

macro_rules! test_entry {
    ($func:expr) => {
        (stringify!($func), $func)
    };
}

/// List of integration tests.
pub const TESTS: &[Test] = &[
    test_entry!(test_syd_version),
    test_entry!(test_syd_export),
    test_entry!(test_syd_true_returns_success),
    test_entry!(test_syd_true_returns_success_with_many_processes),
    test_entry!(test_syd_true_returns_success_with_many_threads),
    test_entry!(test_syd_false_returns_failure),
    test_entry!(test_syd_true_returns_failure_with_many_processes),
    test_entry!(test_syd_true_returns_failure_with_many_threads),
    test_entry!(test_syd_empty_file_returns_enoexec),
    test_entry!(test_syd_non_executable_file_returns_eacces_empty),
    test_entry!(test_syd_non_executable_file_returns_eacces_binary),
    test_entry!(test_syd_non_executable_file_returns_eacces_script),
    test_entry!(test_syd_sigint_returns_130),
    test_entry!(test_syd_sigabrt_returns_134),
    test_entry!(test_syd_sigkill_returns_137),
    test_entry!(test_syd_reap_zombies_bare),
    test_entry!(test_syd_reap_zombies_wrap),
    test_entry!(test_syd_whoami_returns_root_fake),
    test_entry!(test_syd_whoami_returns_root_user),
    test_entry!(test_syd_setuid_nobody_default),
    test_entry!(test_syd_setuid_nobody_safesetid_deny),
    test_entry!(test_syd_setuid_root_safesetid_deny),
    test_entry!(test_syd_setuid_nobody_safesetid_allow),
    test_entry!(test_syd_setgid_nobody_default),
    test_entry!(test_syd_setgid_nobody_safesetid_deny),
    test_entry!(test_syd_setgid_root_safesetid_deny),
    test_entry!(test_syd_setgid_nobody_safesetid_allow),
    test_entry!(test_syd_setreuid_nobody_default_1),
    test_entry!(test_syd_setreuid_nobody_default_2),
    test_entry!(test_syd_setreuid_nobody_default_3),
    test_entry!(test_syd_setreuid_nobody_safesetid_deny_1),
    test_entry!(test_syd_setreuid_nobody_safesetid_deny_2),
    test_entry!(test_syd_setreuid_nobody_safesetid_deny_3),
    test_entry!(test_syd_setreuid_root_safesetid_deny_1),
    test_entry!(test_syd_setreuid_root_safesetid_deny_2),
    test_entry!(test_syd_setreuid_root_safesetid_deny_3),
    test_entry!(test_syd_setreuid_nobody_safesetid_allow_1),
    test_entry!(test_syd_setreuid_nobody_safesetid_allow_2),
    test_entry!(test_syd_setreuid_nobody_safesetid_allow_3),
    test_entry!(test_syd_setregid_nobody_default_1),
    test_entry!(test_syd_setregid_nobody_default_2),
    test_entry!(test_syd_setregid_nobody_default_3),
    test_entry!(test_syd_setregid_nobody_safesetid_deny_1),
    test_entry!(test_syd_setregid_nobody_safesetid_deny_2),
    test_entry!(test_syd_setregid_nobody_safesetid_deny_3),
    test_entry!(test_syd_setregid_root_safesetid_deny_1),
    test_entry!(test_syd_setregid_root_safesetid_deny_2),
    test_entry!(test_syd_setregid_root_safesetid_deny_3),
    test_entry!(test_syd_setregid_nobody_safesetid_allow_1),
    test_entry!(test_syd_setregid_nobody_safesetid_allow_2),
    test_entry!(test_syd_setregid_nobody_safesetid_allow_3),
    test_entry!(test_syd_setresuid_nobody_default_1),
    test_entry!(test_syd_setresuid_nobody_default_2),
    test_entry!(test_syd_setresuid_nobody_default_3),
    test_entry!(test_syd_setresuid_nobody_default_4),
    test_entry!(test_syd_setresuid_nobody_default_5),
    test_entry!(test_syd_setresuid_nobody_default_6),
    test_entry!(test_syd_setresuid_nobody_default_7),
    test_entry!(test_syd_setresuid_nobody_safesetid_deny_1),
    test_entry!(test_syd_setresuid_nobody_safesetid_deny_2),
    test_entry!(test_syd_setresuid_nobody_safesetid_deny_3),
    test_entry!(test_syd_setresuid_nobody_safesetid_deny_4),
    test_entry!(test_syd_setresuid_nobody_safesetid_deny_5),
    test_entry!(test_syd_setresuid_nobody_safesetid_deny_6),
    test_entry!(test_syd_setresuid_nobody_safesetid_deny_7),
    test_entry!(test_syd_setresuid_root_safesetid_deny_1),
    test_entry!(test_syd_setresuid_root_safesetid_deny_2),
    test_entry!(test_syd_setresuid_root_safesetid_deny_3),
    test_entry!(test_syd_setresuid_root_safesetid_deny_4),
    test_entry!(test_syd_setresuid_root_safesetid_deny_5),
    test_entry!(test_syd_setresuid_root_safesetid_deny_6),
    test_entry!(test_syd_setresuid_root_safesetid_deny_7),
    test_entry!(test_syd_setresuid_nobody_safesetid_allow_1),
    test_entry!(test_syd_setresuid_nobody_safesetid_allow_2),
    test_entry!(test_syd_setresuid_nobody_safesetid_allow_3),
    test_entry!(test_syd_setresuid_nobody_safesetid_allow_4),
    test_entry!(test_syd_setresuid_nobody_safesetid_allow_5),
    test_entry!(test_syd_setresuid_nobody_safesetid_allow_6),
    test_entry!(test_syd_setresuid_nobody_safesetid_allow_7),
    test_entry!(test_syd_setresgid_nobody_default_1),
    test_entry!(test_syd_setresgid_nobody_default_2),
    test_entry!(test_syd_setresgid_nobody_default_3),
    test_entry!(test_syd_setresgid_nobody_default_4),
    test_entry!(test_syd_setresgid_nobody_default_5),
    test_entry!(test_syd_setresgid_nobody_default_6),
    test_entry!(test_syd_setresgid_nobody_default_7),
    test_entry!(test_syd_setresgid_nobody_safesetid_deny_1),
    test_entry!(test_syd_setresgid_nobody_safesetid_deny_2),
    test_entry!(test_syd_setresgid_nobody_safesetid_deny_3),
    test_entry!(test_syd_setresgid_nobody_safesetid_deny_4),
    test_entry!(test_syd_setresgid_nobody_safesetid_deny_5),
    test_entry!(test_syd_setresgid_nobody_safesetid_deny_6),
    test_entry!(test_syd_setresgid_nobody_safesetid_deny_7),
    test_entry!(test_syd_setresgid_root_safesetid_deny_1),
    test_entry!(test_syd_setresgid_root_safesetid_deny_2),
    test_entry!(test_syd_setresgid_root_safesetid_deny_3),
    test_entry!(test_syd_setresgid_root_safesetid_deny_4),
    test_entry!(test_syd_setresgid_root_safesetid_deny_5),
    test_entry!(test_syd_setresgid_root_safesetid_deny_6),
    test_entry!(test_syd_setresgid_root_safesetid_deny_7),
    test_entry!(test_syd_setresgid_nobody_safesetid_allow_1),
    test_entry!(test_syd_setresgid_nobody_safesetid_allow_2),
    test_entry!(test_syd_setresgid_nobody_safesetid_allow_3),
    test_entry!(test_syd_setresgid_nobody_safesetid_allow_4),
    test_entry!(test_syd_setresgid_nobody_safesetid_allow_5),
    test_entry!(test_syd_setresgid_nobody_safesetid_allow_6),
    test_entry!(test_syd_setresgid_nobody_safesetid_allow_7),
    test_entry!(test_syd_socket_domain_restrictions),
    test_entry!(test_syd_environment_filter),
    test_entry!(test_syd_environment_harden),
    test_entry!(test_syd_lock),
    test_entry!(test_syd_lock_exec),
    test_entry!(test_syd_lock_prevents_further_cli_args),
    test_entry!(test_syd_lock_prevents_further_cfg_items),
    test_entry!(test_syd_lock_prevents_further_inc_items),
    test_entry!(test_syd_chk),
    test_entry!(test_syd_cmd_exec_with_lock_default),
    test_entry!(test_syd_cmd_exec_with_lock_on),
    test_entry!(test_syd_cmd_exec_with_lock_off_1),
    test_entry!(test_syd_cmd_exec_with_lock_off_2),
    test_entry!(test_syd_cmd_exec_with_lock_exec_1),
    test_entry!(test_syd_cmd_exec_with_lock_exec_2),
    test_entry!(test_syd_parse_config),
    test_entry!(test_syd_include_config),
    test_entry!(test_syd_shellexpand_01),
    test_entry!(test_syd_shellexpand_02),
    test_entry!(test_syd_shellexpand_03),
    test_entry!(test_syd_shellexpand_04),
    test_entry!(test_syd_lock_personality),
    test_entry!(test_syd_mdwe_protections),
    test_entry!(test_syd_mdwe_relax),
    test_entry!(test_syd_mmap_fixed_null),
    test_entry!(test_syd_load_library),
    test_entry!(test_syd_load_library_noexec),
    test_entry!(test_syd_load_library_abort_after_load),
    test_entry!(test_syd_load_library_abort_at_startup),
    test_entry!(test_syd_load_library_check_fd_leaks_bare),
    test_entry!(test_syd_load_library_check_fd_leaks_wrap),
    test_entry!(test_syd_load_library_check_fd_leaks_init_bare),
    test_entry!(test_syd_load_library_check_fd_leaks_init_wrap),
    test_entry!(test_syd_read_sandbox_open_allow),
    test_entry!(test_syd_read_sandbox_open_deny),
    test_entry!(test_syd_stat_sandbox_chdir_allow),
    test_entry!(test_syd_stat_sandbox_chdir_hide),
    test_entry!(test_syd_stat_sandbox_stat_allow),
    test_entry!(test_syd_stat_sandbox_stat_hide),
    test_entry!(test_syd_stat_sandbox_getdents_allow),
    test_entry!(test_syd_stat_sandbox_getdents_hide),
    test_entry!(test_syd_stat_bypass_with_read),
    test_entry!(test_syd_stat_bypass_with_write),
    test_entry!(test_syd_stat_bypass_with_exec),
    test_entry!(test_syd_write_sandbox_open_allow),
    test_entry!(test_syd_write_sandbox_open_deny),
    test_entry!(test_syd_exec_sandbox_open_allow),
    test_entry!(test_syd_exec_sandbox_open_deny),
    test_entry!(test_syd_exec_sandbox_deny_binfmt_script),
    test_entry!(test_syd_exec_sandbox_many_binfmt_script),
    test_entry!(test_syd_exec_sandbox_prevent_library_injection_LD_LIBRARY_PATH),
    test_entry!(test_syd_exec_sandbox_prevent_library_injection_LD_PRELOAD),
    test_entry!(test_syd_network_sandbox_connect_ipv4_allow),
    test_entry!(test_syd_network_sandbox_connect_ipv4_deny),
    test_entry!(test_syd_network_sandbox_connect_ipv6_allow),
    test_entry!(test_syd_network_sandbox_connect_ipv6_deny),
    test_entry!(test_syd_network_sandbox_allow_safe_bind_ipv4_failure),
    test_entry!(test_syd_network_sandbox_allow_safe_bind_ipv4_success),
    test_entry!(test_syd_network_sandbox_allow_safe_bind_ipv6_failure),
    test_entry!(test_syd_network_sandbox_allow_safe_bind_ipv6_success),
    test_entry!(test_syd_handle_toolong_unix_connect),
    test_entry!(test_syd_handle_toolong_unix_sendto),
    test_entry!(test_syd_handle_toolong_unix_sendmsg),
    test_entry!(test_syd_sendmsg_scm_rights_one),
    test_entry!(test_syd_sendmsg_scm_rights_many),
    test_entry!(test_syd_sendmmsg),
    test_entry!(test_syd_mask_simple),
    test_entry!(test_syd_kcapi_hash_block),
    test_entry!(test_syd_kcapi_hash_stream),
    test_entry!(test_syd_kcapi_cipher_block),
    test_entry!(test_syd_exit_wait_default),
    test_entry!(test_syd_exit_wait_default_unsafe_exec),
    test_entry!(test_syd_exit_wait_pid),
    test_entry!(test_syd_exit_wait_pid_unsafe_exec),
    test_entry!(test_syd_exit_wait_pid_with_runaway_cmd_exec_process),
    test_entry!(test_syd_exit_wait_pid_unsafe_exec_with_runaway_cmd_exec_process),
    test_entry!(test_syd_exit_wait_all),
    test_entry!(test_syd_exit_wait_all_unsafe_exec),
    test_entry!(test_syd_exit_wait_all_with_runaway_cmd_exec_process),
    test_entry!(test_syd_exit_wait_all_unsafe_exec_with_runaway_cmd_exec_process),
    test_entry!(test_syd_cli_args_override_user_profile),
    test_entry!(test_syd_ifconfig_lo_bare),
    test_entry!(test_syd_ifconfig_lo_wrap),
    test_entry!(test_syd_binfmt_malformed_too_short),
    test_entry!(test_syd_binfmt_malformed_bad_magic),
    test_entry!(test_syd_binfmt_malformed_invalid_elf_class),
    test_entry!(test_syd_deny_elf32),
    test_entry!(test_syd_deny_elf_dynamic),
    test_entry!(test_syd_deny_elf_static),
    test_entry!(test_syd_deny_script),
    test_entry!(test_syd_force_sandbox),
    test_entry!(test_syd_segvguard_core_safe),
    test_entry!(test_syd_segvguard_core_unsafe),
    test_entry!(test_syd_segvguard_suspension_safe),
    test_entry!(test_syd_segvguard_suspension_unsafe),
    test_entry!(test_syd_symlink_toctou),
    test_entry!(test_syd_symlinkat_toctou),
    test_entry!(test_syd_ptrmod_toctou_chdir),
    test_entry!(test_syd_ptrmod_toctou_exec_fail),
    test_entry!(test_syd_ptrmod_toctou_exec_success_quick),
    test_entry!(test_syd_ptrmod_toctou_exec_success_double_fork),
    test_entry!(test_syd_ptrmod_toctou_exec_success_quick_no_mitigation),
    test_entry!(test_syd_ptrmod_toctou_exec_success_double_fork_no_mitigation),
    test_entry!(test_syd_ptrmod_toctou_open),
    test_entry!(test_syd_ptrmod_toctou_creat),
    test_entry!(test_syd_ptrmod_toctou_opath),
    test_entry!(test_syd_io_uring_escape_strict),
    test_entry!(test_syd_io_uring_escape_unsafe),
    test_entry!(test_syd_opath_escape),
    test_entry!(test_syd_devfd_escape_chdir),
    test_entry!(test_syd_devfd_escape_chdir_relpath_1),
    test_entry!(test_syd_devfd_escape_chdir_relpath_2),
    test_entry!(test_syd_devfd_escape_chdir_relpath_3),
    test_entry!(test_syd_devfd_escape_chdir_relpath_4),
    test_entry!(test_syd_devfd_escape_chdir_relpath_5),
    test_entry!(test_syd_devfd_escape_chdir_relpath_6),
    test_entry!(test_syd_devfd_escape_chdir_relpath_7),
    test_entry!(test_syd_devfd_escape_chdir_relpath_8),
    test_entry!(test_syd_devfd_escape_chdir_relpath_9),
    test_entry!(test_syd_devfd_escape_chdir_relpath_10),
    test_entry!(test_syd_devfd_escape_chdir_relpath_11),
    test_entry!(test_syd_devfd_escape_chdir_relpath_12),
    test_entry!(test_syd_devfd_escape_chdir_relpath_13),
    test_entry!(test_syd_devfd_escape_chdir_relpath_14),
    test_entry!(test_syd_devfd_escape_chdir_relpath_15),
    test_entry!(test_syd_devfd_escape_chdir_relpath_16),
    test_entry!(test_syd_devfd_escape_chdir_relpath_17),
    test_entry!(test_syd_devfd_escape_chdir_relpath_18),
    test_entry!(test_syd_devfd_escape_chdir_relpath_19),
    test_entry!(test_syd_devfd_escape_chdir_relpath_20),
    test_entry!(test_syd_devfd_escape_open),
    test_entry!(test_syd_devfd_escape_open_relpath_1),
    test_entry!(test_syd_devfd_escape_open_relpath_2),
    test_entry!(test_syd_devfd_escape_open_relpath_3),
    test_entry!(test_syd_devfd_escape_open_relpath_4),
    test_entry!(test_syd_devfd_escape_open_relpath_5),
    test_entry!(test_syd_devfd_escape_open_relpath_6),
    test_entry!(test_syd_devfd_escape_open_relpath_7),
    test_entry!(test_syd_devfd_escape_open_relpath_8),
    test_entry!(test_syd_devfd_escape_open_relpath_9),
    test_entry!(test_syd_devfd_escape_open_relpath_10),
    test_entry!(test_syd_devfd_escape_open_relpath_11),
    test_entry!(test_syd_devfd_escape_open_relpath_12),
    test_entry!(test_syd_devfd_escape_open_relpath_13),
    test_entry!(test_syd_devfd_escape_open_relpath_14),
    test_entry!(test_syd_devfd_escape_open_relpath_15),
    test_entry!(test_syd_devfd_escape_open_relpath_16),
    test_entry!(test_syd_devfd_escape_open_relpath_17),
    test_entry!(test_syd_devfd_escape_open_relpath_18),
    test_entry!(test_syd_devfd_escape_open_relpath_19),
    test_entry!(test_syd_devfd_escape_open_relpath_20),
    test_entry!(test_syd_procself_escape_chdir),
    test_entry!(test_syd_procself_escape_chdir_relpath_1),
    test_entry!(test_syd_procself_escape_chdir_relpath_2),
    test_entry!(test_syd_procself_escape_chdir_relpath_3),
    test_entry!(test_syd_procself_escape_chdir_relpath_4),
    test_entry!(test_syd_procself_escape_chdir_relpath_5),
    test_entry!(test_syd_procself_escape_chdir_relpath_6),
    test_entry!(test_syd_procself_escape_chdir_relpath_7),
    test_entry!(test_syd_procself_escape_open),
    test_entry!(test_syd_procself_escape_open_relpath_1),
    test_entry!(test_syd_procself_escape_open_relpath_2),
    test_entry!(test_syd_procself_escape_open_relpath_3),
    test_entry!(test_syd_procself_escape_open_relpath_4),
    test_entry!(test_syd_procself_escape_open_relpath_5),
    test_entry!(test_syd_procself_escape_open_relpath_6),
    test_entry!(test_syd_procself_escape_open_relpath_7),
    test_entry!(test_syd_procself_escape_relpath),
    test_entry!(test_syd_procself_escape_symlink),
    test_entry!(test_syd_procself_escape_symlink_within_container),
    test_entry!(test_syd_umask_bypass_077),
    test_entry!(test_syd_umask_bypass_277),
    test_entry!(test_syd_emulate_opath),
    test_entry!(test_syd_emulate_otmpfile),
    test_entry!(test_syd_honor_umask),
    test_entry!(test_syd_open_utf8_invalid),
    test_entry!(test_syd_exec_in_inaccessible_directory),
    test_entry!(test_syd_fstat_on_pipe),
    test_entry!(test_syd_fstat_on_socket),
    test_entry!(test_syd_fstat_on_deleted_file),
    test_entry!(test_syd_fstat_on_temp_file),
    test_entry!(test_syd_fchmodat_on_proc_fd),
    test_entry!(test_syd_linkat_on_fd),
    test_entry!(test_syd_block_ioctl_tiocsti),
    test_entry!(test_syd_block_prctl_ptrace),
    test_entry!(test_syd_prevent_ptrace_detect),
    test_entry!(test_syd_block_dev_random),
    test_entry!(test_syd_block_dev_urandom),
    test_entry!(test_syd_kill_during_syscall),
    test_entry!(test_syd_open_toolong_path),
    test_entry!(test_syd_open_null_path),
    test_entry!(test_syd_utimensat_null),
    test_entry!(test_syd_normalize_path),
    test_entry!(test_syd_path_resolution),
    test_entry!(test_syd_remove_empty_path),
    test_entry!(test_syd_symlink_readonly_path),
    test_entry!(test_syd_open_trailing_slash),
    test_entry!(test_syd_openat_trailing_slash),
    test_entry!(test_syd_lstat_trailing_slash),
    test_entry!(test_syd_fstatat_trailing_slash),
    test_entry!(test_syd_mkdir_trailing_dot),
    test_entry!(test_syd_mkdirat_trailing_dot),
    test_entry!(test_syd_rmdir_trailing_slashdot),
    test_entry!(test_syd_mkdir_eexist_escape),
    test_entry!(test_syd_mkdirat_eexist_escape),
    test_entry!(test_syd_mknod_eexist_escape),
    test_entry!(test_syd_mknodat_eexist_escape),
    test_entry!(test_syd_fopen_supports_mode_e),
    test_entry!(test_syd_fopen_supports_mode_x),
    test_entry!(test_syd_link_no_symlink_deref),
    test_entry!(test_syd_link_posix),
    test_entry!(test_syd_linkat_posix),
    test_entry!(test_syd_getcwd_long),
    test_entry!(test_syd_creat_thru_dangling),
    test_entry!(test_syd_mkdirat_non_dir_fd),
    test_entry!(test_syd_blocking_udp4),
    test_entry!(test_syd_blocking_udp6),
    test_entry!(test_syd_close_on_exec),
    test_entry!(test_syd_open_exclusive_restart),
    test_entry!(test_syd_open_exclusive_repeat),
    test_entry!(test_syd_pty_io_rust),
    test_entry!(test_syd_pty_io_gawk),
    test_entry!(test_syd_diff_dev_fd),
    test_entry!(test_syd_fifo_multiple_readers),
    test_entry!(test_syd_bind_unix_socket),
    test_entry!(test_syd_signal_protection),
    test_entry!(test_syd_emulate_open_fifo_1),
    test_entry!(test_syd_emulate_open_fifo_2),
    test_entry!(test_syd_deny_magiclinks),
    test_entry!(test_syd_lstat_magiclinks),
    test_entry!(test_syd_access_unsafe_paths),
    test_entry!(test_syd_access_proc_cmdline),
    test_entry!(test_syd_mkdir_with_control_chars_default),
    test_entry!(test_syd_mkdir_with_control_chars_unsafe),
    test_entry!(test_syd_touch_with_control_chars_default),
    test_entry!(test_syd_touch_with_control_chars_unsafe),
    test_entry!(test_syd_unshare_user_bypass_limit),
    test_entry!(test_syd_interrupt_mkdir),
    test_entry!(test_syd_interrupt_bind_ipv4),
    test_entry!(test_syd_interrupt_bind_unix),
    test_entry!(test_syd_interrupt_connect_ipv4),
    //FIXME: This test should be done better.
    //test_entry!(test_syd_repetitive_clone),
    test_entry!(test_syd_pid_fork_deny),
    test_entry!(test_syd_pid_thread_deny),
    test_entry!(test_syd_pid_fork_bomb),
    test_entry!(test_syd_pid_fork_bomb_asm),
    test_entry!(test_syd_pid_thread_bomb),
    test_entry!(test_syd_mem_alloc),
    test_entry!(test_syd_exp_mem_stress_ng_malloc_1),
    test_entry!(test_syd_exp_mem_stress_ng_malloc_2),
    test_entry!(test_syd_exp_mem_stress_ng_mmap),
    test_entry!(test_syd_exp_pid_stress_ng_kill),
    test_entry!(test_syd_exp_pid_stress_ng_allow),
    test_entry!(test_syd_exp_pid_stress_ng_fork),
    test_entry!(test_syd_exp_syscall_fuzz_bare),
    test_entry!(test_syd_exp_syscall_fuzz_wrap),
];

// Tests if syd -V and --version works.
fn test_syd_version() -> TestResult {
    let status = syd().arg("-V").status().expect("execute syd");
    assert_status_ok!(status);

    let status = syd()
        .stdout(Stdio::null())
        .stderr(Stdio::null())
        .arg("--version")
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

// Tests if syd -E works.
fn test_syd_export() -> TestResult {
    if !is_program_available("true") {
        eprintln!("Test requires true, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }

    let status = syd()
        .arg("-Ebpf")
        .stderr(Stdio::null())
        .stdout(Stdio::null())
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    let status = syd()
        .arg("-Epfc")
        .stderr(Stdio::null())
        .stdout(Stdio::null())
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    let status = syd()
        .arg("-EBPF")
        .stderr(Stdio::null())
        .stdout(Stdio::null())
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    let status = syd()
        .arg("-EPFC")
        .stderr(Stdio::null())
        .stdout(Stdio::null())
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    let status = syd()
        .arg("-EbPf")
        .stderr(Stdio::null())
        .stdout(Stdio::null())
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    let status = syd()
        .arg("-EPfc")
        .stderr(Stdio::null())
        .stdout(Stdio::null())
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    let status = syd()
        .arg("-Eb")
        .stderr(Stdio::null())
        .stdout(Stdio::null())
        .status()
        .expect("execute syd");
    assert_status_not_ok!(status);

    let status = syd()
        .arg("-Ef")
        .stderr(Stdio::null())
        .stdout(Stdio::null())
        .status()
        .expect("execute syd");
    assert_status_not_ok!(status);

    Ok(())
}

// Tests if `true` returns success under sandbox.
fn test_syd_true_returns_success() -> TestResult {
    let status = syd()
        .env("SYD_DO", "exit")
        .arg("-mallow/exec,read,stat+/***")
        .args(["--", &SYD_DO, "0"])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

// Tests if `syd` returns success for a sandbox running many processes,
// in case the execve child returns success.
fn test_syd_true_returns_success_with_many_processes() -> TestResult {
    let status = syd()
        .env("SYD_DO", "fork")
        .arg("-mallow/exec,read,stat+/***")
        .args(["--", &SYD_DO, "0", "8"])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

// Tests if `syd` returns success for a sandbox running many threads,
// in case the execve child returns success.
fn test_syd_true_returns_success_with_many_threads() -> TestResult {
    let status = syd()
        .env("SYD_DO", "thread")
        .arg("-mallow/exec,read,stat+/***")
        .args(["--", &SYD_DO, "0", "8"])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

// Tests if `false` returns failure under sandbox.
fn test_syd_false_returns_failure() -> TestResult {
    let status = syd()
        .arg("-mallow/exec,read,stat+/***")
        .arg("--")
        .arg("false")
        .status()
        .expect("execute syd");
    assert_status_not_ok!(status);

    let status = syd()
        .args(["-plib", "--", "false"])
        .status()
        .expect("execute syd");
    assert_status_not_ok!(status);

    Ok(())
}

// Tests if `syd` returns failure for a sandbox running many processes,
// in case the execve child returns failure.
fn test_syd_true_returns_failure_with_many_processes() -> TestResult {
    let status = syd()
        .env("SYD_DO", "fork")
        .arg("-mallow/exec,read,stat+/***")
        .args(["--", &SYD_DO, "7", "8"])
        .status()
        .expect("execute syd");
    assert_status_code!(status, 7);
    Ok(())
}

// Tests if `syd` returns failure for a sandbox running many threads,
// in case the execve child returns failure.
fn test_syd_true_returns_failure_with_many_threads() -> TestResult {
    let status = syd()
        .env("SYD_DO", "thread")
        .arg("-mallow/exec,read,stat+/***")
        .args(["--", &SYD_DO, "7", "8"])
        .status()
        .expect("execute syd");
    assert_status_code!(status, 7);
    Ok(())
}

fn test_syd_empty_file_returns_enoexec() -> TestResult {
    // Step 1: Create a file that's empty called "empty"
    let path = Path::new("empty");
    let file = File::create(path)?;

    // Set permissions to executable
    let mut perms = file.metadata()?.permissions();
    perms.set_mode(0o755); // -rwxr-xr-x
    file.set_permissions(perms)?;
    drop(file); // close the file.

    let status = syd()
        .env("SYD_DO", "exec")
        .arg("-plib")
        .args([
            "-msandbox/exec:on",
            "-mallow/exec+/***",
            "--",
            &SYD_DO,
            "./empty",
        ])
        .status()
        .expect("execute syd");
    assert_status_code!(status, nix::libc::ENOEXEC);

    Ok(())
}

fn test_syd_non_executable_file_returns_eacces_empty() -> TestResult {
    // Create a file that's non-executable called "non-executable"
    let path = Path::new("non-executable");
    let file = File::create(path)?;

    // Set permissions to non-executable
    let mut perms = file.metadata()?.permissions();
    perms.set_mode(0o644); // -rw-r--r--
    file.set_permissions(perms)?;
    drop(file); // close the file.

    let status = syd()
        .env("SYD_DO", "exec")
        .arg("-plib")
        .args([
            "-msandbox/exec:on",
            "-mallow/exec+/***",
            "--",
            &SYD_DO,
            "./non-executable",
        ])
        .status()
        .expect("execute syd");
    // empty & non-executable file must return EACCES not ENOEXEC!
    assert_status_code!(status, nix::libc::EACCES);

    Ok(())
}

fn test_syd_non_executable_file_returns_eacces_binary() -> TestResult {
    // Create a file that's non-executable called "non-executable"
    let path = Path::new("non-executable");
    let mut file = File::create(path)?;

    writeln!(
        file,
        "Heavy is the root of light. Still is the master of moving."
    )?;

    // Set permissions to non-executable
    let mut perms = file.metadata()?.permissions();
    perms.set_mode(0o644); // -rw-r--r--
    file.set_permissions(perms)?;
    drop(file); // close the file.

    let status = syd()
        .env("SYD_DO", "exec")
        .arg("-plib")
        .args([
            "-msandbox/exec:on",
            "-mallow/exec+/***",
            "--",
            &SYD_DO,
            "./non-executable",
        ])
        .status()
        .expect("execute syd");
    assert_status_code!(status, nix::libc::EACCES);

    Ok(())
}

fn test_syd_non_executable_file_returns_eacces_script() -> TestResult {
    // Create a file that's non-executable called "non-executable"
    let path = Path::new("non-executable");
    let mut file = File::create(path)?;

    writeln!(file, "#!/bin/sh\nexit 42")?;

    // Set permissions to non-executable
    let mut perms = file.metadata()?.permissions();
    perms.set_mode(0o644); // -rw-r--r--
    file.set_permissions(perms)?;
    drop(file); // close the file.

    let status = syd()
        .env("SYD_DO", "exec")
        .arg("-plib")
        .args([
            "-msandbox/exec:on",
            "-mallow/exec+/***",
            "--",
            &SYD_DO,
            "./non-executable",
        ])
        .status()
        .expect("execute syd");
    assert_status_code!(status, nix::libc::EACCES);

    Ok(())
}

fn test_syd_sigint_returns_130() -> TestResult {
    if !is_program_available("kill") {
        eprintln!("Test requires kill, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }
    let status = syd()
        .arg("-ppaludis")
        .args(["sh", "-c"])
        .arg(r#"exec kill -INT $$"#)
        .status()
        .expect("execute syd");
    assert_status_code!(status, 130);
    Ok(())
}

fn test_syd_sigabrt_returns_134() -> TestResult {
    if !is_program_available("kill") {
        eprintln!("Test requires kill, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }
    let status = syd()
        .arg("-ppaludis")
        .args(["sh", "-c"])
        .arg(r#"exec kill -ABRT $$"#)
        .status()
        .expect("execute syd");
    assert_status_code!(status, 134);
    Ok(())
}

fn test_syd_sigkill_returns_137() -> TestResult {
    if !is_program_available("kill") {
        eprintln!("Test requires kill, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }
    let status = syd()
        .arg("-ppaludis")
        .args(["sh", "-c"])
        .arg(r#"exec kill -KILL $$"#)
        .status()
        .expect("execute syd");
    assert_status_code!(status, 137);
    Ok(())
}

fn test_syd_reap_zombies_bare() -> TestResult {
    if !is_program_available("bash") {
        eprintln!("Test requires bash, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }
    let status = syd()
        .arg("-ppaludis")
        .args(["bash", "-c"])
        .arg(
            r#"
set -e
for i in {1..10}; do
    ( sleep $i ) &
done
echo >&2 "Spawned 10 processes in the background."
echo >&2 "Disowning and exiting..."
disown
exit 42
        "#,
        )
        .status()
        .expect("execute syd");
    assert_status_code!(status, 42);

    Ok(())
}

fn test_syd_reap_zombies_wrap() -> TestResult {
    if !is_program_available("bash") {
        eprintln!("Test requires bash, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }
    match check_unshare() {
        Some(false) => {
            eprintln!("Test requires Linux namespaces, skipping!");
            env::set_var("SYD_TEST_SOFT_FAIL", "1");
            return Ok(());
        }
        None => {
            return Err(TestError(
                "Failed to test for Linux namespaces!".to_string(),
            ));
        }
        _ => {}
    };
    let status = syd()
        .args(["-ppaludis", "-pcontainer"])
        .args(["bash", "-c"])
        .arg(
            r#"
set -e
for i in {1..10}; do
    ( sleep $i ) &
done
echo >&2 "Spawned 10 processes in the background."
echo >&2 "Disowning and exiting..."
disown
exit 42
        "#,
        )
        .status()
        .expect("execute syd");
    assert_status_code!(status, 42);

    Ok(())
}

// Tests if `whoami` returns `root` with `root/fake:1`
fn test_syd_whoami_returns_root_fake() -> TestResult {
    let status = syd()
        .env("SYD_DO", "getuid")
        .arg("-plib")
        .args(["-mroot/fake:1", "--", &SYD_DO, "0"])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

// Tests if `whoami` returns `root` with `root/map:1`
fn test_syd_whoami_returns_root_user() -> TestResult {
    match check_unshare() {
        Some(false) => {
            eprintln!("Test requires Linux namespaces, skipping!");
            env::set_var("SYD_TEST_SOFT_FAIL", "1");
            return Ok(());
        }
        None => {
            return Err(TestError(
                "Failed to test for Linux namespaces!".to_string(),
            ));
        }
        _ => {}
    };

    let status = syd()
        .env("SYD_DO", "getuid")
        .arg("-plib")
        .args(["-mroot/map:1", "--", &SYD_DO, "0"])
        .stderr(Stdio::inherit())
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_setuid_nobody_default() -> TestResult {
    if !*CAP_SETUID {
        eprintln!("Test requires CAP_SETUID, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }

    // Default filter is a no-op for setuid.
    // Test must abort on uid mismatch after successful call.
    let status = syd()
        .env("SYD_DO", "setuid")
        .arg("-plib")
        .args(["--", &SYD_DO, "65534"])
        .status()
        .expect("execute syd");
    assert_status_code!(status, 128 + nix::libc::SIGABRT);

    Ok(())
}

fn test_syd_setuid_nobody_safesetid_deny() -> TestResult {
    if !*CAP_SETUID {
        eprintln!("Test requires CAP_SETUID, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }

    // SAFETY: Set the comment to test setuid_root_safesetid_deny.
    let (fd_rd, fd_rw) = match pipe() {
        Ok((fd_rd, fd_rw)) => (fd_rd, fd_rw),
        Err(errno) => return Err(TestError(format!("error creating pipe: {errno}!"))),
    };

    // SafeSetID is enabled, but no UID transition defined.
    // The syscall must fail with EPERM _and_ generate an acccess violation.
    let mut child = syd()
        .env("SYD_LOG", "warn")
        .env("SYD_LOG_FD", fd_rw.to_string())
        .env("SYD_DO", "setuid")
        .arg("-plib")
        .args(["-mtrace/allow_safe_setid:1", "--", &SYD_DO, "65534"])
        .spawn()
        .expect("spawn syd");
    let _ = close(fd_rw);

    let status = child.wait().expect("wait for syd");
    assert_status_code!(status, nix::libc::EPERM);

    // Convert raw file descriptor to File, then to BufReader
    let file = unsafe { File::from_raw_fd(fd_rd) };
    let mut reader = BufReader::new(file);

    let mut buffer = String::new();
    match reader.read_to_string(&mut buffer) {
        Ok(0) => {
            eprintln!("No data read from pipe; expected access violation.");
            return Err(TestError(
                "Expected access violation not logged.".to_string(),
            ));
        }
        Ok(_) => {
            // If any data was read, log it
            eprint!("Access violation logged:\n{buffer}");
        }
        Err(e) => {
            // Error reading from the buffer
            eprintln!("Error reading from pipe: {e}");
            return Err(TestError(format!("Error reading from pipe: {e}")));
        }
    }

    Ok(())
}

fn test_syd_setuid_root_safesetid_deny() -> TestResult {
    if !*CAP_SETUID {
        eprintln!("Test requires CAP_SETUID, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }

    // SafeSetID is enabled, but no UID transition defined.
    // The syscall must fail with EPERM even with UID=0.

    // SAFETY: Compared to the test setuid_nobody_safesetid_deny, this
    // test _must not_ generate a syd access violation, because UID
    // change to root must be blocked by the kernel-level parent seccomp
    // filter. This is crucial security-wise so we test it here by
    // overriding SYD_LOG and setting SYD_LOG_FD.
    let (fd_rd, fd_rw) = match pipe() {
        Ok((fd_rd, fd_rw)) => (fd_rd, fd_rw),
        Err(errno) => return Err(TestError(format!("error creating pipe: {errno}!"))),
    };

    let mut child = syd()
        .env("SYD_LOG", "warn")
        .env("SYD_LOG_FD", fd_rw.to_string())
        .env("SYD_DO", "setuid")
        .arg("-plib")
        .args(["-mtrace/allow_safe_setid:1", "--", &SYD_DO, "0"])
        .spawn()
        .expect("spawn syd");
    let _ = close(fd_rw);

    let status = child.wait().expect("wait for syd");
    assert_status_code!(status, nix::libc::EPERM);

    // Convert raw file descriptor to File, then to BufReader
    let file = unsafe { File::from_raw_fd(fd_rd) };
    let mut reader = BufReader::new(file);

    let mut buffer = String::new();
    match reader.read_to_string(&mut buffer) {
        Ok(0) => {
            eprintln!("No data read from pipe; no access violation raised as expected.");
        }
        Ok(_) => {
            // If any data was read, log it
            eprint!("Access violation logged:\n{buffer}");
            return Err(TestError("Unexpected access violation logged.".to_string()));
        }
        Err(e) => {
            // Error reading from the buffer
            eprintln!("Error reading from pipe: {e}");
            return Err(TestError(format!("Error reading from pipe: {e}")));
        }
    }

    Ok(())
}

fn test_syd_setuid_nobody_safesetid_allow() -> TestResult {
    if !*CAP_SETUID {
        eprintln!("Test requires CAP_SETUID, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }

    // SafeSetID is enabled with UID transition defined.
    // The syscall must succeed.
    let uid = Uid::current();
    let status = syd()
        .env("SYD_DO", "setuid")
        .arg("-plib")
        .args([
            "-mtrace/allow_safe_setid:1",
            &format!("-msetuid+{uid}:nobody"),
            "--",
            &SYD_DO,
            "65534",
        ])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_setgid_nobody_default() -> TestResult {
    if !*CAP_SETGID {
        eprintln!("Test requires CAP_SETGID, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }

    // Default filter is a no-op for setuid.
    // Test must abort on gid mismatch after successful call.
    let status = syd()
        .env("SYD_DO", "setgid")
        .arg("-plib")
        .args(["--", &SYD_DO, "65534"])
        .status()
        .expect("execute syd");
    assert_status_code!(status, 128 + nix::libc::SIGABRT);

    Ok(())
}

fn test_syd_setgid_nobody_safesetid_deny() -> TestResult {
    if !*CAP_SETGID {
        eprintln!("Test requires CAP_SETGID, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }

    // SAFETY: Set the comment to test setgid_root_safesetid_deny.
    let (fd_rd, fd_rw) = match pipe() {
        Ok((fd_rd, fd_rw)) => (fd_rd, fd_rw),
        Err(errno) => return Err(TestError(format!("error creating pipe: {errno}!"))),
    };

    // SafeSetID is enabled, but no GID transition defined.
    // The syscall must fail with EPERM _and_ generate an acccess violation.
    let mut child = syd()
        .env("SYD_LOG", "warn")
        .env("SYD_LOG_FD", fd_rw.to_string())
        .env("SYD_DO", "setgid")
        .arg("-plib")
        .args(["-mtrace/allow_safe_setid:1", "--", &SYD_DO, "65534"])
        .spawn()
        .expect("spawn syd");
    let _ = close(fd_rw);

    let status = child.wait().expect("wait for syd");
    assert_status_code!(status, nix::libc::EPERM);

    // Convert raw file descriptor to File, then to BufReader
    let file = unsafe { File::from_raw_fd(fd_rd) };
    let mut reader = BufReader::new(file);

    let mut buffer = String::new();
    match reader.read_to_string(&mut buffer) {
        Ok(0) => {
            eprintln!("No data read from pipe; expected access violation.");
            return Err(TestError(
                "Expected access violation not logged.".to_string(),
            ));
        }
        Ok(_) => {
            // If any data was read, log it
            eprint!("Access violation logged:\n{buffer}");
        }
        Err(e) => {
            // Error reading from the buffer
            eprintln!("Error reading from pipe: {e}");
            return Err(TestError(format!("Error reading from pipe: {e}")));
        }
    }

    Ok(())
}

fn test_syd_setgid_root_safesetid_deny() -> TestResult {
    if !*CAP_SETGID {
        eprintln!("Test requires CAP_SETUID, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }

    // SafeSetID is enabled, but no UID transition defined.
    // The syscall must fail with EPERM even with UID=0.

    // SAFETY: Compared to the test setgid_nobody_safesetid_deny, this
    // test _must not_ generate a syd access violation, because GID
    // change to root must be blocked by the kernel-level parent seccomp
    // filter. This is crucial security-wise so we test it here by
    // overriding SYD_LOG and setting SYD_LOG_FD.
    let (fd_rd, fd_rw) = match pipe() {
        Ok((fd_rd, fd_rw)) => (fd_rd, fd_rw),
        Err(errno) => return Err(TestError(format!("error creating pipe: {errno}!"))),
    };

    let mut child = syd()
        .env("SYD_LOG", "warn")
        .env("SYD_LOG_FD", fd_rw.to_string())
        .env("SYD_DO", "setgid")
        .arg("-plib")
        .args(["-mtrace/allow_safe_setid:1", "--", &SYD_DO, "0"])
        .spawn()
        .expect("spawn syd");
    let _ = close(fd_rw);

    let status = child.wait().expect("wait for syd");
    assert_status_code!(status, nix::libc::EPERM);

    // Convert raw file descriptor to File, then to BufReader
    let file = unsafe { File::from_raw_fd(fd_rd) };
    let mut reader = BufReader::new(file);

    let mut buffer = String::new();
    match reader.read_to_string(&mut buffer) {
        Ok(0) => {
            eprintln!("No data read from pipe; no access violation raised as expected.");
        }
        Ok(_) => {
            // If any data was read, log it
            eprint!("Access violation logged:\n{buffer}");
            return Err(TestError("Unexpected access violation logged.".to_string()));
        }
        Err(e) => {
            // Error reading from the buffer
            eprintln!("Error reading from pipe: {e}");
            return Err(TestError(format!("Error reading from pipe: {e}")));
        }
    }

    Ok(())
}

fn test_syd_setgid_nobody_safesetid_allow() -> TestResult {
    if !*CAP_SETGID {
        eprintln!("Test requires CAP_SETUID, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }

    // SafeSetID is enabled with GID transition defined.
    // The syscall must succeed.
    let uid = Uid::current();
    let status = syd()
        .env("SYD_DO", "setgid")
        .arg("-plib")
        .args([
            "-mtrace/allow_safe_setid:1",
            &format!("-msetgid+{uid}:nobody"),
            "--",
            &SYD_DO,
            "65534",
        ])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_setreuid_nobody_default_1() -> TestResult {
    if !*CAP_SETUID {
        eprintln!("Test requires CAP_SETUID, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }

    // Default filter is a no-op for setuid.
    // Test must abort on uid mismatch after successful call.
    let status = syd()
        .env("SYD_DO", "setreuid")
        .arg("-plib")
        .args(["--", &SYD_DO, "-1", "65534"])
        .status()
        .expect("execute syd");
    assert_status_code!(status, 128 + nix::libc::SIGABRT);

    Ok(())
}

fn test_syd_setreuid_nobody_default_2() -> TestResult {
    if !*CAP_SETUID {
        eprintln!("Test requires CAP_SETUID, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }

    // Default filter is a no-op for setuid.
    // Test must abort on uid mismatch after successful call.
    let status = syd()
        .env("SYD_DO", "setreuid")
        .arg("-plib")
        .args(["--", &SYD_DO, "65534", "-1"])
        .status()
        .expect("execute syd");
    assert_status_code!(status, 128 + nix::libc::SIGABRT);

    Ok(())
}

fn test_syd_setreuid_nobody_default_3() -> TestResult {
    if !*CAP_SETUID {
        eprintln!("Test requires CAP_SETUID, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }

    // Default filter is a no-op for setuid.
    // Test must abort on uid mismatch after successful call.
    let status = syd()
        .env("SYD_DO", "setreuid")
        .arg("-plib")
        .args(["--", &SYD_DO, "65534", "65534"])
        .status()
        .expect("execute syd");
    assert_status_code!(status, 128 + nix::libc::SIGABRT);

    Ok(())
}

fn test_syd_setreuid_nobody_safesetid_deny_1() -> TestResult {
    if !*CAP_SETUID {
        eprintln!("Test requires CAP_SETUID, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }

    // SAFETY: Set the comment to test setreuid_root_safesetid_deny.
    let (fd_rd, fd_rw) = match pipe() {
        Ok((fd_rd, fd_rw)) => (fd_rd, fd_rw),
        Err(errno) => return Err(TestError(format!("error creating pipe: {errno}!"))),
    };

    // SafeSetID is enabled, but no UID transition defined.
    // The syscall must fail with EPERM _and_ generate an acccess violation.
    let mut child = syd()
        .env("SYD_LOG", "warn")
        .env("SYD_LOG_FD", fd_rw.to_string())
        .env("SYD_DO", "setreuid")
        .arg("-plib")
        .args(["-mtrace/allow_safe_setid:1", "--", &SYD_DO, "-1", "65534"])
        .spawn()
        .expect("spawn syd");
    let _ = close(fd_rw);

    let status = child.wait().expect("wait for syd");
    assert_status_code!(status, nix::libc::EPERM);

    // Convert raw file descriptor to File, then to BufReader
    let file = unsafe { File::from_raw_fd(fd_rd) };
    let mut reader = BufReader::new(file);

    let mut buffer = String::new();
    match reader.read_to_string(&mut buffer) {
        Ok(0) => {
            eprintln!("No data read from pipe; expected access violation.");
            return Err(TestError(
                "Expected access violation not logged.".to_string(),
            ));
        }
        Ok(_) => {
            // If any data was read, log it
            eprint!("Access violation logged:\n{buffer}");
        }
        Err(e) => {
            // Error reading from the buffer
            eprintln!("Error reading from pipe: {e}");
            return Err(TestError(format!("Error reading from pipe: {e}")));
        }
    }

    Ok(())
}

fn test_syd_setreuid_nobody_safesetid_deny_2() -> TestResult {
    if !*CAP_SETUID {
        eprintln!("Test requires CAP_SETUID, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }

    // SAFETY: Set the comment to test setreuid_root_safesetid_deny.
    let (fd_rd, fd_rw) = match pipe() {
        Ok((fd_rd, fd_rw)) => (fd_rd, fd_rw),
        Err(errno) => return Err(TestError(format!("error creating pipe: {errno}!"))),
    };

    // SafeSetID is enabled, but no UID transition defined.
    // The syscall must fail with EPERM _and_ generate an acccess violation.
    let mut child = syd()
        .env("SYD_LOG", "warn")
        .env("SYD_LOG_FD", fd_rw.to_string())
        .env("SYD_DO", "setreuid")
        .arg("-plib")
        .args(["-mtrace/allow_safe_setid:1", "--", &SYD_DO, "65534", "-1"])
        .spawn()
        .expect("spawn syd");
    let _ = close(fd_rw);

    let status = child.wait().expect("wait for syd");
    assert_status_code!(status, nix::libc::EPERM);

    // Convert raw file descriptor to File, then to BufReader
    let file = unsafe { File::from_raw_fd(fd_rd) };
    let mut reader = BufReader::new(file);

    let mut buffer = String::new();
    match reader.read_to_string(&mut buffer) {
        Ok(0) => {
            eprintln!("No data read from pipe; expected access violation.");
            return Err(TestError(
                "Expected access violation not logged.".to_string(),
            ));
        }
        Ok(_) => {
            // If any data was read, log it
            eprint!("Access violation logged:\n{buffer}");
        }
        Err(e) => {
            // Error reading from the buffer
            eprintln!("Error reading from pipe: {e}");
            return Err(TestError(format!("Error reading from pipe: {e}")));
        }
    }

    Ok(())
}

fn test_syd_setreuid_nobody_safesetid_deny_3() -> TestResult {
    if !*CAP_SETUID {
        eprintln!("Test requires CAP_SETUID, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }

    // SAFETY: Set the comment to test setreuid_root_safesetid_deny.
    let (fd_rd, fd_rw) = match pipe() {
        Ok((fd_rd, fd_rw)) => (fd_rd, fd_rw),
        Err(errno) => return Err(TestError(format!("error creating pipe: {errno}!"))),
    };

    // SafeSetID is enabled, but no UID transition defined.
    // The syscall must fail with EPERM _and_ generate an acccess violation.
    let mut child = syd()
        .env("SYD_LOG", "warn")
        .env("SYD_LOG_FD", fd_rw.to_string())
        .env("SYD_DO", "setreuid")
        .arg("-plib")
        .args([
            "-mtrace/allow_safe_setid:1",
            "--",
            &SYD_DO,
            "65534",
            "65534",
        ])
        .spawn()
        .expect("spawn syd");
    let _ = close(fd_rw);

    let status = child.wait().expect("wait for syd");
    assert_status_code!(status, nix::libc::EPERM);

    // Convert raw file descriptor to File, then to BufReader
    let file = unsafe { File::from_raw_fd(fd_rd) };
    let mut reader = BufReader::new(file);

    let mut buffer = String::new();
    match reader.read_to_string(&mut buffer) {
        Ok(0) => {
            eprintln!("No data read from pipe; expected access violation.");
            return Err(TestError(
                "Expected access violation not logged.".to_string(),
            ));
        }
        Ok(_) => {
            // If any data was read, log it
            eprint!("Access violation logged:\n{buffer}");
        }
        Err(e) => {
            // Error reading from the buffer
            eprintln!("Error reading from pipe: {e}");
            return Err(TestError(format!("Error reading from pipe: {e}")));
        }
    }

    Ok(())
}

fn test_syd_setreuid_root_safesetid_deny_1() -> TestResult {
    if !*CAP_SETUID {
        eprintln!("Test requires CAP_SETUID, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }

    // SafeSetID is enabled, but no UID transition defined.
    // The syscall must fail with EPERM even with UID=0.

    // SAFETY: Compared to the test setreuid_nobody_safesetid_deny, this
    // test _must not_ generate a syd access violation, because UID
    // change to root must be blocked by the kernel-level parent seccomp
    // filter. This is crucial security-wise so we test it here by
    // overriding SYD_LOG and setting SYD_LOG_FD.
    let (fd_rd, fd_rw) = match pipe() {
        Ok((fd_rd, fd_rw)) => (fd_rd, fd_rw),
        Err(errno) => return Err(TestError(format!("error creating pipe: {errno}!"))),
    };

    let mut child = syd()
        .env("SYD_LOG", "warn")
        .env("SYD_LOG_FD", fd_rw.to_string())
        .env("SYD_DO", "setreuid")
        .arg("-plib")
        .args(["-mtrace/allow_safe_setid:1", "--", &SYD_DO, "-1", "0"])
        .spawn()
        .expect("spawn syd");
    let _ = close(fd_rw);

    let status = child.wait().expect("wait for syd");
    assert_status_code!(status, nix::libc::EPERM);

    // Convert raw file descriptor to File, then to BufReader
    let file = unsafe { File::from_raw_fd(fd_rd) };
    let mut reader = BufReader::new(file);

    let mut buffer = String::new();
    match reader.read_to_string(&mut buffer) {
        Ok(0) => {
            eprintln!("No data read from pipe; no access violation raised as expected.");
        }
        Ok(_) => {
            // If any data was read, log it
            eprint!("Access violation logged:\n{buffer}");
            return Err(TestError("Unexpected access violation logged.".to_string()));
        }
        Err(e) => {
            // Error reading from the buffer
            eprintln!("Error reading from pipe: {e}");
            return Err(TestError(format!("Error reading from pipe: {e}")));
        }
    }

    Ok(())
}

fn test_syd_setreuid_root_safesetid_deny_2() -> TestResult {
    if !*CAP_SETUID {
        eprintln!("Test requires CAP_SETUID, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }

    // SafeSetID is enabled, but no UID transition defined.
    // The syscall must fail with EPERM even with UID=0.

    // SAFETY: Compared to the test setreuid_nobody_safesetid_deny, this
    // test _must not_ generate a syd access violation, because UID
    // change to root must be blocked by the kernel-level parent seccomp
    // filter. This is crucial security-wise so we test it here by
    // overriding SYD_LOG and setting SYD_LOG_FD.
    let (fd_rd, fd_rw) = match pipe() {
        Ok((fd_rd, fd_rw)) => (fd_rd, fd_rw),
        Err(errno) => return Err(TestError(format!("error creating pipe: {errno}!"))),
    };

    let mut child = syd()
        .env("SYD_LOG", "warn")
        .env("SYD_LOG_FD", fd_rw.to_string())
        .env("SYD_DO", "setreuid")
        .arg("-plib")
        .args(["-mtrace/allow_safe_setid:1", "--", &SYD_DO, "0", "-1"])
        .spawn()
        .expect("spawn syd");
    let _ = close(fd_rw);

    let status = child.wait().expect("wait for syd");
    assert_status_code!(status, nix::libc::EPERM);

    // Convert raw file descriptor to File, then to BufReader
    let file = unsafe { File::from_raw_fd(fd_rd) };
    let mut reader = BufReader::new(file);

    let mut buffer = String::new();
    match reader.read_to_string(&mut buffer) {
        Ok(0) => {
            eprintln!("No data read from pipe; no access violation raised as expected.");
        }
        Ok(_) => {
            // If any data was read, log it
            eprint!("Access violation logged:\n{buffer}");
            return Err(TestError("Unexpected access violation logged.".to_string()));
        }
        Err(e) => {
            // Error reading from the buffer
            eprintln!("Error reading from pipe: {e}");
            return Err(TestError(format!("Error reading from pipe: {e}")));
        }
    }

    Ok(())
}

fn test_syd_setreuid_root_safesetid_deny_3() -> TestResult {
    if !*CAP_SETUID {
        eprintln!("Test requires CAP_SETUID, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }

    // SafeSetID is enabled, but no UID transition defined.
    // The syscall must fail with EPERM even with UID=0.

    // SAFETY: Compared to the test setreuid_nobody_safesetid_deny, this
    // test _must not_ generate a syd access violation, because UID
    // change to root must be blocked by the kernel-level parent seccomp
    // filter. This is crucial security-wise so we test it here by
    // overriding SYD_LOG and setting SYD_LOG_FD.
    let (fd_rd, fd_rw) = match pipe() {
        Ok((fd_rd, fd_rw)) => (fd_rd, fd_rw),
        Err(errno) => return Err(TestError(format!("error creating pipe: {errno}!"))),
    };

    let mut child = syd()
        .env("SYD_LOG", "warn")
        .env("SYD_LOG_FD", fd_rw.to_string())
        .env("SYD_DO", "setreuid")
        .arg("-plib")
        .args(["-mtrace/allow_safe_setid:1", "--", &SYD_DO, "0", "0"])
        .spawn()
        .expect("spawn syd");
    let _ = close(fd_rw);

    let status = child.wait().expect("wait for syd");
    assert_status_code!(status, nix::libc::EPERM);

    // Convert raw file descriptor to File, then to BufReader
    let file = unsafe { File::from_raw_fd(fd_rd) };
    let mut reader = BufReader::new(file);

    let mut buffer = String::new();
    match reader.read_to_string(&mut buffer) {
        Ok(0) => {
            eprintln!("No data read from pipe; no access violation raised as expected.");
        }
        Ok(_) => {
            // If any data was read, log it
            eprint!("Access violation logged:\n{buffer}");
            return Err(TestError("Unexpected access violation logged.".to_string()));
        }
        Err(e) => {
            // Error reading from the buffer
            eprintln!("Error reading from pipe: {e}");
            return Err(TestError(format!("Error reading from pipe: {e}")));
        }
    }

    Ok(())
}

fn test_syd_setreuid_nobody_safesetid_allow_1() -> TestResult {
    if !*CAP_SETUID {
        eprintln!("Test requires CAP_SETUID, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }

    // SafeSetID is enabled with UID transition defined.
    // The syscall must succeed.
    let uid = Uid::current();
    let status = syd()
        .env("SYD_DO", "setreuid")
        .arg("-plib")
        .args([
            "-mtrace/allow_safe_setid:1",
            &format!("-msetuid+{uid}:nobody"),
            "--",
            &SYD_DO,
            "-1",
            "65534",
        ])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_setreuid_nobody_safesetid_allow_2() -> TestResult {
    if !*CAP_SETUID {
        eprintln!("Test requires CAP_SETUID, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }

    // SafeSetID is enabled with UID transition defined.
    // The syscall must succeed.
    let uid = Uid::current();
    let status = syd()
        .env("SYD_DO", "setreuid")
        .arg("-plib")
        .args([
            "-mtrace/allow_safe_setid:1",
            &format!("-msetuid+{uid}:nobody"),
            "--",
            &SYD_DO,
            "65534",
            "-1",
        ])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_setreuid_nobody_safesetid_allow_3() -> TestResult {
    if !*CAP_SETUID {
        eprintln!("Test requires CAP_SETUID, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }

    // SafeSetID is enabled with UID transition defined.
    // The syscall must succeed.
    let uid = Uid::current();
    let status = syd()
        .env("SYD_DO", "setreuid")
        .arg("-plib")
        .args([
            "-mtrace/allow_safe_setid:1",
            &format!("-msetuid+{uid}:nobody"),
            "--",
            &SYD_DO,
            "65534",
            "65534",
        ])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_setregid_nobody_default_1() -> TestResult {
    if !*CAP_SETGID {
        eprintln!("Test requires CAP_SETGID, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }

    // Default filter is a no-op for setregid.
    // Test must abort on gid mismatch after successful call.
    let status = syd()
        .env("SYD_DO", "setregid")
        .arg("-plib")
        .args(["--", &SYD_DO, "-1", "65534"])
        .status()
        .expect("execute syd");
    assert_status_code!(status, 128 + nix::libc::SIGABRT);

    Ok(())
}

fn test_syd_setregid_nobody_default_2() -> TestResult {
    if !*CAP_SETGID {
        eprintln!("Test requires CAP_SETGID, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }

    // Default filter is a no-op for setregid.
    // Test must abort on gid mismatch after successful call.
    let status = syd()
        .env("SYD_DO", "setregid")
        .arg("-plib")
        .args(["--", &SYD_DO, "65534", "-1"])
        .status()
        .expect("execute syd");
    assert_status_code!(status, 128 + nix::libc::SIGABRT);

    Ok(())
}

fn test_syd_setregid_nobody_default_3() -> TestResult {
    if !*CAP_SETGID {
        eprintln!("Test requires CAP_SETGID, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }

    // Default filter is a no-op for setregid.
    // Test must abort on gid mismatch after successful call.
    let status = syd()
        .env("SYD_DO", "setregid")
        .arg("-plib")
        .args(["--", &SYD_DO, "65534", "65534"])
        .status()
        .expect("execute syd");
    assert_status_code!(status, 128 + nix::libc::SIGABRT);

    Ok(())
}

fn test_syd_setregid_nobody_safesetid_deny_1() -> TestResult {
    if !*CAP_SETGID {
        eprintln!("Test requires CAP_SETGID, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }

    // SAFETY: Set the comment to test setregid_root_safesetid_deny.
    let (fd_rd, fd_rw) = match pipe() {
        Ok((fd_rd, fd_rw)) => (fd_rd, fd_rw),
        Err(errno) => return Err(TestError(format!("error creating pipe: {errno}!"))),
    };

    // SafeSetID is enabled, but no GID transition defined.
    // The syscall must fail with EPERM _and_ generate an acccess violation.
    let mut child = syd()
        .env("SYD_LOG", "warn")
        .env("SYD_LOG_FD", fd_rw.to_string())
        .env("SYD_DO", "setregid")
        .arg("-plib")
        .args(["-mtrace/allow_safe_setid:1", "--", &SYD_DO, "-1", "65534"])
        .spawn()
        .expect("spawn syd");
    let _ = close(fd_rw);

    let status = child.wait().expect("wait for syd");
    assert_status_code!(status, nix::libc::EPERM);

    // Convert raw file descriptor to File, then to BufReader
    let file = unsafe { File::from_raw_fd(fd_rd) };
    let mut reader = BufReader::new(file);

    let mut buffer = String::new();
    match reader.read_to_string(&mut buffer) {
        Ok(0) => {
            eprintln!("No data read from pipe; expected access violation.");
            return Err(TestError(
                "Expected access violation not logged.".to_string(),
            ));
        }
        Ok(_) => {
            // If any data was read, log it
            eprint!("Access violation logged:\n{buffer}");
        }
        Err(e) => {
            // Error reading from the buffer
            eprintln!("Error reading from pipe: {e}");
            return Err(TestError(format!("Error reading from pipe: {e}")));
        }
    }

    Ok(())
}

fn test_syd_setregid_nobody_safesetid_deny_2() -> TestResult {
    if !*CAP_SETGID {
        eprintln!("Test requires CAP_SETGID, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }

    // SAFETY: Set the comment to test setregid_root_safesetid_deny.
    let (fd_rd, fd_rw) = match pipe() {
        Ok((fd_rd, fd_rw)) => (fd_rd, fd_rw),
        Err(errno) => return Err(TestError(format!("error creating pipe: {errno}!"))),
    };

    // SafeSetID is enabled, but no GID transition defined.
    // The syscall must fail with EPERM _and_ generate an acccess violation.
    let mut child = syd()
        .env("SYD_LOG", "warn")
        .env("SYD_LOG_FD", fd_rw.to_string())
        .env("SYD_DO", "setregid")
        .arg("-plib")
        .args(["-mtrace/allow_safe_setid:1", "--", &SYD_DO, "65534", "-1"])
        .spawn()
        .expect("spawn syd");
    let _ = close(fd_rw);

    let status = child.wait().expect("wait for syd");
    assert_status_code!(status, nix::libc::EPERM);

    // Convert raw file descriptor to File, then to BufReader
    let file = unsafe { File::from_raw_fd(fd_rd) };
    let mut reader = BufReader::new(file);

    let mut buffer = String::new();
    match reader.read_to_string(&mut buffer) {
        Ok(0) => {
            eprintln!("No data read from pipe; expected access violation.");
            return Err(TestError(
                "Expected access violation not logged.".to_string(),
            ));
        }
        Ok(_) => {
            // If any data was read, log it
            eprint!("Access violation logged:\n{buffer}");
        }
        Err(e) => {
            // Error reading from the buffer
            eprintln!("Error reading from pipe: {e}");
            return Err(TestError(format!("Error reading from pipe: {e}")));
        }
    }

    Ok(())
}

fn test_syd_setregid_nobody_safesetid_deny_3() -> TestResult {
    if !*CAP_SETGID {
        eprintln!("Test requires CAP_SETGID, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }

    // SAFETY: Set the comment to test setregid_root_safesetid_deny.
    let (fd_rd, fd_rw) = match pipe() {
        Ok((fd_rd, fd_rw)) => (fd_rd, fd_rw),
        Err(errno) => return Err(TestError(format!("error creating pipe: {errno}!"))),
    };

    // SafeSetID is enabled, but no GID transition defined.
    // The syscall must fail with EPERM _and_ generate an acccess violation.
    let mut child = syd()
        .env("SYD_LOG", "warn")
        .env("SYD_LOG_FD", fd_rw.to_string())
        .env("SYD_DO", "setregid")
        .arg("-plib")
        .args([
            "-mtrace/allow_safe_setid:1",
            "--",
            &SYD_DO,
            "65534",
            "65534",
        ])
        .spawn()
        .expect("spawn syd");
    let _ = close(fd_rw);

    let status = child.wait().expect("wait for syd");
    assert_status_code!(status, nix::libc::EPERM);

    // Convert raw file descriptor to File, then to BufReader
    let file = unsafe { File::from_raw_fd(fd_rd) };
    let mut reader = BufReader::new(file);

    let mut buffer = String::new();
    match reader.read_to_string(&mut buffer) {
        Ok(0) => {
            eprintln!("No data read from pipe; expected access violation.");
            return Err(TestError(
                "Expected access violation not logged.".to_string(),
            ));
        }
        Ok(_) => {
            // If any data was read, log it
            eprint!("Access violation logged:\n{buffer}");
        }
        Err(e) => {
            // Error reading from the buffer
            eprintln!("Error reading from pipe: {e}");
            return Err(TestError(format!("Error reading from pipe: {e}")));
        }
    }

    Ok(())
}

fn test_syd_setregid_root_safesetid_deny_1() -> TestResult {
    if !*CAP_SETGID {
        eprintln!("Test requires CAP_SETGID, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }

    // SafeSetID is enabled, but no UID transition defined.
    // The syscall must fail with EPERM even with UID=0.

    // SAFETY: Compared to the test setregid_nobody_safesetid_deny, this
    // test _must not_ generate a syd access violation, because GID
    // change to root must be blocked by the kernel-level parent seccomp
    // filter. This is crucial security-wise so we test it here by
    // overriding SYD_LOG and setting SYD_LOG_FD.
    let (fd_rd, fd_rw) = match pipe() {
        Ok((fd_rd, fd_rw)) => (fd_rd, fd_rw),
        Err(errno) => return Err(TestError(format!("error creating pipe: {errno}!"))),
    };

    let mut child = syd()
        .env("SYD_LOG", "warn")
        .env("SYD_LOG_FD", fd_rw.to_string())
        .env("SYD_DO", "setregid")
        .arg("-plib")
        .args(["-mtrace/allow_safe_setid:1", "--", &SYD_DO, "-1", "0"])
        .spawn()
        .expect("spawn syd");
    let _ = close(fd_rw);

    // Convert raw file descriptor to File, then to BufReader
    let file = unsafe { File::from_raw_fd(fd_rd) };
    let mut reader = BufReader::new(file);

    let mut buffer = String::new();
    match reader.read_to_string(&mut buffer) {
        Ok(0) => {
            eprintln!("No data read from pipe; no access violation raised as expected.");
        }
        Ok(_) => {
            // If any data was read, log it
            eprint!("Access violation logged:\n{buffer}");
            return Err(TestError("Unexpected access violation logged.".to_string()));
        }
        Err(e) => {
            // Error reading from the buffer
            eprintln!("Error reading from pipe: {e}");
            return Err(TestError(format!("Error reading from pipe: {e}")));
        }
    }

    let status = child.wait().expect("wait for syd");
    assert_status_code!(status, nix::libc::EPERM);

    Ok(())
}

fn test_syd_setregid_root_safesetid_deny_2() -> TestResult {
    if !*CAP_SETGID {
        eprintln!("Test requires CAP_SETGID, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }

    // SafeSetID is enabled, but no UID transition defined.
    // The syscall must fail with EPERM even with UID=0.

    // SAFETY: Compared to the test setregid_nobody_safesetid_deny, this
    // test _must not_ generate a syd access violation, because GID
    // change to root must be blocked by the kernel-level parent seccomp
    // filter. This is crucial security-wise so we test it here by
    // overriding SYD_LOG and setting SYD_LOG_FD.
    let (fd_rd, fd_rw) = match pipe() {
        Ok((fd_rd, fd_rw)) => (fd_rd, fd_rw),
        Err(errno) => return Err(TestError(format!("error creating pipe: {errno}!"))),
    };

    let mut child = syd()
        .env("SYD_LOG", "warn")
        .env("SYD_LOG_FD", fd_rw.to_string())
        .env("SYD_DO", "setregid")
        .arg("-plib")
        .args(["-mtrace/allow_safe_setid:1", "--", &SYD_DO, "0", "-1"])
        .spawn()
        .expect("spawn syd");
    let _ = close(fd_rw);

    // Convert raw file descriptor to File, then to BufReader
    let file = unsafe { File::from_raw_fd(fd_rd) };
    let mut reader = BufReader::new(file);

    let mut buffer = String::new();
    match reader.read_to_string(&mut buffer) {
        Ok(0) => {
            eprintln!("No data read from pipe; no access violation raised as expected.");
        }
        Ok(_) => {
            // If any data was read, log it
            eprint!("Access violation logged:\n{buffer}");
            return Err(TestError("Unexpected access violation logged.".to_string()));
        }
        Err(e) => {
            // Error reading from the buffer
            eprintln!("Error reading from pipe: {e}");
            return Err(TestError(format!("Error reading from pipe: {e}")));
        }
    }

    let status = child.wait().expect("wait for syd");
    assert_status_code!(status, nix::libc::EPERM);

    Ok(())
}

fn test_syd_setregid_root_safesetid_deny_3() -> TestResult {
    if !*CAP_SETGID {
        eprintln!("Test requires CAP_SETGID, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }

    // SafeSetID is enabled, but no UID transition defined.
    // The syscall must fail with EPERM even with UID=0.

    // SAFETY: Compared to the test setregid_nobody_safesetid_deny, this
    // test _must not_ generate a syd access violation, because GID
    // change to root must be blocked by the kernel-level parent seccomp
    // filter. This is crucial security-wise so we test it here by
    // overriding SYD_LOG and setting SYD_LOG_FD.
    let (fd_rd, fd_rw) = match pipe() {
        Ok((fd_rd, fd_rw)) => (fd_rd, fd_rw),
        Err(errno) => return Err(TestError(format!("error creating pipe: {errno}!"))),
    };

    let mut child = syd()
        .env("SYD_LOG", "warn")
        .env("SYD_LOG_FD", fd_rw.to_string())
        .env("SYD_DO", "setregid")
        .arg("-plib")
        .args(["-mtrace/allow_safe_setid:1", "--", &SYD_DO, "0", "0"])
        .spawn()
        .expect("spawn syd");
    let _ = close(fd_rw);

    // Convert raw file descriptor to File, then to BufReader
    let file = unsafe { File::from_raw_fd(fd_rd) };
    let mut reader = BufReader::new(file);

    let mut buffer = String::new();
    match reader.read_to_string(&mut buffer) {
        Ok(0) => {
            eprintln!("No data read from pipe; no access violation raised as expected.");
        }
        Ok(_) => {
            // If any data was read, log it
            eprint!("Access violation logged:\n{buffer}");
            return Err(TestError("Unexpected access violation logged.".to_string()));
        }
        Err(e) => {
            // Error reading from the buffer
            eprintln!("Error reading from pipe: {e}");
            return Err(TestError(format!("Error reading from pipe: {e}")));
        }
    }

    let status = child.wait().expect("wait for syd");
    assert_status_code!(status, nix::libc::EPERM);

    Ok(())
}

fn test_syd_setregid_nobody_safesetid_allow_1() -> TestResult {
    if !*CAP_SETGID {
        eprintln!("Test requires CAP_SETGID, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }

    // SafeSetID is enabled with GID transition defined.
    // The syscall must succeed.
    let uid = Uid::current();
    let status = syd()
        .env("SYD_DO", "setregid")
        .arg("-plib")
        .args([
            "-mtrace/allow_safe_setid:1",
            &format!("-msetgid+{uid}:nobody"),
            "--",
            &SYD_DO,
            "-1",
            "65534",
        ])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_setregid_nobody_safesetid_allow_2() -> TestResult {
    if !*CAP_SETGID {
        eprintln!("Test requires CAP_SETGID, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }

    // SafeSetID is enabled with GID transition defined.
    // The syscall must succeed.
    let uid = Uid::current();
    let status = syd()
        .env("SYD_DO", "setregid")
        .arg("-plib")
        .args([
            "-mtrace/allow_safe_setid:1",
            &format!("-msetgid+{uid}:nobody"),
            "--",
            &SYD_DO,
            "65534",
            "-1",
        ])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_setregid_nobody_safesetid_allow_3() -> TestResult {
    if !*CAP_SETGID {
        eprintln!("Test requires CAP_SETGID, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }

    // SafeSetID is enabled with GID transition defined.
    // The syscall must succeed.
    let uid = Uid::current();
    let status = syd()
        .env("SYD_DO", "setregid")
        .arg("-plib")
        .args([
            "-mtrace/allow_safe_setid:1",
            &format!("-msetgid+{uid}:nobody"),
            "--",
            &SYD_DO,
            "65534",
            "65534",
        ])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_setresuid_nobody_default_1() -> TestResult {
    if !*CAP_SETUID {
        eprintln!("Test requires CAP_SETUID, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }

    // Default filter is a no-op for setresuid.
    // Test must abort on uid mismatch after successful call.
    let status = syd()
        .env("SYD_DO", "setresuid")
        .arg("-plib")
        .args(["--", &SYD_DO, "-1", "-1", "65534"])
        .status()
        .expect("execute syd");
    assert_status_code!(status, 128 + nix::libc::SIGABRT);

    Ok(())
}

fn test_syd_setresuid_nobody_default_2() -> TestResult {
    if !*CAP_SETUID {
        eprintln!("Test requires CAP_SETUID, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }

    // Default filter is a no-op for setresid.
    // Test must abort on uid mismatch after successful call.
    let status = syd()
        .env("SYD_DO", "setresuid")
        .arg("-plib")
        .args(["--", &SYD_DO, "-1", "65534", "-1"])
        .status()
        .expect("execute syd");
    assert_status_code!(status, 128 + nix::libc::SIGABRT);

    Ok(())
}

fn test_syd_setresuid_nobody_default_3() -> TestResult {
    if !*CAP_SETUID {
        eprintln!("Test requires CAP_SETUID, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }

    // Default filter is a no-op for setresuid.
    // Test must abort on uid mismatch after successful call.
    let status = syd()
        .env("SYD_DO", "setresuid")
        .arg("-plib")
        .args(["--", &SYD_DO, "65534", "-1", "-1"])
        .status()
        .expect("execute syd");
    assert_status_code!(status, 128 + nix::libc::SIGABRT);

    Ok(())
}

fn test_syd_setresuid_nobody_default_4() -> TestResult {
    if !*CAP_SETUID {
        eprintln!("Test requires CAP_SETUID, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }

    // Default filter is a no-op for setresuid.
    // Test must abort on uid mismatch after successful call.
    let status = syd()
        .env("SYD_DO", "setresuid")
        .arg("-plib")
        .args(["--", &SYD_DO, "-1", "65534", "65534"])
        .status()
        .expect("execute syd");
    assert_status_code!(status, 128 + nix::libc::SIGABRT);

    Ok(())
}

fn test_syd_setresuid_nobody_default_5() -> TestResult {
    if !*CAP_SETUID {
        eprintln!("Test requires CAP_SETUID, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }

    // Default filter is a no-op for setresuid.
    // Test must abort on uid mismatch after successful call.
    let status = syd()
        .env("SYD_DO", "setresuid")
        .arg("-plib")
        .args(["--", &SYD_DO, "65534", "65534", "-1"])
        .status()
        .expect("execute syd");
    assert_status_code!(status, 128 + nix::libc::SIGABRT);

    Ok(())
}

fn test_syd_setresuid_nobody_default_6() -> TestResult {
    if !*CAP_SETUID {
        eprintln!("Test requires CAP_SETUID, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }

    // Default filter is a no-op for setresuid.
    // Test must abort on uid mismatch after successful call.
    let status = syd()
        .env("SYD_DO", "setresuid")
        .arg("-plib")
        .args(["--", &SYD_DO, "65534", "-1", "65534"])
        .status()
        .expect("execute syd");
    assert_status_code!(status, 128 + nix::libc::SIGABRT);

    Ok(())
}

fn test_syd_setresuid_nobody_default_7() -> TestResult {
    if !*CAP_SETUID {
        eprintln!("Test requires CAP_SETUID, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }

    // Default filter is a no-op for setresuid.
    // Test must abort on uid mismatch after successful call.
    let status = syd()
        .env("SYD_DO", "setresuid")
        .arg("-plib")
        .args(["--", &SYD_DO, "65534", "65534", "65534"])
        .status()
        .expect("execute syd");
    assert_status_code!(status, 128 + nix::libc::SIGABRT);

    Ok(())
}

fn test_syd_setresuid_nobody_safesetid_deny_1() -> TestResult {
    if !*CAP_SETUID {
        eprintln!("Test requires CAP_SETUID, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }

    // SAFETY: Set the comment to test setresuid_root_safesetid_deny.
    let (fd_rd, fd_rw) = match pipe() {
        Ok((fd_rd, fd_rw)) => (fd_rd, fd_rw),
        Err(errno) => return Err(TestError(format!("error creating pipe: {errno}!"))),
    };

    // SafeSetID is enabled, but no UID transition defined.
    // The syscall must fail with EPERM _and_ generate an acccess violation.
    let mut child = syd()
        .env("SYD_LOG", "warn")
        .env("SYD_LOG_FD", fd_rw.to_string())
        .env("SYD_DO", "setresuid")
        .arg("-plib")
        .args([
            "-mtrace/allow_safe_setid:1",
            "--",
            &SYD_DO,
            "-1",
            "-1",
            "65534",
        ])
        .spawn()
        .expect("spawn syd");
    let _ = close(fd_rw);

    // Convert raw file descriptor to File, then to BufReader
    let file = unsafe { File::from_raw_fd(fd_rd) };
    let mut reader = BufReader::new(file);

    let mut buffer = String::new();
    match reader.read_to_string(&mut buffer) {
        Ok(0) => {
            eprintln!("No data read from pipe; expected access violation.");
            return Err(TestError(
                "Expected access violation not logged.".to_string(),
            ));
        }
        Ok(_) => {
            // If any data was read, log it
            eprint!("Access violation logged:\n{buffer}");
        }
        Err(e) => {
            // Error reading from the buffer
            eprintln!("Error reading from pipe: {e}");
            return Err(TestError(format!("Error reading from pipe: {e}")));
        }
    }

    let status = child.wait().expect("wait for syd");
    assert_status_code!(status, nix::libc::EPERM);

    Ok(())
}

fn test_syd_setresuid_nobody_safesetid_deny_2() -> TestResult {
    if !*CAP_SETUID {
        eprintln!("Test requires CAP_SETUID, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }

    // SAFETY: Set the comment to test setresuid_root_safesetid_deny.
    let (fd_rd, fd_rw) = match pipe() {
        Ok((fd_rd, fd_rw)) => (fd_rd, fd_rw),
        Err(errno) => return Err(TestError(format!("error creating pipe: {errno}!"))),
    };

    // SafeSetID is enabled, but no UID transition defined.
    // The syscall must fail with EPERM _and_ generate an acccess violation.
    let mut child = syd()
        .env("SYD_LOG", "warn")
        .env("SYD_LOG_FD", fd_rw.to_string())
        .env("SYD_DO", "setresuid")
        .arg("-plib")
        .args([
            "-mtrace/allow_safe_setid:1",
            "--",
            &SYD_DO,
            "-1",
            "65534",
            "-1",
        ])
        .spawn()
        .expect("spawn syd");
    let _ = close(fd_rw);

    // Convert raw file descriptor to File, then to BufReader
    let file = unsafe { File::from_raw_fd(fd_rd) };
    let mut reader = BufReader::new(file);

    let mut buffer = String::new();
    match reader.read_to_string(&mut buffer) {
        Ok(0) => {
            eprintln!("No data read from pipe; expected access violation.");
            return Err(TestError(
                "Expected access violation not logged.".to_string(),
            ));
        }
        Ok(_) => {
            // If any data was read, log it
            eprint!("Access violation logged:\n{buffer}");
        }
        Err(e) => {
            // Error reading from the buffer
            eprintln!("Error reading from pipe: {e}");
            return Err(TestError(format!("Error reading from pipe: {e}")));
        }
    }

    let status = child.wait().expect("wait for syd");
    assert_status_code!(status, nix::libc::EPERM);

    Ok(())
}

fn test_syd_setresuid_nobody_safesetid_deny_3() -> TestResult {
    if !*CAP_SETUID {
        eprintln!("Test requires CAP_SETUID, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }

    // SAFETY: Set the comment to test setresuid_root_safesetid_deny.
    let (fd_rd, fd_rw) = match pipe() {
        Ok((fd_rd, fd_rw)) => (fd_rd, fd_rw),
        Err(errno) => return Err(TestError(format!("error creating pipe: {errno}!"))),
    };

    // SafeSetID is enabled, but no UID transition defined.
    // The syscall must fail with EPERM _and_ generate an acccess violation.
    let mut child = syd()
        .env("SYD_LOG", "warn")
        .env("SYD_LOG_FD", fd_rw.to_string())
        .env("SYD_DO", "setresuid")
        .arg("-plib")
        .args([
            "-mtrace/allow_safe_setid:1",
            "--",
            &SYD_DO,
            "65534",
            "-1",
            "-1",
        ])
        .spawn()
        .expect("spawn syd");
    let _ = close(fd_rw);

    // Convert raw file descriptor to File, then to BufReader
    let file = unsafe { File::from_raw_fd(fd_rd) };
    let mut reader = BufReader::new(file);

    let mut buffer = String::new();
    match reader.read_to_string(&mut buffer) {
        Ok(0) => {
            eprintln!("No data read from pipe; expected access violation.");
            return Err(TestError(
                "Expected access violation not logged.".to_string(),
            ));
        }
        Ok(_) => {
            // If any data was read, log it
            eprint!("Access violation logged:\n{buffer}");
        }
        Err(e) => {
            // Error reading from the buffer
            eprintln!("Error reading from pipe: {e}");
            return Err(TestError(format!("Error reading from pipe: {e}")));
        }
    }

    let status = child.wait().expect("wait for syd");
    assert_status_code!(status, nix::libc::EPERM);

    Ok(())
}

fn test_syd_setresuid_nobody_safesetid_deny_4() -> TestResult {
    if !*CAP_SETUID {
        eprintln!("Test requires CAP_SETUID, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }

    // SAFETY: Set the comment to test setresuid_root_safesetid_deny.
    let (fd_rd, fd_rw) = match pipe() {
        Ok((fd_rd, fd_rw)) => (fd_rd, fd_rw),
        Err(errno) => return Err(TestError(format!("error creating pipe: {errno}!"))),
    };

    // SafeSetID is enabled, but no UID transition defined.
    // The syscall must fail with EPERM _and_ generate an acccess violation.
    let mut child = syd()
        .env("SYD_LOG", "warn")
        .env("SYD_LOG_FD", fd_rw.to_string())
        .env("SYD_DO", "setresuid")
        .arg("-plib")
        .args([
            "-mtrace/allow_safe_setid:1",
            "--",
            &SYD_DO,
            "-1",
            "65534",
            "65534",
        ])
        .spawn()
        .expect("spawn syd");
    let _ = close(fd_rw);

    // Convert raw file descriptor to File, then to BufReader
    let file = unsafe { File::from_raw_fd(fd_rd) };
    let mut reader = BufReader::new(file);

    let mut buffer = String::new();
    match reader.read_to_string(&mut buffer) {
        Ok(0) => {
            eprintln!("No data read from pipe; expected access violation.");
            return Err(TestError(
                "Expected access violation not logged.".to_string(),
            ));
        }
        Ok(_) => {
            // If any data was read, log it
            eprint!("Access violation logged:\n{buffer}");
        }
        Err(e) => {
            // Error reading from the buffer
            eprintln!("Error reading from pipe: {e}");
            return Err(TestError(format!("Error reading from pipe: {e}")));
        }
    }

    let status = child.wait().expect("wait for syd");
    assert_status_code!(status, nix::libc::EPERM);

    Ok(())
}

fn test_syd_setresuid_nobody_safesetid_deny_5() -> TestResult {
    if !*CAP_SETUID {
        eprintln!("Test requires CAP_SETUID, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }

    // SAFETY: Set the comment to test setresuid_root_safesetid_deny.
    let (fd_rd, fd_rw) = match pipe() {
        Ok((fd_rd, fd_rw)) => (fd_rd, fd_rw),
        Err(errno) => return Err(TestError(format!("error creating pipe: {errno}!"))),
    };

    // SafeSetID is enabled, but no UID transition defined.
    // The syscall must fail with EPERM _and_ generate an acccess violation.
    let mut child = syd()
        .env("SYD_LOG", "warn")
        .env("SYD_LOG_FD", fd_rw.to_string())
        .env("SYD_DO", "setresuid")
        .arg("-plib")
        .args([
            "-mtrace/allow_safe_setid:1",
            "--",
            &SYD_DO,
            "65534",
            "65534",
            "-1",
        ])
        .spawn()
        .expect("spawn syd");
    let _ = close(fd_rw);

    // Convert raw file descriptor to File, then to BufReader
    let file = unsafe { File::from_raw_fd(fd_rd) };
    let mut reader = BufReader::new(file);

    let mut buffer = String::new();
    match reader.read_to_string(&mut buffer) {
        Ok(0) => {
            eprintln!("No data read from pipe; expected access violation.");
            return Err(TestError(
                "Expected access violation not logged.".to_string(),
            ));
        }
        Ok(_) => {
            // If any data was read, log it
            eprint!("Access violation logged:\n{buffer}");
        }
        Err(e) => {
            // Error reading from the buffer
            eprintln!("Error reading from pipe: {e}");
            return Err(TestError(format!("Error reading from pipe: {e}")));
        }
    }

    let status = child.wait().expect("wait for syd");
    assert_status_code!(status, nix::libc::EPERM);

    Ok(())
}

fn test_syd_setresuid_nobody_safesetid_deny_6() -> TestResult {
    if !*CAP_SETUID {
        eprintln!("Test requires CAP_SETUID, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }

    // SAFETY: Set the comment to test setresuid_root_safesetid_deny.
    let (fd_rd, fd_rw) = match pipe() {
        Ok((fd_rd, fd_rw)) => (fd_rd, fd_rw),
        Err(errno) => return Err(TestError(format!("error creating pipe: {errno}!"))),
    };

    // SafeSetID is enabled, but no UID transition defined.
    // The syscall must fail with EPERM _and_ generate an acccess violation.
    let mut child = syd()
        .env("SYD_LOG", "warn")
        .env("SYD_LOG_FD", fd_rw.to_string())
        .env("SYD_DO", "setresuid")
        .arg("-plib")
        .args([
            "-mtrace/allow_safe_setid:1",
            "--",
            &SYD_DO,
            "65534",
            "-1",
            "65534",
        ])
        .spawn()
        .expect("spawn syd");
    let _ = close(fd_rw);

    // Convert raw file descriptor to File, then to BufReader
    let file = unsafe { File::from_raw_fd(fd_rd) };
    let mut reader = BufReader::new(file);

    let mut buffer = String::new();
    match reader.read_to_string(&mut buffer) {
        Ok(0) => {
            eprintln!("No data read from pipe; expected access violation.");
            return Err(TestError(
                "Expected access violation not logged.".to_string(),
            ));
        }
        Ok(_) => {
            // If any data was read, log it
            eprint!("Access violation logged:\n{buffer}");
        }
        Err(e) => {
            // Error reading from the buffer
            eprintln!("Error reading from pipe: {e}");
            return Err(TestError(format!("Error reading from pipe: {e}")));
        }
    }

    let status = child.wait().expect("wait for syd");
    assert_status_code!(status, nix::libc::EPERM);

    Ok(())
}

fn test_syd_setresuid_nobody_safesetid_deny_7() -> TestResult {
    if !*CAP_SETUID {
        eprintln!("Test requires CAP_SETUID, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }

    // SAFETY: Set the comment to test setresuid_root_safesetid_deny.
    let (fd_rd, fd_rw) = match pipe() {
        Ok((fd_rd, fd_rw)) => (fd_rd, fd_rw),
        Err(errno) => return Err(TestError(format!("error creating pipe: {errno}!"))),
    };

    // SafeSetID is enabled, but no UID transition defined.
    // The syscall must fail with EPERM _and_ generate an acccess violation.
    let mut child = syd()
        .env("SYD_LOG", "warn")
        .env("SYD_LOG_FD", fd_rw.to_string())
        .env("SYD_DO", "setresuid")
        .arg("-plib")
        .args([
            "-mtrace/allow_safe_setid:1",
            "--",
            &SYD_DO,
            "65534",
            "65534",
            "65534",
        ])
        .spawn()
        .expect("spawn syd");
    let _ = close(fd_rw);

    // Convert raw file descriptor to File, then to BufReader
    let file = unsafe { File::from_raw_fd(fd_rd) };
    let mut reader = BufReader::new(file);

    let mut buffer = String::new();
    match reader.read_to_string(&mut buffer) {
        Ok(0) => {
            eprintln!("No data read from pipe; expected access violation.");
            return Err(TestError(
                "Expected access violation not logged.".to_string(),
            ));
        }
        Ok(_) => {
            // If any data was read, log it
            eprint!("Access violation logged:\n{buffer}");
        }
        Err(e) => {
            // Error reading from the buffer
            eprintln!("Error reading from pipe: {e}");
            return Err(TestError(format!("Error reading from pipe: {e}")));
        }
    }

    let status = child.wait().expect("wait for syd");
    assert_status_code!(status, nix::libc::EPERM);

    Ok(())
}

fn test_syd_setresuid_root_safesetid_deny_1() -> TestResult {
    if !*CAP_SETUID {
        eprintln!("Test requires CAP_SETUID, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }

    // SafeSetID is enabled, but no UID transition defined.
    // The syscall must fail with EPERM even with UID=0.

    // SAFETY: Compared to the test setresuid_nobody_safesetid_deny, this
    // test _must not_ generate a syd access violation, because UID
    // change to root must be blocked by the kernel-level parent seccomp
    // filter. This is crucial security-wise so we test it here by
    // overriding SYD_LOG and setting SYD_LOG_FD.
    let (fd_rd, fd_rw) = match pipe() {
        Ok((fd_rd, fd_rw)) => (fd_rd, fd_rw),
        Err(errno) => return Err(TestError(format!("error creating pipe: {errno}!"))),
    };

    let mut child = syd()
        .env("SYD_LOG", "warn")
        .env("SYD_LOG_FD", fd_rw.to_string())
        .env("SYD_DO", "setresuid")
        .arg("-plib")
        .args(["-mtrace/allow_safe_setid:1", "--", &SYD_DO, "-1", "-1", "0"])
        .spawn()
        .expect("spawn syd");
    let _ = close(fd_rw);

    // Convert raw file descriptor to File, then to BufReader
    let file = unsafe { File::from_raw_fd(fd_rd) };
    let mut reader = BufReader::new(file);

    let mut buffer = String::new();
    match reader.read_to_string(&mut buffer) {
        Ok(0) => {
            eprintln!("No data read from pipe; no access violation raised as expected.");
        }
        Ok(_) => {
            // If any data was read, log it
            eprint!("Access violation logged:\n{buffer}");
            return Err(TestError("Unexpected access violation logged.".to_string()));
        }
        Err(e) => {
            // Error reading from the buffer
            eprintln!("Error reading from pipe: {e}");
            return Err(TestError(format!("Error reading from pipe: {e}")));
        }
    }

    let status = child.wait().expect("wait for syd");
    assert_status_code!(status, nix::libc::EPERM);

    Ok(())
}

fn test_syd_setresuid_root_safesetid_deny_2() -> TestResult {
    if !*CAP_SETUID {
        eprintln!("Test requires CAP_SETUID, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }

    // SafeSetID is enabled, but no UID transition defined.
    // The syscall must fail with EPERM even with UID=0.

    // SAFETY: Compared to the test setresuid_nobody_safesetid_deny, this
    // test _must not_ generate a syd access violation, because UID
    // change to root must be blocked by the kernel-level parent seccomp
    // filter. This is crucial security-wise so we test it here by
    // overriding SYD_LOG and setting SYD_LOG_FD.
    let (fd_rd, fd_rw) = match pipe() {
        Ok((fd_rd, fd_rw)) => (fd_rd, fd_rw),
        Err(errno) => return Err(TestError(format!("error creating pipe: {errno}!"))),
    };

    let mut child = syd()
        .env("SYD_LOG", "warn")
        .env("SYD_LOG_FD", fd_rw.to_string())
        .env("SYD_DO", "setreuid")
        .arg("-plib")
        .args(["-mtrace/allow_safe_setid:1", "--", &SYD_DO, "-1", "0", "-1"])
        .spawn()
        .expect("spawn syd");
    let _ = close(fd_rw);

    // Convert raw file descriptor to File, then to BufReader
    let file = unsafe { File::from_raw_fd(fd_rd) };
    let mut reader = BufReader::new(file);

    let mut buffer = String::new();
    match reader.read_to_string(&mut buffer) {
        Ok(0) => {
            eprintln!("No data read from pipe; no access violation raised as expected.");
        }
        Ok(_) => {
            // If any data was read, log it
            eprint!("Access violation logged:\n{buffer}");
            return Err(TestError("Unexpected access violation logged.".to_string()));
        }
        Err(e) => {
            // Error reading from the buffer
            eprintln!("Error reading from pipe: {e}");
            return Err(TestError(format!("Error reading from pipe: {e}")));
        }
    }

    let status = child.wait().expect("wait for syd");
    assert_status_code!(status, nix::libc::EPERM);

    Ok(())
}

fn test_syd_setresuid_root_safesetid_deny_3() -> TestResult {
    if !*CAP_SETUID {
        eprintln!("Test requires CAP_SETUID, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }

    // SafeSetID is enabled, but no UID transition defined.
    // The syscall must fail with EPERM even with UID=0.

    // SAFETY: Compared to the test setresuid_nobody_safesetid_deny, this
    // test _must not_ generate a syd access violation, because UID
    // change to root must be blocked by the kernel-level parent seccomp
    // filter. This is crucial security-wise so we test it here by
    // overriding SYD_LOG and setting SYD_LOG_FD.
    let (fd_rd, fd_rw) = match pipe() {
        Ok((fd_rd, fd_rw)) => (fd_rd, fd_rw),
        Err(errno) => return Err(TestError(format!("error creating pipe: {errno}!"))),
    };

    let mut child = syd()
        .env("SYD_LOG", "warn")
        .env("SYD_LOG_FD", fd_rw.to_string())
        .env("SYD_DO", "setresuid")
        .arg("-plib")
        .args(["-mtrace/allow_safe_setid:1", "--", &SYD_DO, "0", "-1", "-1"])
        .spawn()
        .expect("spawn syd");
    let _ = close(fd_rw);

    // Convert raw file descriptor to File, then to BufReader
    let file = unsafe { File::from_raw_fd(fd_rd) };
    let mut reader = BufReader::new(file);

    let mut buffer = String::new();
    match reader.read_to_string(&mut buffer) {
        Ok(0) => {
            eprintln!("No data read from pipe; no access violation raised as expected.");
        }
        Ok(_) => {
            // If any data was read, log it
            eprint!("Access violation logged:\n{buffer}");
            return Err(TestError("Unexpected access violation logged.".to_string()));
        }
        Err(e) => {
            // Error reading from the buffer
            eprintln!("Error reading from pipe: {e}");
            return Err(TestError(format!("Error reading from pipe: {e}")));
        }
    }

    let status = child.wait().expect("wait for syd");
    assert_status_code!(status, nix::libc::EPERM);

    Ok(())
}

fn test_syd_setresuid_root_safesetid_deny_4() -> TestResult {
    if !*CAP_SETUID {
        eprintln!("Test requires CAP_SETUID, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }

    // SafeSetID is enabled, but no UID transition defined.
    // The syscall must fail with EPERM even with UID=0.

    // SAFETY: Compared to the test setresuid_nobody_safesetid_deny, this
    // test _must not_ generate a syd access violation, because UID
    // change to root must be blocked by the kernel-level parent seccomp
    // filter. This is crucial security-wise so we test it here by
    // overriding SYD_LOG and setting SYD_LOG_FD.
    let (fd_rd, fd_rw) = match pipe() {
        Ok((fd_rd, fd_rw)) => (fd_rd, fd_rw),
        Err(errno) => return Err(TestError(format!("error creating pipe: {errno}!"))),
    };

    let mut child = syd()
        .env("SYD_LOG", "warn")
        .env("SYD_LOG_FD", fd_rw.to_string())
        .env("SYD_DO", "setresuid")
        .arg("-plib")
        .args(["-mtrace/allow_safe_setid:1", "--", &SYD_DO, "-1", "0", "0"])
        .spawn()
        .expect("spawn syd");
    let _ = close(fd_rw);

    // Convert raw file descriptor to File, then to BufReader
    let file = unsafe { File::from_raw_fd(fd_rd) };
    let mut reader = BufReader::new(file);

    let mut buffer = String::new();
    match reader.read_to_string(&mut buffer) {
        Ok(0) => {
            eprintln!("No data read from pipe; no access violation raised as expected.");
        }
        Ok(_) => {
            // If any data was read, log it
            eprint!("Access violation logged:\n{buffer}");
            return Err(TestError("Unexpected access violation logged.".to_string()));
        }
        Err(e) => {
            // Error reading from the buffer
            eprintln!("Error reading from pipe: {e}");
            return Err(TestError(format!("Error reading from pipe: {e}")));
        }
    }

    let status = child.wait().expect("wait for syd");
    assert_status_code!(status, nix::libc::EPERM);

    Ok(())
}

fn test_syd_setresuid_root_safesetid_deny_5() -> TestResult {
    if !*CAP_SETUID {
        eprintln!("Test requires CAP_SETUID, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }

    // SafeSetID is enabled, but no UID transition defined.
    // The syscall must fail with EPERM even with UID=0.

    // SAFETY: Compared to the test setresuid_nobody_safesetid_deny, this
    // test _must not_ generate a syd access violation, because UID
    // change to root must be blocked by the kernel-level parent seccomp
    // filter. This is crucial security-wise so we test it here by
    // overriding SYD_LOG and setting SYD_LOG_FD.
    let (fd_rd, fd_rw) = match pipe() {
        Ok((fd_rd, fd_rw)) => (fd_rd, fd_rw),
        Err(errno) => return Err(TestError(format!("error creating pipe: {errno}!"))),
    };

    let mut child = syd()
        .env("SYD_LOG", "warn")
        .env("SYD_LOG_FD", fd_rw.to_string())
        .env("SYD_DO", "setresuid")
        .arg("-plib")
        .args(["-mtrace/allow_safe_setid:1", "--", &SYD_DO, "0", "0", "-1"])
        .spawn()
        .expect("spawn syd");
    let _ = close(fd_rw);

    // Convert raw file descriptor to File, then to BufReader
    let file = unsafe { File::from_raw_fd(fd_rd) };
    let mut reader = BufReader::new(file);

    let mut buffer = String::new();
    match reader.read_to_string(&mut buffer) {
        Ok(0) => {
            eprintln!("No data read from pipe; no access violation raised as expected.");
        }
        Ok(_) => {
            // If any data was read, log it
            eprint!("Access violation logged:\n{buffer}");
            return Err(TestError("Unexpected access violation logged.".to_string()));
        }
        Err(e) => {
            // Error reading from the buffer
            eprintln!("Error reading from pipe: {e}");
            return Err(TestError(format!("Error reading from pipe: {e}")));
        }
    }

    let status = child.wait().expect("wait for syd");
    assert_status_code!(status, nix::libc::EPERM);

    Ok(())
}

fn test_syd_setresuid_root_safesetid_deny_6() -> TestResult {
    if !*CAP_SETUID {
        eprintln!("Test requires CAP_SETUID, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }

    // SafeSetID is enabled, but no UID transition defined.
    // The syscall must fail with EPERM even with UID=0.

    // SAFETY: Compared to the test setresuid_nobody_safesetid_deny, this
    // test _must not_ generate a syd access violation, because UID
    // change to root must be blocked by the kernel-level parent seccomp
    // filter. This is crucial security-wise so we test it here by
    // overriding SYD_LOG and setting SYD_LOG_FD.
    let (fd_rd, fd_rw) = match pipe() {
        Ok((fd_rd, fd_rw)) => (fd_rd, fd_rw),
        Err(errno) => return Err(TestError(format!("error creating pipe: {errno}!"))),
    };

    let mut child = syd()
        .env("SYD_LOG", "warn")
        .env("SYD_LOG_FD", fd_rw.to_string())
        .env("SYD_DO", "setresuid")
        .arg("-plib")
        .args(["-mtrace/allow_safe_setid:1", "--", &SYD_DO, "0", "-1", "0"])
        .spawn()
        .expect("spawn syd");
    let _ = close(fd_rw);

    // Convert raw file descriptor to File, then to BufReader
    let file = unsafe { File::from_raw_fd(fd_rd) };
    let mut reader = BufReader::new(file);

    let mut buffer = String::new();
    match reader.read_to_string(&mut buffer) {
        Ok(0) => {
            eprintln!("No data read from pipe; no access violation raised as expected.");
        }
        Ok(_) => {
            // If any data was read, log it
            eprint!("Access violation logged:\n{buffer}");
            return Err(TestError("Unexpected access violation logged.".to_string()));
        }
        Err(e) => {
            // Error reading from the buffer
            eprintln!("Error reading from pipe: {e}");
            return Err(TestError(format!("Error reading from pipe: {e}")));
        }
    }

    let status = child.wait().expect("wait for syd");
    assert_status_code!(status, nix::libc::EPERM);

    Ok(())
}

fn test_syd_setresuid_root_safesetid_deny_7() -> TestResult {
    if !*CAP_SETUID {
        eprintln!("Test requires CAP_SETUID, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }

    // SafeSetID is enabled, but no UID transition defined.
    // The syscall must fail with EPERM even with UID=0.

    // SAFETY: Compared to the test setresuid_nobody_safesetid_deny, this
    // test _must not_ generate a syd access violation, because UID
    // change to root must be blocked by the kernel-level parent seccomp
    // filter. This is crucial security-wise so we test it here by
    // overriding SYD_LOG and setting SYD_LOG_FD.
    let (fd_rd, fd_rw) = match pipe() {
        Ok((fd_rd, fd_rw)) => (fd_rd, fd_rw),
        Err(errno) => return Err(TestError(format!("error creating pipe: {errno}!"))),
    };

    let mut child = syd()
        .env("SYD_LOG", "warn")
        .env("SYD_LOG_FD", fd_rw.to_string())
        .env("SYD_DO", "setresuid")
        .arg("-plib")
        .args(["-mtrace/allow_safe_setid:1", "--", &SYD_DO, "0", "0", "0"])
        .spawn()
        .expect("spawn syd");
    let _ = close(fd_rw);

    // Convert raw file descriptor to File, then to BufReader
    let file = unsafe { File::from_raw_fd(fd_rd) };
    let mut reader = BufReader::new(file);

    let mut buffer = String::new();
    match reader.read_to_string(&mut buffer) {
        Ok(0) => {
            eprintln!("No data read from pipe; no access violation raised as expected.");
        }
        Ok(_) => {
            // If any data was read, log it
            eprint!("Access violation logged:\n{buffer}");
            return Err(TestError("Unexpected access violation logged.".to_string()));
        }
        Err(e) => {
            // Error reading from the buffer
            eprintln!("Error reading from pipe: {e}");
            return Err(TestError(format!("Error reading from pipe: {e}")));
        }
    }

    let status = child.wait().expect("wait for syd");
    assert_status_code!(status, nix::libc::EPERM);

    Ok(())
}

fn test_syd_setresuid_nobody_safesetid_allow_1() -> TestResult {
    if !*CAP_SETUID {
        eprintln!("Test requires CAP_SETUID, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }

    // SafeSetID is enabled with UID transition defined.
    // The syscall must succeed.
    let uid = Uid::current();
    let status = syd()
        .env("SYD_DO", "setresuid")
        .arg("-plib")
        .args([
            "-mtrace/allow_safe_setid:1",
            &format!("-msetuid+{uid}:nobody"),
            "--",
            &SYD_DO,
            "-1",
            "-1",
            "65534",
        ])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_setresuid_nobody_safesetid_allow_2() -> TestResult {
    if !*CAP_SETUID {
        eprintln!("Test requires CAP_SETUID, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }

    // SafeSetID is enabled with UID transition defined.
    // The syscall must succeed.
    let uid = Uid::current();
    let status = syd()
        .env("SYD_DO", "setresuid")
        .arg("-plib")
        .args([
            "-mtrace/allow_safe_setid:1",
            &format!("-msetuid+{uid}:nobody"),
            "--",
            &SYD_DO,
            "-1",
            "65534",
            "-1",
        ])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_setresuid_nobody_safesetid_allow_3() -> TestResult {
    if !*CAP_SETUID {
        eprintln!("Test requires CAP_SETUID, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }

    // SafeSetID is enabled with UID transition defined.
    // The syscall must succeed.
    let uid = Uid::current();
    let status = syd()
        .env("SYD_DO", "setresuid")
        .arg("-plib")
        .args([
            "-mtrace/allow_safe_setid:1",
            &format!("-msetuid+{uid}:nobody"),
            "--",
            &SYD_DO,
            "65534",
            "-1",
            "-1",
        ])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_setresuid_nobody_safesetid_allow_4() -> TestResult {
    if !*CAP_SETUID {
        eprintln!("Test requires CAP_SETUID, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }

    // SafeSetID is enabled with UID transition defined.
    // The syscall must succeed.
    let uid = Uid::current();
    let status = syd()
        .env("SYD_DO", "setresuid")
        .arg("-plib")
        .args([
            "-mtrace/allow_safe_setid:1",
            &format!("-msetuid+{uid}:nobody"),
            "--",
            &SYD_DO,
            "-1",
            "65534",
            "65534",
        ])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_setresuid_nobody_safesetid_allow_5() -> TestResult {
    if !*CAP_SETUID {
        eprintln!("Test requires CAP_SETUID, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }

    // SafeSetID is enabled with UID transition defined.
    // The syscall must succeed.
    let uid = Uid::current();
    let status = syd()
        .env("SYD_DO", "setresuid")
        .arg("-plib")
        .args([
            "-mtrace/allow_safe_setid:1",
            &format!("-msetuid+{uid}:nobody"),
            "--",
            &SYD_DO,
            "65534",
            "65534",
            "-1",
        ])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_setresuid_nobody_safesetid_allow_6() -> TestResult {
    if !*CAP_SETUID {
        eprintln!("Test requires CAP_SETUID, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }

    // SafeSetID is enabled with UID transition defined.
    // The syscall must succeed.
    let uid = Uid::current();
    let status = syd()
        .env("SYD_DO", "setresuid")
        .arg("-plib")
        .args([
            "-mtrace/allow_safe_setid:1",
            &format!("-msetuid+{uid}:nobody"),
            "--",
            &SYD_DO,
            "65534",
            "-1",
            "65534",
        ])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_setresuid_nobody_safesetid_allow_7() -> TestResult {
    if !*CAP_SETUID {
        eprintln!("Test requires CAP_SETUID, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }

    // SafeSetID is enabled with UID transition defined.
    // The syscall must succeed.
    let uid = Uid::current();
    let status = syd()
        .env("SYD_DO", "setresuid")
        .arg("-plib")
        .args([
            "-mtrace/allow_safe_setid:1",
            &format!("-msetuid+{uid}:nobody"),
            "--",
            &SYD_DO,
            "65534",
            "65534",
            "65534",
        ])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_setresgid_nobody_default_1() -> TestResult {
    if !*CAP_SETGID {
        eprintln!("Test requires CAP_SETGID, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }

    // Default filter is a no-op for setresgid.
    // Test must abort on gid mismatch after successful call.
    let status = syd()
        .env("SYD_DO", "setresgid")
        .arg("-plib")
        .args(["--", &SYD_DO, "-1", "-1", "65534"])
        .status()
        .expect("execute syd");
    assert_status_code!(status, 128 + nix::libc::SIGABRT);

    Ok(())
}

fn test_syd_setresgid_nobody_default_2() -> TestResult {
    if !*CAP_SETGID {
        eprintln!("Test requires CAP_SETGID, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }

    // Default filter is a no-op for setresid.
    // Test must abort on gid mismatch after successful call.
    let status = syd()
        .env("SYD_DO", "setresgid")
        .arg("-plib")
        .args(["--", &SYD_DO, "-1", "65534", "-1"])
        .status()
        .expect("execute syd");
    assert_status_code!(status, 128 + nix::libc::SIGABRT);

    Ok(())
}

fn test_syd_setresgid_nobody_default_3() -> TestResult {
    if !*CAP_SETGID {
        eprintln!("Test requires CAP_SETGID, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }

    // Default filter is a no-op for setresgid.
    // Test must abort on gid mismatch after successful call.
    let status = syd()
        .env("SYD_DO", "setresgid")
        .arg("-plib")
        .args(["--", &SYD_DO, "65534", "-1", "-1"])
        .status()
        .expect("execute syd");
    assert_status_code!(status, 128 + nix::libc::SIGABRT);

    Ok(())
}

fn test_syd_setresgid_nobody_default_4() -> TestResult {
    if !*CAP_SETGID {
        eprintln!("Test requires CAP_SETGID, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }

    // Default filter is a no-op for setresgid.
    // Test must abort on gid mismatch after successful call.
    let status = syd()
        .env("SYD_DO", "setresgid")
        .arg("-plib")
        .args(["--", &SYD_DO, "-1", "65534", "65534"])
        .status()
        .expect("execute syd");
    assert_status_code!(status, 128 + nix::libc::SIGABRT);

    Ok(())
}

fn test_syd_setresgid_nobody_default_5() -> TestResult {
    if !*CAP_SETGID {
        eprintln!("Test requires CAP_SETGID, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }

    // Default filter is a no-op for setresgid.
    // Test must abort on gid mismatch after successful call.
    let status = syd()
        .env("SYD_DO", "setresgid")
        .arg("-plib")
        .args(["--", &SYD_DO, "65534", "65534", "-1"])
        .status()
        .expect("execute syd");
    assert_status_code!(status, 128 + nix::libc::SIGABRT);

    Ok(())
}

fn test_syd_setresgid_nobody_default_6() -> TestResult {
    if !*CAP_SETGID {
        eprintln!("Test requires CAP_SETGID, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }

    // Default filter is a no-op for setresgid.
    // Test must abort on gid mismatch after successful call.
    let status = syd()
        .env("SYD_DO", "setresgid")
        .arg("-plib")
        .args(["--", &SYD_DO, "65534", "-1", "65534"])
        .status()
        .expect("execute syd");
    assert_status_code!(status, 128 + nix::libc::SIGABRT);

    Ok(())
}

fn test_syd_setresgid_nobody_default_7() -> TestResult {
    if !*CAP_SETGID {
        eprintln!("Test requires CAP_SETGID, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }

    // Default filter is a no-op for setresgid.
    // Test must abort on gid mismatch after successful call.
    let status = syd()
        .env("SYD_DO", "setresgid")
        .arg("-plib")
        .args(["--", &SYD_DO, "65534", "65534", "65534"])
        .status()
        .expect("execute syd");
    assert_status_code!(status, 128 + nix::libc::SIGABRT);

    Ok(())
}

fn test_syd_setresgid_nobody_safesetid_deny_1() -> TestResult {
    if !*CAP_SETGID {
        eprintln!("Test requires CAP_SETGID, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }

    // SAFETY: Set the comment to test setresgid_root_safesetid_deny.
    let (fd_rd, fd_rw) = match pipe() {
        Ok((fd_rd, fd_rw)) => (fd_rd, fd_rw),
        Err(errno) => return Err(TestError(format!("error creating pipe: {errno}!"))),
    };

    // SafeSetID is enabled, but no GID transition defined.
    // The syscall must fail with EPERM _and_ generate an acccess violation.
    let mut child = syd()
        .env("SYD_LOG", "warn")
        .env("SYD_LOG_FD", fd_rw.to_string())
        .env("SYD_DO", "setresgid")
        .arg("-plib")
        .args([
            "-mtrace/allow_safe_setid:1",
            "--",
            &SYD_DO,
            "-1",
            "-1",
            "65534",
        ])
        .spawn()
        .expect("spawn syd");
    let _ = close(fd_rw);

    // Convert raw file descriptor to File, then to BufReader
    let file = unsafe { File::from_raw_fd(fd_rd) };
    let mut reader = BufReader::new(file);

    let mut buffer = String::new();
    match reader.read_to_string(&mut buffer) {
        Ok(0) => {
            eprintln!("No data read from pipe; expected access violation.");
            return Err(TestError(
                "Expected access violation not logged.".to_string(),
            ));
        }
        Ok(_) => {
            // If any data was read, log it
            eprint!("Access violation logged:\n{buffer}");
        }
        Err(e) => {
            // Error reading from the buffer
            eprintln!("Error reading from pipe: {e}");
            return Err(TestError(format!("Error reading from pipe: {e}")));
        }
    }

    let status = child.wait().expect("wait for syd");
    assert_status_code!(status, nix::libc::EPERM);

    Ok(())
}

fn test_syd_setresgid_nobody_safesetid_deny_2() -> TestResult {
    if !*CAP_SETGID {
        eprintln!("Test requires CAP_SETGID, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }

    // SAFETY: Set the comment to test setresgid_root_safesetid_deny.
    let (fd_rd, fd_rw) = match pipe() {
        Ok((fd_rd, fd_rw)) => (fd_rd, fd_rw),
        Err(errno) => return Err(TestError(format!("error creating pipe: {errno}!"))),
    };

    // SafeSetID is enabled, but no GID transition defined.
    // The syscall must fail with EPERM _and_ generate an acccess violation.
    let mut child = syd()
        .env("SYD_LOG", "warn")
        .env("SYD_LOG_FD", fd_rw.to_string())
        .env("SYD_DO", "setresgid")
        .arg("-plib")
        .args([
            "-mtrace/allow_safe_setid:1",
            "--",
            &SYD_DO,
            "-1",
            "65534",
            "-1",
        ])
        .spawn()
        .expect("spawn syd");
    let _ = close(fd_rw);

    // Convert raw file descriptor to File, then to BufReader
    let file = unsafe { File::from_raw_fd(fd_rd) };
    let mut reader = BufReader::new(file);

    let mut buffer = String::new();
    match reader.read_to_string(&mut buffer) {
        Ok(0) => {
            eprintln!("No data read from pipe; expected access violation.");
            return Err(TestError(
                "Expected access violation not logged.".to_string(),
            ));
        }
        Ok(_) => {
            // If any data was read, log it
            eprint!("Access violation logged:\n{buffer}");
        }
        Err(e) => {
            // Error reading from the buffer
            eprintln!("Error reading from pipe: {e}");
            return Err(TestError(format!("Error reading from pipe: {e}")));
        }
    }

    let status = child.wait().expect("wait for syd");
    assert_status_code!(status, nix::libc::EPERM);

    Ok(())
}

fn test_syd_setresgid_nobody_safesetid_deny_3() -> TestResult {
    if !*CAP_SETGID {
        eprintln!("Test requires CAP_SETGID, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }

    // SAFETY: Set the comment to test setresgid_root_safesetid_deny.
    let (fd_rd, fd_rw) = match pipe() {
        Ok((fd_rd, fd_rw)) => (fd_rd, fd_rw),
        Err(errno) => return Err(TestError(format!("error creating pipe: {errno}!"))),
    };

    // SafeSetID is enabled, but no GID transition defined.
    // The syscall must fail with EPERM _and_ generate an acccess violation.
    let mut child = syd()
        .env("SYD_LOG", "warn")
        .env("SYD_LOG_FD", fd_rw.to_string())
        .env("SYD_DO", "setresgid")
        .arg("-plib")
        .args([
            "-mtrace/allow_safe_setid:1",
            "--",
            &SYD_DO,
            "65534",
            "-1",
            "-1",
        ])
        .spawn()
        .expect("spawn syd");
    let _ = close(fd_rw);

    // Convert raw file descriptor to File, then to BufReader
    let file = unsafe { File::from_raw_fd(fd_rd) };
    let mut reader = BufReader::new(file);

    let mut buffer = String::new();
    match reader.read_to_string(&mut buffer) {
        Ok(0) => {
            eprintln!("No data read from pipe; expected access violation.");
            return Err(TestError(
                "Expected access violation not logged.".to_string(),
            ));
        }
        Ok(_) => {
            // If any data was read, log it
            eprint!("Access violation logged:\n{buffer}");
        }
        Err(e) => {
            // Error reading from the buffer
            eprintln!("Error reading from pipe: {e}");
            return Err(TestError(format!("Error reading from pipe: {e}")));
        }
    }

    let status = child.wait().expect("wait for syd");
    assert_status_code!(status, nix::libc::EPERM);

    Ok(())
}

fn test_syd_setresgid_nobody_safesetid_deny_4() -> TestResult {
    if !*CAP_SETGID {
        eprintln!("Test requires CAP_SETGID, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }

    // SAFETY: Set the comment to test setresgid_root_safesetid_deny.
    let (fd_rd, fd_rw) = match pipe() {
        Ok((fd_rd, fd_rw)) => (fd_rd, fd_rw),
        Err(errno) => return Err(TestError(format!("error creating pipe: {errno}!"))),
    };

    // SafeSetID is enabled, but no GID transition defined.
    // The syscall must fail with EPERM _and_ generate an acccess violation.
    let mut child = syd()
        .env("SYD_LOG", "warn")
        .env("SYD_LOG_FD", fd_rw.to_string())
        .env("SYD_DO", "setresgid")
        .arg("-plib")
        .args([
            "-mtrace/allow_safe_setid:1",
            "--",
            &SYD_DO,
            "-1",
            "65534",
            "65534",
        ])
        .spawn()
        .expect("spawn syd");
    let _ = close(fd_rw);

    // Convert raw file descriptor to File, then to BufReader
    let file = unsafe { File::from_raw_fd(fd_rd) };
    let mut reader = BufReader::new(file);

    let mut buffer = String::new();
    match reader.read_to_string(&mut buffer) {
        Ok(0) => {
            eprintln!("No data read from pipe; expected access violation.");
            return Err(TestError(
                "Expected access violation not logged.".to_string(),
            ));
        }
        Ok(_) => {
            // If any data was read, log it
            eprint!("Access violation logged:\n{buffer}");
        }
        Err(e) => {
            // Error reading from the buffer
            eprintln!("Error reading from pipe: {e}");
            return Err(TestError(format!("Error reading from pipe: {e}")));
        }
    }

    let status = child.wait().expect("wait for syd");
    assert_status_code!(status, nix::libc::EPERM);

    Ok(())
}

fn test_syd_setresgid_nobody_safesetid_deny_5() -> TestResult {
    if !*CAP_SETGID {
        eprintln!("Test requires CAP_SETGID, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }

    // SAFETY: Set the comment to test setresgid_root_safesetid_deny.
    let (fd_rd, fd_rw) = match pipe() {
        Ok((fd_rd, fd_rw)) => (fd_rd, fd_rw),
        Err(errno) => return Err(TestError(format!("error creating pipe: {errno}!"))),
    };

    // SafeSetID is enabled, but no GID transition defined.
    // The syscall must fail with EPERM _and_ generate an acccess violation.
    let mut child = syd()
        .env("SYD_LOG", "warn")
        .env("SYD_LOG_FD", fd_rw.to_string())
        .env("SYD_DO", "setresgid")
        .arg("-plib")
        .args([
            "-mtrace/allow_safe_setid:1",
            "--",
            &SYD_DO,
            "65534",
            "65534",
            "-1",
        ])
        .spawn()
        .expect("spawn syd");
    let _ = close(fd_rw);

    // Convert raw file descriptor to File, then to BufReader
    let file = unsafe { File::from_raw_fd(fd_rd) };
    let mut reader = BufReader::new(file);

    let mut buffer = String::new();
    match reader.read_to_string(&mut buffer) {
        Ok(0) => {
            eprintln!("No data read from pipe; expected access violation.");
            return Err(TestError(
                "Expected access violation not logged.".to_string(),
            ));
        }
        Ok(_) => {
            // If any data was read, log it
            eprint!("Access violation logged:\n{buffer}");
        }
        Err(e) => {
            // Error reading from the buffer
            eprintln!("Error reading from pipe: {e}");
            return Err(TestError(format!("Error reading from pipe: {e}")));
        }
    }

    let status = child.wait().expect("wait for syd");
    assert_status_code!(status, nix::libc::EPERM);

    Ok(())
}

fn test_syd_setresgid_nobody_safesetid_deny_6() -> TestResult {
    if !*CAP_SETGID {
        eprintln!("Test requires CAP_SETGID, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }

    // SAFETY: Set the comment to test setresgid_root_safesetid_deny.
    let (fd_rd, fd_rw) = match pipe() {
        Ok((fd_rd, fd_rw)) => (fd_rd, fd_rw),
        Err(errno) => return Err(TestError(format!("error creating pipe: {errno}!"))),
    };

    // SafeSetID is enabled, but no GID transition defined.
    // The syscall must fail with EPERM _and_ generate an acccess violation.
    let mut child = syd()
        .env("SYD_LOG", "warn")
        .env("SYD_LOG_FD", fd_rw.to_string())
        .env("SYD_DO", "setresgid")
        .arg("-plib")
        .args([
            "-mtrace/allow_safe_setid:1",
            "--",
            &SYD_DO,
            "65534",
            "-1",
            "65534",
        ])
        .spawn()
        .expect("spawn syd");
    let _ = close(fd_rw);

    // Convert raw file descriptor to File, then to BufReader
    let file = unsafe { File::from_raw_fd(fd_rd) };
    let mut reader = BufReader::new(file);

    let mut buffer = String::new();
    match reader.read_to_string(&mut buffer) {
        Ok(0) => {
            eprintln!("No data read from pipe; expected access violation.");
            return Err(TestError(
                "Expected access violation not logged.".to_string(),
            ));
        }
        Ok(_) => {
            // If any data was read, log it
            eprint!("Access violation logged:\n{buffer}");
        }
        Err(e) => {
            // Error reading from the buffer
            eprintln!("Error reading from pipe: {e}");
            return Err(TestError(format!("Error reading from pipe: {e}")));
        }
    }

    let status = child.wait().expect("wait for syd");
    assert_status_code!(status, nix::libc::EPERM);

    Ok(())
}

fn test_syd_setresgid_nobody_safesetid_deny_7() -> TestResult {
    if !*CAP_SETGID {
        eprintln!("Test requires CAP_SETGID, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }

    // SAFETY: Set the comment to test setresgid_root_safesetid_deny.
    let (fd_rd, fd_rw) = match pipe() {
        Ok((fd_rd, fd_rw)) => (fd_rd, fd_rw),
        Err(errno) => return Err(TestError(format!("error creating pipe: {errno}!"))),
    };

    // SafeSetID is enabled, but no GID transition defined.
    // The syscall must fail with EPERM _and_ generate an acccess violation.
    let mut child = syd()
        .env("SYD_LOG", "warn")
        .env("SYD_LOG_FD", fd_rw.to_string())
        .env("SYD_DO", "setresgid")
        .arg("-plib")
        .args([
            "-mtrace/allow_safe_setid:1",
            "--",
            &SYD_DO,
            "65534",
            "65534",
            "65534",
        ])
        .spawn()
        .expect("spawn syd");
    let _ = close(fd_rw);

    // Convert raw file descriptor to File, then to BufReader
    let file = unsafe { File::from_raw_fd(fd_rd) };
    let mut reader = BufReader::new(file);

    let mut buffer = String::new();
    match reader.read_to_string(&mut buffer) {
        Ok(0) => {
            eprintln!("No data read from pipe; expected access violation.");
            return Err(TestError(
                "Expected access violation not logged.".to_string(),
            ));
        }
        Ok(_) => {
            // If any data was read, log it
            eprint!("Access violation logged:\n{buffer}");
        }
        Err(e) => {
            // Error reading from the buffer
            eprintln!("Error reading from pipe: {e}");
            return Err(TestError(format!("Error reading from pipe: {e}")));
        }
    }

    let status = child.wait().expect("wait for syd");
    assert_status_code!(status, nix::libc::EPERM);

    Ok(())
}

fn test_syd_setresgid_root_safesetid_deny_1() -> TestResult {
    if !*CAP_SETGID {
        eprintln!("Test requires CAP_SETGID, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }

    // SafeSetID is enabled, but no GID transition defined.
    // The syscall must fail with EPERM even with GID=0.

    // SAFETY: Compared to the test setresgid_nobody_safesetid_deny, this
    // test _must not_ generate a syd access violation, because GID
    // change to root must be blocked by the kernel-level parent seccomp
    // filter. This is crucial security-wise so we test it here by
    // overriding SYD_LOG and setting SYD_LOG_FD.
    let (fd_rd, fd_rw) = match pipe() {
        Ok((fd_rd, fd_rw)) => (fd_rd, fd_rw),
        Err(errno) => return Err(TestError(format!("error creating pipe: {errno}!"))),
    };

    let mut child = syd()
        .env("SYD_LOG", "warn")
        .env("SYD_LOG_FD", fd_rw.to_string())
        .env("SYD_DO", "setresgid")
        .arg("-plib")
        .args(["-mtrace/allow_safe_setid:1", "--", &SYD_DO, "-1", "-1", "0"])
        .spawn()
        .expect("spawn syd");
    let _ = close(fd_rw);

    // Convert raw file descriptor to File, then to BufReader
    let file = unsafe { File::from_raw_fd(fd_rd) };
    let mut reader = BufReader::new(file);

    let mut buffer = String::new();
    match reader.read_to_string(&mut buffer) {
        Ok(0) => {
            eprintln!("No data read from pipe; no access violation raised as expected.");
        }
        Ok(_) => {
            // If any data was read, log it
            eprint!("Access violation logged:\n{buffer}");
            return Err(TestError("Unexpected access violation logged.".to_string()));
        }
        Err(e) => {
            // Error reading from the buffer
            eprintln!("Error reading from pipe: {e}");
            return Err(TestError(format!("Error reading from pipe: {e}")));
        }
    }

    let status = child.wait().expect("wait for syd");
    assert_status_code!(status, nix::libc::EPERM);

    Ok(())
}

fn test_syd_setresgid_root_safesetid_deny_2() -> TestResult {
    if !*CAP_SETGID {
        eprintln!("Test requires CAP_SETGID, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }

    // SafeSetID is enabled, but no GID transition defined.
    // The syscall must fail with EPERM even with GID=0.

    // SAFETY: Compared to the test setresgid_nobody_safesetid_deny, this
    // test _must not_ generate a syd access violation, because GID
    // change to root must be blocked by the kernel-level parent seccomp
    // filter. This is crucial security-wise so we test it here by
    // overriding SYD_LOG and setting SYD_LOG_FD.
    let (fd_rd, fd_rw) = match pipe() {
        Ok((fd_rd, fd_rw)) => (fd_rd, fd_rw),
        Err(errno) => return Err(TestError(format!("error creating pipe: {errno}!"))),
    };

    let mut child = syd()
        .env("SYD_LOG", "warn")
        .env("SYD_LOG_FD", fd_rw.to_string())
        .env("SYD_DO", "setreuid")
        .arg("-plib")
        .args(["-mtrace/allow_safe_setid:1", "--", &SYD_DO, "-1", "0", "-1"])
        .spawn()
        .expect("spawn syd");
    let _ = close(fd_rw);

    // Convert raw file descriptor to File, then to BufReader
    let file = unsafe { File::from_raw_fd(fd_rd) };
    let mut reader = BufReader::new(file);

    let mut buffer = String::new();
    match reader.read_to_string(&mut buffer) {
        Ok(0) => {
            eprintln!("No data read from pipe; no access violation raised as expected.");
        }
        Ok(_) => {
            // If any data was read, log it
            eprint!("Access violation logged:\n{buffer}");
            return Err(TestError("Unexpected access violation logged.".to_string()));
        }
        Err(e) => {
            // Error reading from the buffer
            eprintln!("Error reading from pipe: {e}");
            return Err(TestError(format!("Error reading from pipe: {e}")));
        }
    }

    let status = child.wait().expect("wait for syd");
    assert_status_code!(status, nix::libc::EPERM);

    Ok(())
}

fn test_syd_setresgid_root_safesetid_deny_3() -> TestResult {
    if !*CAP_SETGID {
        eprintln!("Test requires CAP_SETGID, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }

    // SafeSetID is enabled, but no GID transition defined.
    // The syscall must fail with EPERM even with GID=0.

    // SAFETY: Compared to the test setresgid_nobody_safesetid_deny, this
    // test _must not_ generate a syd access violation, because GID
    // change to root must be blocked by the kernel-level parent seccomp
    // filter. This is crucial security-wise so we test it here by
    // overriding SYD_LOG and setting SYD_LOG_FD.
    let (fd_rd, fd_rw) = match pipe() {
        Ok((fd_rd, fd_rw)) => (fd_rd, fd_rw),
        Err(errno) => return Err(TestError(format!("error creating pipe: {errno}!"))),
    };

    let mut child = syd()
        .env("SYD_LOG", "warn")
        .env("SYD_LOG_FD", fd_rw.to_string())
        .env("SYD_DO", "setresgid")
        .arg("-plib")
        .args(["-mtrace/allow_safe_setid:1", "--", &SYD_DO, "0", "-1", "-1"])
        .spawn()
        .expect("spawn syd");
    let _ = close(fd_rw);

    // Convert raw file descriptor to File, then to BufReader
    let file = unsafe { File::from_raw_fd(fd_rd) };
    let mut reader = BufReader::new(file);

    let mut buffer = String::new();
    match reader.read_to_string(&mut buffer) {
        Ok(0) => {
            eprintln!("No data read from pipe; no access violation raised as expected.");
        }
        Ok(_) => {
            // If any data was read, log it
            eprint!("Access violation logged:\n{buffer}");
            return Err(TestError("Unexpected access violation logged.".to_string()));
        }
        Err(e) => {
            // Error reading from the buffer
            eprintln!("Error reading from pipe: {e}");
            return Err(TestError(format!("Error reading from pipe: {e}")));
        }
    }

    let status = child.wait().expect("wait for syd");
    assert_status_code!(status, nix::libc::EPERM);

    Ok(())
}

fn test_syd_setresgid_root_safesetid_deny_4() -> TestResult {
    if !*CAP_SETGID {
        eprintln!("Test requires CAP_SETGID, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }

    // SafeSetID is enabled, but no GID transition defined.
    // The syscall must fail with EPERM even with GID=0.

    // SAFETY: Compared to the test setresgid_nobody_safesetid_deny, this
    // test _must not_ generate a syd access violation, because GID
    // change to root must be blocked by the kernel-level parent seccomp
    // filter. This is crucial security-wise so we test it here by
    // overriding SYD_LOG and setting SYD_LOG_FD.
    let (fd_rd, fd_rw) = match pipe() {
        Ok((fd_rd, fd_rw)) => (fd_rd, fd_rw),
        Err(errno) => return Err(TestError(format!("error creating pipe: {errno}!"))),
    };

    let mut child = syd()
        .env("SYD_LOG", "warn")
        .env("SYD_LOG_FD", fd_rw.to_string())
        .env("SYD_DO", "setresgid")
        .arg("-plib")
        .args(["-mtrace/allow_safe_setid:1", "--", &SYD_DO, "-1", "0", "0"])
        .spawn()
        .expect("spawn syd");
    let _ = close(fd_rw);

    // Convert raw file descriptor to File, then to BufReader
    let file = unsafe { File::from_raw_fd(fd_rd) };
    let mut reader = BufReader::new(file);

    let mut buffer = String::new();
    match reader.read_to_string(&mut buffer) {
        Ok(0) => {
            eprintln!("No data read from pipe; no access violation raised as expected.");
        }
        Ok(_) => {
            // If any data was read, log it
            eprint!("Access violation logged:\n{buffer}");
            return Err(TestError("Unexpected access violation logged.".to_string()));
        }
        Err(e) => {
            // Error reading from the buffer
            eprintln!("Error reading from pipe: {e}");
            return Err(TestError(format!("Error reading from pipe: {e}")));
        }
    }

    let status = child.wait().expect("wait for syd");
    assert_status_code!(status, nix::libc::EPERM);

    Ok(())
}

fn test_syd_setresgid_root_safesetid_deny_5() -> TestResult {
    if !*CAP_SETGID {
        eprintln!("Test requires CAP_SETGID, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }

    // SafeSetID is enabled, but no GID transition defined.
    // The syscall must fail with EPERM even with GID=0.

    // SAFETY: Compared to the test setresgid_nobody_safesetid_deny, this
    // test _must not_ generate a syd access violation, because GID
    // change to root must be blocked by the kernel-level parent seccomp
    // filter. This is crucial security-wise so we test it here by
    // overriding SYD_LOG and setting SYD_LOG_FD.
    let (fd_rd, fd_rw) = match pipe() {
        Ok((fd_rd, fd_rw)) => (fd_rd, fd_rw),
        Err(errno) => return Err(TestError(format!("error creating pipe: {errno}!"))),
    };

    let mut child = syd()
        .env("SYD_LOG", "warn")
        .env("SYD_LOG_FD", fd_rw.to_string())
        .env("SYD_DO", "setresgid")
        .arg("-plib")
        .args(["-mtrace/allow_safe_setid:1", "--", &SYD_DO, "0", "0", "-1"])
        .spawn()
        .expect("spawn syd");
    let _ = close(fd_rw);

    // Convert raw file descriptor to File, then to BufReader
    let file = unsafe { File::from_raw_fd(fd_rd) };
    let mut reader = BufReader::new(file);

    let mut buffer = String::new();
    match reader.read_to_string(&mut buffer) {
        Ok(0) => {
            eprintln!("No data read from pipe; no access violation raised as expected.");
        }
        Ok(_) => {
            // If any data was read, log it
            eprint!("Access violation logged:\n{buffer}");
            return Err(TestError("Unexpected access violation logged.".to_string()));
        }
        Err(e) => {
            // Error reading from the buffer
            eprintln!("Error reading from pipe: {e}");
            return Err(TestError(format!("Error reading from pipe: {e}")));
        }
    }

    let status = child.wait().expect("wait for syd");
    assert_status_code!(status, nix::libc::EPERM);

    Ok(())
}

fn test_syd_setresgid_root_safesetid_deny_6() -> TestResult {
    if !*CAP_SETGID {
        eprintln!("Test requires CAP_SETGID, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }

    // SafeSetID is enabled, but no GID transition defined.
    // The syscall must fail with EPERM even with GID=0.

    // SAFETY: Compared to the test setresgid_nobody_safesetid_deny, this
    // test _must not_ generate a syd access violation, because GID
    // change to root must be blocked by the kernel-level parent seccomp
    // filter. This is crucial security-wise so we test it here by
    // overriding SYD_LOG and setting SYD_LOG_FD.
    let (fd_rd, fd_rw) = match pipe() {
        Ok((fd_rd, fd_rw)) => (fd_rd, fd_rw),
        Err(errno) => return Err(TestError(format!("error creating pipe: {errno}!"))),
    };

    let mut child = syd()
        .env("SYD_LOG", "warn")
        .env("SYD_LOG_FD", fd_rw.to_string())
        .env("SYD_DO", "setresgid")
        .arg("-plib")
        .args(["-mtrace/allow_safe_setid:1", "--", &SYD_DO, "0", "-1", "0"])
        .spawn()
        .expect("spawn syd");
    let _ = close(fd_rw);

    // Convert raw file descriptor to File, then to BufReader
    let file = unsafe { File::from_raw_fd(fd_rd) };
    let mut reader = BufReader::new(file);

    let mut buffer = String::new();
    match reader.read_to_string(&mut buffer) {
        Ok(0) => {
            eprintln!("No data read from pipe; no access violation raised as expected.");
        }
        Ok(_) => {
            // If any data was read, log it
            eprint!("Access violation logged:\n{buffer}");
            return Err(TestError("Unexpected access violation logged.".to_string()));
        }
        Err(e) => {
            // Error reading from the buffer
            eprintln!("Error reading from pipe: {e}");
            return Err(TestError(format!("Error reading from pipe: {e}")));
        }
    }

    let status = child.wait().expect("wait for syd");
    assert_status_code!(status, nix::libc::EPERM);

    Ok(())
}

fn test_syd_setresgid_root_safesetid_deny_7() -> TestResult {
    if !*CAP_SETGID {
        eprintln!("Test requires CAP_SETGID, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }

    // SafeSetID is enabled, but no GID transition defined.
    // The syscall must fail with EPERM even with GID=0.

    // SAFETY: Compared to the test setresgid_nobody_safesetid_deny, this
    // test _must not_ generate a syd access violation, because GID
    // change to root must be blocked by the kernel-level parent seccomp
    // filter. This is crucial security-wise so we test it here by
    // overriding SYD_LOG and setting SYD_LOG_FD.
    let (fd_rd, fd_rw) = match pipe() {
        Ok((fd_rd, fd_rw)) => (fd_rd, fd_rw),
        Err(errno) => return Err(TestError(format!("error creating pipe: {errno}!"))),
    };

    let mut child = syd()
        .env("SYD_LOG", "warn")
        .env("SYD_LOG_FD", fd_rw.to_string())
        .env("SYD_DO", "setresgid")
        .arg("-plib")
        .args(["-mtrace/allow_safe_setid:1", "--", &SYD_DO, "0", "0", "0"])
        .spawn()
        .expect("spawn syd");
    let _ = close(fd_rw);

    // Convert raw file descriptor to File, then to BufReader
    let file = unsafe { File::from_raw_fd(fd_rd) };
    let mut reader = BufReader::new(file);

    let mut buffer = String::new();
    match reader.read_to_string(&mut buffer) {
        Ok(0) => {
            eprintln!("No data read from pipe; no access violation raised as expected.");
        }
        Ok(_) => {
            // If any data was read, log it
            eprint!("Access violation logged:\n{buffer}");
            return Err(TestError("Unexpected access violation logged.".to_string()));
        }
        Err(e) => {
            // Error reading from the buffer
            eprintln!("Error reading from pipe: {e}");
            return Err(TestError(format!("Error reading from pipe: {e}")));
        }
    }

    let status = child.wait().expect("wait for syd");
    assert_status_code!(status, nix::libc::EPERM);

    Ok(())
}

fn test_syd_setresgid_nobody_safesetid_allow_1() -> TestResult {
    if !*CAP_SETGID {
        eprintln!("Test requires CAP_SETGID, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }

    // SafeSetID is enabled with GID transition defined.
    // The syscall must succeed.
    let uid = Uid::current();
    let status = syd()
        .env("SYD_DO", "setresgid")
        .arg("-plib")
        .args([
            "-mtrace/allow_safe_setid:1",
            &format!("-msetgid+{uid}:nobody"),
            "--",
            &SYD_DO,
            "-1",
            "-1",
            "65534",
        ])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_setresgid_nobody_safesetid_allow_2() -> TestResult {
    if !*CAP_SETGID {
        eprintln!("Test requires CAP_SETGID, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }

    // SafeSetID is enabled with GID transition defined.
    // The syscall must succeed.
    let uid = Uid::current();
    let status = syd()
        .env("SYD_DO", "setresgid")
        .arg("-plib")
        .args([
            "-mtrace/allow_safe_setid:1",
            &format!("-msetgid+{uid}:nobody"),
            "--",
            &SYD_DO,
            "-1",
            "65534",
            "-1",
        ])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_setresgid_nobody_safesetid_allow_3() -> TestResult {
    if !*CAP_SETGID {
        eprintln!("Test requires CAP_SETGID, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }

    // SafeSetID is enabled with GID transition defined.
    // The syscall must succeed.
    let uid = Uid::current();
    let status = syd()
        .env("SYD_DO", "setresgid")
        .arg("-plib")
        .args([
            "-mtrace/allow_safe_setid:1",
            &format!("-msetgid+{uid}:nobody"),
            "--",
            &SYD_DO,
            "65534",
            "-1",
            "-1",
        ])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_setresgid_nobody_safesetid_allow_4() -> TestResult {
    if !*CAP_SETGID {
        eprintln!("Test requires CAP_SETGID, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }

    // SafeSetID is enabled with GID transition defined.
    // The syscall must succeed.
    let uid = Uid::current();
    let status = syd()
        .env("SYD_DO", "setresgid")
        .arg("-plib")
        .args([
            "-mtrace/allow_safe_setid:1",
            &format!("-msetgid+{uid}:nobody"),
            "--",
            &SYD_DO,
            "-1",
            "65534",
            "65534",
        ])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_setresgid_nobody_safesetid_allow_5() -> TestResult {
    if !*CAP_SETGID {
        eprintln!("Test requires CAP_SETGID, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }

    // SafeSetID is enabled with GID transition defined.
    // The syscall must succeed.
    let uid = Uid::current();
    let status = syd()
        .env("SYD_DO", "setresgid")
        .arg("-plib")
        .args([
            "-mtrace/allow_safe_setid:1",
            &format!("-msetgid+{uid}:nobody"),
            "--",
            &SYD_DO,
            "65534",
            "65534",
            "-1",
        ])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_setresgid_nobody_safesetid_allow_6() -> TestResult {
    if !*CAP_SETGID {
        eprintln!("Test requires CAP_SETGID, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }

    // SafeSetID is enabled with GID transition defined.
    // The syscall must succeed.
    let uid = Uid::current();
    let status = syd()
        .env("SYD_DO", "setresgid")
        .arg("-plib")
        .args([
            "-mtrace/allow_safe_setid:1",
            &format!("-msetgid+{uid}:nobody"),
            "--",
            &SYD_DO,
            "65534",
            "-1",
            "65534",
        ])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_setresgid_nobody_safesetid_allow_7() -> TestResult {
    if !*CAP_SETGID {
        eprintln!("Test requires CAP_SETGID, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }

    // SafeSetID is enabled with GID transition defined.
    // The syscall must succeed.
    let uid = Uid::current();
    let status = syd()
        .env("SYD_DO", "setresgid")
        .arg("-plib")
        .args([
            "-mtrace/allow_safe_setid:1",
            &format!("-msetgid+{uid}:nobody"),
            "--",
            &SYD_DO,
            "65534",
            "65534",
            "65534",
        ])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

// Checks socket domain restrictions
fn test_syd_socket_domain_restrictions() -> TestResult {
    let allows = [
        (nix::libc::AF_UNIX, nix::libc::SOCK_DGRAM, 0),
        (nix::libc::AF_UNIX, nix::libc::SOCK_STREAM, 0),
        (nix::libc::AF_INET, nix::libc::SOCK_DGRAM, 0),
        (nix::libc::AF_INET, nix::libc::SOCK_STREAM, 0),
        (nix::libc::AF_INET6, nix::libc::SOCK_DGRAM, 0),
        (nix::libc::AF_INET6, nix::libc::SOCK_STREAM, 0),
    ];
    let denies = [
        // Do not add privileged sockets here.
        (
            nix::libc::AF_NETLINK,
            nix::libc::SOCK_DGRAM,
            nix::libc::NETLINK_GENERIC,
        ),
        (
            nix::libc::AF_NETLINK,
            nix::libc::SOCK_DGRAM,
            nix::libc::NETLINK_ROUTE,
        ),
    ];
    let kcapis = [(nix::libc::AF_ALG, nix::libc::SOCK_SEQPACKET, 0)];

    for (domain, ty, proto) in &allows {
        let status = syd()
            .env("SYD_DO", "socket")
            .arg("-plib")
            .args([
                "--",
                &SYD_DO,
                &format!("{domain}"),
                &format!("{ty}"),
                &format!("{proto}"),
            ])
            .status()
            .expect("execute syd");
        assert_status_ok!(status);
    }

    for (domain, ty, proto) in &denies {
        let status = syd()
            .env("SYD_DO", "socket")
            .arg("-plib")
            .args([
                "--",
                &SYD_DO,
                &format!("{domain}"),
                &format!("{ty}"),
                &format!("{proto}"),
            ])
            .status()
            .expect("execute syd");
        assert_status_code!(status, nix::libc::EAFNOSUPPORT);

        let status = syd()
            .env("SYD_DO", "socket")
            .arg("-plib")
            .arg("-mtrace/allow_unsupp_socket:1")
            .args([
                "--",
                &SYD_DO,
                &format!("{domain}"),
                &format!("{ty}"),
                &format!("{proto}"),
            ])
            .status()
            .expect("execute syd");
        assert_status_ok!(status);
    }

    for (domain, ty, proto) in &kcapis {
        let status = syd()
            .env("SYD_DO", "socket")
            .arg("-plib")
            .args([
                "--",
                &SYD_DO,
                &format!("{domain}"),
                &format!("{ty}"),
                &format!("{proto}"),
            ])
            .status()
            .expect("execute syd");
        assert_status_code!(status, nix::libc::EAFNOSUPPORT);

        let status = syd()
            .env("SYD_DO", "socket")
            .arg("-plib")
            .arg("-mtrace/allow_safe_kcapi:1")
            .args([
                "--",
                &SYD_DO,
                &format!("{domain}"),
                &format!("{ty}"),
                &format!("{proto}"),
            ])
            .status()
            .expect("execute syd");
        // Careful, kernel may return EAFNOSUPPORT
        // if CONFIG_CRYPTO_USER_API is either not
        // enabled or compiled as a module and the
        // module is not yet loaded.
        assert_status_code_matches!(status, 0 | nix::libc::EAFNOSUPPORT);

        let status = syd()
            .env("SYD_DO", "socket")
            .arg("-plib")
            .arg("-mtrace/allow_unsupp_socket:1")
            .arg("-mtrace/allow_safe_kcapi:1")
            .args([
                "--",
                &SYD_DO,
                &format!("{domain}"),
                &format!("{ty}"),
                &format!("{proto}"),
            ])
            .status()
            .expect("execute syd");
        // Careful, kernel may return EAFNOSUPPORT
        // if CONFIG_CRYPTO_USER_API is either not
        // enabled or compiled as a module and the
        // module is not yet loaded.
        assert_status_code_matches!(status, 0 | nix::libc::EAFNOSUPPORT);
    }

    Ok(())
}

// Checks environment filtering
fn test_syd_environment_filter() -> TestResult {
    if !is_program_available("sh") {
        eprintln!("Test requires sh, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }
    const ENV: &str = "SAFE";
    env::set_var(ENV, "/var/empty");

    // Step 1: Allow by default
    let output = syd()
        .arg("-plib")
        .arg("--")
        .args(["/bin/sh", "-c", &format!("echo ${ENV}")])
        .output()
        .expect("execute syd");
    let output = String::from_utf8_lossy(&output.stdout);
    let output = output.trim_end();
    assert!(output == "/var/empty", "output1:{output}");

    // Step 2: Override with -evar=val
    let output = syd()
        .arg("-plib")
        .arg(&format!("-e{ENV}=/var/empty:/var/empty"))
        .arg("--")
        .args(["/bin/sh", "-c", &format!("echo ${ENV}")])
        .output()
        .expect("execute syd");
    let output = String::from_utf8_lossy(&output.stdout);
    let output = output.trim_end();
    assert!(output == "/var/empty:/var/empty", "output2:{output}");

    // Step 3: Unset with -evar
    let output = syd()
        .arg("-plib")
        .arg(&format!("-e{ENV}"))
        .arg("--")
        .args(["/bin/sh", "-c", &format!("echo ${ENV}")])
        .output()
        .expect("execute syd");
    let output = String::from_utf8_lossy(&output.stdout);
    let output = output.trim_end();
    assert!(output.is_empty(), "output3:{output}");

    // Step 4: Pass-through with -evar=
    let output = syd()
        .arg("-plib")
        .arg(&format!("-e{ENV}="))
        .arg("--")
        .args(["/bin/sh", "-c", &format!("echo ${ENV}")])
        .output()
        .expect("execute syd");
    let output = String::from_utf8_lossy(&output.stdout);
    let output = output.trim_end();
    assert!(output == "/var/empty", "output4:{output}");

    env::remove_var(ENV);

    Ok(())
}

// Checks environment hardening and -e pass-through.
fn test_syd_environment_harden() -> TestResult {
    if !is_program_available("sh") {
        eprintln!("Test requires sh, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }
    const ENV: &str = "LD_LIBRARY_PATH";
    env::set_var(ENV, "/var/empty");

    // Step 1: Deny by default
    let output = syd()
        .arg("-plib")
        .arg("--")
        .args(["/bin/sh", "-c", &format!("echo ${ENV}")])
        .output()
        .expect("execute syd");
    let output = String::from_utf8_lossy(&output.stdout);
    let output = output.trim_end();
    assert!(output.is_empty(), "output1:{output}");

    // Step 2: Override with -evar=val
    let output = syd()
        .arg("-plib")
        .arg(&format!("-e{ENV}=/var/empty:/var/empty"))
        .arg("--")
        .args(["/bin/sh", "-c", &format!("echo ${ENV}")])
        .output()
        .expect("execute syd");
    let output = String::from_utf8_lossy(&output.stdout);
    let output = output.trim_end();
    assert!(output == "/var/empty:/var/empty", "output2:{output}");

    // Step 3: Unset with -evar
    let output = syd()
        .arg("-plib")
        .arg(&format!("-e{ENV}"))
        .arg("--")
        .args(["/bin/sh", "-c", &format!("echo ${ENV}")])
        .output()
        .expect("execute syd");
    let output = String::from_utf8_lossy(&output.stdout);
    let output = output.trim_end();
    assert!(output.is_empty(), "output3:{output}");

    // Step 4: Pass-through with -evar=
    let output = syd()
        .arg("-plib")
        .arg(&format!("-e{ENV}="))
        .arg("--")
        .args(["/bin/sh", "-c", &format!("echo ${ENV}")])
        .output()
        .expect("execute syd");
    let output = String::from_utf8_lossy(&output.stdout);
    let output = output.trim_end();
    assert!(output == "/var/empty", "output4:{output}");

    // Step 5: Allow with -m trace/allow_unsafe_env:1
    let output = syd()
        .arg("-plib")
        .arg("-mtrace/allow_unsafe_env:1")
        .arg("--")
        .args(["/bin/sh", "-c", &format!("echo ${ENV}")])
        .output()
        .expect("execute syd");
    let output = String::from_utf8_lossy(&output.stdout);
    let output = output.trim_end();
    assert!(output == "/var/empty", "output:{output}");

    // Step 6: Toggle -m trace/allow_unsafe_env
    let output = syd()
        .arg("-plib")
        .arg("-mtrace/allow_unsafe_env:1")
        .arg("-mtrace/allow_unsafe_env:0")
        .arg("--")
        .args(["/bin/sh", "-c", &format!("echo ${ENV}")])
        .output()
        .expect("execute syd");
    let output = String::from_utf8_lossy(&output.stdout);
    let output = output.trim_end();
    assert!(output.is_empty(), "output1:{output}");

    env::remove_var(ENV);

    Ok(())
}

// Tests if `lock:on` command disables access to `/dev/syd`.
fn test_syd_lock() -> TestResult {
    if !is_program_available("bash") {
        eprintln!("Test requires bash, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }
    eprintln!("+ bash -c \"test -e /dev/syd\"");
    let status = syd()
        .arg("-plib")
        .args(["--", "/bin/bash", "-c", "test -c /dev/syd"])
        .status()
        .expect("execute syd");
    assert_status_not_ok!(status);

    eprintln!("+ bash -c \"test -e /dev/syd\"");
    let status = syd()
        .arg("-plib")
        .args(["-mlock:on", "--", "/bin/bash", "-c", "test -c /dev/syd"])
        .status()
        .expect("execute syd");
    assert_status_not_ok!(status);

    Ok(())
}

// Tests if `lock:exec` locks the sandbox for all except the exec child.
fn test_syd_lock_exec() -> TestResult {
    if !is_program_available("bash") {
        eprintln!("Test requires bash, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }
    let status = syd()
        .arg("-plib")
        .args(["-mlock:exec", "--", "/bin/bash", "-c", "test -e /dev/syd"])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    let status = syd()
        .arg("-plib")
        .args([
            "-mlock:exec",
            "--",
            "/bin/bash",
            "-c",
            "( test -e /dev/syd )",
        ])
        .status()
        .expect("execute syd");
    assert_status_not_ok!(status);

    Ok(())
}

// Tests if `-mlock:on` prevents subsequent -m CLI args.
fn test_syd_lock_prevents_further_cli_args() -> TestResult {
    if !is_program_available("true") {
        eprintln!("Test requires true, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }

    let status = syd()
        .arg("-plib")
        .args(["-mlock:on", "-mlock:off", "--", "true"])
        .status()
        .expect("execute syd");
    assert_status_not_ok!(status);

    let status = syd()
        .arg("-plib")
        .args([
            "-mtrace/allow_unsafe_exec:0",
            "-mlock:on",
            "-mlock:off",
            "--",
            "true",
        ])
        .status()
        .expect("execute syd");
    assert_status_not_ok!(status);

    let status = syd()
        .arg("-plib")
        .args(["-mlock:on", "-mtrace/allow_unsafe_exec:1", "--", "true"])
        .status()
        .expect("execute syd");
    assert_status_not_ok!(status);

    // Check for warning that paludis profile was not parsed.
    // Syd does not bail out in this case, just warns and continues.
    let (fd_rd, fd_rw) = match pipe() {
        Ok((fd_rd, fd_rw)) => (fd_rd, fd_rw),
        Err(errno) => return Err(TestError(format!("error creating pipe: {errno}!"))),
    };

    let mut child = syd()
        .env("SYD_LOG", "warn")
        .env("SYD_LOG_FD", fd_rw.to_string())
        .arg("-plib")
        .args(["-mlock:on", "-ppaludis", "--", "true"])
        .spawn()
        .expect("execute syd");
    let _ = close(fd_rw);

    let status = child.wait().expect("wait for syd");
    assert_status_ok!(status);

    // Convert raw file descriptor to File, then to BufReader
    let file = unsafe { File::from_raw_fd(fd_rd) };
    let mut reader = BufReader::new(file);

    let mut buffer = String::new();
    match reader.read_to_string(&mut buffer) {
        Ok(0) => {
            eprintln!("No data read from pipe; expected access violation.");
            return Err(TestError("Expected parse error not logged.".to_string()));
        }
        Ok(_) => {
            // If any data was read, log it
            eprint!("Read data from pipe:\n{buffer}");
            if buffer.contains("sandbox is locked") {
                eprintln!("Sandbox was locked as expected.");
            } else {
                return Err(TestError("Expected parse error not logged".to_string()));
            }
        }
        Err(e) => {
            // Error reading from the buffer
            eprintln!("Error reading from pipe: {e}");
            return Err(TestError(format!("Error reading from pipe: {e}")));
        }
    }

    // Check for warning that /dev/null file was not parsed.
    // Syd does not bail out in this case, just warns and continues.
    let (fd_rd, fd_rw) = match pipe() {
        Ok((fd_rd, fd_rw)) => (fd_rd, fd_rw),
        Err(errno) => return Err(TestError(format!("error creating pipe: {errno}!"))),
    };

    let mut child = syd()
        .env("SYD_LOG", "warn")
        .env("SYD_LOG_FD", fd_rw.to_string())
        .arg("-plib")
        .args(["-mlock:on", "-P/dev/null", "--", "true"])
        .spawn()
        .expect("execute syd");
    let _ = close(fd_rw);

    let status = child.wait().expect("wait for syd");
    assert_status_ok!(status);

    // Convert raw file descriptor to File, then to BufReader
    let file = unsafe { File::from_raw_fd(fd_rd) };
    let mut reader = BufReader::new(file);

    let mut buffer = String::new();
    match reader.read_to_string(&mut buffer) {
        Ok(0) => {
            eprintln!("No data read from pipe; expected access violation.");
            return Err(TestError("Expected parse error not logged.".to_string()));
        }
        Ok(_) => {
            // If any data was read, log it
            eprint!("Read data from pipe:\n{buffer}");
            if buffer.contains("sandbox is locked") {
                eprintln!("Sandbox was locked as expected.");
            } else {
                return Err(TestError("Expected parse error not logged".to_string()));
            }
        }
        Err(e) => {
            // Error reading from the buffer
            eprintln!("Error reading from pipe: {e}");
            return Err(TestError(format!("Error reading from pipe: {e}")));
        }
    }

    Ok(())
}

// Tests if `-mlock:on` prevents subsequent configuration file edits.
fn test_syd_lock_prevents_further_cfg_items() -> TestResult {
    if !is_program_available("true") {
        eprintln!("Test requires true, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }

    let conf = "lock:on\nlock:off\n";
    let mut file = File::create("conf.syd-3")?;
    write!(file, "{conf}")?;
    drop(file);

    // Check for warning that include line was not parsed.
    // Syd does not bail out in this case, just warns and continues.
    let (fd_rd, fd_rw) = match pipe() {
        Ok((fd_rd, fd_rw)) => (fd_rd, fd_rw),
        Err(errno) => return Err(TestError(format!("error creating pipe: {errno}!"))),
    };

    let mut child = syd()
        .env("SYD_LOG", "warn")
        .env("SYD_LOG_FD", fd_rw.to_string())
        .arg("-plib")
        .args(["-P./conf.syd-3", "--", "true"])
        .spawn()
        .expect("execute syd");
    let _ = close(fd_rw);

    let status = child.wait().expect("wait for syd");
    assert_status_ok!(status);

    // Convert raw file descriptor to File, then to BufReader
    let file = unsafe { File::from_raw_fd(fd_rd) };
    let mut reader = BufReader::new(file);

    let mut buffer = String::new();
    match reader.read_to_string(&mut buffer) {
        Ok(0) => {
            eprintln!("No data read from pipe; expected access violation.");
            return Err(TestError("Expected parse error not logged.".to_string()));
        }
        Ok(_) => {
            // If any data was read, log it
            eprint!("Read data from pipe:\n{buffer}");
            if buffer.contains("sandbox is locked") {
                eprintln!("Sandbox was locked as expected.");
            } else {
                return Err(TestError("Expected parse error not logged".to_string()));
            }
        }
        Err(e) => {
            // Error reading from the buffer
            eprintln!("Error reading from pipe: {e}");
            return Err(TestError(format!("Error reading from pipe: {e}")));
        }
    }

    Ok(())
}

// Tests if `-mlock:on` prevents subsequent file includes.
fn test_syd_lock_prevents_further_inc_items() -> TestResult {
    if !is_program_available("true") {
        eprintln!("Test requires true, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }

    let conf = "sandbox/read:on\n";
    let mut file = File::create("conf-1.syd-3")?;
    write!(file, "{conf}")?;
    drop(file);

    let conf = "lock:on\ninclude conf-1.syd-3\n";
    let mut file = File::create("conf-2.syd-3")?;
    write!(file, "{conf}")?;
    drop(file);

    // Check for warning that include line was not parsed.
    // Syd does not bail out in this case, just warns and continues.
    let (fd_rd, fd_rw) = match pipe() {
        Ok((fd_rd, fd_rw)) => (fd_rd, fd_rw),
        Err(errno) => return Err(TestError(format!("error creating pipe: {errno}!"))),
    };

    let mut child = syd()
        .env("SYD_LOG", "warn")
        .env("SYD_LOG_FD", fd_rw.to_string())
        .arg("-plib")
        .args(["-P./conf-2.syd-3", "--", "true"])
        .spawn()
        .expect("execute syd");
    let _ = close(fd_rw);

    let status = child.wait().expect("wait for syd");
    assert_status_ok!(status);

    // Convert raw file descriptor to File, then to BufReader
    let file = unsafe { File::from_raw_fd(fd_rd) };
    let mut reader = BufReader::new(file);

    let mut buffer = String::new();
    match reader.read_to_string(&mut buffer) {
        Ok(0) => {
            eprintln!("No data read from pipe; expected access violation.");
            return Err(TestError("Expected parse error not logged.".to_string()));
        }
        Ok(_) => {
            // If any data was read, log it
            eprint!("Read data from pipe:\n{buffer}");
            if buffer.contains("sandbox is locked") {
                eprintln!("Sandbox was locked as expected.");
            } else {
                return Err(TestError("Expected parse error not logged".to_string()));
            }
        }
        Err(e) => {
            // Error reading from the buffer
            eprintln!("Error reading from pipe: {e}");
            return Err(TestError(format!("Error reading from pipe: {e}")));
        }
    }

    Ok(())
}

// Tests if syd-chk works to check if process is running under syd.
fn test_syd_chk() -> TestResult {
    eprintln!("+ syd-chk");
    let status = std::process::Command::new(&*SYD_CHK)
        .status()
        .expect("execute syd");
    assert_status_not_ok!(status);

    let status = syd()
        .arg("-plib")
        .args(["--", &SYD_CHK])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_cmd_exec_with_lock_default() -> TestResult {
    if !is_program_available("bash") {
        eprintln!("Test requires sh, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }

    let status = syd()
        .args(["-plib", "--", "bash", "-cex"])
        .arg(format!(
            "
#!/bin/bash
: > test
cat > exec.sh <<EOF
#!/bin/sh -ex
# Careful here, cmd/exec changes CWD to /.
echo OK > $PWD/test
exit 42
EOF
chmod +x exec.sh
test -c \"$({} $PWD/exec.sh)\"
sleep 5
test -s ./test
true
",
            *SYD_EXEC
        ))
        .status()
        .expect("execute syd");
    assert_status_not_ok!(status);

    Ok(())
}

fn test_syd_cmd_exec_with_lock_on() -> TestResult {
    if !is_program_available("bash") {
        eprintln!("Test requires sh, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }

    let status = syd()
        .args(["-plib", "-mlock:on", "--", "bash", "-cex"])
        .arg(format!(
            "
#!/bin/bash
: > test
cat > exec.sh <<EOF
#!/bin/sh -ex
# Careful here, cmd/exec changes CWD to /.
echo OK > $PWD/test
exit 42
EOF
chmod +x exec.sh
test -c \"$({} $PWD/exec.sh)\"
sleep 5
test -s ./test
true
",
            *SYD_EXEC
        ))
        .status()
        .expect("execute syd");
    assert_status_not_ok!(status);

    Ok(())
}

fn test_syd_cmd_exec_with_lock_off_1() -> TestResult {
    if !is_program_available("bash") {
        eprintln!("Test requires sh, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }

    let status = syd()
        .args(["-plib", "-mlock:off", "--", "bash", "-cex"])
        .arg(format!(
            "
#!/bin/bash
: > test
cat > exec.sh <<EOF
#!/bin/sh -ex
# Careful here, cmd/exec changes CWD to /.
echo OK > $PWD/test
exit 42
EOF
chmod +x exec.sh
test -c \"$({} $PWD/exec.sh)\"
sleep 5
test -s ./test
true
",
            *SYD_EXEC
        ))
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_cmd_exec_with_lock_off_2() -> TestResult {
    if !is_program_available("bash") {
        eprintln!("Test requires sh, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }

    let status = syd()
        .args(["-plib", "-mlock:off", "--", "bash", "-cex"])
        .arg(format!(
            "
#!/bin/bash
: > test
cat > exec.sh <<EOF
#!/bin/sh -ex
# Careful here, cmd/exec changes CWD to /.
echo OK > $PWD/test
exit 42
EOF
chmod +x exec.sh
(
    test -c \"$({} $PWD/exec.sh)\"
) &
wait
sleep 5
test -s ./test
true
",
            *SYD_EXEC
        ))
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_cmd_exec_with_lock_exec_1() -> TestResult {
    if !is_program_available("bash") {
        eprintln!("Test requires sh, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }

    let status = syd()
        .args(["-plib", "-mlock:exec", "--", "bash", "-cex"])
        .arg(format!(
            "
#!/bin/bash
: > test
cat > exec.sh <<EOF
#!/bin/sh -ex
# Careful here, cmd/exec changes CWD to /.
echo OK > $PWD/test
exit 42
EOF
chmod +x exec.sh
test -c \"$({} $PWD/exec.sh)\"
sleep 5
test -s ./test
true
",
            *SYD_EXEC
        ))
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_cmd_exec_with_lock_exec_2() -> TestResult {
    if !is_program_available("bash") {
        eprintln!("Test requires sh, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }

    let status = syd()
        .args(["-plib", "-mlock:exec", "--", "bash", "-cex"])
        .arg(format!(
            "
#!/bin/bash
: > test
cat > exec.sh <<EOF
#!/bin/sh -ex
# Careful here, cmd/exec changes CWD to /.
echo OK > $PWD/test
exit 42
EOF
chmod +x exec.sh
(
    test -c \"$({} $PWD/exec.sh)\"
) &
wait
sleep 5
test -s ./test
true
",
            *SYD_EXEC
        ))
        .status()
        .expect("execute syd");
    assert_status_not_ok!(status);

    Ok(())
}

fn test_syd_parse_config() -> TestResult {
    if !is_program_available("true") {
        eprintln!("Test requires true, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }

    let conf = "lock:on\n";
    let mut file = File::create("conf.syd-3")?;
    write!(file, "{conf}")?;
    drop(file);

    let status = syd()
        .arg("-plib")
        .arg("-P./conf.syd-3")
        .args(["--", "true"])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    let status = syd()
        .arg("-plib")
        .arg("-P./conf.syd-3")
        .arg("-mlock:exec")
        .args(["--", "true"])
        .status()
        .expect("execute syd");
    assert_status_not_ok!(status);

    Ok(())
}

fn test_syd_include_config() -> TestResult {
    if !is_program_available("true") {
        eprintln!("Test requires true, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }

    let idir = " Heavy\tis\tthe\troot\tof\tlight.\tStill\tis\tthe\tmaster\tof\tmoving ";
    create_dir_all(idir)?;

    let name = " Heavy is the root of light.\tStill is the master of moving";
    let conf = "lock:${SYD_LOCK_STATE}\n";
    let mut file = File::create(format!("./{idir}/{name}.syd-3"))?;
    write!(file, "{conf}")?;
    drop(file);

    let conf = format!("include {name}\n");
    let mut file = File::create(format!("./{idir}/conf.syd-3"))?;
    write!(file, "{conf}")?;
    drop(file);

    let status = syd()
        .env("SYD_LOCK_STATE", "on")
        .arg("-plib")
        .arg(&format!("-P./{idir}/conf.syd-3"))
        .args(["--", "true"])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    let status = syd()
        .env("SYD_LOCK_STATE", "on")
        .arg("-plib")
        .arg(&format!("-P./{idir}/conf.syd-3"))
        .arg("-mlock:exec")
        .args(["--", "true"])
        .status()
        .expect("execute syd");
    assert_status_not_ok!(status);

    Ok(())
}

fn test_syd_shellexpand_01() -> TestResult {
    if !is_program_available("true") {
        eprintln!("Test requires true, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }

    let conf = "allow/write+${SYD_TEST_OOPS}/***\n";
    let mut file = File::create("conf.syd-3")?;
    write!(file, "{conf}")?;
    drop(file);

    let status = syd()
        .env("SYD_TEST_OOPS", "/home")
        .arg("-plib")
        .arg("-P./conf.syd-3")
        .args(["--", "true"])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_shellexpand_02() -> TestResult {
    if !is_program_available("true") {
        eprintln!("Test requires true, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }

    let conf = "allow/write+${SYD_TEST_OOPS}/***\n";
    let mut file = File::create("conf.syd-3")?;
    write!(file, "{conf}")?;
    drop(file);

    let status = syd()
        .env_remove("SYD_TEST_OOPS")
        .arg("-plib")
        .arg("-P./conf.syd-3")
        .args(["--", "true"])
        .status()
        .expect("execute syd");
    assert_status_not_ok!(status);

    Ok(())
}

fn test_syd_shellexpand_03() -> TestResult {
    if !is_program_available("true") {
        eprintln!("Test requires true, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }

    let conf = "allow/write+${SYD_TEST_OOPS:-/home}/***\n";
    let mut file = File::create("conf.syd-3")?;
    write!(file, "{conf}")?;
    drop(file);

    let status = syd()
        .env_remove("SYD_TEST_OOPS")
        .arg("-plib")
        .arg("-P./conf.syd-3")
        .args(["--", "true"])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_shellexpand_04() -> TestResult {
    if !is_program_available("true") {
        eprintln!("Test requires true, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }

    let conf = "allow/write+${SYD_TEST_OOPS:-}/***\n";
    let mut file = File::create("conf.syd-3")?;
    write!(file, "{conf}")?;
    drop(file);

    let status = syd()
        .env_remove("SYD_TEST_OOPS")
        .arg("-plib")
        .arg("-P./conf.syd-3")
        .args(["--", "true"])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

// Tests if personality(2) locking works with trace/lock_personality.
fn test_syd_lock_personality() -> TestResult {
    // Check if the target architecture is 32-bit and exit if true
    #[cfg(not(target_pointer_width = "64"))]
    {
        eprintln!("Test requires 64-bit target, skippping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }

    let status = syd()
        .env("SYD_DO", "personality")
        .arg("-plib")
        .arg("-mtrace/lock_personality:1")
        .args(["--", &SYD_DO, "false"])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    let status = syd()
        .env("SYD_DO", "personality")
        .arg("-plib")
        .arg("-mtrace/lock_personality:0")
        .arg("-mtrace/lock_personality:1")
        .args(["--", &SYD_DO, "false"])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    let status = syd()
        .env("SYD_DO", "personality")
        .arg("-plib")
        .args(["--", &SYD_DO, "true"])
        .status()
        .expect("execute syd");
    // This and the rest of the asserts fail on GitLab CI.
    if !*GL_BUILD {
        assert_status_ok!(status);
    } else {
        ignore!(status.success(), "status:{status:?}");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }

    let status = syd()
        .env("SYD_DO", "personality")
        .arg("-plib")
        .arg("-mtrace/lock_personality:0")
        .args(["--", &SYD_DO, "true"])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    let status = syd()
        .env("SYD_DO", "personality")
        .arg("-plib")
        .arg("-mtrace/lock_personality:1")
        .arg("-mtrace/lock_personality:0")
        .args(["--", &SYD_DO, "true"])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

// Tests if MDWE protections work as expected.
fn test_syd_mdwe_protections() -> TestResult {
    let status = syd()
        .env("SYD_DO", "mdwe")
        .arg("-plib")
        .args(["--", &SYD_DO, "false"])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    let status = syd()
        .env("SYD_DO", "mdwe")
        .arg("-plib")
        .arg("-mtrace/allow_unsafe_memory:0")
        .args(["--", &SYD_DO, "false"])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    let status = syd()
        .env("SYD_DO", "mdwe")
        .arg("-plib")
        .arg("-mtrace/allow_unsafe_memory:1")
        .arg("-mtrace/allow_unsafe_memory:0")
        .args(["--", &SYD_DO, "false"])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    let status = syd()
        .env("SYD_DO", "mdwe")
        .arg("-plib")
        .arg("-mtrace/allow_unsafe_memory:1")
        .args(["--", &SYD_DO, "true"])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    let status = syd()
        .env("SYD_DO", "mdwe")
        .arg("-plib")
        .arg("-mtrace/allow_unsafe_memory:1")
        .arg("-mtrace/allow_unsafe_memory:0")
        .arg("-mtrace/allow_unsafe_memory:1")
        .args(["--", &SYD_DO, "true"])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

// Test if MDWE can be relaxed as expected.
fn test_syd_mdwe_relax() -> TestResult {
    if !is_program_available("luajit") {
        eprintln!("Test requires luajit, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }

    // Execute with default restrictions.
    // Expect LuaJIT to segfault.
    let status = syd()
        .arg("-plib")
        .args(["--", "luajit", "-e", "for i=1,1e5 do local a=i*2 end"])
        .status()
        .expect("execute syd");
    // This segfaults on x86-64, but exits with 1 on arm64.
    assert!(
        matches!(status.code().unwrap_or(127), 1 | EX_SIGSEGV),
        "status:{status:?}"
    );

    // Relax restrictions.
    // Expect LuaJIT to succeed.
    let status = syd()
        .arg("-plib")
        .arg("-mtrace/allow_unsafe_memory:1")
        .args(["--", "luajit", "-e", "for i=1,1e5 do local a=i*2 end"])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

// Test if mmap(NULL, MMAP_FIXED) is prevented.
fn test_syd_mmap_fixed_null() -> TestResult {
    let status = syd()
        .env("SYD_DO", "mmap_fixed_null")
        .arg("-plib")
        .args(["--", &SYD_DO])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_load_library() -> TestResult {
    if !is_program_available("cc") {
        eprintln!("Test requires cc, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }
    // Write code.
    let code = r#"
#include <stdio.h>
#include <unistd.h>

__attribute__((constructor))
void syd_init(void)
{
    printf("library initialized at pid %d\n", getpid());
}

int syd_main(void)
{
    printf("library loaded at pid %d\n", getpid());
    return 42;
}
"#;
    let mut file = File::create("load.c")?;
    write!(file, "{code}")?;

    // Compile code.
    Command::new("cc")
        .args([
            "-Wall", "-Wextra", "load.c", "-shared", "-o", "load.so", "-fPIC",
        ])
        .status()?;

    // Load code.
    // ENOSYS = Dynamic linking not supported.
    let status = syd()
        .args(["-plib", "--", "./load.so"])
        .status()
        .expect("execute syd");
    let code = status.code().unwrap_or(127);
    if code != nix::libc::ENOSYS {
        assert_status_code!(status, 42);
    } else {
        eprintln!("Dynamic linking not supported, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(()); // Skip rest of the tests.
    }

    // Try to load nonexisting library.
    let status = syd()
        .args(["-plib", "--", "./nolib.so"])
        .status()
        .expect("execute syd");
    assert_status_code!(status, nix::libc::ENOENT);

    // Try to load a library without the syd_main symbol.
    let code = r"int oops(void) { return 42; }";
    let mut file = File::create("load.c")?;
    write!(file, "{code}")?;
    Command::new("cc")
        .args([
            "-Wall", "-Wextra", "load.c", "-shared", "-o", "load.so", "-fPIC",
        ])
        .status()?;
    let status = syd()
        .args(["-plib", "--", "./load.so"])
        .status()
        .expect("execute syd");
    assert_status_code!(status, 128);

    Ok(())
}

fn test_syd_load_library_noexec() -> TestResult {
    if !is_program_available("cc") {
        eprintln!("Test requires cc, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }
    // Check for namespaces.
    match check_unshare() {
        Some(false) => {
            eprintln!("Test requires Linux namespaces, skipping!");
            env::set_var("SYD_TEST_SOFT_FAIL", "1");
            return Ok(());
        }
        None => {
            return Err(TestError(
                "Failed to test for Linux namespaces!".to_string(),
            ));
        }
        _ => {}
    };

    // Write code.
    let code = r#"
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

__attribute__((constructor))
void syd_init(void)
{
    printf("library initialized at pid %d\n", getpid());
}

int syd_main(void)
{
    printf("library loaded at pid %d\n", getpid());
    return 42;
}
"#;
    let mut file = File::create("load.c")?;
    write!(file, "{code}")?;

    // Compile code.
    Command::new("cc")
        .args([
            "-Wall", "-Wextra", "load.c", "-shared", "-o", "load.so", "-fPIC",
        ])
        .status()?;

    // Load code.
    // ENOSYS = Dynamic linking not supported.
    let status = syd()
        .args([
            "-plib",
            "-munshare/user:1",
            "-munshare/pid:1",
            "-mbind+/:/:noexec",
            "--",
            "./load.so",
        ])
        .status()
        .expect("execute syd");
    let code = status.code().unwrap_or(127);
    if code != nix::libc::ENOSYS {
        assert_status_code!(status, 42);
    } else {
        eprintln!("Dynamic linking not supported, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
    }

    Ok(())
}

fn test_syd_load_library_abort_after_load() -> TestResult {
    if !is_program_available("cc") {
        eprintln!("Test requires cc, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }

    // Write code.
    let code = r#"
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

__attribute__((constructor))
void syd_init(void)
{
    printf("library initialized at pid %d\n", getpid());
}

int syd_main(void)
{
    printf("library loaded at pid %d\n", getpid());
    abort();
}
"#;
    let mut file = File::create("load.c")?;
    write!(file, "{code}")?;

    // Compile code.
    Command::new("cc")
        .args([
            "-Wall", "-Wextra", "load.c", "-shared", "-o", "load.so", "-fPIC",
        ])
        .status()?;

    // Load code.
    // ENOSYS = Dynamic linking not supported.
    const XABORT: i32 = 128 + nix::libc::SIGABRT;
    const XSEGV: i32 = 128 + nix::libc::SIGSEGV;
    let status = syd()
        .args(["-plib", "--", "./load.so"])
        .status()
        .expect("execute syd");
    let code = status.code().unwrap_or(127);
    if code != nix::libc::ENOSYS {
        assert!(
            matches!(code, XABORT | XSEGV),
            "code:{code} status:{status:?}"
        );
    } else {
        eprintln!("Dynamic linking not supported, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
    }

    Ok(())
}

fn test_syd_load_library_abort_at_startup() -> TestResult {
    if !is_program_available("cc") {
        eprintln!("Test requires cc, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }

    // Write code.
    let code = r#"
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

__attribute__((constructor))
void syd_init(void)
{
    printf("library initialized at pid %d\n", getpid());
    abort(); // This takes down syd with it...
}

int syd_main(void)
{
    printf("library loaded at pid %d\n", getpid());
    return 42;
}
"#;
    let mut file = File::create("load.c")?;
    write!(file, "{code}")?;

    // Compile code.
    Command::new("cc")
        .args([
            "-Wall", "-Wextra", "load.c", "-shared", "-o", "load.so", "-fPIC",
        ])
        .status()?;

    // Load code.
    // ENOSYS = Dynamic linking not supported.
    let status = syd()
        .args(["-plib", "--", "./load.so"])
        .status()
        .expect("execute syd");
    let code = status.code().unwrap_or(127);
    if code != nix::libc::ENOSYS {
        let sign = status.signal().unwrap_or(0);
        assert!(
            matches!(sign, nix::libc::SIGABRT | nix::libc::SIGSEGV),
            "code:{code} status:{status:?}"
        );
    } else {
        eprintln!("Dynamic linking not supported, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
    }

    Ok(())
}

fn test_syd_load_library_check_fd_leaks_bare() -> TestResult {
    if !is_program_available("cc") {
        eprintln!("Test requires cc, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }
    if !is_program_available("sh") {
        eprintln!("Test requires sh, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }

    let status = Command::new("sh")
        .arg("-cex")
        .arg(
            r##"
#!/bin/sh
cat > load.c <<EOF
#include <dirent.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

int syd_main(void) {
    DIR *dir;
    struct dirent *entry;
    int fd_leaks = 0, dir_fd;

    // Open the directory containing file descriptors
    dir = opendir("/proc/self/fd");
    if (!dir) {
        perror("Failed to open /proc/self/fd");
        return -1; // Return -1 in case of error
    }

    // Get the file descriptor for the directory stream
    dir_fd = dirfd(dir);
    if (dir_fd == -1) {
        perror("Failed to get file descriptor for directory");
        closedir(dir);
        return -1; // Return -1 in case of error
    }

    // Iterate over all entries in the directory
    while ((entry = readdir(dir)) != NULL) {
        int fd;
        char *end;

        // Convert the name of the entry to an integer
        fd = strtol(entry->d_name, &end, 10);
        if (*end != '\0' || entry->d_name == end) continue; // Skip non-integer entries

        // Build the path to the symbolic link for the file descriptor
        char link_path[4096];
        char target_path[4096];
        snprintf(link_path, sizeof(link_path), "/proc/self/fd/%d", fd);

        ssize_t len = readlink(link_path, target_path, sizeof(target_path) - 1);
        if (len > 0) {
            target_path[len] = '\0'; // Ensure null termination
            // We ignore standard input, output, and error which are 0, 1, and 2
            if (fd <= 2) {
                printf("Ignoring standard open fd %d -> %s...\n", fd, target_path);
            } else if (fd == dir_fd) {
                printf("Ignoring fd to current directory fd %d -> %s...\n", fd, target_path);
            } else {
                printf("!!! Leaked file descriptor %d -> %s !!!\n", fd, target_path);
                fd_leaks++;
            }
        } else {
            perror("Failed to read link");
            fd_leaks++;
        }
    }

    closedir(dir);
    return fd_leaks;
}
EOF

cc -Wall -Wextra load.c -shared -o load.so -fPIC
    "##,
        )
        .status()
        .expect("execute sh");
    assert_status_ok!(status);

    // Load code.
    // ENOSYS = Dynamic linking not supported.
    let status = syd()
        .args(["-plib", "--", "./load.so"])
        .status()
        .expect("execute syd");
    let code = status.code().unwrap_or(127);
    if code != nix::libc::ENOSYS {
        assert_status_ok!(status);
    } else {
        eprintln!("Dynamic linking not supported, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
    }

    Ok(())
}

fn test_syd_load_library_check_fd_leaks_wrap() -> TestResult {
    match check_unshare() {
        Some(false) => {
            eprintln!("Test requires Linux namespaces, skipping!");
            env::set_var("SYD_TEST_SOFT_FAIL", "1");
            return Ok(());
        }
        None => {
            return Err(TestError(
                "Failed to test for Linux namespaces!".to_string(),
            ));
        }
        _ => {}
    };
    if !is_program_available("cc") {
        eprintln!("Test requires cc, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }
    if !is_program_available("sh") {
        eprintln!("Test requires sh, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }

    let status = Command::new("sh")
        .arg("-cex")
        .arg(
            r##"
#!/bin/sh
cat > load.c <<EOF
#include <dirent.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

int syd_main(void) {
    DIR *dir;
    struct dirent *entry;
    int fd_leaks = 0, dir_fd;

    // Open the directory containing file descriptors
    dir = opendir("/proc/self/fd");
    if (!dir) {
        perror("Failed to open /proc/self/fd");
        return -1; // Return -1 in case of error
    }

    // Get the file descriptor for the directory stream
    dir_fd = dirfd(dir);
    if (dir_fd == -1) {
        perror("Failed to get file descriptor for directory");
        closedir(dir);
        return -1; // Return -1 in case of error
    }

    // Iterate over all entries in the directory
    while ((entry = readdir(dir)) != NULL) {
        int fd;
        char *end;

        // Convert the name of the entry to an integer
        fd = strtol(entry->d_name, &end, 10);
        if (*end != '\0' || entry->d_name == end) continue; // Skip non-integer entries

        // Build the path to the symbolic link for the file descriptor
        char link_path[4096];
        char target_path[4096];
        snprintf(link_path, sizeof(link_path), "/proc/self/fd/%d", fd);

        ssize_t len = readlink(link_path, target_path, sizeof(target_path) - 1);
        if (len > 0) {
            target_path[len] = '\0'; // Ensure null termination
            // We ignore standard input, output, and error which are 0, 1, and 2
            if (fd <= 2) {
                printf("Ignoring standard open fd %d -> %s...\n", fd, target_path);
            } else if (fd == dir_fd) {
                printf("Ignoring fd to current directory fd %d -> %s...\n", fd, target_path);
            } else {
                printf("!!! Leaked file descriptor %d -> %s !!!\n", fd, target_path);
                fd_leaks++;
            }
        } else {
            perror("Failed to read link");
            fd_leaks++;
        }
    }

    closedir(dir);
    return fd_leaks;
}
EOF

cc -Wall -Wextra load.c -shared -o load.so -fPIC
    "##,
        )
        .status()
        .expect("execute sh");
    assert_status_ok!(status);

    // Load code.
    // ENOSYS = Dynamic linking not supported.
    let status = syd()
        .args([
            "-plib",
            "-munshare/user:1",
            "-munshare/pid:1",
            "--",
            "./load.so",
        ])
        .status()
        .expect("execute syd");
    let code = status.code().unwrap_or(127);
    if code != nix::libc::ENOSYS {
        assert_status_ok!(status);
    } else {
        eprintln!("Dynamic linking not supported, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
    }

    Ok(())
}

fn test_syd_load_library_check_fd_leaks_init_bare() -> TestResult {
    if !is_program_available("cc") {
        eprintln!("Test requires cc, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }
    if !is_program_available("sh") {
        eprintln!("Test requires sh, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }

    let status = Command::new("sh")
        .arg("-cex")
        .arg(
            r##"
#!/bin/sh
cat > load.c <<EOF
#include <dirent.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

static int fd_leaks;

__attribute__((constructor))
void syd_init(void)
{
    DIR *dir;
    struct dirent *entry;
    int dir_fd;

    // Open the directory containing file descriptors
    dir = opendir("/proc/self/fd");
    if (!dir) {
        perror("Failed to open /proc/self/fd");
        fd_leaks = -1;
        return;
    }

    // Get the file descriptor for the directory stream
    dir_fd = dirfd(dir);
    if (dir_fd == -1) {
        perror("Failed to get file descriptor for directory");
        closedir(dir);
        fd_leaks = -1;
        return;
    }

    // Iterate over all entries in the directory
    while ((entry = readdir(dir)) != NULL) {
        int fd;
        char *end;

        // Convert the name of the entry to an integer
        fd = strtol(entry->d_name, &end, 10);
        if (*end != '\0' || entry->d_name == end) continue; // Skip non-integer entries

        // Build the path to the symbolic link for the file descriptor
        char link_path[4096];
        char target_path[4096];
        snprintf(link_path, sizeof(link_path), "/proc/self/fd/%d", fd);

        ssize_t len = readlink(link_path, target_path, sizeof(target_path) - 1);
        if (len > 0) {
            target_path[len] = '\0'; // Ensure null termination
            // We ignore standard input, output, and error which are 0, 1, and 2
            if (fd <= 2) {
                printf("Ignoring standard open fd %d -> %s...\n", fd, target_path);
            } else if (fd == dir_fd) {
                printf("Ignoring fd to current directory fd %d -> %s...\n", fd, target_path);
            } else {
                printf("!!! Leaked file descriptor %d -> %s !!!\n", fd, target_path);
                fd_leaks++;
            }
        } else {
            perror("Failed to read link");
            fd_leaks++;
        }
    }

    closedir(dir);
}

int syd_main(void)
{
    if (fd_leaks > 0) {
        printf("Detected %d file descriptor leaks during init!\n", fd_leaks);
    } else {
        printf("No file descriptor leaks detected during init!\n");
    }
    return fd_leaks;
}
EOF

cc -Wall -Wextra load.c -shared -o load.so -fPIC
    "##,
        )
        .status()
        .expect("execute sh");
    assert_status_ok!(status);

    // Load code.
    // ENOSYS = Dynamic linking not supported.
    let status = syd()
        .args(["-plib", "--", "./load.so"])
        .status()
        .expect("execute syd");
    let code = status.code().unwrap_or(127);
    if code != nix::libc::ENOSYS {
        assert_status_ok!(status);
    } else {
        eprintln!("Dynamic linking not supported, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
    }

    Ok(())
}

fn test_syd_load_library_check_fd_leaks_init_wrap() -> TestResult {
    match check_unshare() {
        Some(false) => {
            eprintln!("Test requires Linux namespaces, skipping!");
            env::set_var("SYD_TEST_SOFT_FAIL", "1");
            return Ok(());
        }
        None => {
            return Err(TestError(
                "Failed to test for Linux namespaces!".to_string(),
            ));
        }
        _ => {}
    };
    if !is_program_available("cc") {
        eprintln!("Test requires cc, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }
    if !is_program_available("sh") {
        eprintln!("Test requires sh, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }

    let status = Command::new("sh")
        .arg("-cex")
        .arg(
            r##"
#!/bin/sh
cat > load.c <<EOF
#include <dirent.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

static int fd_leaks;

__attribute__((constructor))
void syd_init(void)
{
    DIR *dir;
    struct dirent *entry;
    int dir_fd;

    // Open the directory containing file descriptors
    dir = opendir("/proc/self/fd");
    if (!dir) {
        perror("Failed to open /proc/self/fd");
        fd_leaks = -1;
        return;
    }

    // Get the file descriptor for the directory stream
    dir_fd = dirfd(dir);
    if (dir_fd == -1) {
        perror("Failed to get file descriptor for directory");
        closedir(dir);
        fd_leaks = -1;
        return;
    }

    // Iterate over all entries in the directory
    while ((entry = readdir(dir)) != NULL) {
        int fd;
        char *end;

        // Convert the name of the entry to an integer
        fd = strtol(entry->d_name, &end, 10);
        if (*end != '\0' || entry->d_name == end) continue; // Skip non-integer entries

        // Build the path to the symbolic link for the file descriptor
        char link_path[4096];
        char target_path[4096];
        snprintf(link_path, sizeof(link_path), "/proc/self/fd/%d", fd);

        ssize_t len = readlink(link_path, target_path, sizeof(target_path) - 1);
        if (len > 0) {
            target_path[len] = '\0'; // Ensure null termination
            // We ignore standard input, output, and error which are 0, 1, and 2
            if (fd <= 2) {
                printf("Ignoring standard open fd %d -> %s...\n", fd, target_path);
            } else if (fd == dir_fd) {
                printf("Ignoring fd to current directory fd %d -> %s...\n", fd, target_path);
            } else {
                printf("!!! Leaked file descriptor %d -> %s !!!\n", fd, target_path);
                fd_leaks++;
            }
        } else {
            perror("Failed to read link");
            fd_leaks++;
        }
    }

    closedir(dir);
}

int syd_main(void)
{
    if (fd_leaks > 0) {
        printf("Detected %d file descriptor leaks during init!\n", fd_leaks);
    } else {
        printf("No file descriptor leaks detected during init!\n");
    }
    return fd_leaks;
}
EOF

cc -Wall -Wextra load.c -shared -o load.so -fPIC
    "##,
        )
        .status()
        .expect("execute sh");
    assert_status_ok!(status);

    // Load code.
    // ENOSYS = Dynamic linking not supported.
    let status = syd()
        .args([
            "-plib",
            "-munshare/user:1",
            "-munshare/pid:1",
            "--",
            "./load.so",
        ])
        .status()
        .expect("execute syd");
    let code = status.code().unwrap_or(127);
    if code != nix::libc::ENOSYS {
        assert_status_ok!(status);
    } else {
        eprintln!("Dynamic linking not supported, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
    }

    Ok(())
}

// Tests if read sandboxing for open works to allow.
fn test_syd_read_sandbox_open_allow() -> TestResult {
    if !is_program_available("dd") {
        eprintln!("Test requires dd, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }
    eprintln!("+ dd if=/dev/null");
    let status = syd()
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-msandbox/stat:on",
            "-mallow/stat+/***",
            "-mallow/read+/***",
            "-mdeny/read+/dev/***",
            "-mallow/read+/dev/null",
            "--",
            "dd",
            "if=/dev/null",
        ])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

// Tests if read sandboxing for open works to deny.
fn test_syd_read_sandbox_open_deny() -> TestResult {
    if !is_program_available("cat") {
        eprintln!("Test requires cat, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }
    let status = syd()
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-mallow/read+/***",
            "-mdeny/read+/dev/null",
            "--",
            "cat",
            "/dev/null",
        ])
        .status()
        .expect("execute syd");
    assert_status_not_ok!(status);
    Ok(())
}

// Tests if stat sandboxing for chdir works to allow.
fn test_syd_stat_sandbox_chdir_allow() -> TestResult {
    if !is_program_available("bash") {
        eprintln!("Test requires bash, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }
    let status = syd()
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/stat:on",
            "-mdeny/stat+/dev",
            "-mallow/stat+/dev",
            "--",
            "/bin/sh",
            "-c",
            "cd /dev",
        ])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

// Tests if stat sandboxing for stat works to hide.
fn test_syd_stat_sandbox_chdir_hide() -> TestResult {
    if !is_program_available("sh") {
        eprintln!("Test requires sh, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }
    eprintln!("+ cd /dev");
    let status = syd()
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/stat:on",
            "-mallow/stat+/***",
            "-mdeny/stat+/dev",
            "--",
            "/bin/sh",
            "-c",
            "cd /dev || exit 42",
        ])
        .status()
        .expect("execute syd");
    assert_status_code!(status, 42);
    Ok(())
}

// Tests if stat sandboxing for stat works to allow.
fn test_syd_stat_sandbox_stat_allow() -> TestResult {
    if !is_program_available("ls") {
        eprintln!("Test requires ls, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }
    eprintln!("+ ls /dev/null");
    let status = syd()
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/stat:on",
            "-mdeny/stat+/dev/null",
            "-mallow/stat+/dev/null",
            "--",
            "ls",
            "/dev/null",
        ])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

// Tests if stat sandboxing for stat works to hide.
fn test_syd_stat_sandbox_stat_hide() -> TestResult {
    if !is_program_available("ls") {
        eprintln!("Test requires ls, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }
    eprintln!("+ ls /dev/null");
    let status = syd()
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/stat:on",
            "-mallow/stat+/***",
            "-mdeny/stat+/dev/null",
            "--",
            "ls",
            "/dev/null",
        ])
        .status()
        .expect("execute syd");
    assert_status_not_ok!(status);

    Ok(())
}

// Tests if stat sandboxing for getdents works to allow.
fn test_syd_stat_sandbox_getdents_allow() -> TestResult {
    if !is_program_available("ls") {
        eprintln!("Test requires ls, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }
    eprintln!("+ ls /dev");
    let output = syd()
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/stat:on",
            "-mdeny/stat+/dev/null",
            "-mallow/stat+/dev/null",
            "--",
            "ls",
            "/dev",
        ])
        .output()
        .expect("execute syd");
    assert!(
        output
            .stdout
            .windows(b"null".len())
            .any(|window| window == b"null"),
        "Stdout:\n{:?}",
        output.stdout
    );

    Ok(())
}

// Tests if stat sandboxing for getdents works to hide.
fn test_syd_stat_sandbox_getdents_hide() -> TestResult {
    if !is_program_available("ls") {
        eprintln!("Test requires ls, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }
    eprintln!("+ ls /dev");
    let output = syd()
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/stat:on",
            "-mallow/stat+/***",
            "-mdeny/stat+/dev/null",
            "--",
            "ls",
            "/dev",
        ])
        .output()
        .expect("execute syd");
    assert!(
        output
            .stdout
            .windows(b"null".len())
            .any(|window| window != b"null"),
        "Stdout:{:?}",
        output.stdout
    );

    Ok(())
}

// Tests if stat sandboxing can be bypassed by read attempt
fn test_syd_stat_bypass_with_read() -> TestResult {
    let status = syd()
        .env("SYD_DO", "stat_bypass_with_read")
        .args([
            "-ppaludis",
            "-msandbox/read:on",
            "-msandbox/stat:on",
            "-mallow/read+/***",
            "-mallow/stat+/***",
            "-mdeny/read+/etc/***",
            "-mdeny/stat+/etc/***",
            "-mallow/read+/etc/ld*/***",
            "-mallow/stat+/etc/ld*/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

// Tests if stat sandboxing can be bypassed by write attempt
fn test_syd_stat_bypass_with_write() -> TestResult {
    let status = syd()
        .env("SYD_DO", "stat_bypass_with_write")
        .args([
            "-ppaludis",
            "-msandbox/stat:on",
            "-msandbox/write:on",
            "-mallow/stat+/***",
            "-mallow/write+/***",
            "-mdeny/stat+/etc/***",
            "-mdeny/write+/etc/***",
            "-mallow/stat+/etc/ld*/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

// Tests if stat sandboxing can be bypassed by exec attempt
fn test_syd_stat_bypass_with_exec() -> TestResult {
    let status = syd()
        .env("SYD_DO", "stat_bypass_with_exec")
        .args([
            "-ppaludis",
            "-msandbox/exec:on",
            "-msandbox/stat:on",
            "-mallow/exec+/***",
            "-mallow/stat+/***",
            "-mdeny/exec+/**/z?sh",
            "-mdeny/stat+/**/z?sh",
            "-mdeny/exec+/**/[bd]ash",
            "-mdeny/stat+/**/[bd]ash",
            "-mdeny/exec+/**/busybox",
            "-mdeny/stat+/**/busybox",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

// Tests if write sandboxing for open works to allow.
fn test_syd_write_sandbox_open_allow() -> TestResult {
    if !is_program_available("bash") {
        eprintln!("Test requires bash, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }
    eprintln!("+ bash -c \"echo welcome to the machine >> /dev/null\"");
    let status = syd()
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/write:on",
            "-mdeny/write+/dev/***",
            "-mallow/write+/dev/null",
            "--",
            "/bin/bash",
            "-c",
            "echo welcome to the machine >> /dev/null",
        ])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

// Tests if write sandboxing for open works to deny.
fn test_syd_write_sandbox_open_deny() -> TestResult {
    if !is_program_available("bash") {
        eprintln!("Test requires bash, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }
    eprintln!("+ bash -c \"echo welcome to the machine >> /dev/null\"");
    let status = syd()
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/write:on",
            "-mallow/write+/***",
            "-mdeny/write+/dev/null",
            "--",
            "/bin/bash",
            "-c",
            "echo welcome to the machine >> /dev/null",
        ])
        .status()
        .expect("execute syd");
    assert_status_not_ok!(status);

    Ok(())
}

// Tests if exec sandboxing works to allow.
fn test_syd_exec_sandbox_open_allow() -> TestResult {
    if !is_program_available("true") {
        eprintln!("Test requires true, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }
    let bin = which("true")?;
    let status = syd()
        .args([
            "-plib",
            "-msandbox/lock:off",
            "-msandbox/exec:on",
            "-mdeny/exec+/***",
            "-mallow/exec+/**/*.so*",
            &format!("-mallow/exec+{bin}"),
            "-atrue", // this may be busybox
            "--",
            &bin.to_string(),
        ])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

// Tests if exec sandboxing works to deny.
fn test_syd_exec_sandbox_open_deny() -> TestResult {
    if !is_program_available("true") {
        eprintln!("Test requires true, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }
    let bin = which("true")?;
    let status = syd()
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/exec:on",
            "-mallow/exec+/***",
            &format!("-mdeny/exec+{bin}"),
            "-atrue", // this may be busybox
            "--",
            &bin.to_string(),
        ])
        .status()
        .expect("execute syd");
    assert_status_not_ok!(status);

    Ok(())
}

// Check if #! interpreter path of scripts are properly sandboxed.
fn test_syd_exec_sandbox_deny_binfmt_script() -> TestResult {
    if !is_program_available("sh") {
        eprintln!("Test requires sh, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }

    // Write script.
    let path = "./script.sh";
    let script = r#"#!/bin/sh -ex
exit 42
"#;
    let mut file = File::create(path)?;
    write!(file, "{script}")?;
    drop(file); // Close the file to avoid ETXTBUSY.

    syd::fs::chmod_x(path).expect("Failed to make file executable");

    // Step 1: Allow both the interpreter and the script.
    let status = syd()
        .args([
            "-plib",
            "-msandbox/exec:on",
            "-mallow/exec+/***",
            "--",
            "./script.sh",
        ])
        .status()
        .expect("execute syd");
    assert_status_code!(status, 42);

    // Step 2: Allow the interpreter but disable the script.
    let status = syd()
        .args([
            "-plib",
            "-msandbox/exec:on",
            "-mallow/exec+/***",
            "-mdeny/exec+/**/script.sh",
            "--",
            "./script.sh",
        ])
        .status()
        .expect("execute syd");
    assert_status_code!(status, nix::libc::EACCES);

    // Step 3: Allow the script but disable the interpreter.
    let status = syd()
        .args([
            "-plib",
            "-msandbox/exec:on",
            "-mdeny/exec+/***",
            "-mallow/exec+/**/*.so*",
            "-mallow/exec+/**/script.sh",
            "--",
            "./script.sh",
        ])
        .status()
        .expect("execute syd");
    assert_status_killed!(status);

    Ok(())
}

// Check if a script which has an interpreter that itself is a script is properly sandboxed.
fn test_syd_exec_sandbox_many_binfmt_script() -> TestResult {
    if !is_program_available("sh") {
        eprintln!("Test requires sh, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }

    // Write script1 whose interpreter points to script2.
    let path1 = "./script1.sh";
    let script1 = r#"#!./script2.sh
"#;
    let mut file1 = File::create(path1)?;
    write!(file1, "{script1}")?;

    // Write script2 whole interpreter points to /bin/sh.
    let path2 = "./script2.sh";
    let script2 = r#"#!/bin/sh -ex
exit 42
"#;
    let mut file2 = File::create(path2)?;
    write!(file2, "{script2}")?;

    // Close the files to avoid ETXTBUSY.
    drop(file1);
    drop(file2);

    // Set permissions to make the scripts executable.
    for path in [path1, path2] {
        syd::fs::chmod_x(path).expect("Failed to set file executable");
    }

    // Step 1: Allow both the interpreter and the script.
    let status = syd()
        .args([
            "-plib",
            "-msandbox/exec:on",
            "-mallow/exec+/***",
            "--",
            "./script1.sh",
        ])
        .status()
        .expect("execute syd");
    assert_status_code!(status, 42);

    // Step 2: Allow the interpreter but disable the script2.
    let status = syd()
        .args([
            "-plib",
            "-msandbox/exec:on",
            "-mallow/exec+/***",
            "-mdeny/exec+/**/script2.sh",
            "--",
            "./script1.sh",
        ])
        .status()
        .expect("execute syd");
    fixup!(
        status.code().unwrap_or(127) == nix::libc::EACCES,
        "status:{status:?}"
    );

    // Step 3: Allow the scripts but disable the interpreter.
    // This will slip through the seccomp sandbox
    // but it's caught by the exec-TOCTOU mitigator.
    let status = syd()
        .args([
            "-plib",
            "-msandbox/exec:on",
            "-mdeny/exec+/***",
            "-mallow/exec+/**/*.so*",
            "-mallow/exec+/**/script[1-2].sh",
            "--",
            "./script1.sh",
        ])
        .status()
        .expect("execute syd");
    assert_status_killed!(status);

    Ok(())
}

// Check if a denylisted library can be injected using LD_LIBRARY_PATH.
// Note the seccomp sandbox is not able to catch this.
// This is prevented by the TOCTOU-mitigator on exec(2) exit.
fn test_syd_exec_sandbox_prevent_library_injection_LD_LIBRARY_PATH() -> TestResult {
    if !cfg!(feature = "elf") {
        eprintln!("Test requires syd built with elf feature, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }
    if !is_program_available("cc") {
        eprintln!("Test requires cc, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }
    if !is_program_available("sh") {
        eprintln!("Test requires sh, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }

    let status = syd()
        .env("LD_TRACE_LOADED_OBJECTS", "YesPlease")
        .env("LD_VERBOSE", "YesPlease")
        .args([
            "-plib",
            "-msandbox/exec:on",
            "-mallow/exec+/***",
            "-mdeny/exec+/**/lib-bad/*.so",
            "--",
            "sh",
            "-cex",
        ])
        .arg(
            r##"
#!/bin/sh -ex

# Ensure syd's CWD does not match our CWD
mkdir -m700 -p foo
cd foo

cat > lib-good.c <<EOF
int func(void) { return 0; }
EOF

cat > lib-bad.c <<EOF
int func(void) { return 42; }
EOF

cat > bin.c <<EOF
extern int func(void);
int main(void) { return func(); }
EOF

mkdir -m700 -p lib-good lib-bad
cc -Wall -Wextra lib-good.c -shared -o lib-good/libext.so -fPIC
cc -Wall -Wextra lib-bad.c -shared -o lib-bad/libext.so -fPIC

cc -Wall -Wextra bin.c -L./lib-good -lext -obin
r=0
env LD_LIBRARY_PATH="./lib-good:$LD_LIBRARY_PATH" ./bin || r=$?
echo >&2 "Good returned: $r"
test $r -eq 0

r=0
env LD_LIBRARY_PATH="./lib-bad:$LD_LIBRARY_PATH" ./bin || r=$?
echo >&2 "Bad returned: $r"
if test $r -eq 42; then
    echo >&2 "Library injection succeded!"
    false
else
    echo >&2 "Library injection failed!"
    true
fi
    "##,
        )
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

// Check if a denylisted library can be injected using LD_PRELOAD.
// Note the seccomp sandbox is not able to catch this.
// This is prevented by the TOCTOU-mitigator on exec(2) exit.
fn test_syd_exec_sandbox_prevent_library_injection_LD_PRELOAD() -> TestResult {
    if !cfg!(feature = "elf") {
        eprintln!("Test requires syd built with elf feature, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }
    if !is_program_available("cc") {
        eprintln!("Test requires cc, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }
    if !is_program_available("sh") {
        eprintln!("Test requires sh, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }

    let status = syd()
        .env("LD_TRACE_LOADED_OBJECTS", "YesPlease")
        .env("LD_VERBOSE", "YesPlease")
        .args([
            "-plib",
            "-msandbox/exec:on",
            "-mallow/exec+/***",
            "-mdeny/exec+/**/lib-bad/*.so",
            "--",
            "sh",
            "-cex",
        ])
        .arg(
            r##"
#!/bin/sh -ex

# Ensure syd's CWD does not match our CWD
mkdir -m700 -p foo
cd foo

cat > lib-good.c <<EOF
#include <sys/types.h>
pid_t getpid(void) { return 0; }
EOF

cat > lib-bad.c <<EOF
#include <sys/types.h>
pid_t getpid(void) { return 0; }
EOF

cat > bin.c <<EOF
#include <sys/types.h>
#include <unistd.h>
int main(void) { return getpid(); }
EOF

mkdir -m700 -p lib-good lib-bad
cc -Wall -Wextra lib-good.c -shared -o lib-good/libext.so -fPIC
cc -Wall -Wextra lib-bad.c -shared -o lib-bad/libext.so -fPIC

cc -Wall -Wextra bin.c -obin
r=0
env LD_PRELOAD="./lib-good/libext.so" ./bin || r=$?
echo >&2 "Good returned: $r"
test $r -eq 0

r=0
env LD_PRELOAD="./lib-bad/libext.so" ./bin || r=$?
echo >&2 "Bad returned: $r"
if test $r -eq 1; then
    echo >&2 "Library injection succeded!"
    false
else
    echo >&2 "Library injection failed!"
fi

r=0
env LD_PRELOAD="foo bar baz ./lib-bad/libext.so" ./bin || r=$?
echo >&2 "Bad returned: $r"
if test $r -eq 1; then
    echo >&2 "Library injection succeded!"
    false
else
    echo >&2 "Library injection failed!"
    true
fi

r=0
env LD_PRELOAD="foo:bar:baz:./lib-bad/libext.so:a:b:c" ./bin || r=$?
echo >&2 "Bad returned: $r"
if test $r -eq 1; then
    echo >&2 "Library injection succeded!"
    false
else
    echo >&2 "Library injection failed!"
    true
fi
    "##,
        )
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

// Tests if network connect sandboxing works to allow.
fn test_syd_network_sandbox_connect_ipv4_allow() -> TestResult {
    let status = syd()
        .env("SYD_DO", "connect4")
        .args([
            "-mallow/exec,read,stat+/***",
            "-mtrace/allow_safe_bind:0",
            "-mallow/net/bind+127.0.0.1!4242",
            "-mallow/net/connect+127.0.0.1!4242",
            "--",
            &SYD_DO,
            "127.0.0.1",
            "4242",
        ])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

// Tests if network connect sandboxing works to deny.
fn test_syd_network_sandbox_connect_ipv4_deny() -> TestResult {
    let status = syd()
        .env("SYD_DO", "connect4")
        .args([
            "-mallow/exec,read,stat+/***",
            "-mtrace/allow_safe_bind:0",
            "-mallow/net/bind+127.0.0.1!4242",
            "-mdeny/net/connect+127.0.0.1!4242",
            "--",
            &SYD_DO,
            "127.0.0.1",
            "4242",
        ])
        .status()
        .expect("execute syd");
    assert_status_code!(status, nix::libc::ECONNREFUSED);
    Ok(())
}

// Tests if network connect sandboxing works to allow.
fn test_syd_network_sandbox_connect_ipv6_allow() -> TestResult {
    if !check_ipv6() {
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }
    let status = syd()
        .env("SYD_DO", "connect6")
        .args([
            "-mallow/exec,read,stat+/***",
            "-mtrace/allow_safe_bind:0",
            "-mallow/net/bind+::1!4242",
            "-mallow/net/connect+::1!4242",
            "--",
            &SYD_DO,
            "::1",
            "4242",
        ])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

// Tests if network connect sandboxing works to deny.
fn test_syd_network_sandbox_connect_ipv6_deny() -> TestResult {
    if !check_ipv6() {
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }
    let status = syd()
        .env("SYD_DO", "connect6")
        .args([
            "-mallow/exec,read,stat+/***",
            "-mtrace/allow_safe_bind:0",
            "-mallow/net/bind+::1!4242",
            "-mdeny/net/connect+::1!4242",
            "--",
            &SYD_DO,
            "::1",
            "4242",
        ])
        .status()
        .expect("execute syd");
    assert_status_code!(status, nix::libc::ECONNREFUSED);
    Ok(())
}

fn test_syd_network_sandbox_allow_safe_bind_ipv4_failure() -> TestResult {
    let status = syd()
        .env("SYD_DO", "connect4_0")
        .args([
            "-mallow/exec,read,stat+/***",
            "-mtrace/allow_safe_bind:0",
            "-mallow/net/bind+127.0.0.1!0",
            "--",
            &SYD_DO,
            "127.0.0.1",
        ])
        .status()
        .expect("execute syd");
    assert_status_code!(status, nix::libc::ECONNREFUSED);
    Ok(())
}

fn test_syd_network_sandbox_allow_safe_bind_ipv4_success() -> TestResult {
    let status = syd()
        .env("SYD_DO", "connect4_0")
        .args([
            "-mallow/exec,read,stat+/***",
            "-mtrace/allow_safe_bind:1",
            "-mallow/net/bind+127.0.0.1!0",
            "--",
            &SYD_DO,
            "127.0.0.1",
        ])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_network_sandbox_allow_safe_bind_ipv6_failure() -> TestResult {
    if !check_ipv6() {
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }
    let status = syd()
        .env("SYD_DO", "connect6_0")
        .args([
            "-mallow/exec,read,stat+/***",
            "-mtrace/allow_safe_bind:0",
            "-mallow/net/bind+::1!0",
            "--",
            &SYD_DO,
            "::1",
        ])
        .status()
        .expect("execute syd");
    assert_status_code!(status, nix::libc::ECONNREFUSED);
    Ok(())
}

fn test_syd_network_sandbox_allow_safe_bind_ipv6_success() -> TestResult {
    if !check_ipv6() {
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }
    let status = syd()
        .env("SYD_DO", "connect6_0")
        .args([
            "-mallow/exec,read,stat+/***",
            "-mtrace/allow_safe_bind:1",
            "-mallow/net/bind+::1!0",
            "--",
            &SYD_DO,
            "::1",
        ])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_handle_toolong_unix_connect() -> TestResult {
    let status = syd()
        .env("SYD_DO", "toolong_unix_connect")
        .args([
            "-mallow/exec,read,stat,write+/***",
            "-mallow/net/bind+/***",
            "-mtrace/allow_safe_bind:1",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_sendmsg_scm_rights_one() -> TestResult {
    let status = syd()
        .env("SYD_DO", "sendmsg_scm_rights_one")
        .args([
            "-mallow/exec,read,stat,write+/***",
            "-mallow/net/bind+/***",
            "-mtrace/allow_safe_bind:1",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_sendmsg_scm_rights_many() -> TestResult {
    let status = syd()
        .env("SYD_DO", "sendmsg_scm_rights_many")
        .args([
            "-mallow/exec,read,stat,write+/***",
            "-mallow/net/bind+/***",
            "-mtrace/allow_safe_bind:1",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_sendmmsg() -> TestResult {
    let status = syd()
        .env("SYD_DO", "sendmmsg")
        .args([
            "-mallow/exec,read,stat,write+/***",
            "-mallow/net/bind+/***",
            "-mtrace/allow_safe_bind:1",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_handle_toolong_unix_sendto() -> TestResult {
    let status = syd()
        .env("SYD_DO", "toolong_unix_sendto")
        .args([
            "-mallow/exec,read,stat,write+/***",
            "-mallow/net/bind+/***",
            "-mtrace/allow_safe_bind:1",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_handle_toolong_unix_sendmsg() -> TestResult {
    let status = syd()
        .env("SYD_DO", "toolong_unix_sendmsg")
        .args([
            "-mallow/exec,read,stat,write+/***",
            "-mallow/net/bind+/***",
            "-mtrace/allow_safe_bind:1",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_mask_simple() -> TestResult {
    for p in &["cmp", "readlink", "sh"] {
        if !is_program_available(p) {
            eprintln!("Test requires {p}, skipping!");
            env::set_var("SYD_TEST_SOFT_FAIL", "1");
            return Ok(());
        }
    }

    let status = syd()
        .args([
            "-plib",
            "-mlock:exec",
            "-msandbox/read:on",
            "-msandbox/write:on",
            "-mallow/read,write+/***",
            "--",
            "sh",
            "-cx",
        ])
        .arg(
            r##"
echo 'Heavy is the root of light. Still is the master of moving.' > tao.orig
echo 'Heavy is the root of light. Still is the master of moving.' > tao.mask
abs=$(readlink -f tao.mask)
test -f "$abs" || exit 1
test -c "/dev/syd/mask+${abs}" || exit 2
test -f "$abs" || exit 3
cat tao.mask || exit 4
echo > tao.mask || exit 5
cmp tao.orig tao.mask && exit 6
test -c "/dev/syd/mask-${abs}" || exit 7
cmp tao.orig tao.mask || exit 8
true
    "##,
        )
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_kcapi_hash_block() -> TestResult {
    let status = syd()
        .env("SYD_DO", "kcapi_hash_block")
        .args(["-plib", "--", &SYD_DO, "0"])
        .status()
        .expect("execute syd");
    assert_status_ok_if_supported!(status);

    let status = syd()
        .env("SYD_DO", "kcapi_hash_block")
        .args(["-plib", "-mtrace/allow_safe_kcapi:1", "--", &SYD_DO, "0"])
        .status()
        .expect("execute syd");
    assert_status_ok_if_supported!(status);

    let status = syd()
        .env("SYD_DO", "kcapi_hash_block")
        .args(["-plib", "-msandbox/net:on", "--", &SYD_DO, "0"])
        .status()
        .expect("execute syd");
    assert_status_not_supported!(status);

    let status = syd()
        .env("SYD_DO", "kcapi_hash_block")
        .args([
            "-plib",
            "-msandbox/net:on",
            "-mtrace/allow_safe_kcapi:1",
            "--",
            &SYD_DO,
            "0",
        ])
        .status()
        .expect("execute syd");
    assert_status_ok_if_supported!(status);

    Ok(())
}

fn test_syd_kcapi_hash_stream() -> TestResult {
    let status = syd()
        .env("SYD_DO", "kcapi_hash_stream")
        .args(["-plib", "--", &SYD_DO, "0"])
        .status()
        .expect("execute syd");
    assert_status_ok_if_supported!(status);

    let status = syd()
        .env("SYD_DO", "kcapi_hash_stream")
        .args(["-plib", "-mtrace/allow_safe_kcapi:1", "--", &SYD_DO, "0"])
        .status()
        .expect("execute syd");
    assert_status_ok_if_supported!(status);

    let status = syd()
        .env("SYD_DO", "kcapi_hash_stream")
        .args(["-plib", "-msandbox/net:on", "--", &SYD_DO, "0"])
        .status()
        .expect("execute syd");
    assert_status_not_supported!(status);

    let status = syd()
        .env("SYD_DO", "kcapi_hash_stream")
        .args([
            "-plib",
            "-msandbox/net:on",
            "-mtrace/allow_safe_kcapi:1",
            "--",
            &SYD_DO,
            "0",
        ])
        .status()
        .expect("execute syd");
    assert_status_ok_if_supported!(status);

    Ok(())
}

fn test_syd_kcapi_cipher_block() -> TestResult {
    let status = syd()
        .env("SYD_DO", "kcapi_cipher_block")
        .args(["-plib", "--", &SYD_DO, "0"])
        .status()
        .expect("execute syd");
    assert_status_ok_if_supported!(status);

    let status = syd()
        .env("SYD_DO", "kcapi_cipher_block")
        .args(["-plib", "-mtrace/allow_safe_kcapi:1", "--", &SYD_DO, "0"])
        .status()
        .expect("execute syd");
    assert_status_ok_if_supported!(status);

    let status = syd()
        .env("SYD_DO", "kcapi_cipher_block")
        .args(["-plib", "-msandbox/net:on", "--", &SYD_DO, "0"])
        .status()
        .expect("execute syd");
    assert_status_not_supported!(status);

    let status = syd()
        .env("SYD_DO", "kcapi_cipher_block")
        .args([
            "-plib",
            "-msandbox/net:on",
            "-mtrace/allow_safe_kcapi:1",
            "--",
            &SYD_DO,
            "0",
        ])
        .status()
        .expect("execute syd");
    assert_status_ok_if_supported!(status);

    Ok(())
}

fn test_syd_exit_wait_default() -> TestResult {
    if !is_program_available("sh") {
        eprintln!("Test requires sh, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }
    let _ = remove_file("./test");
    let status = syd()
        .args(["-plib", "--", "sh", "-cex"])
        .arg(
            r##"
#!/bin/sh
: > test
cat > exec.sh <<'EOF'
#!/bin/sh -ex
sleep 5
echo OK > test
exit 42
EOF
chmod +x exec.sh
nohup ./exec.sh &
true
"##,
        )
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    let path = Path::new("./test");
    assert!(path.exists());
    let data = metadata(path).expect("Unable to access test file metadata");
    assert_eq!(data.len(), 0, "Unexpected file metadata: {data:?}");

    Ok(())
}

fn test_syd_exit_wait_default_unsafe_exec() -> TestResult {
    if !is_program_available("sh") {
        eprintln!("Test requires sh, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }
    let _ = remove_file("./test");
    let status = syd()
        .args(["-plib", "-mtrace/allow_unsafe_exec:1", "--", "sh", "-cex"])
        .arg(
            r##"
#!/bin/sh
: > test
cat > exec.sh <<'EOF'
#!/bin/sh -ex
sleep 5
echo OK > test
exit 42
EOF
chmod +x exec.sh
nohup ./exec.sh &
true
"##,
        )
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    let path = Path::new("./test");
    assert!(path.exists());
    let data = metadata(path).expect("Unable to access test file metadata");
    assert_eq!(data.len(), 0, "Unexpected file metadata: {data:?}");

    Ok(())
}

fn test_syd_exit_wait_pid() -> TestResult {
    if !is_program_available("sh") {
        eprintln!("Test requires sh, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }
    let _ = remove_file("./test");
    let status = syd()
        .args(["-plib", "-mtrace/exit_wait_all:0", "--", "sh", "-cex"])
        .arg(
            r##"
#!/bin/sh
: > test
cat > exec.sh <<'EOF'
#!/bin/sh -ex
sleep 5
echo OK > test
exit 42
EOF
chmod +x exec.sh
nohup ./exec.sh &
true
"##,
        )
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    let path = Path::new("./test");
    assert!(path.exists());
    let data = metadata(path).expect("Unable to access test file metadata");
    assert_eq!(data.len(), 0, "Unexpected file metadata: {data:?}");

    Ok(())
}

fn test_syd_exit_wait_pid_unsafe_exec() -> TestResult {
    if !is_program_available("sh") {
        eprintln!("Test requires sh, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }
    let _ = remove_file("./test");
    let status = syd()
        .args([
            "-plib",
            "-mtrace/exit_wait_all:0",
            "-mtrace/allow_unsafe_exec:1",
            "--",
            "sh",
            "-cex",
        ])
        .arg(
            r##"
#!/bin/sh
: > test
cat > exec.sh <<'EOF'
#!/bin/sh -ex
sleep 5
echo OK > test
exit 42
EOF
chmod +x exec.sh
nohup ./exec.sh &
true
"##,
        )
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    let path = Path::new("./test");
    assert!(path.exists());
    let data = metadata(path).expect("Unable to access test file metadata");
    assert_eq!(data.len(), 0, "Unexpected file metadata: {data:?}");

    Ok(())
}

fn test_syd_exit_wait_pid_with_runaway_cmd_exec_process() -> TestResult {
    if !is_program_available("sh") {
        eprintln!("Test requires sh, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }
    let _ = remove_file("./test");
    let status = syd()
        .args([
            "-plib",
            "-mlock:exec",
            "-mtrace/exit_wait_all:0",
            "--",
            "sh",
            "-cex",
        ])
        .arg(format!(
            "
#!/bin/sh
: > test
chmod 600 test
cat > exec.sh <<EOF
#!/bin/sh -ex
sleep \\$1
echo \\$2 > $PWD/test
exit 42
EOF
chmod +x exec.sh
nohup ./exec.sh 5 OK &
# Careful here, cmd/exec changes CWD to /.
test -c \"$({} $PWD/exec.sh 15 RUN)\"
true
",
            *SYD_EXEC
        ))
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    let path = Path::new("./test");
    assert!(path.exists());
    let data = metadata(path).expect("Unable to access test file metadata");
    assert_eq!(data.len(), 0, "Unexpected file metadata: {data:?}");

    // Wait for the runaway process to change the file.
    eprintln!("Waiting for 20 seconds for the runaway process to write to the file");
    sleep(Duration::from_secs(20));
    let data = metadata(path).expect("Unable to access test file metadata");
    assert_eq!(data.len(), 4, "Unexpected file metadata: {data:?}");

    Ok(())
}

fn test_syd_exit_wait_pid_unsafe_exec_with_runaway_cmd_exec_process() -> TestResult {
    if !is_program_available("sh") {
        eprintln!("Test requires sh, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }
    let _ = remove_file("./test");
    let status = syd()
        .args([
            "-plib",
            "-mlock:exec",
            "-mtrace/exit_wait_all:0",
            "-mtrace/allow_unsafe_exec:1",
            "--",
            "sh",
            "-cex",
        ])
        .arg(format!(
            "
#!/bin/sh
: > test
chmod 600 test
cat > exec.sh <<EOF
#!/bin/sh -ex
sleep \\$1
echo \\$2 > $PWD/test
exit 42
EOF
chmod +x exec.sh
nohup ./exec.sh 5 OK &
# Careful here, cmd/exec changes CWD to /.
test -c \"$({} $PWD/exec.sh 15 RUN)\"
true
",
            *SYD_EXEC
        ))
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    let path = Path::new("./test");
    assert!(path.exists());
    let data = metadata(path).expect("Unable to access test file metadata");
    assert_eq!(data.len(), 0, "Unexpected file metadata: {data:?}");

    // Wait for the runaway process to change the file.
    eprintln!("Waiting for 20 seconds for the runaway process to write to the file");
    sleep(Duration::from_secs(20));
    let data = metadata(path).expect("Unable to access test file metadata");
    assert_eq!(data.len(), 4, "Unexpected file metadata: {data:?}");

    Ok(())
}

fn test_syd_exit_wait_all() -> TestResult {
    if !is_program_available("sh") {
        eprintln!("Test requires sh, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }
    let _ = remove_file("./test");
    let status = syd()
        .args(["-plib", "-mtrace/exit_wait_all:1", "--", "sh", "-cex"])
        .arg(
            r##"
#!/bin/sh
: > test
cat > exec.sh <<'EOF'
#!/bin/sh -ex
sleep 5
echo OK > test
exit 42
EOF
chmod +x exec.sh
nohup ./exec.sh &
true
"##,
        )
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    let path = Path::new("./test");
    assert!(path.exists());
    let data = metadata(path).expect("Unable to access test file metadata");
    assert_eq!(data.len(), 3, "Unexpected file metadata: {data:?}");

    Ok(())
}

fn test_syd_exit_wait_all_unsafe_exec() -> TestResult {
    if !is_program_available("sh") {
        eprintln!("Test requires sh, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }
    let _ = remove_file("./test");
    let status = syd()
        .args([
            "-plib",
            "-mtrace/exit_wait_all:1",
            "-mtrace/allow_unsafe_exec:1",
            "--",
            "sh",
            "-cex",
        ])
        .arg(
            r##"
#!/bin/sh
: > test
cat > exec.sh <<'EOF'
#!/bin/sh -ex
sleep 5
echo OK > test
exit 42
EOF
chmod +x exec.sh
nohup ./exec.sh &
true
"##,
        )
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    let path = Path::new("./test");
    assert!(path.exists());
    let data = metadata(path).expect("Unable to access test file metadata");
    assert_eq!(data.len(), 3, "Unexpected file metadata: {data:?}");

    Ok(())
}

fn test_syd_exit_wait_all_with_runaway_cmd_exec_process() -> TestResult {
    if !is_program_available("sh") {
        eprintln!("Test requires sh, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }
    let _ = remove_file("./test");
    let status = syd()
        .args([
            "-plib",
            "-mlock:exec",
            "-mtrace/exit_wait_all:1",
            "--",
            "sh",
            "-cex",
        ])
        .arg(format!(
            "
#!/bin/sh
: > test
chmod 600 test
cat > exec.sh <<EOF
#!/bin/sh -ex
sleep \\$1
echo \\$2 > $PWD/test
exit 42
EOF
chmod +x exec.sh
nohup ./exec.sh 5 OK &
# Careful here, cmd/exec changes CWD to /.
test -c \"$({} $PWD/exec.sh 15 RUN)\"
true
",
            *SYD_EXEC
        ))
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    let path = Path::new("./test");
    assert!(path.exists());
    let data = metadata(path).expect("Unable to access test file metadata");
    assert_eq!(data.len(), 3, "Unexpected file metadata: {data:?}");

    // Wait for the runaway process to change the file.
    eprintln!("Waiting for 20 seconds for the runaway process to write to the file");
    sleep(Duration::from_secs(20));
    let data = metadata(path).expect("Unable to access test file metadata");
    assert_eq!(data.len(), 4, "Unexpected file metadata: {data:?}");

    Ok(())
}

fn test_syd_exit_wait_all_unsafe_exec_with_runaway_cmd_exec_process() -> TestResult {
    if !is_program_available("sh") {
        eprintln!("Test requires sh, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }
    let _ = remove_file("./test");
    let status = syd()
        .args([
            "-plib",
            "-mlock:exec",
            "-mtrace/exit_wait_all:1",
            "-mtrace/allow_unsafe_exec:1",
            "--",
            "sh",
            "-cex",
        ])
        .arg(format!(
            "
#!/bin/sh
: > test
chmod 600 test
cat > exec.sh <<EOF
#!/bin/sh -ex
sleep \\$1
echo \\$2 > $PWD/test
exit 42
EOF
chmod +x exec.sh
nohup ./exec.sh 5 OK &
# Careful here, cmd/exec changes CWD to /.
test -c \"$({} $PWD/exec.sh 15 RUN)\"
true
",
            *SYD_EXEC
        ))
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    let path = Path::new("./test");
    assert!(path.exists());
    let data = metadata(path).expect("Unable to access test file metadata");
    assert_eq!(data.len(), 3, "Unexpected file metadata: {data:?}");

    // Wait for the runaway process to change the file.
    eprintln!("Waiting for 20 seconds for the runaway process to write to the file");
    sleep(Duration::from_secs(20));
    let data = metadata(path).expect("Unable to access test file metadata");
    assert_eq!(data.len(), 4, "Unexpected file metadata: {data:?}");

    Ok(())
}

fn test_syd_cli_args_override_user_profile() -> TestResult {
    if !is_program_available("sh") {
        eprintln!("Test requires sh, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }
    let _ = unlink(".user.syd-3");
    let mut file = File::create(".user.syd-3").expect("Failed to create .user.syd-3");
    file.write_all(b"mem/max:4242\npid/max:2525\n")
        .expect("Failed to write to .user.syd-3");

    let mut child = syd()
        .args(["-mpid/max:4242", "-mstat", "-ctrue"])
        .stderr(Stdio::piped())
        .spawn()
        .expect("execute syd");

    // Read the output from the child process
    let child_stderr = child.stderr.as_mut().expect("child stderr");
    let mut output = String::new();
    let raw_fd = child_stderr.as_raw_fd();
    let mut file = unsafe { File::from_raw_fd(raw_fd) };
    if let Err(error) = file.read_to_string(&mut output) {
        return Err(TestError(format!(
            "Failed to read output of child process: {error}"
        )));
    }
    print!("Child output:\n{output}");

    assert!(output.contains("Pid Max: 4242"));
    //This may fail if the site-wide config file has lock:on.
    //assert!(output.contains("Memory Max: 4242"));

    Ok(())
}

fn test_syd_ifconfig_lo_bare() -> TestResult {
    let status = syd()
        .env("SYD_DO", "ifconfig_lo")
        .args(["-plib", "--", &SYD_DO])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_ifconfig_lo_wrap() -> TestResult {
    match check_unshare() {
        Some(false) => {
            eprintln!("Test requires Linux namespaces, skipping!");
            env::set_var("SYD_TEST_SOFT_FAIL", "1");
            return Ok(());
        }
        None => {
            return Err(TestError(
                "Failed to test for Linux namespaces!".to_string(),
            ));
        }
        _ => {}
    };

    let status = syd()
        .env("SYD_DO", "ifconfig_lo")
        .args([
            "-plib",
            "-munshare/user:1",
            "-munshare/net:1",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_binfmt_malformed_too_short() -> TestResult {
    if !is_program_available("python") {
        eprintln!("Test requires Python, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }

    let cwd = current_dir(true)?.display().to_string();

    syd::fs::cat("interpreter", "π")?;
    syd::fs::cat("script1", "#!interpreter")?;
    syd::fs::cat("script2", "#!./interpreter")?;
    syd::fs::cat("script3", "#!./././././././interpreter")?;
    syd::fs::cat("script4", "#! ./interpreter")?;
    syd::fs::cat("script5", "#!\t./interpreter")?;
    syd::fs::cat(
        "script6",
        format!("#! \t  \t   \t    \t../{cwd}/./../{cwd}/./././interpreter"),
    )?;
    syd::fs::chmod_x("interpreter")?;
    for i in 1..=6 {
        syd::fs::chmod_x(&format!("script{i}"))?;
    }

    let status = syd()
        .args([
            "-plib",
            "-mlock:exec",
            "-msandbox/exec:on",
            "-mallow/exec+/***",
            "--",
            "python",
            "-c",
        ])
        .arg(
            r##"
import os
import sys
import errno

def say(msg):
    print(msg, file=sys.stderr)

def run(script):
    pid = os.fork()

    if pid == 0:
        # Child process
        say(f"Executing script {script}")
        try:
            os.execvp(script, [script])
        except OSError as e:
            exit(e.errno)
    elif pid > 0:
        # Parent process
        _, status = os.waitpid(pid, 0)
        code = os.WEXITSTATUS(status)
        say(f"Script {script} returned {code}")
        return code
    else:
        raise Exception("Failed to fork")

# syd delivering ENOEXEC
# TODO: Distinguish from kernel
for i in range(1, 7):
    if run(f"./script{i}") != errno.ENOEXEC:
        exit(1)

say("Turn allow_unsupp_binfmt on")
os.lstat("/dev/syd/trace/allow_unsupp_binfmt:1")

# kernel delivering ENOEXEC
# TODO: Distinguish from syd
for i in range(1, 7):
    if run(f"./script{i}") != errno.ENOEXEC:
        exit(2)
    "##,
        )
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_binfmt_malformed_bad_magic() -> TestResult {
    if !is_program_available("python") {
        eprintln!("Test requires Python, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    } else if !cfg!(feature = "elf") {
        eprintln!("Test requires syd built with elf feature, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }

    // Number of bytes in an identifier.
    const SIZE_OF_IDENT: usize = 16;
    let cwd = current_dir(true)?.display().to_string();

    syd::fs::cat("interpreter", "#".repeat(SIZE_OF_IDENT))?;
    syd::fs::cat("script1", "#!interpreter")?;
    syd::fs::cat("script2", "#!./interpreter")?;
    syd::fs::cat("script3", "#!./././././././interpreter")?;
    syd::fs::cat("script4", "#! ./interpreter")?;
    syd::fs::cat("script5", "#!\t./interpreter")?;
    syd::fs::cat(
        "script6",
        format!("#! \t  \t   \t    \t../{cwd}/./../{cwd}/./././interpreter"),
    )?;
    syd::fs::chmod_x("interpreter")?;
    for i in 1..=6 {
        syd::fs::chmod_x(&format!("script{i}"))?;
    }

    let status = syd()
        .args([
            "-plib",
            "-mlock:exec",
            "-msandbox/exec:on",
            "-mallow/exec+/***",
            "--",
            "python",
            "-c",
        ])
        .arg(
            r##"
import os
import sys
import errno

def say(msg):
    print(msg, file=sys.stderr)

def run(script):
    pid = os.fork()

    if pid == 0:
        # Child process
        say(f"Executing script {script}")
        try:
            os.execvp(script, [script])
        except OSError as e:
            exit(e.errno)
    elif pid > 0:
        # Parent process
        _, status = os.waitpid(pid, 0)
        code = os.WEXITSTATUS(status)
        say(f"Script {script} returned {code}")
        return code
    else:
        raise Exception("Failed to fork")

for i in range(1, 7):
    if run(f"./script{i}") != errno.ENOEXEC:
        exit(1)

say("Turn allow_unsupp_binfmt on")
os.lstat("/dev/syd/trace/allow_unsupp_binfmt:1")

for i in range(1, 7):
    if run(f"./script{i}") != errno.ENOEXEC:
        exit(2)

say("Turn allow_unsupp_binfmt off")
os.lstat("/dev/syd/trace/allow_unsupp_binfmt:0")

for i in range(1, 7):
    if run(f"./script{i}") != errno.ENOEXEC:
        exit(3)
    "##,
        )
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_binfmt_malformed_invalid_elf_class() -> TestResult {
    if !is_program_available("python") {
        eprintln!("Test requires Python, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    } else if !cfg!(feature = "elf") {
        eprintln!("Test requires syd built with elf feature, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }

    // Define a byte array `elf` to be used in testing ELF header parsing.
    // This array is structured to mimic an ELF header but includes an intentionally invalid ELF class to trigger an error.
    // The ELF header starts with a magic number 0x7F followed by "ELF" (45 4C 46 in hexadecimal),
    // which is correctly set to indicate that this is an ELF file.
    let elf = b"\x7FELF\x03\x01\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00";
    // ELF header with a malformed class
    // The byte at index 4 (0-based index) represents the EI_CLASS:
    // - 0 (ELFCLASSNONE) means invalid class (not used),
    // - 1 (ELFCLASS32) means 32-bit objects,
    // - 2 (ELFCLASS64) means 64-bit objects.
    // Here, the byte at index 4 is set to 3, which is not a valid ELF class value,
    // as the defined valid values are 0 for ELFCLASSNONE, 1 for ELFCLASS32, and 2 for ELFCLASS64.
    // The value 3 does not correspond to any recognized or valid ELF class, which is intended to simulate
    // a corrupted or non-standard ELF file and is expected to trigger a 'Malformed' error in the parsing logic
    // indicating an "invalid ELF class".

    let cwd = current_dir(true)?.display().to_string();

    syd::fs::cat("interpreter", elf)?;
    syd::fs::cat("script1", "#!interpreter")?;
    syd::fs::cat("script2", "#!./interpreter")?;
    syd::fs::cat("script3", "#!./././././././interpreter")?;
    syd::fs::cat("script4", "#! ./interpreter")?;
    syd::fs::cat("script5", "#!\t./interpreter")?;
    syd::fs::cat(
        "script6",
        format!("#! \t  \t   \t    \t../{cwd}/./../{cwd}/./././interpreter"),
    )?;
    syd::fs::chmod_x("interpreter")?;
    for i in 1..=6 {
        syd::fs::chmod_x(&format!("script{i}"))?;
    }

    let status = syd()
        .args([
            "-plib",
            "-mlock:exec",
            "-msandbox/exec:on",
            "-mallow/exec+/***",
            "--",
            "python",
            "-c",
        ])
        .arg(
            r##"
import os
import sys
import errno

def say(msg):
    print(msg, file=sys.stderr)

def run(script):
    pid = os.fork()

    if pid == 0:
        # Child process
        say(f"Executing script {script}")
        try:
            os.execvp(script, [script])
        except OSError as e:
            exit(e.errno)
    elif pid > 0:
        # Parent process
        _, status = os.waitpid(pid, 0)
        code = os.WEXITSTATUS(status)
        say(f"Script {script} returned {code}")
        return code
    else:
        raise Exception("Failed to fork")

# syd delivering ENOEXEC
# TODO: Distinguish from kernel
# One idea is to check JSON seccomp logs for flags:1 aka continue,
# which indicates it is the kernel doing the error return.
for i in range(1, 7):
    if run(f"./script{i}") != errno.ENOEXEC:
        exit(1)

say("Turn allow_unsupp_binfmt on")
os.lstat("/dev/syd/trace/allow_unsupp_binfmt:1")

# kernel delivering ENOEXEC
# TODO: Distinguish from syd
for i in range(1, 7):
    if run(f"./script{i}") != errno.ENOEXEC:
        exit(2)

say("Turn allow_unsupp_binfmt off")
os.lstat("/dev/syd/trace/allow_unsupp_binfmt:0")

# syd delivering ENOEXEC
# TODO: Distinguish from kernel
for i in range(1, 7):
    if run(f"./script{i}") != errno.ENOEXEC:
        exit(3)
    "##,
        )
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_deny_elf32() -> TestResult {
    if cfg!(target_pointer_width = "32") {
        eprintln!("Test requires a 64-bit syd, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    } else if !cfg!(feature = "elf") {
        eprintln!("Test requires syd built with elf feature, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }
    if !is_program_available("cc") {
        eprintln!("Test requires cc, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }
    if !is_program_available("sh") {
        eprintln!("Test requires sh, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }

    let status = syd()
        .args([
            "-plib",
            "-mlock:exec",
            "-msandbox/exec:on",
            "-mallow/exec+/***",
            "--",
            "sh",
            "-cx",
        ])
        .arg(
            r##"
#!/bin/sh
cat > exit.c <<EOF
#include <stdlib.h>
int main() {
    exit(0);
}
EOF

cc -m32 exit.c -o exit
test $? -eq 0 || exit 128
./exit || exit 1
test -c /dev/syd/trace/deny_elf32:1 || exit 2
./exit && exit 3
test -c /dev/syd/trace/deny_elf32:0 || exit 2
./exit || exit 4
true
    "##,
        )
        .status()
        .expect("execute syd");
    let code = status.code().unwrap_or(127);
    assert!(code == 0 || code == 128, "code:{code} status:{status:?}");
    if code == 128 {
        // Compilation failed, test skipped.
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
    }
    Ok(())
}

fn test_syd_deny_elf_dynamic() -> TestResult {
    if !cfg!(feature = "elf") {
        eprintln!("Test requires syd built with elf feature, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }
    if !is_program_available("cc") {
        eprintln!("Test requires cc, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }
    if !is_program_available("sh") {
        eprintln!("Test requires sh, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }

    let status = syd()
        .args([
            "-plib",
            "-mlock:exec",
            "-msandbox/exec:on",
            "-mallow/exec+/***",
            "--",
            "sh",
            "-cx",
        ])
        .arg(
            r##"
#!/bin/sh
cat > exit.c <<EOF
#include <stdlib.h>
int main() {
    exit(0);
}
EOF

cc exit.c -o exit
test $? -eq 0 || exit 128
./exit || exit 1
test -c /dev/syd/trace/deny_elf_dynamic:1 || exit 2
./exit && exit 3
test -c /dev/syd/trace/deny_elf_dynamic:0 || exit 2
./exit || exit 4
true
    "##,
        )
        .status()
        .expect("execute syd");
    let code = status.code().unwrap_or(127);
    assert!(code == 0 || code == 128, "code:{code} status:{status:?}");
    if code == 128 {
        // Compilation failed, test skipped.
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
    }
    Ok(())
}

fn test_syd_deny_elf_static() -> TestResult {
    if !cfg!(feature = "elf") {
        eprintln!("Test requires syd built with elf feature, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }
    if !is_program_available("cc") {
        eprintln!("Test requires cc, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }
    if !is_program_available("sh") {
        eprintln!("Test requires sh, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }

    let status = syd()
        .args([
            "-plib",
            "-mlock:exec",
            "-msandbox/exec:on",
            "-mallow/exec+/***",
            "--",
            "sh",
            "-cx",
        ])
        .arg(
            r##"
#!/bin/sh
cat > exit.c <<EOF
#include <stdlib.h>
int main() {
    exit(0);
}
EOF

cc exit.c -o exit -static
test $? -eq 0 || exit 128
./exit || exit 1
test -c /dev/syd/trace/deny_elf_static:1 || exit 2
./exit && exit 3
test -c /dev/syd/trace/deny_elf_static:0 || exit 2
./exit || exit 4
true
    "##,
        )
        .status()
        .expect("execute syd");
    let code = status.code().unwrap_or(127);
    assert!(code == 0 || code == 128, "code:{code} status:{status:?}");
    if code == 128 {
        // Compilation failed, test skipped.
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
    }
    Ok(())
}

fn test_syd_deny_script() -> TestResult {
    syd::fs::cat("script", "#!/bin/sh\nexit 42")?;
    syd::fs::chmod_x("script")?;

    // Scripts are allowed by default.
    let status = syd()
        .args(["-plib", "-msandbox/exec:on", "-mallow/exec+/***"])
        .args(["--", "./script"])
        .status()
        .expect("execute syd");
    assert_status_code!(status, 42);

    // Scripts are denied with deny_script:1.
    // syd uses ENOSYS to distinguish from a kernel deny.
    let status = syd()
        .args([
            "-plib",
            "-msandbox/exec:on",
            "-mallow/exec+/***",
            "-mtrace/deny_script:1",
        ])
        .args(["--", "./script"])
        .status()
        .expect("execute syd");
    assert_status_code!(status, nix::libc::ENOSYS);

    Ok(())
}

fn test_syd_force_sandbox() -> TestResult {
    if !is_program_available("true") {
        eprintln!("Test requires true, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }

    let bad_md5 = "a".repeat(32);
    let bad_sha1 = "a".repeat(40);
    let bad_sha256 = "a".repeat(64);
    let bad_sha384 = "a".repeat(96);
    let bad_sha512 = "a".repeat(128);

    // Note, `which" returns canonicalized path.
    let bin_true = which("true").expect("syd-test-do in PATH");

    let md5_sum = const_hex::encode(syd::hash::hash(
        BufReader::new(File::open(&bin_true).expect("open syd-test-do")),
        HashAlgorithm::Md5,
    )?);
    let sha1_sum = const_hex::encode(syd::hash::hash(
        BufReader::new(File::open(&bin_true).expect("open syd-test-do")),
        HashAlgorithm::Sha1,
    )?);
    let sha256_sum = const_hex::encode(syd::hash::hash(
        BufReader::new(File::open(&bin_true).expect("open syd-test-do")),
        HashAlgorithm::Sha256,
    )?);
    let sha384_sum = const_hex::encode(syd::hash::hash(
        BufReader::new(File::open(&bin_true).expect("open syd-test-do")),
        HashAlgorithm::Sha384,
    )?);
    let sha512_sum = const_hex::encode(syd::hash::hash(
        BufReader::new(File::open(&bin_true).expect("open syd-test-do")),
        HashAlgorithm::Sha512,
    )?);

    // Test 1: Force sandboxing defaults.
    let status = syd()
        .args(["-plib", "-msandbox/force:on", "--", "true"])
        .status()
        .expect("execute syd");
    assert_status_killed!(status);

    let status = syd()
        .args([
            "-plib",
            "-msandbox/force:on",
            "-mforce/default:allow",
            "--",
            "true",
        ])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    let status = syd()
        .args([
            "-plib",
            "-msandbox/force:on",
            "-mforce/default:warn",
            "--",
            "true",
        ])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    let status = syd()
        .args([
            "-plib",
            "-msandbox/force:on",
            "-mforce/default:kill",
            "--",
            "true",
        ])
        .status()
        .expect("execute syd");
    assert_status_killed!(status);

    // Test 2: Force sandboxing allow with MD5 & SHA1,3-{256,384,512}
    // We set force/default:warn so as not to care about dynamic libraries.
    let status = syd()
        .args([
            "-plib",
            "-msandbox/force:on",
            "-mforce/default:warn",
            &format!("-mforce+{bin_true}:{md5_sum}:kill"),
            "--",
            "true",
        ])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    let status = syd()
        .args([
            "-plib",
            "-msandbox/force:on",
            "-mforce/default:warn",
            &format!("-mforce+{bin_true}:{sha1_sum}:kill"),
            "--",
            "true",
        ])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    let status = syd()
        .args([
            "-plib",
            "-msandbox/force:on",
            "-mforce/default:warn",
            &format!("-mforce+{bin_true}:{sha256_sum}:kill"),
            "--",
            "true",
        ])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    let status = syd()
        .args([
            "-plib",
            "-msandbox/force:on",
            "-mforce/default:warn",
            &format!("-mforce+{bin_true}:{sha384_sum}:kill"),
            "--",
            "true",
        ])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    let status = syd()
        .args([
            "-plib",
            "-msandbox/force:on",
            "-mforce/default:warn",
            &format!("-mforce+{bin_true}:{sha512_sum}:kill"),
            "--",
            "true",
        ])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    let status = syd()
        .args([
            "-plib",
            "-msandbox/force:on",
            "-mforce/default:warn",
            &format!("-mforce+{bin_true}:{bad_md5}:kill"),
            "--",
            "true",
        ])
        .status()
        .expect("execute syd");
    assert_status_killed!(status);

    let status = syd()
        .args([
            "-plib",
            "-msandbox/force:on",
            "-mforce/default:warn",
            &format!("-mforce+{bin_true}:{bad_sha1}:kill"),
            "--",
            "true",
        ])
        .status()
        .expect("execute syd");
    assert_status_killed!(status);

    let status = syd()
        .args([
            "-plib",
            "-msandbox/force:on",
            "-mforce/default:warn",
            &format!("-mforce+{bin_true}:{bad_sha256}:kill"),
            "--",
            "true",
        ])
        .status()
        .expect("execute syd");
    assert_status_killed!(status);

    let status = syd()
        .args([
            "-plib",
            "-msandbox/force:on",
            "-mforce/default:warn",
            &format!("-mforce+{bin_true}:{bad_sha384}:kill"),
            "--",
            "true",
        ])
        .status()
        .expect("execute syd");
    assert_status_killed!(status);

    let status = syd()
        .args([
            "-plib",
            "-msandbox/force:on",
            "-mforce/default:warn",
            &format!("-mforce+{bin_true}:{bad_sha512}:kill"),
            "--",
            "true",
        ])
        .status()
        .expect("execute syd");
    assert_status_killed!(status);

    Ok(())
}

fn test_syd_segvguard_core_safe() -> TestResult {
    if !is_program_available("perl") {
        eprintln!("Test requires Perl, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }
    if let Err(errno) = enable_coredumps() {
        eprintln!("Failed to enable coredumps: {errno}");
        eprintln!("Skipping test!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }
    let syd_do = &SYD_DO.to_string();
    let status = syd()
        .env("SYD_DO", "segv")
        .args([
            "-plib",
            "-mlock:exec",
            "-mtrace/allow_unsafe_prlimit:0",
            "-msandbox/exec:on",
            "-mallow/exec+/***",
            "--",
            "perl",
            "-e",
        ])
        .arg(format!(
            r#"
use strict;
use warnings;
use POSIX qw(:errno_h :signal_h :sys_wait_h);

# WCOREDUMP is not POSIX.
sub wcoredump {{
    my $status = shift;
    return ($status & 128) != 0;
}}

# Common variables.
my $pid;
my $sig;
my $code;
my $status;

# segvguard/maxcrashes is set to 5 by default.
my $coredumps = 0; # safe mode, may never trigger segvguard.
for my $i (1..5) {{
    $pid = fork();
    if ($pid == 0) {{ # Child process
        exec('{syd_do}') or exit($!); # Exit with the actual errno
    }}

    waitpid($pid, 0);
    $status = $?;
    if (WIFEXITED($status)) {{
        $code = WEXITSTATUS($status);
        die "process $i did not dump core but exited with code $code\n";
    }} elsif (WIFSIGNALED($status)) {{
        warn "process $i was terminated by signal " . WTERMSIG($status) . "\n";
        if (wcoredump($status)) {{
            $coredumps += 1;
            warn "process $i dumped core.\n";
        }} else {{
            warn "process $i did not dump core.\n";
        }}
    }} else {{
        die "process $i exited unexpectedly with status $status\n";
    }}
}}

if (!$coredumps) {{
    warn "failed to trigger segvguard due to safe mode, skipping test!\n";
    exit 0;
}} else {{
    warn "caused $coredumps coredumps, proceeding...\n";
}}

# Now segvguard must block
$pid = fork();
if ($pid == 0) {{
    exec('{syd_do}') or exit($!); # Exit with the actual errno
}}
waitpid($pid, 0);
$status = $?;
if (WIFSIGNALED($status)) {{
    $sig = WTERMSIG($status);
    if ($sig == SIGKILL) {{
        warn "execution was prevented by segvguard\n";
    }} else {{
        die "process was terminated with unexpected signal $sig\n";
    }}
}} else {{
    die "process exited unexpectedly with status $status\n";
}}

# Ensure segvguard allows everything else.
$pid = fork();
if ($pid == 0) {{
    exec('true') or exit($!); # Exit with the actual errno
}}
waitpid($pid, 0);
$status = $?;
if (WIFEXITED($status)) {{
    $code = WEXITSTATUS($status);
    if ($code == 0) {{
        warn "execution was allowed by segvguard\n";
    }} else {{
        die "process exited with unexpected code $code\n";
    }}
}} else {{
    die "process exited unexpectedly with status $status\n";
}}

# Disable segvguard and retry!
if (-c '/dev/syd/segvguard/expiry:0') {{
    warn "segvguard disabled"
}} else {{
    die "failed to disable segvguard"
}}

# Now segvguard must allow
$pid = fork();
if ($pid == 0) {{
    exec('{syd_do}') or exit($!); # Exit with the actual errno
}}
waitpid($pid, 0);
$status = $?;
if (WIFEXITED($status)) {{
    $code = WEXITSTATUS($status);
    die "process was not terminated but exited with code $code\n";
}} elsif (WIFSIGNALED($status)) {{
    warn "process was terminated by signal " . WTERMSIG($status) . "\n";
    if (wcoredump($status)) {{
        warn "process dumped core.\n";
    }} else {{
        warn "process did not dump core.\n";
    }}
}} else {{
    die "process exited unexpectedly with status $status\n";
}}

1;
    "#,
        ))
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_segvguard_core_unsafe() -> TestResult {
    if !is_program_available("perl") {
        eprintln!("Test requires Perl, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }
    if let Err(errno) = enable_coredumps() {
        eprintln!("Failed to enable coredumps: {errno}");
        eprintln!("Skipping test!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }
    let syd_do = &SYD_DO.to_string();
    let status = syd()
        .env("SYD_DO", "segv")
        .args([
            "-plib",
            "-mlock:exec",
            "-mtrace/allow_unsafe_prlimit:1",
            "-msandbox/exec:on",
            "-mallow/exec+/***",
            "--",
            "perl",
            "-e",
        ])
        .arg(format!(
            r#"
use strict;
use warnings;
use POSIX qw(:errno_h :signal_h :sys_wait_h);

# WCOREDUMP is not POSIX.
sub wcoredump {{
    my $status = shift;
    return ($status & 128) != 0;
}}

# Common variables.
my $pid;
my $sig;
my $code;
my $status;

# segvguard/maxcrashes is set to 5 by default.
for my $i (1..5) {{
    $pid = fork();
    if ($pid == 0) {{ # Child process
        exec('{syd_do}') or exit($!); # Exit with the actual errno
    }}

    waitpid($pid, 0);
    $status = $?;
    if (WIFEXITED($status)) {{
        $code = WEXITSTATUS($status);
        die "process $i did not dump core but exited with code $code\n";
    }} elsif (WIFSIGNALED($status) && wcoredump($status)) {{
        warn "process $i dumped core as expected\n";
    }} else {{
        die "process $i exited unexpectedly with status $status\n";
    }}
}}

# Now segvguard must block
$pid = fork();
if ($pid == 0) {{
    exec('{syd_do}') or exit($!); # Exit with the actual errno
}}
waitpid($pid, 0);
$status = $?;
if (WIFSIGNALED($status)) {{
    $sig = WTERMSIG($status);
    if ($sig == SIGKILL) {{
        warn "execution was prevented by segvguard\n";
    }} else {{
        die "process was terminated with unexpected signal $sig\n";
    }}
}} else {{
    die "process exited unexpectedly with status $status\n";
}}

# Ensure segvguard allows everything else.
$pid = fork();
if ($pid == 0) {{
    exec('true') or exit($!); # Exit with the actual errno
}}
waitpid($pid, 0);
$status = $?;
if (WIFEXITED($status)) {{
    $code = WEXITSTATUS($status);
    if ($code == 0) {{
        warn "execution was allowed by segvguard\n";
    }} else {{
        die "process exited with unexpected code $code\n";
    }}
}} else {{
    die "process exited unexpectedly with status $status\n";
}}

# Disable segvguard and retry!
if (-c '/dev/syd/segvguard/expiry:0') {{
    warn "segvguard disabled"
}} else {{
    die "failed to disable segvguard"
}}

# Now segvguard must allow
$pid = fork();
if ($pid == 0) {{
    exec('{syd_do}') or exit($!); # Exit with the actual errno
}}
waitpid($pid, 0);
$status = $?;
if (WIFEXITED($status)) {{
    $code = WEXITSTATUS($status);
    die "process did not dump core but exited with code $code\n";
}} elsif (WIFSIGNALED($status) && wcoredump($status)) {{
    warn "process dumped core as expected\n";
}} else {{
    die "process exited unexpectedly with status $status\n";
}}

1;
    "#,
        ))
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_segvguard_suspension_safe() -> TestResult {
    if !is_program_available("perl") {
        eprintln!("Test requires Perl, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }
    if let Err(errno) = enable_coredumps() {
        eprintln!("Failed to enable coredumps: {errno}");
        eprintln!("Skipping test!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }
    let syd_do = &SYD_DO.to_string();
    let status = syd()
        .env("SYD_DO", "segv")
        .args([
            "-plib",
            "-mlock:exec",
            "-mtrace/allow_unsafe_prlimit:0",
            "-msandbox/exec:on",
            "-mallow/exec+/***",
            "--",
            "perl",
            "-e",
        ])
        .arg(format!(
            r#"
use strict;
use warnings;
use POSIX qw(:errno_h :sys_wait_h);

# WCOREDUMP is not POSIX.
sub wcoredump {{
    my $status = shift;
    return ($status & 128) != 0;
}}

# Common variables.
my $pid;
my $code;
my $status;

# Set segvguard suspension to 10 seconds.
if (-c '/dev/syd/segvguard/suspension:10') {{
    warn "segvguard suspension set to 10 seconds"
}} else {{
    die "failed to set segvguard suspension"
}}

my $coredumps = 0; # safe mode, may never trigger segvguard.
for my $i (1..5) {{
    $pid = fork();
    if ($pid == 0) {{ # Child process
        exec('{syd_do}') or exit($!); # Exit with the actual errno
    }}

    waitpid($pid, 0);
    $status = $?;
    if (WIFEXITED($status)) {{
        $code = WEXITSTATUS($status);
        die "process $i did not dump core but exited with code $code\n";
    }} elsif (WIFSIGNALED($status)) {{
        warn "process $i was terminated by signal " . WTERMSIG($status) . "\n";
        if (wcoredump($status)) {{
            $coredumps += 1;
            warn "process $i dumped core.\n";
        }} else {{
            warn "process $i did not dump core.\n";
        }}
    }} else {{
        die "process $i exited unexpectedly with status $status\n";
    }}
}}

if (!$coredumps) {{
    warn "failed to trigger segvguard due to safe mode, skipping test!\n";
    exit 0;
}} else {{
    warn "caused $coredumps coredumps, proceeding...\n";
}}

# Ensure segvguard suspension expires.
warn "waiting 10 seconds for segvguard suspension to expire...";
sleep 10;

# Now segvguard must allow.
$pid = fork();
if ($pid == 0) {{
    exec('{syd_do}') or exit($!); # Exit with the actual errno
}}
waitpid($pid, 0);
$status = $?;
if (WIFEXITED($status)) {{
    $code = WEXITSTATUS($status);
    die "process did not dump core but exited with code $code\n";
}} elsif (WIFSIGNALED($status) && wcoredump($status)) {{
    warn "process dumped core as expected\n";
}} else {{
    die "process exited unexpectedly with status $status\n";
}}

1;
    "#,
        ))
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_segvguard_suspension_unsafe() -> TestResult {
    if !is_program_available("perl") {
        eprintln!("Test requires Perl, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }
    if let Err(errno) = enable_coredumps() {
        eprintln!("Failed to enable coredumps: {errno}");
        eprintln!("Skipping test!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }
    let syd_do = &SYD_DO.to_string();
    let status = syd()
        .env("SYD_DO", "segv")
        .args([
            "-plib",
            "-mlock:exec",
            "-mtrace/allow_unsafe_prlimit:1",
            "-msandbox/exec:on",
            "-mallow/exec+/***",
            "--",
            "perl",
            "-e",
        ])
        .arg(format!(
            r#"
use strict;
use warnings;
use POSIX qw(:errno_h :sys_wait_h);

# WCOREDUMP is not POSIX.
sub wcoredump {{
    my $status = shift;
    return ($status & 128) != 0;
}}

# Common variables.
my $pid;
my $code;
my $status;

# Set segvguard suspension to 10 seconds.
if (-c '/dev/syd/segvguard/suspension:10') {{
    warn "segvguard suspension set to 10 seconds"
}} else {{
    die "failed to set segvguard suspension"
}}

# segvguard/maxcrashes is set to 5 by default.
for my $i (1..5) {{
    $pid = fork();
    if ($pid == 0) {{ # Child process
        exec('{syd_do}') or exit($!); # Exit with the actual errno
    }}

    waitpid($pid, 0);
    $status = $?;
    if (WIFEXITED($status)) {{
        $code = WEXITSTATUS($status);
        die "process $i did not dump core but exited with code $code\n";
    }} elsif (WIFSIGNALED($status) && wcoredump($status)) {{
        warn "process $i dumped core as expected\n";
    }} else {{
        die "process $i exited unexpectedly with status $status\n";
    }}
}}

# Ensure segvguard suspension expires.
warn "waiting 10 seconds for segvguard suspension to expire...";
sleep 10;

# Now segvguard must allow.
$pid = fork();
if ($pid == 0) {{
    exec('{syd_do}') or exit($!); # Exit with the actual errno
}}
waitpid($pid, 0);
$status = $?;
if (WIFEXITED($status)) {{
    $code = WEXITSTATUS($status);
    die "process did not dump core but exited with code $code\n";
}} elsif (WIFSIGNALED($status) && wcoredump($status)) {{
    warn "process dumped core as expected\n";
}} else {{
    die "process exited unexpectedly with status $status\n";
}}

1;
    "#,
        ))
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_symlink_toctou() -> TestResult {
    let status = syd()
        .env("SYD_DO", "symlink_toctou")
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-msandbox/stat:on",
            "-msandbox/write:on",
            "-mallow/read+/***",
            "-mdeny/stat+/etc/***",
            "-mallow/stat+/etc/ld*",
            "-mdeny/read+/etc/passwd",
            "-mdeny/write+/etc/passwd",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_symlinkat_toctou() -> TestResult {
    let status = syd()
        .env("SYD_DO", "symlinkat_toctou")
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-msandbox/stat:on",
            "-msandbox/write:on",
            "-mallow/read+/***",
            "-mdeny/stat+/etc/***",
            "-mallow/stat+/etc/ld*",
            "-mdeny/read+/etc/passwd",
            "-mdeny/write+/etc/passwd",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_ptrmod_toctou_chdir() -> TestResult {
    let status = syd()
        .env("SYD_DO", "ptrmod_toctou_chdir")
        .args([
            "-plib",
            "-msandbox/stat:on",
            "-mallow/stat+/***",
            "-mdeny/stat+/var/empty/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    // FIXME: https://bugzilla.kernel.org/show_bug.cgi?id=218501
    // Note for some reason this test reliably passes on Ubuntu CI,
    // so we do not use fixup! here.
    ignore!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_ptrmod_toctou_exec_fail() -> TestResult {
    let status = syd()
        .env("SYD_DO", "ptrmod_toctou_exec_fail")
        .args([
            "-plib",
            "-msandbox/exec:on",
            "-mallow/exec+/***",
            "-mdeny/exec+/**/toctou_exec",
            "-mfilter/exec+/**/toctou_exec",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    // FIXME: https://bugzilla.kernel.org/show_bug.cgi?id=218501
    ignore!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_ptrmod_toctou_exec_success_quick() -> TestResult {
    // Test requires /bin/false to be denylisted.
    // false may point to various alternatives such
    // as gfalse, coreutils, busybox etc.
    let f = Path::new("/bin/false").canonicalize().expect("/bin/false");
    let status = syd()
        .env("SYD_DO", "ptrmod_toctou_exec_success_quick")
        .args([
            "-plib",
            "-msandbox/exec:on",
            "-mallow/exec+/***",
            "-mdeny/exec+/**/toctou_exec",
            "-mfilter/exec+/**/toctou_exec",
            &format!("-mdeny/exec+{}", f.display()),
            &format!("-mfilter/exec+{}", f.display()),
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    const EXKILL: i32 = 128 + nix::libc::SIGKILL;
    assert!(
        matches!(status.code().unwrap_or(127), 0 | EXKILL),
        "status:{status:?}"
    );
    Ok(())
}

fn test_syd_ptrmod_toctou_exec_success_double_fork() -> TestResult {
    // Test requires /bin/false to be denylisted.
    // false may point to various alternatives such
    // as gfalse, coreutils, busybox etc.
    let f = Path::new("/bin/false").canonicalize().expect("/bin/false");
    let status = syd()
        .env("SYD_DO", "ptrmod_toctou_exec_success_double_fork")
        .args([
            "-plib",
            "-msandbox/exec:on",
            "-mallow/exec+/***",
            "-mdeny/exec+/**/toctou_exec",
            "-mfilter/exec+/**/toctou_exec",
            &format!("-mdeny/exec+{}", f.display()),
            &format!("-mfilter/exec+{}", f.display()),
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    const EXKILL: i32 = 128 + nix::libc::SIGKILL;
    assert!(
        matches!(status.code().unwrap_or(127), 0 | EXKILL),
        "status:{status:?}"
    );
    Ok(())
}

fn test_syd_ptrmod_toctou_exec_success_quick_no_mitigation() -> TestResult {
    // Test requires /bin/false to be denylisted.
    // false may point to various alternatives such
    // as gfalse, coreutils, busybox etc.
    let f = Path::new("/bin/false").canonicalize().expect("/bin/false");
    let status = syd()
        .env("SYD_DO", "ptrmod_toctou_exec_success_quick")
        .args([
            "-plib",
            "-mtrace/allow_unsafe_exec:1",
            "-msandbox/exec:on",
            "-mallow/exec+/***",
            "-mdeny/exec+/**/toctou_exec",
            "-mfilter/exec+/**/toctou_exec",
            &format!("-mdeny/exec+{}", f.display()),
            &format!("-mfilter/exec+{}", f.display()),
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    const EXKILL: i32 = 128 + nix::libc::SIGKILL;
    ignore!(
        matches!(status.code().unwrap_or(127), 0 | EXKILL),
        "status:{status:?}"
    );
    Ok(())
}

fn test_syd_ptrmod_toctou_exec_success_double_fork_no_mitigation() -> TestResult {
    // Test requires /bin/false to be denylisted.
    // false may point to various alternatives such
    // as gfalse, coreutils, busybox etc.
    let f = Path::new("/bin/false").canonicalize().expect("/bin/false");
    let status = syd()
        .env("SYD_DO", "ptrmod_toctou_exec_success_double_fork")
        .args([
            "-plib",
            "-mtrace/allow_unsafe_exec:1",
            "-msandbox/exec:on",
            "-mallow/exec+/***",
            "-mdeny/exec+/**/toctou_exec",
            "-mfilter/exec+/**/toctou_exec",
            &format!("-mdeny/exec+{}", f.display()),
            &format!("-mfilter/exec+{}", f.display()),
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    const EXKILL: i32 = 128 + nix::libc::SIGKILL;
    ignore!(
        matches!(status.code().unwrap_or(127), 0 | EXKILL),
        "status:{status:?}"
    );
    Ok(())
}

fn test_syd_ptrmod_toctou_open() -> TestResult {
    let status = syd()
        .env("SYD_DO", "ptrmod_toctou_open")
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-msandbox/stat:on",
            "-msandbox/write:on",
            "-mallow/read+/***",
            "-mdeny/stat+/etc/***",
            "-mallow/stat+/etc/ld*",
            "-mdeny/read+/etc/passwd",
            "-mdeny/write+/etc/passwd",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_ptrmod_toctou_creat() -> TestResult {
    let status = syd()
        .env("SYD_DO", "ptrmod_toctou_creat")
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-msandbox/stat:on",
            "-msandbox/write:on",
            "-mallow/read+/***",
            "-mdeny/stat+/etc/***",
            "-mallow/stat+/etc/ld*",
            "-mdeny/write+/**/deny.syd-tmp",
            "-mfilter/write+/**/deny.syd-tmp",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_ptrmod_toctou_opath() -> TestResult {
    let status = syd()
        .env("SYD_DO", "ptrmod_toctou_opath")
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-msandbox/stat:on",
            "-msandbox/write:on",
            "-mallow/read+/***",
            "-mallow/stat+/etc",
            "-mdeny/stat+/etc/**",
            "-mallow/stat+/etc/ld*",
            "-mdeny/read,stat,write+/etc/passwd",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    // FIXME: https://bugzilla.kernel.org/show_bug.cgi?id=218501
    fixup!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_io_uring_escape_strict() -> TestResult {
    #[cfg(feature = "uring")]
    {
        // Step 1: Default is strict.
        let status = syd()
            .env("SYD_DO", "io_uring_escape")
            .args([
                "-ppaludis",
                "-msandbox/lock:off",
                "-msandbox/read:on",
                "-msandbox/stat:on",
                "-msandbox/write:on",
                "-mallow/read+/***",
                "-mdeny/stat+/etc/***",
                "-mallow/stat+/etc/ld*",
                "-mdeny/read,write+/etc/passwd",
                "--",
                &SYD_DO,
                "0",
            ])
            .status()
            .expect("execute syd");
        assert_status_ok!(status);
    }
    Ok(())
}

fn test_syd_io_uring_escape_unsafe() -> TestResult {
    #[cfg(feature = "uring")]
    {
        // Step 2: Relax uring restriction.
        let status = syd()
            .env("SYD_DO", "io_uring_escape")
            .args([
                "-ppaludis",
                "-msandbox/lock:off",
                "-msandbox/read:on",
                "-msandbox/stat:on",
                "-msandbox/write:on",
                "-mallow/read+/***",
                "-mdeny/stat+/etc/***",
                "-mallow/stat+/etc/ld*",
                "-mdeny/read,write+/etc/passwd",
                "-mtrace/allow_unsafe_uring:1",
                "--",
                &SYD_DO,
                "1",
            ])
            .status()
            .expect("execute syd");
        assert_status_ok!(status);
    }

    Ok(())
}

fn test_syd_opath_escape() -> TestResult {
    let status = syd()
        .env("SYD_DO", "opath_escape")
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-msandbox/stat:on",
            "-msandbox/write:on",
            "-mallow/read+/***",
            "-mdeny/stat+/etc/***",
            "-mallow/stat+/etc/ld*",
            "-mdeny/read+/etc/passwd",
            "-mdeny/write+/etc/passwd",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_devfd_escape_chdir() -> TestResult {
    let status = syd()
        .env("SYD_DO", "devfd_escape_chdir")
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-msandbox/write:on",
            "-mallow/read+/***",
            "-mallow/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_devfd_escape_chdir_relpath_1() -> TestResult {
    let status = syd()
        .env("SYD_DO", "devfd_escape_chdir_relpath_1")
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-msandbox/write:on",
            "-mallow/read+/***",
            "-mallow/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_devfd_escape_chdir_relpath_2() -> TestResult {
    let status = syd()
        .env("SYD_DO", "devfd_escape_chdir_relpath_2")
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-msandbox/write:on",
            "-mallow/read+/***",
            "-mallow/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_devfd_escape_chdir_relpath_3() -> TestResult {
    let status = syd()
        .env("SYD_DO", "devfd_escape_chdir_relpath_3")
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-msandbox/write:on",
            "-mallow/read+/***",
            "-mallow/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_devfd_escape_chdir_relpath_4() -> TestResult {
    let status = syd()
        .env("SYD_DO", "devfd_escape_chdir_relpath_4")
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-msandbox/write:on",
            "-mallow/read+/***",
            "-mallow/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_devfd_escape_chdir_relpath_5() -> TestResult {
    let status = syd()
        .env("SYD_DO", "devfd_escape_chdir_relpath_5")
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-msandbox/write:on",
            "-mallow/read+/***",
            "-mallow/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_devfd_escape_chdir_relpath_6() -> TestResult {
    let status = syd()
        .env("SYD_DO", "devfd_escape_chdir_relpath_6")
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-msandbox/write:on",
            "-mallow/read+/***",
            "-mallow/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_devfd_escape_chdir_relpath_7() -> TestResult {
    let status = syd()
        .env("SYD_DO", "devfd_escape_chdir_relpath_7")
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-msandbox/write:on",
            "-mallow/read+/***",
            "-mallow/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_devfd_escape_chdir_relpath_8() -> TestResult {
    let status = syd()
        .env("SYD_DO", "devfd_escape_chdir_relpath_8")
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-msandbox/write:on",
            "-mallow/read+/***",
            "-mallow/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_devfd_escape_chdir_relpath_9() -> TestResult {
    let status = syd()
        .env("SYD_DO", "devfd_escape_chdir_relpath_9")
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-msandbox/write:on",
            "-mallow/read+/***",
            "-mallow/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_devfd_escape_chdir_relpath_10() -> TestResult {
    let status = syd()
        .env("SYD_DO", "devfd_escape_chdir_relpath_10")
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-msandbox/write:on",
            "-mallow/read+/***",
            "-mallow/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_devfd_escape_chdir_relpath_11() -> TestResult {
    let status = syd()
        .env("SYD_DO", "devfd_escape_chdir_relpath_11")
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-msandbox/write:on",
            "-mallow/read+/***",
            "-mallow/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_devfd_escape_chdir_relpath_12() -> TestResult {
    let status = syd()
        .env("SYD_DO", "devfd_escape_chdir_relpath_12")
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-msandbox/write:on",
            "-mallow/read+/***",
            "-mallow/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_devfd_escape_chdir_relpath_13() -> TestResult {
    let status = syd()
        .env("SYD_DO", "devfd_escape_chdir_relpath_13")
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-msandbox/write:on",
            "-mallow/read+/***",
            "-mallow/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_devfd_escape_chdir_relpath_14() -> TestResult {
    let status = syd()
        .env("SYD_DO", "devfd_escape_chdir_relpath_14")
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-msandbox/write:on",
            "-mallow/read+/***",
            "-mallow/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_devfd_escape_chdir_relpath_15() -> TestResult {
    let status = syd()
        .env("SYD_DO", "devfd_escape_chdir_relpath_15")
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-msandbox/write:on",
            "-mallow/read+/***",
            "-mallow/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_devfd_escape_chdir_relpath_16() -> TestResult {
    let status = syd()
        .env("SYD_DO", "devfd_escape_chdir_relpath_16")
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-msandbox/write:on",
            "-mallow/read+/***",
            "-mallow/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_devfd_escape_chdir_relpath_17() -> TestResult {
    let status = syd()
        .env("SYD_DO", "devfd_escape_chdir_relpath_17")
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-msandbox/write:on",
            "-mallow/read+/***",
            "-mallow/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_devfd_escape_chdir_relpath_18() -> TestResult {
    let status = syd()
        .env("SYD_DO", "devfd_escape_chdir_relpath_18")
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-msandbox/write:on",
            "-mallow/read+/***",
            "-mallow/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_devfd_escape_chdir_relpath_19() -> TestResult {
    let status = syd()
        .env("SYD_DO", "devfd_escape_chdir_relpath_19")
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-msandbox/write:on",
            "-mallow/read+/***",
            "-mallow/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_devfd_escape_chdir_relpath_20() -> TestResult {
    let status = syd()
        .env("SYD_DO", "devfd_escape_chdir_relpath_20")
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-msandbox/write:on",
            "-mallow/read+/***",
            "-mallow/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_devfd_escape_open() -> TestResult {
    let status = syd()
        .env("SYD_DO", "devfd_escape_open")
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-msandbox/write:on",
            "-mallow/read+/***",
            "-mallow/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_devfd_escape_open_relpath_1() -> TestResult {
    let status = syd()
        .env("SYD_DO", "devfd_escape_open_relpath_1")
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-msandbox/write:on",
            "-mallow/read+/***",
            "-mallow/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_devfd_escape_open_relpath_2() -> TestResult {
    let status = syd()
        .env("SYD_DO", "devfd_escape_open_relpath_2")
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-msandbox/write:on",
            "-mallow/read+/***",
            "-mallow/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_devfd_escape_open_relpath_3() -> TestResult {
    let status = syd()
        .env("SYD_DO", "devfd_escape_open_relpath_3")
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-msandbox/write:on",
            "-mallow/read+/***",
            "-mallow/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_devfd_escape_open_relpath_4() -> TestResult {
    let status = syd()
        .env("SYD_DO", "devfd_escape_open_relpath_4")
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-msandbox/write:on",
            "-mallow/read+/***",
            "-mallow/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_devfd_escape_open_relpath_5() -> TestResult {
    let status = syd()
        .env("SYD_DO", "devfd_escape_open_relpath_5")
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-msandbox/write:on",
            "-mallow/read+/***",
            "-mallow/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_devfd_escape_open_relpath_6() -> TestResult {
    let status = syd()
        .env("SYD_DO", "devfd_escape_open_relpath_6")
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-msandbox/write:on",
            "-mallow/read+/***",
            "-mallow/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_devfd_escape_open_relpath_7() -> TestResult {
    let status = syd()
        .env("SYD_DO", "devfd_escape_open_relpath_7")
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-msandbox/write:on",
            "-mallow/read+/***",
            "-mallow/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_devfd_escape_open_relpath_8() -> TestResult {
    let status = syd()
        .env("SYD_DO", "devfd_escape_open_relpath_8")
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-msandbox/write:on",
            "-mallow/read+/***",
            "-mallow/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_devfd_escape_open_relpath_9() -> TestResult {
    let status = syd()
        .env("SYD_DO", "devfd_escape_open_relpath_9")
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-msandbox/write:on",
            "-mallow/read+/***",
            "-mallow/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_devfd_escape_open_relpath_10() -> TestResult {
    let status = syd()
        .env("SYD_DO", "devfd_escape_open_relpath_10")
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-msandbox/write:on",
            "-mallow/read+/***",
            "-mallow/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_devfd_escape_open_relpath_11() -> TestResult {
    let status = syd()
        .env("SYD_DO", "devfd_escape_open_relpath_11")
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-msandbox/write:on",
            "-mallow/read+/***",
            "-mallow/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_devfd_escape_open_relpath_12() -> TestResult {
    let status = syd()
        .env("SYD_DO", "devfd_escape_open_relpath_12")
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-msandbox/write:on",
            "-mallow/read+/***",
            "-mallow/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_devfd_escape_open_relpath_13() -> TestResult {
    let status = syd()
        .env("SYD_DO", "devfd_escape_open_relpath_13")
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-msandbox/write:on",
            "-mallow/read+/***",
            "-mallow/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_devfd_escape_open_relpath_14() -> TestResult {
    let status = syd()
        .env("SYD_DO", "devfd_escape_open_relpath_14")
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-msandbox/write:on",
            "-mallow/read+/***",
            "-mallow/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_devfd_escape_open_relpath_15() -> TestResult {
    let status = syd()
        .env("SYD_DO", "devfd_escape_open_relpath_15")
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-msandbox/write:on",
            "-mallow/read+/***",
            "-mallow/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_devfd_escape_open_relpath_16() -> TestResult {
    let status = syd()
        .env("SYD_DO", "devfd_escape_open_relpath_16")
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-msandbox/write:on",
            "-mallow/read+/***",
            "-mallow/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_devfd_escape_open_relpath_17() -> TestResult {
    let status = syd()
        .env("SYD_DO", "devfd_escape_open_relpath_17")
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-msandbox/write:on",
            "-mallow/read+/***",
            "-mallow/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_devfd_escape_open_relpath_18() -> TestResult {
    let status = syd()
        .env("SYD_DO", "devfd_escape_open_relpath_18")
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-msandbox/write:on",
            "-mallow/read+/***",
            "-mallow/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_devfd_escape_open_relpath_19() -> TestResult {
    let status = syd()
        .env("SYD_DO", "devfd_escape_open_relpath_19")
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-msandbox/write:on",
            "-mallow/read+/***",
            "-mallow/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_devfd_escape_open_relpath_20() -> TestResult {
    let status = syd()
        .env("SYD_DO", "devfd_escape_open_relpath_20")
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-msandbox/write:on",
            "-mallow/read+/***",
            "-mallow/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_procself_escape_chdir() -> TestResult {
    let status = syd()
        .env("SYD_DO", "procself_escape_chdir")
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-mallow/read+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_procself_escape_chdir_relpath_1() -> TestResult {
    let status = syd()
        .env("SYD_DO", "procself_escape_chdir_relpath_1")
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-mallow/read+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_procself_escape_chdir_relpath_2() -> TestResult {
    let status = syd()
        .env("SYD_DO", "procself_escape_chdir_relpath_2")
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-mallow/read+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_procself_escape_chdir_relpath_3() -> TestResult {
    let status = syd()
        .env("SYD_DO", "procself_escape_chdir_relpath_3")
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-mallow/read+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_procself_escape_chdir_relpath_4() -> TestResult {
    let status = syd()
        .env("SYD_DO", "procself_escape_chdir_relpath_4")
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-mallow/read+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_procself_escape_chdir_relpath_5() -> TestResult {
    let status = syd()
        .env("SYD_DO", "procself_escape_chdir_relpath_5")
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-mallow/read+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_procself_escape_chdir_relpath_6() -> TestResult {
    let status = syd()
        .env("SYD_DO", "procself_escape_chdir_relpath_6")
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-mallow/read+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_procself_escape_chdir_relpath_7() -> TestResult {
    let status = syd()
        .env("SYD_DO", "procself_escape_chdir_relpath_7")
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-mallow/read+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_procself_escape_open() -> TestResult {
    let status = syd()
        .env("SYD_DO", "procself_escape_open")
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-mallow/read+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_procself_escape_open_relpath_1() -> TestResult {
    let status = syd()
        .env("SYD_DO", "procself_escape_open_relpath_1")
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-mallow/read+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_procself_escape_open_relpath_2() -> TestResult {
    let status = syd()
        .env("SYD_DO", "procself_escape_open_relpath_2")
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-mallow/read+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_procself_escape_open_relpath_3() -> TestResult {
    let status = syd()
        .env("SYD_DO", "procself_escape_open_relpath_3")
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-mallow/read+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_procself_escape_open_relpath_4() -> TestResult {
    let status = syd()
        .env("SYD_DO", "procself_escape_open_relpath_4")
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-mallow/read+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_procself_escape_open_relpath_5() -> TestResult {
    let status = syd()
        .env("SYD_DO", "procself_escape_open_relpath_5")
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-mallow/read+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_procself_escape_open_relpath_6() -> TestResult {
    let status = syd()
        .env("SYD_DO", "procself_escape_open_relpath_6")
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-mallow/read+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_procself_escape_open_relpath_7() -> TestResult {
    let status = syd()
        .env("SYD_DO", "procself_escape_open_relpath_7")
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-mallow/read+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_procself_escape_relpath() -> TestResult {
    if !is_program_available("grep") {
        eprintln!("Test requires grep, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }
    eprintln!("+ grep 'Name:[[:space:]]syd' /proc/./self/status");
    let status = syd()
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-mallow/read+/***",
            "--",
            "grep",
            "Name:[[:space:]]syd",
            "/proc/./self/status",
        ])
        .status()
        .expect("execute syd");
    assert!(
        status.code().unwrap_or(127) == 1,
        "code:{:?}",
        status.code()
    );
    Ok(())
}

fn test_syd_procself_escape_symlink() -> TestResult {
    let status = syd()
        .env("SYD_DO", "procself_escape_symlink")
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-mallow/read+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_procself_escape_symlink_within_container() -> TestResult {
    match check_unshare() {
        Some(false) => {
            eprintln!("Test requires Linux namespaces, skipping!");
            env::set_var("SYD_TEST_SOFT_FAIL", "1");
            return Ok(());
        }
        None => {
            return Err(TestError(
                "Failed to test for Linux namespaces!".to_string(),
            ));
        }
        _ => {}
    };

    let status = syd()
        .env("SYD_DO", "procself_escape_symlink")
        .args([
            "-mallow/read+/***",
            "-ppaludis",
            "-pcontainer",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_umask_bypass_077() -> TestResult {
    // Set a liberal umask as the test expects.
    let prev_umask = umask(Mode::from_bits_truncate(0o022));
    let status = syd()
        .env("SYD_DO", "umask_bypass_077")
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/write:on",
            "-mallow/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    let _ = umask(prev_umask);

    assert_status_ok!(status);
    Ok(())
}

fn test_syd_umask_bypass_277() -> TestResult {
    // Set a liberal umask as the test expects.
    let prev_umask = umask(Mode::from_bits_truncate(0o022));
    let status = syd()
        .env("SYD_DO", "umask_bypass_277")
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/write:on",
            "-mallow/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    let _ = umask(prev_umask);

    assert_status_ok!(status);
    Ok(())
}

fn test_syd_emulate_opath() -> TestResult {
    let status = syd()
        .env("SYD_DO", "emulate_opath")
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-mallow/read+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");

    assert_status_ok!(status);
    Ok(())
}

fn test_syd_emulate_otmpfile() -> TestResult {
    let status = syd()
        .env("SYD_DO", "emulate_otmpfile")
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/write:on",
            "-mallow/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");

    assert_status_ok!(status);
    Ok(())
}

fn test_syd_honor_umask() -> TestResult {
    let prev_umask = umask(Mode::from_bits_truncate(0o077));
    let status = syd()
        .env("SYD_DO", "honor_umask")
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/write:on",
            "-mallow/write+/***",
            "--",
            &SYD_DO,
            "0600",
        ])
        .status()
        .expect("execute syd");
    let _ = umask(prev_umask);
    assert_status_ok!(status);

    let prev_umask = umask(Mode::from_bits_truncate(0o022));
    let status = syd()
        .env("SYD_DO", "honor_umask")
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/write:on",
            "-mallow/write+/***",
            "--",
            &SYD_DO,
            "0644",
        ])
        .status()
        .expect("execute syd");
    let _ = umask(prev_umask);
    assert_status_ok!(status);

    let prev_umask = umask(Mode::from_bits_truncate(0));
    let status = syd()
        .env("SYD_DO", "honor_umask")
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/write:on",
            "-mallow/write+/***",
            "--",
            &SYD_DO,
            "0666",
        ])
        .status()
        .expect("execute syd");
    let _ = umask(prev_umask);
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_open_utf8_invalid() -> TestResult {
    let status = syd()
        .env("SYD_DO", "open_utf8_invalid")
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/write:on",
            "-mallow/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_exec_in_inaccessible_directory() -> TestResult {
    if !is_program_available("bash") {
        eprintln!("Test requires bash, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }
    let status = syd()
        .env("SYD_DO", "exec_in_inaccessible_directory")
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/exec:on",
            "-msandbox/write:on",
            "-mallow/exec+/***",
            "-mallow/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");

    assert_status_ok!(status);
    Ok(())
}

fn test_syd_fstat_on_pipe() -> TestResult {
    let status = syd()
        .env("SYD_DO", "fstat_on_pipe")
        .args([
            "-ppaludis",
            "-msandbox/stat:on",
            "-mallow/stat+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_fstat_on_socket() -> TestResult {
    let status = syd()
        .env("SYD_DO", "fstat_on_socket")
        .args([
            "-ppaludis",
            "-msandbox/stat:on",
            "-mallow/stat+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_fstat_on_deleted_file() -> TestResult {
    let status = syd()
        .env("SYD_DO", "fstat_on_deleted_file")
        .args([
            "-ppaludis",
            "-msandbox/read:on",
            "-msandbox/stat:on",
            "-msandbox/write:on",
            "-mallow/read+/***",
            "-mallow/stat+/***",
            "-mallow/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_fstat_on_temp_file() -> TestResult {
    let status = syd()
        .env("SYD_DO", "fstat_on_temp_file")
        .args([
            "-ppaludis",
            "-msandbox/read:on",
            "-msandbox/stat:on",
            "-msandbox/write:on",
            "-mallow/read+/***",
            "-mallow/stat+/***",
            "-mallow/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_fchmodat_on_proc_fd() -> TestResult {
    let status = syd()
        .env("SYD_DO", "fchmodat_on_proc_fd")
        .args([
            "-ppaludis",
            "-msandbox/read:on",
            "-msandbox/stat:on",
            "-msandbox/write:on",
            "-mallow/read+/***",
            "-mallow/stat+/***",
            "-mallow/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_linkat_on_fd() -> TestResult {
    let status = syd()
        .env("SYD_DO", "linkat_on_fd")
        .args([
            "-ppaludis",
            "-msandbox/read:on",
            "-msandbox/stat:on",
            "-msandbox/write:on",
            "-mallow/read+/***",
            "-mallow/stat+/***",
            "-mallow/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_block_ioctl_tiocsti() -> TestResult {
    let status = syd()
        .env("SYD_DO", "block_ioctl_tiocsti")
        .args([
            "-ppaludis",
            "-msandbox/read:on",
            "-msandbox/stat:on",
            "-msandbox/write:on",
            "-mallow/read+/***",
            "-mallow/stat+/***",
            "-mallow/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_block_prctl_ptrace() -> TestResult {
    let status = syd()
        .env("SYD_DO", "block_prctl_ptrace")
        .args([
            "-ppaludis",
            "-msandbox/read:on",
            "-msandbox/stat:on",
            "-msandbox/write:on",
            "-mallow/read+/***",
            "-mallow/stat+/***",
            "-mallow/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_prevent_ptrace_detect() -> TestResult {
    let status = syd()
        .env("SYD_DO", "detect_ptrace")
        .args(["-plib", "--", &SYD_DO])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_block_dev_random() -> TestResult {
    let status = syd()
        .env("SYD_DO", "block_dev_random")
        .args([
            "-ppaludis",
            "-msandbox/read:on",
            "-msandbox/stat:on",
            "-msandbox/write:on",
            "-mallow/read,stat,write+/***",
            "-mdeny/read+/dev/random",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_block_dev_urandom() -> TestResult {
    let status = syd()
        .env("SYD_DO", "block_dev_urandom")
        .args([
            "-ppaludis",
            "-msandbox/read:on",
            "-msandbox/stat:on",
            "-msandbox/write:on",
            "-mallow/read,stat,write+/***",
            "-mdeny/read+/dev/urandom",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_kill_during_syscall() -> TestResult {
    let status = syd()
        .env("SYD_DO", "kill_during_syscall")
        .args([
            "-ppaludis",
            "-msandbox/read:on",
            "-msandbox/stat:on",
            "-msandbox/write:on",
            "-mallow/read+/***",
            "-mallow/stat+/***",
            "-mallow/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_open_toolong_path() -> TestResult {
    let status = syd()
        .env("SYD_DO", "open_toolong_path")
        .args([
            "-ppaludis",
            "-msandbox/read:on",
            "-msandbox/stat:on",
            "-msandbox/write:on",
            "-mallow/read+/***",
            "-mallow/stat+/***",
            "-mallow/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_open_null_path() -> TestResult {
    let status = syd()
        .env("SYD_DO", "open_null_path")
        .args([
            "-ppaludis",
            "-msandbox/read:on",
            "-msandbox/stat:on",
            "-msandbox/write:on",
            "-mallow/read+/***",
            "-mallow/stat+/***",
            "-mallow/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_utimensat_null() -> TestResult {
    let status = syd()
        .env("SYD_DO", "utimensat_null")
        .args([
            "-ppaludis",
            "-msandbox/read:on",
            "-msandbox/stat:on",
            "-msandbox/write:on",
            "-mallow/read+/***",
            "-mallow/stat+/***",
            "-mallow/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");

    // Fails on GitLab CI.
    if !*GL_BUILD {
        assert_status_ok!(status);
    } else {
        ignore!(status.success(), "status:{status:?}");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
    }
    Ok(())
}

fn test_syd_normalize_path() -> TestResult {
    if !is_program_available("sh") {
        eprintln!("Test requires sh, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }
    const NORMALIZE_PATH_TESTS: &[&str] = &[
        "null",
        "./null",
        ".////null",
        ".///.////.///./null",
        "./././././././null",
        "./././.././././dev/null",
        "../dev/././../dev/././null",
    ];

    for path in NORMALIZE_PATH_TESTS {
        let status = syd()
            .args([
                "-ppaludis",
                "-msandbox/read:off",
                "-msandbox/stat:off",
                "-msandbox/write:on",
                "-msandbox/exec:off",
                "-mdeny/write+/***",
                "-mallow/write+/dev/null",
                "--",
                "sh",
                "-c",
                &format!("cd /dev; :> {path}"),
            ])
            .status()
            .expect("execute syd");
        assert_eq!(
            status.code().unwrap_or(127),
            0,
            "path:{path}, status:{status:?}"
        );
    }

    Ok(())
}

fn test_syd_path_resolution() -> TestResult {
    let cwd = syd::fs::canonicalize(
        Pid::this(),
        "/proc/self/cwd",
        None,
        true,
        false,
        syd::fs::MissingHandling::Normal,
    )?
    .to_string_lossy()
    .to_string();

    let status = syd()
        .env("SYD_DO", "path_resolution")
        .args([
            "-ppaludis",
            "-msandbox/read:on",
            "-msandbox/stat:on",
            "-msandbox/write:on",
            "-mallow/read+/",
            "-mallow/read+/etc/***",
            "-mallow/read+/lib/***",
            "-mallow/read+/proc/***",
            "-mallow/read+/usr/***",
            "-mallow/stat+/etc/***",
            "-mallow/stat+/lib/***",
            "-mallow/stat+/proc/***",
            "-mallow/stat+/usr/***",
            &format!("-mallow/read+{cwd}/test_file.txt"),
            &format!("-mallow/stat+{cwd}/test_file.txt"),
            &format!("-mallow/write+{cwd}/test_file.txt"),
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_symlink_readonly_path() -> TestResult {
    if !is_program_available("sh") {
        eprintln!("Test requires sh, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }
    let status = syd()
        .args([
            "-ppaludis",
            "-msandbox/read:on",
            "-msandbox/stat:on",
            "-msandbox/write:on",
            "-mallow/read+/***",
            "-mallow/stat+/***",
            "-mallow/write+/***",
            "-mdeny/write+/",
            "--",
            "sh",
            "-c",
            "ln -s / test_syd_symlink_readonly_path && unlink test_syd_symlink_readonly_path",
        ])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_remove_empty_path() -> TestResult {
    if !is_program_available("sh") {
        eprintln!("Test requires sh, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }
    let status = syd()
        .args([
            "-ppaludis",
            "-msandbox/read:on",
            "-msandbox/stat:on",
            "-msandbox/write:on",
            "-mallow/read+/***",
            "-mallow/stat+/***",
            "-mallow/write+/***",
            "--",
            "sh",
            "-c",
            "env LC_ALL=C LANG=C LANGUAGE=C rm '' 2>&1 | tee /dev/stderr | grep -qi 'No such file or directory'"
        ])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_open_trailing_slash() -> TestResult {
    let status = syd()
        .env("SYD_DO", "open_trailing_slash")
        .args([
            "-ppaludis",
            "-msandbox/read:on",
            "-msandbox/stat:on",
            "-msandbox/write:on",
            "-mallow/read+/***",
            "-mallow/stat+/***",
            "-mallow/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_openat_trailing_slash() -> TestResult {
    let status = syd()
        .env("SYD_DO", "openat_trailing_slash")
        .args([
            "-ppaludis",
            "-msandbox/read:on",
            "-msandbox/stat:on",
            "-msandbox/write:on",
            "-mallow/read+/***",
            "-mallow/stat+/***",
            "-mallow/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_lstat_trailing_slash() -> TestResult {
    let status = syd()
        .env("SYD_DO", "lstat_trailing_slash")
        .args([
            "-ppaludis",
            "-msandbox/read:on",
            "-msandbox/stat:on",
            "-msandbox/write:on",
            "-mallow/read+/***",
            "-mallow/stat+/***",
            "-mallow/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_fstatat_trailing_slash() -> TestResult {
    let status = syd()
        .env("SYD_DO", "fstatat_trailing_slash")
        .args([
            "-ppaludis",
            "-msandbox/read:on",
            "-msandbox/stat:on",
            "-msandbox/write:on",
            "-mallow/read+/***",
            "-mallow/stat+/***",
            "-mallow/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_mkdir_trailing_dot() -> TestResult {
    let status = syd()
        .env("SYD_DO", "mkdir_trailing_dot")
        .args([
            "-ppaludis",
            "-msandbox/read:on",
            "-msandbox/stat:on",
            "-msandbox/write:on",
            "-mallow/read+/***",
            "-mallow/stat+/***",
            "-mallow/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_mkdirat_trailing_dot() -> TestResult {
    let status = syd()
        .env("SYD_DO", "mkdirat_trailing_dot")
        .args([
            "-ppaludis",
            "-msandbox/read:on",
            "-msandbox/stat:on",
            "-msandbox/write:on",
            "-mallow/read+/***",
            "-mallow/stat+/***",
            "-mallow/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_rmdir_trailing_slashdot() -> TestResult {
    let status = syd()
        .env("SYD_DO", "rmdir_trailing_slashdot")
        .args([
            "-ppaludis",
            "-msandbox/read:on",
            "-msandbox/stat:on",
            "-msandbox/write:on",
            "-mallow/read+/***",
            "-mallow/stat+/***",
            "-mallow/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_mkdir_eexist_escape() -> TestResult {
    let status = syd()
        .env("SYD_DO", "mkdir_eexist_escape")
        .args([
            "-ppaludis",
            "-msandbox/read:on",
            "-msandbox/stat:on",
            "-msandbox/write:on",
            "-mallow/read+/***",
            "-mallow/stat+/***",
            "-mallow/write+/***",
            "-mdeny/read,stat,write+/boot/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_mkdirat_eexist_escape() -> TestResult {
    let status = syd()
        .env("SYD_DO", "mkdirat_eexist_escape")
        .args([
            "-ppaludis",
            "-msandbox/read:on",
            "-msandbox/stat:on",
            "-msandbox/write:on",
            "-mallow/read+/***",
            "-mallow/stat+/***",
            "-mallow/write+/***",
            "-mdeny/read,stat,write+/boot/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_mknod_eexist_escape() -> TestResult {
    let status = syd()
        .env("SYD_DO", "mknod_eexist_escape")
        .args([
            "-ppaludis",
            "-msandbox/read:on",
            "-msandbox/stat:on",
            "-msandbox/write:on",
            "-mallow/read+/***",
            "-mallow/stat+/***",
            "-mallow/write+/***",
            "-mdeny/read,stat,write+/boot/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_mknodat_eexist_escape() -> TestResult {
    let status = syd()
        .env("SYD_DO", "mknodat_eexist_escape")
        .args([
            "-ppaludis",
            "-msandbox/read:on",
            "-msandbox/stat:on",
            "-msandbox/write:on",
            "-mallow/read+/***",
            "-mallow/stat+/***",
            "-mallow/write+/***",
            "-mdeny/read,stat,write+/boot/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_fopen_supports_mode_x() -> TestResult {
    let status = syd()
        .env("SYD_DO", "fopen_supports_mode_x")
        .args([
            "-ppaludis",
            "-msandbox/read:on",
            "-msandbox/stat:on",
            "-msandbox/write:on",
            "-mallow/read+/***",
            "-mallow/stat+/***",
            "-mallow/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_fopen_supports_mode_e() -> TestResult {
    let status = syd()
        .env("SYD_DO", "fopen_supports_mode_e")
        .args([
            "-ppaludis",
            "-msandbox/read:on",
            "-msandbox/stat:on",
            "-msandbox/write:on",
            "-mallow/read+/***",
            "-mallow/stat+/***",
            "-mallow/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_link_no_symlink_deref() -> TestResult {
    let status = syd()
        .env("SYD_DO", "link_no_symlink_deref")
        .args([
            "-ppaludis",
            "-msandbox/read:on",
            "-msandbox/stat:on",
            "-msandbox/write:on",
            "-mallow/read+/***",
            "-mallow/stat+/***",
            "-mallow/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_link_posix() -> TestResult {
    let status = syd()
        .env("SYD_DO", "link_posix")
        .args([
            "-ppaludis",
            "-msandbox/read:on",
            "-msandbox/stat:on",
            "-msandbox/write:on",
            "-mallow/read+/***",
            "-mallow/stat+/***",
            "-mallow/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_linkat_posix() -> TestResult {
    let status = syd()
        .env("SYD_DO", "linkat_posix")
        .args([
            "-ppaludis",
            "-msandbox/read:on",
            "-msandbox/stat:on",
            "-msandbox/write:on",
            "-mallow/read+/***",
            "-mallow/stat+/***",
            "-mallow/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_getcwd_long() -> TestResult {
    let status = syd()
        .env("SYD_DO", "getcwd_long")
        .args([
            "-ppaludis",
            "-msandbox/read:on",
            "-msandbox/stat:on",
            "-msandbox/write:on",
            "-mallow/read,stat,write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");

    // FIXME: Fails on musl, possibly not something to fix.
    ignore!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_creat_thru_dangling() -> TestResult {
    let status = syd()
        .env("SYD_DO", "creat_thru_dangling")
        .args([
            "-ppaludis",
            "-msandbox/read:on",
            "-msandbox/stat:on",
            "-msandbox/write:on",
            "-mallow/read+/***",
            "-mallow/stat+/***",
            "-mallow/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_mkdirat_non_dir_fd() -> TestResult {
    let status = syd()
        .env("SYD_DO", "mkdirat_non_dir_fd")
        .args([
            "-ppaludis",
            "-msandbox/read:on",
            "-msandbox/stat:on",
            "-msandbox/write:on",
            "-mallow/read+/***",
            "-mallow/stat+/***",
            "-mallow/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_blocking_udp4() -> TestResult {
    let status = syd()
        .env("SYD_DO", "blocking_udp4")
        .args([
            "-ppaludis",
            "-msandbox/read:on",
            "-msandbox/stat:on",
            "-msandbox/write:on",
            "-msandbox/net:on",
            "-mtrace/allow_safe_bind:0",
            "-mallow/read+/***",
            "-mallow/stat+/***",
            "-mallow/write+/***",
            "-mallow/net/bind+loopback!65432",
            "-mallow/net/connect+loopback!65432",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_blocking_udp6() -> TestResult {
    // Gitlab CI uses docker which has no IPv6.
    if *GL_BUILD {
        eprintln!("IPv6 not available on CI!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }
    let status = syd()
        .env("SYD_DO", "blocking_udp6")
        .args([
            "-ppaludis",
            "-msandbox/read:on",
            "-msandbox/stat:on",
            "-msandbox/write:on",
            "-msandbox/net:on",
            "-mtrace/allow_safe_bind:0",
            "-mallow/read,stat,write+/***",
            "-mallow/net/bind+loopback6!65432",
            "-mallow/net/connect+loopback6!65432",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_close_on_exec() -> TestResult {
    let status = syd()
        .env("SYD_DO", "close_on_exec")
        .args([
            "-ppaludis",
            "-msandbox/read:on",
            "-msandbox/stat:on",
            "-msandbox/write:on",
            "-msandbox/net:on",
            "-mallow/read+/***",
            "-mallow/stat+/***",
            "-mallow/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_open_exclusive_restart() -> TestResult {
    let status = syd()
        .env("SYD_DO", "open_exclusive_restart")
        .args([
            "-plib",
            "-msandbox/read:on",
            "-msandbox/stat:on",
            "-msandbox/write:on",
            "-msandbox/net:on",
            "-mallow/read,stat,write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    // FIXME: This is a kernel bug, mixi will report it, check dev/seccomp_poc_no_lib.c
    ignore!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_open_exclusive_repeat() -> TestResult {
    let status = syd()
        .env("SYD_DO", "open_exclusive_repeat")
        .args([
            "-plib",
            "-msandbox/read:on",
            "-msandbox/stat:on",
            "-msandbox/write:on",
            "-msandbox/net:on",
            "-mallow/read,stat,write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_pty_io_rust() -> TestResult {
    env::set_var("SYD_TEST_TIMEOUT", "30s");
    let status = syd()
        .env("SYD_DO", "pty_io_rust")
        .args([
            "-ppaludis",
            "-msandbox/read:on",
            "-msandbox/stat:on",
            "-msandbox/write:on",
            "-msandbox/net:on",
            "-mallow/read,stat,write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    env::remove_var("SYD_TEST_TIMEOUT");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_pty_io_gawk() -> TestResult {
    if !is_program_available("gawk") {
        eprintln!("Test requires gawk, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }
    env::set_var("SYD_TEST_TIMEOUT", "30s");
    let status = syd()
        .env("SYD_DO", "pty_io_gawk")
        .args([
            "-ppaludis",
            "-msandbox/read:on",
            "-msandbox/stat:on",
            "-msandbox/write:on",
            "-msandbox/net:on",
            "-mallow/read+/***",
            "-mallow/stat+/***",
            "-mallow/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    env::remove_var("SYD_TEST_TIMEOUT");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_diff_dev_fd() -> TestResult {
    if !is_program_available("diff") {
        eprintln!("Test requires diff, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }
    env::set_var("SYD_TEST_TIMEOUT", "30s");
    let status = syd()
        .env("SYD_DO", "diff_dev_fd")
        .args([
            "-ppaludis",
            "-msandbox/read:on",
            "-msandbox/stat:on",
            "-msandbox/write:on",
            "-msandbox/net:on",
            "-mallow/read+/***",
            "-mallow/stat+/***",
            "-mallow/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    env::remove_var("SYD_TEST_TIMEOUT");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_fifo_multiple_readers() -> TestResult {
    if !is_program_available("bash") {
        eprintln!("Test requires bash, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }
    if !is_program_available("nproc") {
        eprintln!("Test requires nproc, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }
    env::set_var("SYD_TEST_TIMEOUT", "15s");
    let status = syd()
        .args([
            "-plib",
            "-msandbox/read:on",
            "-msandbox/stat:on",
            "-msandbox/write:on",
            "-mallow/read,stat,write+/***",
            "--",
            "bash",
            "-c",
        ])
        .arg(
            r#"
# Attempt to DOS syd by spawning multiple FIFO readers in the background.
set -ex
nreaders=$(expr $(nproc) '*' 2)
fifo=$(env TMPDIR=. mktemp -u)
mkfifo "$fifo"
for i in $(eval echo {1..${nreaders}}); do
    cat "$fifo" &
done
# Give the cats a little time to settle and potentially block syd.
sleep 3
# Execute a system call that syd must intervene, this must not block.
touch "$fifo".done
rm -f "$fifo".done
# All good, unblock the cats and wait.
:>"$fifo"
wait
rm -f "$fifo" || true
"#,
        )
        .status()
        .expect("execute syd");
    env::remove_var("SYD_TEST_TIMEOUT");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_bind_unix_socket() -> TestResult {
    let status = syd()
        .env("SYD_DO", "bind_unix_socket")
        .args([
            "-ppaludis",
            "-msandbox/read:on",
            "-msandbox/stat:on",
            "-msandbox/write:on",
            "-msandbox/net:on",
            "-mallow/read+/***",
            "-mallow/stat+/***",
            "-mallow/write+/***",
            "-mallow/net/bind+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_signal_protection() -> TestResult {
    if !is_program_available("bash") {
        eprintln!("Test requires bash, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }
    match check_unshare() {
        Some(false) => {
            eprintln!("Test requires Linux namespaces, skipping!");
            env::set_var("SYD_TEST_SOFT_FAIL", "1");
            return Ok(());
        }
        None => {
            return Err(TestError(
                "Failed to test for Linux namespaces!".to_string(),
            ));
        }
        _ => {}
    };

    let status = syd()
        .args([
            "-ppaludis",
            "-pcontainer",
            "-mdeny/read-/proc/1/**",
            "-mdeny/stat-/proc/1/**",
            "-mdeny/write-/proc/1/***",
            "--",
            "bash",
            "-c",
        ])
        .arg(
            r#"
set -ex
pid=( $(pgrep -w syd) )
# Sending signal 0 must work.
for tid in "${pid[@]}"; do
    kill -0 ${tid}
    sleep 1
done
# Sending other signals are not permitted.
for sig in INT ABRT STOP KILL; do
    for tid in "${pid[@]}"; do
        kill -${sig} ${tid} && exit 1
        sleep 1
    done
done
"#,
        )
        .status()
        .expect("execute syd");

    assert_status_ok!(status);
    Ok(())
}

fn test_syd_emulate_open_fifo_1() -> TestResult {
    let status = syd()
        .env("SYD_DO", "emulate_open_fifo_1")
        .args([
            "-plib",
            "-msandbox/read:on",
            "-msandbox/stat:on",
            "-msandbox/write:on",
            "-mallow/read,stat,write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_emulate_open_fifo_2() -> TestResult {
    let status = syd()
        .env("SYD_DO", "emulate_open_fifo_2")
        .args([
            "-plib",
            "-msandbox/read:on",
            "-msandbox/stat:on",
            "-msandbox/write:on",
            "-mallow/read,stat,write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_deny_magiclinks() -> TestResult {
    // Test expects PID namespace.
    match check_unshare() {
        Some(false) => {
            eprintln!("Test requires Linux namespaces, skipping!");
            env::set_var("SYD_TEST_SOFT_FAIL", "1");
            return Ok(());
        }
        None => {
            return Err(TestError(
                "Failed to test for Linux namespaces!".to_string(),
            ));
        }
        _ => {}
    }

    // Check protections with stat sandboxing on.
    eprintln!("\x1b[36m<<< paludis >>>\x1b[0m");
    let status = syd()
        .env("SYD_DO", "deny_magiclinks")
        .args([
            "-ppaludis",
            "-munshare/user:1",
            "-munshare/pid:1",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    // Check protections with stat sandboxing off.
    eprintln!("\x1b[36m<<< lib >>>\x1b[0m");
    let status = syd()
        .env("SYD_DO", "deny_magiclinks")
        .args([
            "-plib",
            "-munshare/user:1",
            "-munshare/pid:1",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    // Check protections with stat sandboxing off and lock on.
    eprintln!("\x1b[36m<<< lib with lock on >>>\x1b[0m");
    let status = syd()
        .env("SYD_DO", "deny_magiclinks")
        .args([
            "-plib",
            "-munshare/user:1",
            "-munshare/pid:1",
            "-mlock:on",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_lstat_magiclinks() -> TestResult {
    // Test expects PID namespace.
    match check_unshare() {
        Some(false) => {
            eprintln!("Test requires Linux namespaces, skipping!");
            env::set_var("SYD_TEST_SOFT_FAIL", "1");
            return Ok(());
        }
        None => {
            return Err(TestError(
                "Failed to test for Linux namespaces!".to_string(),
            ));
        }
        _ => {}
    }

    // Check protections with stat sandboxing on.
    eprintln!("\x1b[36m<<< paludis >>>\x1b[0m");
    let status = syd()
        .env("SYD_DO", "lstat_magiclinks")
        .args([
            "-ppaludis",
            "-munshare/user:1",
            "-munshare/pid:1",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    // Check protections with stat sandboxing off.
    eprintln!("\x1b[36m<<< lib >>>\x1b[0m");
    let status = syd()
        .env("SYD_DO", "lstat_magiclinks")
        .args([
            "-plib",
            "-munshare/user:1",
            "-munshare/pid:1",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    // Check protections with stat sandboxing off and lock on.
    eprintln!("\x1b[36m<<< lib with lock on >>>\x1b[0m");
    let status = syd()
        .env("SYD_DO", "lstat_magiclinks")
        .args([
            "-plib",
            "-munshare/user:1",
            "-munshare/pid:1",
            "-mlock:on",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_access_unsafe_paths() -> TestResult {
    // Check protections with read, stat, write sandboxing on.
    eprintln!("\x1b[36m<<< paludis >>>\x1b[0m");
    let status = syd()
        .env("SYD_DO", "access_unsafe_paths")
        .args(["-ppaludis", "--", &SYD_DO])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    // Check protections with sandboxing off.
    eprintln!("\x1b[36m<<< lib >>>\x1b[0m");
    let status = syd()
        .env("SYD_DO", "access_unsafe_paths")
        .args(["-plib", "--", &SYD_DO])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    // Check protections with sandboxing off and lock on.
    eprintln!("\x1b[36m<<< lib with lock on >>>\x1b[0m");
    let status = syd()
        .env("SYD_DO", "access_unsafe_paths")
        .args(["-plib", "-mlock:on", "--", &SYD_DO])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_access_proc_cmdline() -> TestResult {
    for p in &["cat", "sh"] {
        if !is_program_available(p) {
            eprintln!("Test requires {p}, skipping!");
            env::set_var("SYD_TEST_SOFT_FAIL", "1");
            return Ok(());
        }
    }
    let status = syd()
        .arg("-plib")
        .args(["sh", "-cx"])
        .arg(
            r#"
cmdline=$(cat /proc/cmdline)
if test -n "$cmdline"; then
    echo >&2 "/proc/cmdline leaked with sandboxing off."
    false
else
    echo >&2 "/proc/cmdline is empty as expected."
    true
fi
        "#,
        )
        .status()
        .expect("execute syd");
    assert_status_code!(status, 1);

    let status = syd()
        .arg("-plib")
        .arg("-msandbox/read:on")
        .arg("-mallow/read+/***")
        .args(["sh", "-cx"])
        .arg(
            r#"
cmdline=$(cat /proc/cmdline)
if test -n "$cmdline"; then
    echo >&2 "/proc/cmdline leaked with sandboxing off."
    false
else
    echo >&2 "/proc/cmdline is empty as expected."
    true
fi
        "#,
        )
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_mkdir_with_control_chars_default() -> TestResult {
    if !is_program_available("bash") {
        eprintln!("Test requires bash, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }

    // Create pipe for communication and check if syd raises an access violation.
    let (fd_rd, fd_rw) = match pipe() {
        Ok((fd_rd, fd_rw)) => (fd_rd, fd_rw),
        Err(errno) => return Err(TestError(format!("error creating pipe: {errno}!"))),
    };

    let mut child = syd()
        .env("SYD_LOG", "warn")
        .env("SYD_LOG_FD", fd_rw.to_string())
        .args([
            "-plib",
            "-msandbox/read:on",
            "-msandbox/stat:on",
            "-msandbox/write:on",
            "-mallow/read,stat,write+/***",
            "--",
            "bash",
            "-cx",
        ])
        .arg(
            r##"
#!/bin/bash
r=0
mkdir mccd || exit 127
mkdir $'./mccd/test_alert_dir\a' && r=1
test -e $'./mccd/test_alert_dir\a' && r=2
mkdir $'./mccd/test_vertical_tab_dir\v' && r=3
test -e $'./mccd/test_vertical_tab_dir\v' && r=4
mkdir $'./mccd/test_form_feed_dir\f' && r=5
test -e $'./mccd/test_form_feed_dir\f' && r=6
mkdir $'./mccd/test_multi_control_dir\x01\x02\x03' && r=7
test -e $'./mccd/test_multi_control_dir\x01\x02\x03' && r=8
mkdir $'./mccd/test_\x1F_unit_sep_dir' && r=9
test -e $'./mccd/test_\x1F_unit_sep_dir' && r=10
exit $r
    "##,
        )
        .spawn()
        .expect("spawn syd");
    let _ = close(fd_rw);

    let status = child.wait().expect("wait for syd");
    assert_status_ok!(status);

    // Convert raw file descriptor to File, then to BufReader
    let file = unsafe { File::from_raw_fd(fd_rd) };
    let mut reader = BufReader::new(file);

    let mut buffer = String::new();
    match reader.read_to_string(&mut buffer) {
        Ok(0) => {
            eprintln!("No data read from pipe; expected access violation.");
            return Err(TestError(
                "Expected access violation not logged.".to_string(),
            ));
        }
        Ok(_) if buffer.contains("\"access\"") => {
            // If any data was read, log it
            eprint!("Access violation logged:\n{buffer}");
        }
        Ok(_) => {
            eprint!("Unexpected data, expected access violation:\n{buffer}");
            return Err(TestError(
                "Expected access violation not logged.".to_string(),
            ));
        }
        Err(e) => {
            // Error reading from the buffer
            eprintln!("Error reading from pipe: {e}");
            return Err(TestError(format!("Error reading from pipe: {e}")));
        }
    }

    Ok(())
}

fn test_syd_mkdir_with_control_chars_unsafe() -> TestResult {
    if !is_program_available("bash") {
        eprintln!("Test requires bash, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }

    let status = syd()
        .args([
            "-plib",
            "-mtrace/allow_unsafe_filename:1",
            "-msandbox/read:on",
            "-msandbox/stat:on",
            "-msandbox/write:on",
            "-mallow/read,stat,write+/***",
            "--",
            "bash",
            "-cx",
        ])
        .arg(
            r##"
#!/bin/bash
r=0
mkdir mccu || exit 127
mkdir $'./mccu/test_alert_dir\a' || r=1
test -e $'./mccu/test_alert_dir\a' || r=2
mkdir $'./mccu/test_vertical_tab_dir\v' || r=3
test -e $'./mccu/test_vertical_tab_dir\v' || r=4
mkdir $'./mccu/test_form_feed_dir\f' || r=5
test -e $'./mccu/test_form_feed_dir\f' || r=6
mkdir $'./mccu/test_multi_control_dir\x01\x02\x03' || r=7
test -e $'./mccu/test_multi_control_dir\x01\x02\x03' || r=8
mkdir $'./mccu/test_\x1F_unit_sep_dir' || r=9
test -e $'./mccu/test_\x1F_unit_sep_dir' || r=10
exit $r
    "##,
        )
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_touch_with_control_chars_default() -> TestResult {
    if !is_program_available("bash") {
        eprintln!("Test requires bash, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }

    // Create pipe for communication and check if syd raises an access violation.
    let (fd_rd, fd_rw) = match pipe() {
        Ok((fd_rd, fd_rw)) => (fd_rd, fd_rw),
        Err(errno) => return Err(TestError(format!("error creating pipe: {errno}!"))),
    };

    let mut child = syd()
        .env("SYD_LOG", "warn")
        .env("SYD_LOG_FD", fd_rw.to_string())
        .args([
            "-plib",
            "-msandbox/read:on",
            "-msandbox/stat:on",
            "-msandbox/write:on",
            "-mallow/read,stat,write+/***",
            "--",
            "bash",
            "-cx",
        ])
        .arg(
            r##"
#!/bin/bash
r=0
mkdir tccd || exit 127
touch $'./tccd/test_alert_file\a' && r=1
test -e $'./tccd/test_alert_file\a' && r=2
touch $'./tccd/test_vertical_tab_file\v' && r=3
test -e $'./tccd/test_vertical_tab_file\v' && r=4
touch $'./tccd/test_form_feed_file\f' && r=5
test -e $'./tccd/test_form_feed_file\f' && r=6
touch $'./tccd/test_multi_control_file\x01\x02\x03' && r=7
test -e $'./tccd/test_multi_control_file\x01\x02\x03' && r=8
touch $'./tccd/test_\x1F_unit_sep_file' && r=9
test -e $'./tccd/test_\x1F_unit_sep_file' && r=10
exit $r
    "##,
        )
        .spawn()
        .expect("spawn syd");
    let _ = close(fd_rw);

    let status = child.wait().expect("wait for syd");
    assert_status_ok!(status);

    // Convert raw file descriptor to File, then to BufReader
    let file = unsafe { File::from_raw_fd(fd_rd) };
    let mut reader = BufReader::new(file);

    let mut buffer = String::new();
    match reader.read_to_string(&mut buffer) {
        Ok(0) => {
            eprintln!("No data read from pipe; expected access violation.");
            return Err(TestError(
                "Expected access violation not logged.".to_string(),
            ));
        }
        Ok(_) if buffer.contains("\"access\"") => {
            // If any data was read, log it
            eprint!("Access violation logged:\n{buffer}");
        }
        Ok(_) => {
            eprint!("Unexpected data, expected access violation:\n{buffer}");
            return Err(TestError(
                "Expected access violation not logged.".to_string(),
            ));
        }
        Err(e) => {
            // Error reading from the buffer
            eprintln!("Error reading from pipe: {e}");
            return Err(TestError(format!("Error reading from pipe: {e}")));
        }
    }

    Ok(())
}

fn test_syd_touch_with_control_chars_unsafe() -> TestResult {
    if !is_program_available("bash") {
        eprintln!("Test requires bash, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }

    let status = syd()
        .args([
            "-plib",
            "-mtrace/allow_unsafe_filename:true",
            "-msandbox/read:on",
            "-msandbox/stat:on",
            "-msandbox/write:on",
            "-mallow/read,stat,write+/***",
            "--",
            "bash",
            "-cx",
        ])
        .arg(
            r##"
#!/bin/bash
r=0
mkdir tccu || exit 127
touch $'./tccu/test_alert_file\a' || r=1
test -e $'./tccu/test_alert_file\a' || r=2
touch $'./tccu/test_vertical_tab_file\v' || r=3
test -e $'./tccu/test_vertical_tab_file\v' || r=4
touch $'./tccu/test_form_feed_file\f' || r=5
test -e $'./tccu/test_form_feed_file\f' || r=6
touch $'./tccu/test_multi_control_file\x01\x02\x03' || r=7
test -e $'./tccu/test_multi_control_file\x01\x02\x03' || r=8
touch $'./tccu/test_\x1F_unit_sep_file' || r=9
test -e $'./tccu/test_\x1F_unit_sep_file' || r=10
exit $r
    "##,
        )
        .status()
        .expect("spawn syd");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_unshare_user_bypass_limit() -> TestResult {
    match check_unshare() {
        Some(false) => {
            eprintln!("Test requires Linux namespaces, skipping!");
            env::set_var("SYD_TEST_SOFT_FAIL", "1");
            return Ok(());
        }
        None => {
            return Err(TestError(
                "Failed to test for Linux namespaces!".to_string(),
            ));
        }
        _ => {}
    }

    let status = syd()
        .env("SYD_DO", "unshare_user_bypass_limit")
        .args(["-plib", "-pcontainer", "--", &SYD_DO])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_interrupt_mkdir() -> TestResult {
    let status = syd()
        .env("SYD_DO", "interrupt_mkdir")
        .args([
            "-plib",
            "-msandbox/read:on",
            "-msandbox/stat:on",
            "-msandbox/write:on",
            "-msandbox/net:on",
            "-mallow/read,stat,write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    // FIXME: This is a kernel bug, mixi will report it, check dev/seccomp_poc_no_lib.c
    ignore!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_interrupt_bind_ipv4() -> TestResult {
    let status = syd()
        .env("SYD_DO", "interrupt_bind_ipv4")
        .args([
            "-plib",
            "-msandbox/read:on",
            "-msandbox/stat:on",
            "-msandbox/write:on",
            "-msandbox/net:on",
            "-mallow/read,stat,write+/***",
            "-mallow/net/bind+loopback!65432",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    // FIXME: This is a kernel bug, mixi will report it, they have a POC.
    ignore!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_interrupt_bind_unix() -> TestResult {
    let status = syd()
        .env("SYD_DO", "interrupt_bind_unix")
        .args([
            "-plib",
            "-msandbox/read:on",
            "-msandbox/stat:on",
            "-msandbox/write:on",
            "-msandbox/net:on",
            "-mallow/read,stat,write+/***",
            "-mallow/net/bind+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    // FIXME: This is a kernel bug, mixi will report it, check dev/seccomp_poc_no_lib.c
    ignore!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_interrupt_connect_ipv4() -> TestResult {
    let status = syd()
        .env("SYD_DO", "interrupt_connect_ipv4")
        .args([
            "-plib",
            "-msandbox/read:on",
            "-msandbox/stat:on",
            "-msandbox/write:on",
            "-msandbox/net:on",
            "-mallow/read,stat,write+/***",
            "-mallow/net/bind+loopback!65432",
            "-mallow/net/connect+loopback!65432",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    // FIXME: This is a kernel bug, mixi will report it, they have a POC.
    ignore!(status.success(), "status:{status:?}");
    Ok(())
}

/*
fn test_syd_repetitive_clone() -> TestResult {
    if !is_program_available("bash") {
        eprintln!("Test requires bash, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    };
    let status = syd()
        .env("SYD_DO", "repetitive_clone")
        .args(["-ppaludis", "--", &SYD_DO])
        .status()
        .expect("execute syd");
    assert!(
        status.code().unwrap_or(0) == 128 + nix::libc::SIGKILL,
        "status:{status:?}"
    );
    Ok(())
}
*/

fn test_syd_exp_syscall_fuzz_bare() -> TestResult {
    let epoch = std::time::Instant::now();
    let status = syd()
        .env("SYD_DO", "syscall_fuzz")
        .args([
            "-ppaludis",
            "-mlock:off",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-msandbox/stat:on",
            "-msandbox/write:on",
            "-msandbox/exec:on",
            "-msandbox/net:on",
            "-mallow/read+/***",
            "-mallow/stat+/***",
            "-mallow/exec+/***",
            "-mallow/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    let code = status.code().unwrap_or(127);
    let time = format_duration(epoch.elapsed());
    println!("# fuzz completed in {time} with code {code}.");
    ignore!(code == 0, "status:{status:?}");
    Ok(())
}

fn test_syd_exp_syscall_fuzz_wrap() -> TestResult {
    match check_unshare() {
        Some(false) => {
            eprintln!("Test requires Linux namespaces, skipping!");
            env::set_var("SYD_TEST_SOFT_FAIL", "1");
            return Ok(());
        }
        None => {
            return Err(TestError(
                "Failed to test for Linux namespaces!".to_string(),
            ));
        }
        _ => {}
    }

    let epoch = std::time::Instant::now();
    let status = syd()
        .env("SYD_DO", "syscall_fuzz")
        .args([
            "-ppaludis",
            "-pcontainer",
            "-mlock:off",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-msandbox/stat:on",
            "-msandbox/write:on",
            "-msandbox/exec:on",
            "-msandbox/net:on",
            "-mallow/read+/***",
            "-mallow/stat+/***",
            "-mallow/exec+/***",
            "-mallow/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    let code = status.code().unwrap_or(127);
    let time = format_duration(epoch.elapsed());
    println!("# fuzz completed in {time} with code {code}.");
    ignore!(code == 0, "status:{status:?}");
    Ok(())
}

fn test_syd_pid_thread_deny() -> TestResult {
    match check_pid() {
        Some(false) => {
            eprintln!("Test requires PID sandboxing to work in a container, skipping!");
            env::set_var("SYD_TEST_SOFT_FAIL", "1");
            return Ok(());
        }
        None => {
            return Err(TestError("Failed to test for Pid Sanboxing!".to_string()));
        }
        _ => {}
    }

    let status = syd()
        .env("SYD_DO", "thread")
        .args([
            "-ppaludis",
            "-pcontainer",
            "-msandbox/pid:on",
            "-mpid/max:1",
            "--",
            &SYD_DO,
            "0",
            "24",
        ])
        .status()
        .expect("execute syd");
    // Early fails get EACCES ie 13.
    // In debug mode, this fails with 101.
    // In release mode, this fails with 134.
    // Both indicate thread::spawn paniced.
    assert!(
        matches!(status.code().unwrap_or(127), 13 | 101 | 134),
        "status:{status:?}"
    );

    Ok(())
}

fn test_syd_pid_fork_deny() -> TestResult {
    match check_pid() {
        Some(false) => {
            eprintln!("Test requires PID sandboxing to work in a container, skipping!");
            env::set_var("SYD_TEST_SOFT_FAIL", "1");
            return Ok(());
        }
        None => {
            return Err(TestError("Failed to test for Pid Sanboxing!".to_string()));
        }
        _ => {}
    }

    let status = syd()
        .env("SYD_DO", "fork")
        .args([
            "-ppaludis",
            "-pcontainer",
            "-msandbox/pid:on",
            "-mpid/max:16",
            "--",
            &SYD_DO,
            "0",
            "24",
        ])
        .status()
        .expect("execute syd");
    env::remove_var("SYD_DO");
    assert!(
        status.code().unwrap_or(127) == nix::libc::EACCES,
        "status:{status:?}"
    );

    Ok(())
}

fn test_syd_pid_fork_bomb() -> TestResult {
    match check_pid() {
        Some(false) => {
            eprintln!("Test requires PID sandboxing to work in a container, skipping!");
            env::set_var("SYD_TEST_SOFT_FAIL", "1");
            return Ok(());
        }
        None => {
            return Err(TestError("Failed to test for Pid Sanboxing!".to_string()));
        }
        _ => {
            eprintln!("PID sandboxing works in a container!");
            eprintln!("Proceeding with test...");
        }
    }

    eprintln!("Running the fork bomb test for 10 seconds...");
    env::set_var("SYD_TEST_TIMEOUT", "10s");
    let child = syd()
        .env("SYD_DO", "fork_bomb")
        .env("SYD_TEST_FORCE", "IKnowWhatIAmDoing")
        .args([
            "-plib",
            "-pcontainer",
            "-msandbox/pid:on",
            "-mpid/max:16",
            "--",
            &SYD_DO,
        ])
        .stdout(Stdio::inherit())
        .stderr(Stdio::piped())
        .spawn()
        .expect("execute syd");
    env::remove_var("SYD_TEST_TIMEOUT");

    eprintln!("Done spawning test, waiting test to complete...");
    let output = child.wait_with_output().expect("wait syd");
    let output = String::from_utf8_lossy(&output.stderr);
    eprintln!("{output}");

    // Check if the output contains the specified string and assert based on it
    assert!(
        output.contains(r#""cap":"p""#),
        "syd has not raised a pid violation!"
    );

    Ok(())
}

fn test_syd_pid_fork_bomb_asm() -> TestResult {
    match check_pid() {
        Some(false) => {
            eprintln!("Test requires PID sandboxing to work in a container, skipping!");
            env::set_var("SYD_TEST_SOFT_FAIL", "1");
            return Ok(());
        }
        None => {
            return Err(TestError("Failed to test for Pid Sanboxing!".to_string()));
        }
        _ => {
            eprintln!("PID sandboxing works in a container!");
            eprintln!("Proceeding with test...");
        }
    }

    eprintln!("Running the fork bomb test for 10 seconds...");
    env::set_var("SYD_TEST_TIMEOUT", "10s");
    let child = syd()
        .env("SYD_DO", "fork_bomb_asm")
        .env("SYD_TEST_FORCE", "IKnowWhatIAmDoing")
        .args([
            "-plib",
            "-pcontainer",
            "-msandbox/pid:on",
            "-mpid/max:16",
            "--",
            &SYD_DO,
        ])
        .stdout(Stdio::inherit())
        .stderr(Stdio::piped())
        .spawn()
        .expect("execute syd");
    env::remove_var("SYD_TEST_TIMEOUT");

    eprintln!("Done spawning test, waiting test to complete...");
    let output = child.wait_with_output().expect("wait syd");
    let output = String::from_utf8_lossy(&output.stderr);
    eprintln!("{output}");

    // Check if the output contains the specified string and assert based on it
    assert!(
        output.contains(r#""cap":"p""#),
        "syd has not raised a pid violation!"
    );

    Ok(())
}

fn test_syd_pid_thread_bomb() -> TestResult {
    match check_pid() {
        Some(false) => {
            eprintln!("Test requires PID sandboxing to work in a container, skipping!");
            env::set_var("SYD_TEST_SOFT_FAIL", "1");
            return Ok(());
        }
        None => {
            return Err(TestError("Failed to test for Pid Sanboxing!".to_string()));
        }
        _ => {
            eprintln!("PID sandboxing works in a container!");
            eprintln!("Proceeding with test...");
        }
    }

    eprintln!("Running the thread bomb test for 10 seconds...");
    env::set_var("SYD_TEST_TIMEOUT", "10s");
    let child = syd()
        .env("SYD_DO", "thread_bomb")
        .env("SYD_TEST_FORCE", "IKnowWhatIAmDoing")
        .args([
            "-plib",
            "-pcontainer",
            "-msandbox/pid:on",
            "-mpid/max:16",
            "--",
            &SYD_DO,
        ])
        .stdout(Stdio::inherit())
        .stderr(Stdio::piped())
        .spawn()
        .expect("execute syd");
    env::remove_var("SYD_TEST_TIMEOUT");

    eprintln!("Done spawning test, waiting test to complete...");
    let output = child.wait_with_output().expect("wait syd");
    let output = String::from_utf8_lossy(&output.stderr);
    eprintln!("{output}");

    // Check if the output contains the specified string and assert based on it
    assert!(
        output.contains(r#""cap":"p""#),
        "syd has not raised a pid violation!"
    );

    Ok(())
}

fn test_syd_exp_pid_stress_ng_kill() -> TestResult {
    if !is_program_available("stress-ng") {
        eprintln!("Test requires stress-ng, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    };
    match check_pid() {
        Some(false) => {
            eprintln!("Test requires PID sandboxing to work in a container, skipping!");
            env::set_var("SYD_TEST_SOFT_FAIL", "1");
            return Ok(());
        }
        None => {
            return Err(TestError("Failed to test for Pid Sanboxing!".to_string()));
        }
        _ => {}
    }

    let status = syd()
        .args([
            "-ppaludis",
            "-pcontainer",
            "-msandbox/pid:on",
            "-mpid/max:1",
            "-mpid/kill:t",
            "--",
            "stress-ng",
            "-c",
            "1",
            "-t",
            "7",
        ])
        .status()
        .expect("execute syd");
    assert!(
        status.code().unwrap_or(127) == 128 + nix::libc::SIGKILL,
        "status:{status:?}"
    );

    Ok(())
}

fn test_syd_exp_pid_stress_ng_allow() -> TestResult {
    if !is_program_available("stress-ng") {
        eprintln!("Test requires stress-ng, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    };
    match check_pid() {
        Some(false) => {
            eprintln!("Test requires PID sandboxing to work in a container, skipping!");
            env::set_var("SYD_TEST_SOFT_FAIL", "1");
            return Ok(());
        }
        None => {
            return Err(TestError("Failed to test for Pid Sanboxing!".to_string()));
        }
        _ => {}
    }

    let _ = remove_file("log");
    let status = syd()
        .args([
            "-ppaludis",
            "-pcontainer",
            "-msandbox/pid:on",
            "-mpid/max:2",
            "--",
            "stress-ng",
            "--log-file",
            "log",
            "-c",
            "1",
            "-t",
            "7",
        ])
        .status()
        .expect("execute syd");
    // Fails on CI.
    if !*CI_BUILD {
        assert_status_ok!(status);
    } else {
        ignore!(status.success(), "status:{status:?}");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }

    let mut file = File::open("log")?;
    let mut logs = String::new();
    file.read_to_string(&mut logs)?;

    assert!(!logs.contains("errno="), "logs:{logs:?}");

    Ok(())
}

fn test_syd_exp_pid_stress_ng_fork() -> TestResult {
    if !is_program_available("stress-ng") {
        eprintln!("Test requires stress-ng, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    };
    match check_pid() {
        Some(false) => {
            eprintln!("Test requires PID sandboxing to work in a container, skipping!");
            env::set_var("SYD_TEST_SOFT_FAIL", "1");
            return Ok(());
        }
        None => {
            return Err(TestError("Failed to test for Pid Sanboxing!".to_string()));
        }
        _ => {}
    }

    let _ = remove_file("log");
    let status = syd()
        .args([
            "-ppaludis",
            "-pcontainer",
            "-msandbox/pid:on",
            "-mpid/max:128",
            "-mfilter/pid:1",
            "--",
            "stress-ng",
            "--log-file",
            "log",
            "-f",
            "4",
            "-t",
            "15",
            "--fork-max",
            "1024",
        ])
        .status()
        .expect("execute syd");
    // Fails on CI.
    if !*CI_BUILD {
        assert_status_ok!(status);
    } else {
        ignore!(status.success(), "status:{status:?}");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }

    let mut file = File::open("log")?;
    let mut logs = String::new();
    file.read_to_string(&mut logs)?;
    assert!(!logs.contains("errno="), "logs:{logs:?}");

    Ok(())
}

fn test_syd_mem_alloc() -> TestResult {
    let status = syd()
        .env("SYD_DO", "alloc")
        .env("SYD_TEST_FORCE", "IKnowWhatIAmDoing")
        .args([
            "-ppaludis",
            "-msandbox/mem:on",
            "-mmem/max:128M",
            "-mmem/vm_max:256M",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    // Segmentation fault is expected.
    // Iot is confusing but happens on alpine+musl.
    // Otherwise we require ENOMEM.
    assert!(
        matches!(
            status.code().unwrap_or(127),
            nix::libc::ENOMEM | EX_SIGIOT | EX_SIGSEGV
        ),
        "status:{status:?}"
    );
    Ok(())
}

fn test_syd_exp_mem_stress_ng_malloc_1() -> TestResult {
    if !is_program_available("stress-ng") {
        eprintln!("Test requires stress-ng, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    };
    match check_unshare() {
        Some(false) => {
            eprintln!("Test requires Linux namespaces, skipping!");
            env::set_var("SYD_TEST_SOFT_FAIL", "1");
            return Ok(());
        }
        None => {
            return Err(TestError(
                "Failed to test for Linux namespaces!".to_string(),
            ));
        }
        _ => {}
    };

    let command = syd()
        .args([
            "-ppaludis",
            "-pcontainer",
            "-msandbox/mem:on",
            "-mmem/max:32M",
            "-mmem/vm_max:256M",
            "--",
            "stress-ng",
            "-v",
            "-t",
            "5",
            "--malloc",
            "4",
            "--malloc-bytes",
            "128M",
        ])
        .stdout(Stdio::inherit())
        .stderr(Stdio::piped())
        .spawn()
        .expect("spawn syd");

    let output = command.wait_with_output().expect("wait syd");
    let output = String::from_utf8_lossy(&output.stderr);
    eprintln!("{output}");
    assert!(output.contains(r#""cap":"m""#), "out:{output:?}");

    Ok(())
}

fn test_syd_exp_mem_stress_ng_malloc_2() -> TestResult {
    if !is_program_available("stress-ng") {
        eprintln!("Test requires stress-ng, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    };
    match check_unshare() {
        Some(false) => {
            eprintln!("Test requires Linux namespaces, skipping!");
            env::set_var("SYD_TEST_SOFT_FAIL", "1");
            return Ok(());
        }
        None => {
            return Err(TestError(
                "Failed to test for Linux namespaces!".to_string(),
            ));
        }
        _ => {}
    };

    let command = syd()
        .args([
            "-ppaludis",
            "-pcontainer",
            "-msandbox/mem:on",
            "-mmem/max:32M",
            "-mmem/vm_max:256M",
            "--",
            "stress-ng",
            "-v",
            "-t",
            "5",
            "--malloc",
            "4",
            "--malloc-bytes",
            "128M",
            "--malloc-touch",
        ])
        .stdout(Stdio::inherit())
        .stderr(Stdio::piped())
        .spawn()
        .expect("spawn syd");

    let output = command.wait_with_output().expect("wait syd");
    let output = String::from_utf8_lossy(&output.stderr);
    eprintln!("{output}");
    assert!(output.contains(r#""cap":"m""#), "out:{output:?}");

    Ok(())
}

fn test_syd_exp_mem_stress_ng_mmap() -> TestResult {
    if !is_program_available("stress-ng") {
        eprintln!("Test requires stress-ng, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    };
    match check_unshare() {
        Some(false) => {
            eprintln!("Test requires Linux namespaces, skipping!");
            env::set_var("SYD_TEST_SOFT_FAIL", "1");
            return Ok(());
        }
        None => {
            return Err(TestError(
                "Failed to test for Linux namespaces!".to_string(),
            ));
        }
        _ => {}
    };

    let command = syd()
        .args([
            "-ppaludis",
            "-pcontainer",
            "-msandbox/mem:on",
            "-mmem/max:16M",
            "-mmem/vm_max:64M",
            "--",
            "stress-ng",
            "-v",
            "-t",
            "5",
            "--mmap",
            "4",
            "--mmap-bytes",
            "1G",
        ])
        .stdout(Stdio::inherit())
        .stderr(Stdio::piped())
        .spawn()
        .expect("spawn syd");

    let output = command.wait_with_output().expect("wait syd");
    let output = String::from_utf8_lossy(&output.stderr);
    eprintln!("{output}");
    fixup!(output.contains(r#""cap":"m""#), "out:{output:?}");

    Ok(())
}
